grammar SysY;

compUnit: (decl | funcDef) +;
decl: constDecl | varDecl;
constDecl: 'const' bType constDef (',' constDef)* ';' ;

constDef: Ident ('[' constExp ']')* '=' constInitVal;
constInitVal:
        constExp
        | '{'( constInitVal (',' constInitVal)* )?'}';
varDecl: bType varDef (',' varDef)* ';';
varDef:
        (Ident ('[' constExp ']')* '=' initVal)
        | Ident ('[' constExp ']')*;

initVal: exp
        | '{'( initVal (',' initVal)* )?'}';

funcDef: funcType Ident '(' funcFParams? ')' block;
funcType: INT | FLOAT | VOID;

funcFParams: funcFParam (',' funcFParam)*;
funcFParam: bType Ident funcFArrParam?;
funcFArrParam : ('['']'('['exp']')*) ;

block: '{'blockItem*'}';
blockItem: decl|stmt;
stmt:
      lVal '=' exp';' # assign
    | exp? ';' # expr
    | block # blockStmt
    | 'if' '('cond')' stmt ('else' stmt)? # if
    | 'while' '('cond')' stmt # while
    | 'break' ';' # break
    | 'continue' ';' # continue
    | 'return' exp? ';' # return;

exp : addExp;
cond: lOrExp;
lVal: Ident ('['exp']')*;
primaryExp: '(' exp ')' | lVal | number | String;
number: intConst | floatConst;
unaryExp: primaryExp
        | Ident '(' funcRParams? ')'
        |  unaryOp unaryExp;

funcRParams: exp (',' exp)*;

mulExp: unaryExp  (mulOp unaryExp)*;
addExp: mulExp (addOp mulExp)*;
relExp: addExp (relOp addExp)*;
eqExp: relExp (eqOp relExp)*;
lAndExp: eqExp ('&&' eqExp)*;
lOrExp: lAndExp ('||' lAndExp)*;

constExp: addExp;

mulOp : (MULT|DIV|MOD);
addOp : (ADD|SUB);
relOp : (LT|GT|LE|GE);
eqOp : (EQ|NE);
bType: INT | FLOAT | VOID;

DIV : '/';
MULT : '*';
MOD: '%';

ADD: '+';
SUB: '-';

LT: '<';
GT: '>';
LE: '<=';
GE: '>=';

EQ: '==';
NE: '!=';

INT: 'int';
FLOAT: 'float';
VOID: 'void';

unaryOp: ADD|SUB|EXCL;

EXCL: '!';
Ident: [_a-zA-Z] [_a-zA-Z0-9]*;
intConst: Dec | Oct | Hex;
Dec: [1-9][0-9]*;
Oct: '0'[0-7]*;
Hex: ('0x'|'0X') [a-fA-F0-9]+;
floatConst:  DecF | HexF;// 浮点的词法，注意可能出问题，16进制浮点不支持了?

DecF: (([0-9]*'.'[0-9]+ | [0-9]+'.')|([0-9]+)) (('e'|'E') ('+'|'-')? [0-9]+)?;
//DecExp: ('e'|'E') ('+'|'-')? [0-9]+;
//DecFrac: [0-9]*'.'[0-9]+ | [0-9]+'.';
//DigiSeq: [0-9]+;


HexF: ((('0x'|'0X') [a-fA-F0-9]+) | (('0x'|'0X') (([a-fA-F0-9]*'.'[a-fA-F0-9]+)|[a-fA-F0-9]+'.'))) (('P'|'p') ('+'|'-')? [0-9]+)?;
//BinExp: ('P'|'p') ('+'|'-')? [0-9]+;
//HexFrac: ('0x'|'0X') (([a-fA-F0-9]*'.'[a-fA-F0-9]+)|[a-fA-F0-9]+'.');



String: '"' (.)*? '"'; //TODO

WS: [ \r\n\t]+ -> skip;

LineComment: '//' ~ [\r\n]* -> skip;

MultiLineComment: '/*' .*? '*/' -> skip;