//
// Created by shliba on 2023/6/21.
//

#include "ast_visitor.h"
#include "cmath"
#include "sstream"


int str2double(string s)
{
    double d;
    std::stringstream ss;
    ss<<s;
    ss>>std::setprecision(16)>>d;
    ss.clear();
    return (int)d;
}


int string2num(const std::string& s) {
    if (s.size() > 1 && s[0] == '0' && (s[1] == 'x' || s[1] == 'X')) {
        int ans = 0;
        for (int i = 2; i < s.length(); ++i) {
            if (std::isdigit(s[i])) {
                ans = ans * 16 + s[i] - '0';
            } else if (islower(s[i])) {
                ans = ans * 16 + s[i] - 'a' + 10;
            } else {
                ans = ans * 16 + s[i] - 'A' + 10;
            }
        }
        return ans;
    } else if (s.size() > 1 && s[0] == '0') {
        int ans = 0;
        for (int i = 1; i < s.length(); ++i) {
            if (std::isdigit(s[i])) {
                ans = ans * 8 + s[i] - '0';
            }
        }
        return ans;
    } else if (s.find('e')){
        return str2double(s);
    } else {
        return std::stoi(s);
    }
}

bool is_zero(std::shared_ptr<Ir::Reg>& reg) {
    if (reg->is_constant() && ((reg->get_type() == Ir::Int && reg->get_num().int_num == 0)
                               || (reg->get_type() == Ir::Float && reg->get_num().float_num == 0))) {
        return true;
    }
    return false;
}

bool is_zero_array(SysYParser::InitValContext* ctx) {
    if (ctx->initVal().empty() || (ctx->initVal().size() == 1 && ctx->initVal()[0]->getText() == "0")) {
        return true;
    }
    return false;
}

std::any AstVisitor::visitCompUnit(SysYParser::CompUnitContext *ctx) {

    cu = new Ir::CompileUnit();
    add_lib();
    set_cu(cu);
    create_scope();
    visit_first_pass(ctx);
    for (auto &item : ctx->children) {
        Instanceof(funcdef, SysYParser::FuncDefContext *, item) {
            visitFuncDef(funcdef);
        } else Instanceof(decl, SysYParser::DeclContext*, item) {
            visitDecl(decl);
        } else {
            assert(false);
        }
    }
    for (const auto &item : mem_stack) {

    }
    return cu;
}

std::any AstVisitor::visitDecl(SysYParser::DeclContext *ctx) {
    if (ctx->constDecl()) {
        visitConstDecl(ctx->constDecl());
    } else if(ctx->varDecl()) {
        visitVarDecl(ctx->varDecl());
    }
    return nullptr;
}

Ir::ScalaType type_trans(SysYParser::BTypeContext *ctx) {
    if (ctx->FLOAT()) {
        return Ir::Float;
    } else if (ctx->INT()) {
        return Ir::Int;
    } else if (ctx->VOID()){
        return Ir::Void;
    } else {
        assert(false);
    }
}

std::any AstVisitor::visitFuncDef(SysYParser::FuncDefContext *ctx) {

    auto func = get_func(ctx->Ident()->getText());

    ret_flag = false;
    Ir::Factory::reset(func);
    current_func = func;
    ret_bb = Ir::Factory::get_noadd_new_bb(nullptr, current_func);
    auto first_bb = Ir::Factory::get_new_bb(nullptr, current_func);
    current_func->set_first_bb(first_bb);
    current_bb = first_bb;
    create_scope();
    auto &params = func->get_params();
    if (ctx->funcFParams()) {
        for (auto &param_ctx : ctx->funcFParams()->funcFParam()) {

            auto param = visitFuncFParam(param_ctx);
            auto reg = std::any_cast<std::shared_ptr<Ir::Type>>(param);

            params.push_back(reg);
        }
    }


    if (!func->is_void()) {
        auto alloca_instr = Ir::Factory::get_alloca(func->get_type());
        ret_mem = dynamic_cast<Ir::AllocaInstr*>(alloca_instr.get())->get_addr();
        func->ret_mem = ret_mem;
        first_bb->push_back(alloca_instr);
    }
    current_bb = Ir::Factory::get_new_bb(nullptr, func);
    Ir::Factory::build_branch(current_bb, *first_bb);
    visitBlock(ctx->block());
    Ir::Factory::build_branch(ret_bb.get(), *current_bb);
    if (!func->is_void()) {
        auto reg = Ir::Factory::get_reg(ret_mem->get_type());
        Ir::Factory::build_load(reg, ret_mem, *ret_bb);
        Ir::Factory::build_ret(reg, *ret_bb);
    } else {
        Ir::Factory::build_ret(*ret_bb);
    }
    func->set_reg_max_num(Ir::Factory::reg_num);
    func->set_bb_max_num(Ir::Factory::bb_num);
    func->get_bbs().push_back(std::move(ret_bb));
    delete_scope();
    return nullptr;
}

Ir::BinaryType rel_type_trans(SysYParser::RelOpContext *ctx) {
    if (ctx->GE()) {
        return Ir::SGE;
    } else if (ctx->GT()) {
        return Ir::SGT;
    } else if (ctx->LE()) {
        return Ir::SLE;
    } else {
        return Ir::SLT;
    }
}




Ir::ScalaType func_type_trans(SysYParser::FuncTypeContext *ctx) {
    if (ctx->FLOAT()) {
        return Ir::Float;
    } else if (ctx->INT()) {
        return Ir::Int;
    } else if (ctx->VOID()){
        return Ir::Void;
    } else {
        assert(false);
    }
}

Ir::ScalaType AstVisitor::current_type = Ir::Void;
Ir::Function* AstVisitor::current_func = nullptr;
Ir::Loop* AstVisitor::current_loop = nullptr;
Ir::BasicBlock* AstVisitor::current_bb = nullptr;
Ir::CompileUnit* AstVisitor::cu = nullptr;
std::unique_ptr<Ir::BasicBlock> AstVisitor::ret_bb = nullptr;
std::shared_ptr<Ir::MemObject> AstVisitor::ret_mem = nullptr;
bool AstVisitor::ret_flag = false;
static struct CondSturct{
    Ir::BasicBlock* true_block;
    Ir::BasicBlock* false_block;
    Ir::BasicBlock* loop_header;
} current_cond_struct;


std::any AstVisitor::visitConstDecl(SysYParser::ConstDeclContext *ctx) {
    current_type = type_trans(ctx->bType());
    for (auto &def : ctx->constDef()) {
        visitConstDef(def);
    }
    return nullptr;
}

std::any AstVisitor::visitVarDecl(SysYParser::VarDeclContext *ctx) {
    current_type = type_trans(ctx->bType());
    for (auto &def : ctx->varDef()) {
        if (is_global_scope()) {
            visit_global_def(def);
        } else {
            visitVarDef(def);
        }
    }
    return nullptr;
}




std::any AstVisitor::visitConstDef(SysYParser::ConstDefContext *ctx) {
    auto ident = ctx->Ident()->getText();
    if (ctx->constExp().empty()) {
        if (ident == "MAX") {
            int z = 10;
        }
        auto reg = std::any_cast<std::shared_ptr<Ir::Reg>>(visitConstExp(ctx->constInitVal()->constExp()));
        if (is_global_scope()) {
            reg->set_global_name(ident + "_" + std::to_string(Ir::Factory::global_num++));
            add_global(reg);
            reg->set_global(true);

            add_new_reg(ident, reg);
        } else {
            auto alloca_instr = Ir::Factory::get_alloca(reg->get_type());
            auto mem = dynamic_cast<Ir::AllocaInstr*>(alloca_instr.get())->get_addr();
            add_new_reg(ident, reg);
            reg->set_global(false);
            add_new_mem(ident, mem);
            current_func->get_first_bb()->get_instrs().push_front(std::move(alloca_instr));
            Ir::Factory::build_store(reg, mem, *current_bb);

        }
        //        if (current_type == Ir::Float) {
        //            add_global_float(ident, reg->get_num().float_num);
        //        } else {
        //            add_global_int(ident, reg->get_num().int_num);
        //        }
    } else {
        auto array = Ir::Factory::get_vector(current_type, ident, true);
        add_new_mem(ident, array);
        add_global(array);
        for (auto &const_exp : ctx->constExp()) {
            auto size = visit_int_const(const_exp);
            array->get_dims().push_back(size);
        }

        visit_const_array(ctx->constInitVal(), array);

    }
    return nullptr;
}

void AstVisitor::visit_global_def(SysYParser::VarDefContext *ctx) {
    auto ident = ctx->Ident()->getText();
    if (ctx->constExp().empty()) {
        std::shared_ptr<Ir::Reg> reg;
        if (ctx->initVal()) {
            reg = visit_global_initval(ctx->initVal());
        } else {
            Ir::RegConst zero;
            zero.float_num = 0;
            zero.int_num = 0;
            reg = Ir::Factory::get_global_reg(current_type, zero, ident);
        }
        reg->set_constant(false);
        add_new_reg(ident, reg);
        //        if (current_type == Ir::Float) {
        //            add_global_float(ident, reg->get_num().float_num);
        //        } else {
        //            add_global_int(ident, reg->get_num().int_num);
        //        }
    } else {
        auto array = Ir::Factory::get_vector(current_type, ident, false);
        add_new_mem(ident, array);
        for (auto &const_exp : ctx->constExp()) {
            auto size = visit_int_const(const_exp);
            array->get_dims().push_back(size);
        }

        visit_global_array(ctx->initVal(), array);

    }
}

std::shared_ptr<Ir::Reg> AstVisitor::visit_global_initval(SysYParser::InitValContext *ctx) {
    std::shared_ptr<Ir::Reg> reg;
    Ir::RegConst reg_const;
    if (current_type == Ir::Int) {
        reg_const.int_num = visit_int_add(ctx->exp()->addExp());
        reg = Ir::Factory::get_const_reg(Ir::Int, reg_const);
        reg->set_global(true);
    } else if (current_type == Ir::Float) {
        reg_const.float_num = visit_float_add(ctx->exp()->addExp());
        reg = Ir::Factory::get_const_reg(Ir::Float, reg_const);
        reg->set_global(true);
    } else {
        assert(false);
    }
    return reg;
}



std::any AstVisitor::visitConstExp(SysYParser::ConstExpContext *ctx) {
    std::shared_ptr<Ir::Reg> reg;
    Ir::RegConst reg_const;
    if (current_type == Ir::Int) {
        reg_const.int_num = visit_int_const(ctx);
        reg = Ir::Factory::get_const_reg(Ir::Int, reg_const);
        reg->set_global(true);
    } else if (current_type == Ir::Float) {
        reg_const.float_num = visit_float_const(ctx);
        reg = Ir::Factory::get_const_reg(Ir::Float, reg_const);
        reg->set_global(true);
    } else {
        assert(false);
    }
    return reg;
}



int AstVisitor::visit_int_const(SysYParser::ConstExpContext *ctx) {
    return visit_int_add(ctx->addExp());
}

int AstVisitor::visit_int_add(SysYParser::AddExpContext *ctx) {
    auto add = ctx;
    int num = visit_int_mul(add->mulExp(0));
    int i = 1;
    for (auto &op : add->addOp()) {
        if (op->ADD()) {
            num += visit_int_mul(add->mulExp(i++));
        } else {
            num -= visit_int_mul(add->mulExp(i++));
        }
    }
    return num;
}


int AstVisitor::visit_int_mul(SysYParser::MulExpContext *ctx) {
    auto mul = ctx;
    int num = visit_int_unary(mul->unaryExp(0));
    int i = 1;
    for (auto &op : mul->mulOp()) {
        if (op->MULT()) {
            num *= visit_int_unary(mul->unaryExp(i++));
        } else if (op->DIV()){
            num /= visit_int_unary(mul->unaryExp(i++));
        } else if (op->MOD()) {
            num %= visit_int_unary(mul->unaryExp(i++));
        }
    }
    return num;
}

int AstVisitor::visit_int_unary(SysYParser::UnaryExpContext *ctx) {
    auto unary = ctx;
    bool positive = true;
    while (auto op = unary->unaryOp()) {
        unary = unary->unaryExp();
        assert(unary != nullptr);
        if (op->ADD()) {
            continue;
        } else if (op->SUB()) {
            positive = !positive;
        } else {
            assert(false);
        }
    }
    auto factor = 1;
    if (!positive) {
        factor = -1;
    }
    if (auto primary = unary->primaryExp()) {
        if (primary->exp()) {
            return factor * visit_int_add(primary->exp()->addExp());
        } else if(auto l = primary->lVal()) {
            if (l->exp().empty()) {
                return factor * get_const_int(l->Ident()->getText());
            } else {
                //TODO 数组初值
                assert(false);
            }
        } else if(auto number = primary->number()) {
            //TODO 可能涉及隐式类型转换
            auto ident = number->getText();
            return factor * string2num(ident);
        } else {
            assert(false);
        }
    }
    return 0;
}



float AstVisitor::visit_float_const(SysYParser::ConstExpContext *ctx) {
    return visit_float_add(ctx->addExp());
}


float AstVisitor::visit_float_add(SysYParser::AddExpContext *ctx) {
    auto add = ctx;
    float num = visit_float_mul(add->mulExp(0));
    int i = 1;
    for (auto &op : add->addOp()) {
        if (op->ADD()) {
            num += visit_float_mul(add->mulExp(i++));
        } else {
            num -= visit_float_mul(add->mulExp(i++));
        }
    }
    return num;
}


float AstVisitor::visit_float_mul(SysYParser::MulExpContext *ctx) {
    auto mul = ctx;
    float num = visit_float_unary(mul->unaryExp(0));
    int i = 1;
    for (auto &op : mul->mulOp()) {
        if (op->MULT()) {
            num *= visit_float_unary(mul->unaryExp(i++));
        } else if (op->DIV()){
            num /= visit_float_unary(mul->unaryExp(i++));
        }
    }
    return num;
}

float AstVisitor::visit_float_unary(SysYParser::UnaryExpContext *ctx) {
    auto unary = ctx;
    bool positive = true;
    while (auto op = unary->unaryOp()) {
        unary = unary->unaryExp();
        assert(unary != nullptr);
        if (op->ADD()) {
            continue;
        } else if (op->SUB()) {
            positive = !positive;
        } else {
            assert(false);
        }
    }
    auto factor = 1;
    if (!positive) {
        factor = -1;
    }
    if (auto primary = unary->primaryExp()) {
        if (primary->exp()) {
            return factor * visit_float_add(primary->exp()->addExp());
        } else if(auto l = primary->lVal()) {
            if (l->exp().empty()) {
                return factor * get_const_float(l->Ident()->getText());
            } else {
                //TODO 数组初值
                assert(false);
            }
        } else if(auto number = primary->number()) {
            //TODO 可能涉及隐式类型转换
            auto ident = number->getText();
            return factor * std::stod(ident);
        } else {
            assert(false);
        }
    } else {
        assert(false);
    }
}

void visit_recur_func(antlr4::tree::ParseTree* ctx, Ir::Function* func) {
    for (auto &item : ctx->children) {

        Instanceof(unary, SysYParser::UnaryExpContext *,item) {
            if (unary->Ident())
            auto name =unary->Ident()->getText();
            if (unary->Ident() && unary->Ident()->getText() == func->get_name()) {
                func->set_recur(true);
                return;
            }
        }
        visit_recur_func(item, func);

    }
}


void AstVisitor::visit_first_pass(SysYParser::CompUnitContext *ctx) {
    for (const auto &func_def : ctx->funcDef()) {

        auto name = func_def->Ident()->getText();
        auto func = Ir::Factory::get_func(name, func_type_trans(func_def->funcType()), cu);
        add_func(name, func);
        visit_recur_func(func_def, func);
    }
}



//TODO 这两个方法过于相似，可以考虑后续优化为一个
void AstVisitor::visit_global_array(SysYParser::InitValContext *ctx, std::shared_ptr<Ir::Vector> array) {
    std::vector<Ir::RegConst> seq;
    if (ctx && !is_zero_array(ctx)) {
        visit_global_array_sub(ctx, array, seq, 0);
    }
    add_global_array(array->get_name(), seq, array->get_type());
}

int AstVisitor::visit_global_array_sub(SysYParser::InitValContext *ctx, std::shared_ptr<Ir::Vector> array, std::vector<Ir::RegConst> & seq, int deep) {
    if (ctx->exp()) {
        Ir::RegConst reg_const;
        if (array->get_type() == Ir::Int) {
            reg_const.int_num = visit_int_add(ctx->exp()->addExp());
        } else {
            reg_const.float_num = visit_float_add(ctx->exp()->addExp());
        }
        seq.push_back(reg_const);
        return 1;
    } else {
        int sum = 0;
        for (size_t i = 0; i < ctx->initVal().size(); ++i) {
            sum += visit_global_array_sub(ctx->initVal(i), array, seq, deep + 1);
        }
        int all_size = 1;
        for (size_t i = deep; i < array->get_dims().size(); ++i) {
            all_size *= array->get_dims()[i];
        }
        if (sum < all_size) {
            Ir::RegConst zero;
            zero.int_num = 0;
            std::vector<Ir::RegConst> zero_v = std::vector( all_size - sum, zero);
            seq.insert(seq.end(), zero_v.begin(), zero_v.end());
            sum = all_size;
        }
        return sum;
    }

}


void AstVisitor::visit_const_array(SysYParser::ConstInitValContext *ctx, std::shared_ptr<Ir::Vector> array) {
    std::vector<Ir::RegConst> seq;
    visit_const_array_sub(ctx, array, seq, 0);
    add_global_array(array->get_name(), seq, array->get_type());
}

int AstVisitor::visit_array_initval_sub(SysYParser::InitValContext *ctx, std::shared_ptr<Ir::Vector> array, std::vector<std::shared_ptr<Ir::Reg>> &seq, int deep) {
    if (ctx->exp()) {
        seq.push_back(std::any_cast<std::shared_ptr<Ir::Reg>>(visitExp(ctx->exp())));
        return 1;
    } else {
        int sum = 0;
        for (size_t i = 0; i < ctx->initVal().size(); ++i) {
            sum += visit_array_initval_sub(ctx->initVal(i), array, seq, deep + 1);
        }
        int all_size = 1;
        for (size_t i = deep; i < array->get_dims().size(); ++i) {
            all_size *= array->get_dims()[i];
        }
        if (sum < all_size) {
            auto zero = Ir::Factory::get_const_reg(Ir::Int, Ir::RegConst());
            std::vector<std::shared_ptr<Ir::Reg>> zero_v = std::vector( all_size - sum, zero);
            seq.insert(seq.end(), zero_v.begin(), zero_v.end());
            sum = all_size;
        }
        return sum;
    }
}


int AstVisitor::visit_const_array_sub(SysYParser::ConstInitValContext *ctx, std::shared_ptr<Ir::Vector> array,
                                      std::vector<Ir::RegConst> &seq, int deep) {
    if (ctx->constExp()) {
        Ir::RegConst reg_const;
        if (array->get_type() == Ir::Int) {
            reg_const.int_num = visit_int_const(ctx->constExp());
        } else {
            reg_const.float_num = visit_float_const(ctx->constExp());
        }
        seq.push_back(reg_const);
        return 1;
    } else {
        int sum = 0;
        for (size_t i = 0; i < ctx->constInitVal().size(); ++i) {
            sum += visit_const_array_sub(ctx->constInitVal(i), array, seq, deep + 1);
        }
        int all_size = 1;
        for (size_t i = deep; i < array->get_dims().size(); ++i) {
            all_size *= array->get_dims()[i];
        }
        if (sum < all_size) {
            Ir::RegConst zero;
            zero.int_num = 0;
            std::vector<Ir::RegConst> zero_v = std::vector( all_size - sum, zero);
            seq.insert(seq.end(), zero_v.begin(), zero_v.end());
            sum = all_size;
        }
        return sum;
    }
}

int get_sub_size(std::vector<int>& dims, int index) {
    int ans = 1;
    for (int i = index + 1; i < dims.size(); ++i) {
        ans *= dims[i];
    }
    return ans;
}

std::any AstVisitor::visitVarDef(SysYParser::VarDefContext *ctx) {
    auto name = ctx->Ident()->getText();

    if (ctx->constExp().empty()) {
        auto alloca_instr = Ir::Factory::get_alloca(current_type);
        auto mem = dynamic_cast<Ir::AllocaInstr*>(alloca_instr.get())->get_addr();
        add_new_mem(name, mem);
        current_func->get_first_bb()->get_instrs().push_front(std::move(alloca_instr));
        if (ctx->initVal()) {

            auto reg = std::any_cast<std::shared_ptr<Ir::Reg>>(visitInitVal(ctx->initVal()));
            Ir::Factory::build_store(reg, mem, *current_bb);
        }
    } else {
        if (!current_func->is_recur()) {
            auto local_name = "local_array" + std::to_string(get_localarray());
            auto array = Ir::Factory::get_vector(current_type,local_name , false);
            int sum = 1;
            for (const auto &item : ctx->constExp()) {
                auto dim = visit_int_const(item);
                sum *= dim;
                array->get_dims().push_back(dim);
            }
            add_new_mem(name, array);
            add_global( array);
            std::vector<Ir::RegConst> array_value;
            if (ctx->initVal() && !is_zero_array(ctx->initVal())) {
                std::vector<std::shared_ptr<Ir::Reg>> seq;
                visit_array_initval_sub(ctx->initVal(), array, seq, 0);
                for (int i = 0; i < sum; ++i) {
                    if (i < (int)seq.size()) {
                        if (seq[i]->is_constant()) {
                            array_value.push_back(seq[i]->get_num());
                        } else {
                            Ir::RegConst zero;
                            zero.int_num = 0;
                            zero.float_num = 0;
                            array_value.push_back(zero);
                            std::vector<std::shared_ptr<Ir::Reg>> indexs;
                            auto current = i;
                            for (int i = 0; i < array->get_dims().size(); ++i) {
                                auto dim = get_sub_size(array->get_dims(), i);
                                int range = current / dim;
                                Ir::RegConst reg_const;
                                reg_const.int_num = range;
                                indexs.push_back(Ir::Factory::get_const_reg(Ir::Int, reg_const));
                                current = current % dim;
                            }
                            Ir::Factory::build_store_array(seq[i], array, indexs, *current_bb);
                        }
                    }
                }
            }
            add_global_array(array->get_name(), array_value, array->get_type());
        } else {
            auto array = Ir::Factory::get_vector(current_type);
            Ir::Factory::build_alloca(array, *current_bb);
            int sum = 1;
            for (const auto &item : ctx->constExp()) {
                auto dim = visit_int_const(item);
                sum *= dim;
                array->get_dims().push_back(dim);
            }
            add_new_mem(name, array);
            if (ctx->initVal()) {
                std::vector<std::shared_ptr<Ir::Reg>> seq;
                visit_array_initval_sub(ctx->initVal(), array, seq, 0);
                for (int i = 0; i < sum; ++i) {
                    std::shared_ptr<Ir::Reg> value;
                    if (i < (int)seq.size()) {
                        value = seq[i];
                    } else {
                        value = Ir::Factory::get_zero_reg(current_type);
                    }
                    std::vector<std::shared_ptr<Ir::Reg>> indexs;
                    auto current = i;
                    for (int i = 0; i < array->get_dims().size(); ++i) {
                        auto dim = get_sub_size(array->get_dims(), i);
                        int range = current / dim;
                        Ir::RegConst reg_const;
                        reg_const.int_num = range;
                        indexs.push_back(Ir::Factory::get_const_reg(Ir::Int, reg_const));
                        current = current % dim;
                    }

                    Ir::Factory::build_store_array(value, array, indexs, *current_bb);
                }
            }
        }
    }
    return nullptr;
}

std::any AstVisitor::visitFuncFParam(SysYParser::FuncFParamContext *ctx) {
    auto type = type_trans(ctx->bType());
    auto name = ctx->Ident()->getText();
    std::shared_ptr<Ir::Type> sysType;

    if (ctx->funcFArrParam()) {
        auto array = Ir::Factory::get_vector(type);
        array->set_param(true);
        add_new_mem(name, array);
        array->get_dims().push_back(0);
        for (const auto &item : ctx->funcFArrParam()->exp()) {
            int size = visit_int_add(item->addExp());
            array->get_dims().push_back(size);
        }
        sysType = array;
        return sysType;
    } else {
        auto reg = Ir::Factory::get_reg(type);
        reg->set_param(true);
        auto instr = Ir::Factory::get_alloca(type);
        auto mem = dynamic_cast<Ir::AllocaInstr*>(instr.get())->get_addr();
        add_new_mem(name, mem);
        current_func->get_first_bb()->get_instrs().push_front(std::move(instr));
        Ir::Factory::build_store(reg, mem, *current_bb);
        sysType = reg;
        return sysType;
    }

}
std::any AstVisitor::visitBlock(SysYParser::BlockContext *ctx) {
    create_scope();


    for (auto &item : ctx->blockItem()) {
        if (ret_flag) {
            delete_scope();
            return nullptr;
        }
        visitBlockItem(item);
    }
    delete_scope();
    return nullptr;
}
std::any AstVisitor::visitBlockItem(SysYParser::BlockItemContext *ctx) {
    if (ctx->decl()) {
        visitDecl(ctx->decl());
    } else {
        visitStmt(ctx->stmt());

    }
    return nullptr;
}
std::any AstVisitor::visitInitVal(SysYParser::InitValContext *ctx) {
    if (ctx->exp()) {
        return visitExp(ctx->exp());
    } else {
        assert(false);
    }
}
std::any AstVisitor::visitExp(SysYParser::ExpContext *ctx) {
    return visitAddExp(ctx->addExp());
}
std::any AstVisitor::visitAddExp(SysYParser::AddExpContext *ctx) {
    auto ans = visitMulExp(ctx->mulExp(0));
    if (ctx->addOp().empty()){
        return ans;
    }
    auto result = std::any_cast<std::shared_ptr<Ir::Reg>>(ans);
    int i = 1;
    for (auto &op : ctx->addOp()) {
        auto reg = std::any_cast<std::shared_ptr<Ir::Reg>>(visitMulExp(ctx->mulExp(i++)));
        auto new_result = Ir::Factory::get_reg(current_type);
        if (op->ADD()) {
            Ir::Factory::build_binary(new_result, result, reg, Ir::ADD, *current_bb);
        } else {
            Ir::Factory::build_binary(new_result, result, reg, Ir::SUB, *current_bb);
        }
        result = new_result;
    }
    return result;
}


std::any AstVisitor::visitMulExp(SysYParser::MulExpContext *ctx) {
    auto temp_reg = visitUnaryExp(ctx->unaryExp(0));
    if (ctx->mulOp().empty()) {
        return temp_reg;
    }
    auto result = std::any_cast<std::shared_ptr<Ir::Reg>>(temp_reg);
    int i = 1;
    for (auto &op : ctx->mulOp()) {
        auto reg = std::any_cast<std::shared_ptr<Ir::Reg>>(visitUnaryExp(ctx->unaryExp(i++)));
        auto new_result = Ir::Factory::get_reg(reg->get_type() > result->get_type() ? reg->get_type() : result->get_type());
        if (op->MULT()) {
            Ir::Factory::build_binary(new_result, result, reg, Ir::MUL, *current_bb);
        } else if (op->DIV()) {
            Ir::Factory::build_binary(new_result, result, reg, Ir::SDIV, *current_bb);
        } else {
            Ir::Factory::build_binary(new_result, result, reg, Ir::SREM, *current_bb);
        }
        result = new_result;
    }
    return result;
}

std::any AstVisitor::visitUnaryExp(SysYParser::UnaryExpContext *ctx) {
    bool positive = true;

    if (ctx->primaryExp()) {
        return visitPrimaryExp(ctx->primaryExp());
    } else if (ctx->Ident()) {
        auto func_name = ctx->Ident()->getText();
        auto func = get_func(func_name);
        if (func_name == "getfarray") {
            int x = 10;
        }
        std::vector<std::shared_ptr<Ir::Type>> params;
        if (ctx->funcRParams()) {
            auto p_it = func->get_params().begin();
            for (auto &item : ctx->funcRParams()->exp()) {
                auto last = current_type;
                current_type = p_it->get()->get_type();
                auto exp = visitExp(item);
                std::shared_ptr<Ir::Type> param;
                if (exp.type() == typeid(std::shared_ptr<Ir::Reg>)) {
                    auto temp_reg = std::any_cast<std::shared_ptr<Ir::Reg>>(exp);
                    param = temp_reg;
                } else {
                    auto array = std::any_cast<std::shared_ptr<Ir::Vector>>(exp);
                    param = array;
                }
                params.push_back(param);
                p_it++;
                current_type = last;
            }
        } else if (func_name == "starttime" || func_name == "stoptime") {
            auto param = Ir::Factory::get_zero_reg(Ir::Int);
            params.push_back(param);
        }
        if (!func->is_void()) {
            auto reg = Ir::Factory::get_reg(func->get_type());

            Ir::Factory::build_call(reg, func, params, *current_bb);
            return reg;
        } else {
            Ir::Factory::build_call(func, params, *current_bb);
            return nullptr;
        }
    } else {
        auto unary = ctx;
        while (unary->unaryOp()) {
            if (unary->unaryOp()->SUB()) {
                positive = !positive;
            } else if (unary->unaryOp()->EXCL()) {
                auto reg = std::any_cast<std::shared_ptr<Ir::Reg>>(visitUnaryExp(unary->unaryExp()));
                auto dst = Ir::Factory::get_reg(Ir::Bool);
                Ir::Factory::build_binary(dst, reg, Ir::Factory::get_zero_reg(reg->get_type()), Ir::EQ, *current_bb);
                if (!positive) {
                    auto new_re  = Ir::Factory::get_reg(Ir::Int);
                    Ir::RegConst zero;
                    zero.int_num = 0;
                    zero.float_num = 0;
                    Ir::Factory::build_binary(new_re, Ir::Factory::get_const_reg(dst->get_type(),zero), dst, Ir::SUB, *current_bb);
                    return new_re;
                }
                return dst;
            }
            unary = unary->unaryExp();
        }
        if (positive) {
            return visitUnaryExp(unary);
        } else {
            auto reg = std::any_cast<std::shared_ptr<Ir::Reg>>(visitUnaryExp(unary));
            auto dst = Ir::Factory::get_reg(reg->get_type());
            Ir::RegConst zero;
            zero.int_num = 0;
            zero.float_num = 0;
            Ir::Factory::build_binary(dst, Ir::Factory::get_const_reg(reg->get_type(),zero), reg, Ir::SUB, *current_bb);
            return dst;
        }
    }
}



std::any AstVisitor::visitPrimaryExp(SysYParser::PrimaryExpContext *ctx) {
    if (ctx->exp()) {
        return visitExp(ctx->exp());
    } else if (ctx->lVal()) {
        return visitLVal(ctx->lVal());
    } else if (ctx->number()) {
        auto number = ctx->number();
        Ir::RegConst reg_const;
        auto type = current_type;
        if (number->intConst()) {
            //            if (current_type == Ir::Int) {
            type = Ir::Int;
            auto num = number->intConst()->getText();
            reg_const.int_num = string2num(number->intConst()->getText());
            //            } else if (current_type == Ir::Float){
            //                reg_const.float_num = std::stod(number->intConst()->getText());
            //            } else {
            //                type = Ir::Int;
            //                reg_const.int_num = string2num(number->intConst()->getText());
            //            }
        } else {
            //            if (current_type == Ir::Int) {
            //                reg_const.int_num = (int)std::stod(number->floatConst()->getText());
            //            } else if (current_type == Ir::Float){
            //                reg_const.float_num = std::stod(number->floatConst()->getText());
            //            } else {
            type = Ir::Float;
            reg_const.float_num = std::stod(number->floatConst()->getText());
            reg_const.float_num = std::stod(number->floatConst()->getText());
        }
        auto reg = Ir::Factory::get_const_reg(type, reg_const);
        return reg;
    } else {
        assert(false);
    }
}



std::any AstVisitor::visitLVal(SysYParser::LValContext *ctx) {
    std::string name = ctx->Ident()->getText();
    auto mem = get_mem(name);
    Instanceof(vec, Ir::Vector*, mem.get()) {
        std::vector<std::shared_ptr<Ir::Reg>> indexs;

        for (auto &exp : ctx->exp()) {
            auto last = current_type;
            current_type = Ir::Int;
            indexs.push_back(std::any_cast<std::shared_ptr<Ir::Reg>>(visitExp(exp)));
            current_type = last;
        }
        Instanceof(array, Ir::Vector*, mem.get()) {
            std::shared_ptr<Ir::Vector> array_addr(array);
            if (array->get_dims().size() == indexs.size()) {
                auto dst = Ir::Factory::get_reg(array_addr->get_type());
                Ir::Factory::build_load_array(dst, array_addr, indexs, *current_bb);
                return dst;
            } else {
                auto dst = Ir::Factory::get_vector(array->get_type());
                dst->get_dims().push_back(0);
                for (auto i = indexs.size() + 1; i < array->get_dims().size(); ++i) {
                    dst->get_dims().push_back(array->get_dims()[i]);
                }
                Ir::Factory::build_load_vector(dst, array_addr, indexs, *current_bb);
                return dst;
            }
        } else {
            assert(false);
        }
    } else {
        auto dst = Ir::Factory::get_reg(mem->get_type());
        if (mem->get_id() != -1) {
            Ir::Factory::build_load(dst, mem, *current_bb);
        } else {
            auto global = get_reg(name);
            dst->set_type(global->get_type());
            Ir::Factory::build_load_global(dst, global, *current_bb);
        }
        return dst;
    }

}
std::any AstVisitor::visitStmt(SysYParser::StmtContext *ctx) {


    Instanceof(if_ctx, SysYParser::IfContext*, ctx){
        return visitIf(if_ctx);
        // if
    } else Instanceof(while_ctx, SysYParser::WhileContext*, ctx){
        return visitWhile(while_ctx);
        // while
    }  else Instanceof(break_ctx, SysYParser::BreakContext*, ctx){
        ret_flag = true;
        return visitBreak(break_ctx);
        // break
    }  else Instanceof(continue_ctx, SysYParser::ContinueContext*, ctx){
        ret_flag = true;
        return visitContinue(continue_ctx);
        // continue
    } Instanceof(return_ctx, SysYParser::ReturnContext*, ctx){
        ret_flag = true;
        return visitReturn(return_ctx);
        // return
    } else Instanceof(assign, SysYParser::AssignContext*, ctx) {
        return visitAssign(assign);
    }
    else Instanceof(expr, SysYParser::ExprContext *, ctx) {
        return visitExpr(expr);
        // expr
    }
    else Instanceof(block, SysYParser::BlockStmtContext *, ctx) {
        return visitBlockStmt(block);
        // blockstmt
    } else {
        assert(false);
    }
}


std::any AstVisitor::visitIf(SysYParser::IfContext *ctx) {
    auto true_block = Ir::Factory::get_noadd_new_bb(current_loop, current_func);
    auto false_block = Ir::Factory::get_noadd_new_bb(current_loop, current_func);
    auto last_true = current_cond_struct.true_block ;
    auto last_false = current_cond_struct.false_block;
    current_cond_struct.true_block = true_block.get();
    current_cond_struct.false_block = false_block.get();
    visitCond(ctx->cond());
    current_bb = current_cond_struct.true_block;
    current_func->get_bbs().push_back(std::move(true_block));
    visitStmt(ctx->stmt(0));
    ret_flag = false;
    if (ctx->stmt().size() > 1) {
        auto new_bb = Ir::Factory::get_new_bb(current_loop, current_func);
        Ir::Factory::build_branch(new_bb, *current_bb);
        current_bb = current_cond_struct.false_block;
        current_func->get_bbs().push_back(std::move(false_block));
        visitStmt(ctx->stmt(1));
        ret_flag = false;
        Ir::Factory::build_branch(new_bb, *current_bb);
        current_bb = new_bb;
    } else {
        Ir::Factory::build_branch(current_cond_struct.false_block, *current_bb);
        current_bb = current_cond_struct.false_block;
        current_func->get_bbs().push_back(std::move(false_block));
    }
    current_cond_struct.true_block = last_true;
    current_cond_struct.false_block = last_false;
    return nullptr;
}
std::any AstVisitor::visitWhile(SysYParser::WhileContext *ctx) {
    auto true_block = Ir::Factory::get_noadd_new_bb(nullptr, current_func);
    auto false_block = Ir::Factory::get_noadd_new_bb(current_loop, current_func);
    auto exit_prev_block = Ir::Factory::get_noadd_new_bb(current_loop, current_func);
    auto header = Ir::Factory::get_new_bb(nullptr, current_func);
    auto last_true = current_cond_struct.true_block ;
    auto last_false = current_cond_struct.false_block;
    current_cond_struct.true_block = true_block.get();
    current_cond_struct.false_block = false_block.get();
    current_cond_struct.loop_header = header;
    int depth = 1;
    if (current_loop != nullptr) {
        depth = current_loop->get_depth() + 1;
    }
    current_loop = Ir::Factory::get_loop(exit_prev_block.get(), false_block.get(), depth, current_func, current_loop);

    true_block->set_belong_to_loop(current_loop);
    true_block->set_depth(depth);
    current_loop->get_all_bbs().insert(true_block.get());


    Ir::Factory::build_branch(header, *current_bb);
    current_bb = header;
    visitCond(ctx->cond());
    current_bb = current_cond_struct.true_block;
    current_func->get_bbs().push_back(std::move(true_block));
    visitStmt(ctx->stmt());
    ret_flag = false;
    // stmt结束后,切换到exit prev,变成do while 形式

    Ir::Factory::build_branch(exit_prev_block.get(), *current_bb);
    current_bb = exit_prev_block.get();
    current_func->get_bbs().push_back(std::move(exit_prev_block));
    visitCond(ctx->cond());

    // Ir::Factory::build_branch(header, *current_bb);
    current_bb = current_cond_struct.false_block;
    current_func->get_bbs().push_back(std::move(false_block));
    current_loop = current_loop->get_parent();
    current_cond_struct.true_block = last_true;
    current_cond_struct.false_block = last_false;
    return nullptr;
}

std::any AstVisitor::visitBreak(SysYParser::BreakContext *ctx) {
    Ir::Factory::build_branch(current_loop->get_exiting_bb(), *current_bb);
    return nullptr;
}

std::any AstVisitor::visitContinue(SysYParser::ContinueContext *ctx) {
    Ir::Factory::build_branch(current_loop->get_header(), *current_bb);
    return SysYBaseVisitor::visitContinue(ctx);
}

std::any AstVisitor::visitReturn(SysYParser::ReturnContext *ctx) {
    if (ctx->exp()) {
        auto last = current_type;
        current_type = current_func->get_type();
        auto reg = std::any_cast<std::shared_ptr<Ir::Reg>>(visitExp(ctx->exp()));
        last = current_type;
        Ir::Factory::build_store(reg, ret_mem, *current_bb);
        Ir::Factory::build_branch(ret_bb.get(), *current_bb);
    } else {
        Ir::Factory::build_branch(ret_bb.get(), *current_bb);
    }

    return nullptr;
}

std::any AstVisitor::visitCond(SysYParser::CondContext *ctx) {
    auto lor = ctx->lOrExp();
    auto last_false = current_cond_struct.false_block;
    auto len = lor->lAndExp().size();
    for (size_t i = 0; i < len; ++i) {
        if (i != len - 1) {
            auto new_bb = Ir::Factory::get_new_bb(current_loop, current_func);
            current_cond_struct.false_block = new_bb;
            visitLAndExp(lor->lAndExp(i));
            current_bb = new_bb;
        } else {
            current_cond_struct.false_block = last_false;
            visitLAndExp(lor->lAndExp(i));
        }
    }
    return nullptr;
}
std::any AstVisitor::visitAssign(SysYParser::AssignContext *ctx) {
    auto lval = ctx->lVal();
    // assign
    if (auto exp = ctx->exp()) {
        auto name = lval->Ident()->getText();
        if (lval->exp().empty()) {
            auto mem = get_mem(name);
            current_type = mem->get_type();
            auto reg
                = std::any_cast<std::shared_ptr<Ir::Reg>>(visitExp(exp));
            if (mem->get_id() != -1) {
                Ir::Factory::build_store(reg, mem, *current_bb);
            } else {
                auto glo = get_reg(name);
                Ir::Factory::build_store_global(reg, glo, *current_bb);
            }
        } else {
            std::vector<std::shared_ptr<Ir::Reg>> indexs;
            for (auto &sub_exp : lval->exp()) {
                auto last = current_type;
                current_type = Ir::Int;
                indexs.push_back(std::any_cast<std::shared_ptr<Ir::Reg>>(
                    visitExp(sub_exp)));
                current_type = last;

            }
            auto mem = get_mem(name);
            current_type = mem->get_type();
            auto reg
                = std::any_cast<std::shared_ptr<Ir::Reg>>(visitExp(exp));
            Ir::Factory::build_store_array(reg, std::shared_ptr<Ir::Vector>(dynamic_cast<Ir::Vector*>(mem.get())), indexs, *current_bb);
        }
        return nullptr;
    } else {
        assert(false);
    }

}
std::any AstVisitor::visitExpr(SysYParser::ExprContext *ctx) {
    if (ctx->exp()) {
        return visitExp(ctx->exp());
    } else {
        return nullptr;
    }
}
std::any AstVisitor::visitBlockStmt(SysYParser::BlockStmtContext *ctx) {
    auto new_bb = Ir::Factory::get_new_bb(nullptr, current_func);
    if (current_bb != nullptr) {
        Ir::Factory::build_branch(new_bb, *current_bb);
    }
    current_bb = new_bb;
    return visitBlock(ctx->block());
}
std::any AstVisitor::visitRelExp(SysYParser::RelExpContext *ctx) {
    auto result = std::any_cast<std::shared_ptr<Ir::Reg>>(visitAddExp(ctx->addExp(0)));
    auto len = ctx->addExp().size();
    for (size_t i = 1; i < len; ++i) {
        auto reg = std::any_cast<std::shared_ptr<Ir::Reg>>(visitAddExp(ctx->addExp(i)));
        auto new_re = Ir::Factory::get_reg(Ir::Bool);
        Ir::Factory::build_binary(new_re, result, reg, rel_type_trans(ctx->relOp(i - 1)), *current_bb);
        result = new_re;
    }
    return result;
}

std::any AstVisitor::visitEqExp(SysYParser::EqExpContext *ctx) {
    auto result = std::any_cast<std::shared_ptr<Ir::Reg>>(visitRelExp(ctx->relExp(0)));
    auto len = ctx->relExp().size();
    if (result->get_type() != Ir::Bool && len == 1) {
        auto new_reg = Ir::Factory::get_reg(Ir::Bool);
        Ir::Factory::build_binary(new_reg, result, Ir::Factory::get_zero_reg(result->get_type()), Ir::NE, *current_bb);
        result = new_reg;
    }
    for (size_t i = 1; i < len; ++i) {
        auto reg = std::any_cast<std::shared_ptr<Ir::Reg>>(visitRelExp(ctx->relExp(i)));
        auto new_re = Ir::Factory::get_reg(Ir::Bool);
        if (ctx->eqOp(i - 1)->EQ()) {
            Ir::Factory::build_binary(new_re, result, reg, Ir::EQ, *current_bb);
        } else {
            Ir::Factory::build_binary(new_re, result, reg, Ir::NE, *current_bb);
        }
        result = new_re;
    }
    return result;
}

std::any AstVisitor::visitLAndExp(SysYParser::LAndExpContext *ctx) {

    auto result
        = std::any_cast<std::shared_ptr<Ir::Reg>>(visitEqExp(ctx->eqExp(0)));
    auto len = ctx->eqExp().size();

    for (size_t i = 1; i < len; ++i) {
        auto new_bb = Ir::Factory::get_new_bb(current_loop, current_func);
        Ir::Factory::build_branch(result, new_bb, current_cond_struct.false_block, *current_bb);
        current_bb = new_bb;
        result = std::any_cast<std::shared_ptr<Ir::Reg>>(visitEqExp(ctx->eqExp(i)));
    }
    Ir::Factory::build_branch(result, current_cond_struct.true_block, current_cond_struct.false_block, *current_bb);
    return nullptr;
}
void AstVisitor::add_lib() {
    auto name = std::string ("getint");
    auto func = Ir::Factory::get_func(name, Ir::Int, cu);
    add_func(name, func);
    func->set_lib(true);

    name = std::string ("getch");
    func = Ir::Factory::get_func(name, Ir::Int, cu);
    add_func(name, func);
    func->set_lib(true);

    name = std::string ("getfloat");
    func = Ir::Factory::get_func(name, Ir::Float, cu);
    add_func(name, func);
    func->set_lib(true);


    name = std::string ("_sysy_starttime");
    func = Ir::Factory::get_func(name, Ir::Void, cu);
    name = std::string ("starttime");
    add_func(name, func);
    func->set_lib(true);
    std::shared_ptr<Ir::Type> param = Ir::Factory::get_reg(Ir::Int);
    func->get_params().push_back(param);

    name = std::string ("_sysy_stoptime");
    func = Ir::Factory::get_func(name, Ir::Void, cu);
    name = std::string ("stoptime");
    add_func(name, func);
    func->set_lib(true);
    Ir::Factory::reset(func);
    param = Ir::Factory::get_reg(Ir::Int);
    func->get_params().push_back(param);


    name = std::string ("putint");
    func = Ir::Factory::get_func(name, Ir::Void, cu);
    add_func(name, func);
    func->set_lib(true);
    Ir::Factory::reset(func);
    param = Ir::Factory::get_reg(Ir::Int);
    func->get_params().push_back(param);



    name = std::string ("putch");
    func = Ir::Factory::get_func(name, Ir::Void, cu);
    add_func(name, func);
    func->set_lib(true);
    Ir::Factory::reset(func);
    param = Ir::Factory::get_reg(Ir::Int);
    func->get_params().push_back(param);

    name = std::string ("putfloat");
    func = Ir::Factory::get_func(name, Ir::Void, cu);
    add_func(name, func);
    func->set_lib(true);
    Ir::Factory::reset(func);
    param = Ir::Factory::get_reg(Ir::Float);
    func->get_params().push_back(param);

    name = std::string ("putarray");
    func = Ir::Factory::get_func(name, Ir::Void, cu);
    add_func(name, func);
    func->set_lib(true);
    Ir::Factory::reset(func);
    param = Ir::Factory::get_reg(Ir::Int);
    func->get_params().push_back(param);
    auto array = Ir::Factory::get_vector(Ir::Int);
    array->get_dims().push_back(0);
    func->get_params().push_back(array);

    name = std::string ("putfarray");
    func = Ir::Factory::get_func(name, Ir::Void, cu);
    add_func(name, func);
    func->set_lib(true);
    Ir::Factory::reset(func);
    param = Ir::Factory::get_reg(Ir::Int);
    func->get_params().push_back(param);
    array = Ir::Factory::get_vector(Ir::Float);
    array->get_dims().push_back(0);
    func->get_params().push_back(array);

    name = std::string ("getarray");
    func = Ir::Factory::get_func(name, Ir::Int, cu);
    add_func(name, func);
    func->set_lib(true);
    Ir::Factory::reset(func);
    array = Ir::Factory::get_vector(Ir::Int);
    array->get_dims().push_back(0);
    func->get_params().push_back(array);

    name = std::string ("getfarray");
    func = Ir::Factory::get_func(name, Ir::Int, cu);
    add_func(name, func);
    func->set_lib(true);
    Ir::Factory::reset(func);
    array = Ir::Factory::get_vector(Ir::Float);
    array->get_dims().push_back(0);
    func->get_params().push_back(array);

    name = std::string ("_atri_memset");
    func = Ir::Factory::get_func(name, Ir::Void, cu);
    add_func(name, func);
    func->set_lib(true);
    Ir::Factory::reset(func);
    array = Ir::Factory::get_vector(Ir::Int);
    array->get_dims().push_back(0);
    func->get_params().push_back(array);
    auto size = Ir::Factory::get_reg(Ir::Int);
//    auto val = Ir::Factory::get_reg(Ir::Int);
    func->get_params().push_back(size);
//    func->get_params().push_back(val);

}

