//
// Created by shliba on 2023/6/21.
//

#ifndef COMPILER_AST_VISITOR_H
#define COMPILER_AST_VISITOR_H

#include "SysYBaseVisitor.h"
#include "symbol_table.h"
#include "../optimization/ir.h"
#include <cassert>




class AstVisitor : public SysYBaseVisitor {
    static Ir::ScalaType current_type;   //当前遍历的类型
    static Ir::Function* current_func;
    static Ir::Loop* current_loop;
    static Ir::BasicBlock* current_bb;
    static Ir::CompileUnit* cu;
    static std::unique_ptr<Ir::BasicBlock> ret_bb;
    static std::shared_ptr<Ir::MemObject> ret_mem ;
    static bool ret_flag;
public:

    std::any visitCompUnit(SysYParser::CompUnitContext *ctx) override;
    std::any visitFuncDef(SysYParser::FuncDefContext *ctx) override;
    std::any visitDecl(SysYParser::DeclContext *ctx) override;
    std::any visitConstDecl(SysYParser::ConstDeclContext *ctx) override;
    std::any visitVarDecl(SysYParser::VarDeclContext *ctx) override;
    std::any visitConstDef(SysYParser::ConstDefContext *ctx) override;
    std::any visitConstExp(SysYParser::ConstExpContext *ctx) override;
    std::any visitVarDef(SysYParser::VarDefContext *ctx) override;
    std::any visitFuncFParam(SysYParser::FuncFParamContext *ctx) override;
    std::any visitBlock(SysYParser::BlockContext *ctx) override;
    std::any visitInitVal(SysYParser::InitValContext *ctx) override;
    std::any visitExp(SysYParser::ExpContext *ctx) override;
    std::any visitMulExp(SysYParser::MulExpContext *ctx) override;
    std::any visitUnaryExp(SysYParser::UnaryExpContext *ctx) override;
    std::any visitPrimaryExp(SysYParser::PrimaryExpContext *ctx) override;
    std::any visitAddExp(SysYParser::AddExpContext *ctx) override;
    std::any visitLVal(SysYParser::LValContext *ctx) override;
    std::any visitStmt(SysYParser::StmtContext *ctx);
    std::any visitIf(SysYParser::IfContext *ctx) override;
    std::any visitWhile(SysYParser::WhileContext *ctx) override;
    std::any visitBreak(SysYParser::BreakContext *ctx) override;
    std::any visitContinue(SysYParser::ContinueContext *ctx) override;
    std::any visitReturn(SysYParser::ReturnContext *ctx) override;
    std::any visitCond(SysYParser::CondContext *ctx) override;
    std::any visitAssign(SysYParser::AssignContext *ctx) override;
    std::any visitExpr(SysYParser::ExprContext *ctx) override;
    std::any visitBlockStmt(SysYParser::BlockStmtContext *ctx) override;
    std::any visitRelExp(SysYParser::RelExpContext *ctx) override;
    std::any visitEqExp(SysYParser::EqExpContext *ctx) override;
    std::any visitLAndExp(SysYParser::LAndExpContext *ctx) override;

    void visit_global_def(SysYParser::VarDefContext *ctx);


    void visit_first_pass(SysYParser::CompUnitContext *ctx);

    int visit_int_mul(SysYParser::MulExpContext *ctx);

    int visit_int_unary(SysYParser::UnaryExpContext *ctx);

    int visit_int_const(SysYParser::ConstExpContext *ctx);

    int visit_int_add(SysYParser::AddExpContext *ctx);

    float visit_float_const(SysYParser::ConstExpContext *ctx);
    
    float visit_float_mul(SysYParser::MulExpContext *ctx);
    
    float visit_float_unary(SysYParser::UnaryExpContext *ctx);

    float visit_float_add(SysYParser::AddExpContext *ctx);

    void visit_const_array(SysYParser::ConstInitValContext *ctx, std::shared_ptr<Ir::Vector> array);

    int visit_const_array_sub(SysYParser::ConstInitValContext *ctx, std::shared_ptr<Ir::Vector> array,
                               std::vector<Ir::RegConst>&, int);

    int visit_array_initval_sub(SysYParser::InitValContext *ctx, std::shared_ptr<Ir::Vector> array,
                              std::vector<std::shared_ptr<Ir::Reg>>& seq, int deep);

    std::shared_ptr<Ir::Reg> visit_global_initval(SysYParser::InitValContext*);

    void visit_global_array(SysYParser::InitValContext *ctx, std::shared_ptr<Ir::Vector> array);

    int visit_global_array_sub(SysYParser::InitValContext *ctx, std::shared_ptr<Ir::Vector> array,
                              std::vector<Ir::RegConst>&, int);

    std::any visitBlockItem(SysYParser::BlockItemContext *ctx) override;

    static void add_lib();
};





#endif // COMPILER_AST_VISITOR_H

