//
// Created by shliba on 2023/6/21.
//

#include "symbol_table.h"

static int localarray_num = 0;


std::vector<Ir::RegConst> & get_global_array(std::string &name) {
    if (global_arrays.count(name)) {
        return global_arrays[name];
    } else {
        assert(false);
    }
}

void add_global_array(const std::string &name, std::vector<Ir::RegConst>& array, Ir::ScalaType type) {
    global_arrays[name] = array;
    if (type == Ir::Int) {
        table_cu->get_global_int_arrays()[name];
        for ( auto &item : array) {
            table_cu->get_global_int_arrays()[name].push_back(item.int_num);
        }
    } else {
        table_cu->get_global_float_arrays()[name];
        for ( auto &item : array) {
            table_cu->get_global_float_arrays()[name].push_back(item.float_num);
        }
    }
}




int get_localarray() {
    return localarray_num++;
}


void add_func(const string &name,  Ir::Function *func) {
    functions[name] = func ;
}

Ir::Function* get_func(const std::string& name) {
    return functions.find(name)->second;
}

void create_scope() {
    scope_stack.push_back(std::make_shared<RegMap>());
    mem_stack.push_back(std::make_shared<MemObjectMap>());
}

void delete_scope() {
    scope_stack.pop_back();
    mem_stack.pop_back();
}



void add_new_reg(const std::string& name, const std::shared_ptr<Ir::Reg> & reg) {
    auto scope = scope_stack.back();
    if (scope_stack.size() == 1) {
        reg->set_global_name(name);
        table_cu->get_globals().push_back(reg);
    }
    (*scope)[name] = reg;
}

void add_global( const std::shared_ptr<Ir::Type> & t) {
    if (is_global_scope()) {
        return ;
    }
    table_cu->get_globals().push_back(t);
}

void set_cu(Ir::CompileUnit *cu) {
    table_cu = cu;
}

void add_new_mem(const std::string& name, const std::shared_ptr<Ir::MemObject> & vector) {
    auto scope = mem_stack.back();
    if (mem_stack.size() == 1) {

        table_cu->get_globals().push_back(vector);
    }
    (*scope)[name] = vector;
}

void add_global_mem(const std::string& name, const std::shared_ptr<Ir::MemObject> & vector) {
    auto scope = mem_stack.front();
    (*scope)[name] = vector;
}




int get_const_int(const std::string& name) {
    auto reg = get_reg(name);
    if (!reg->is_constant()) {
        assert(false);
    } else if (reg->get_type() == Ir::Float){
        return (int)reg->get_num().float_num;
    } else {
        return reg->get_num().int_num;
    }
}

float get_const_float(const std::string& name) {
    auto reg = get_reg(name);
    if (!reg->is_constant()) {
        assert(false);
    } else if (reg->get_type() == Ir::Float){
        return reg->get_num().float_num;
    } else {
        return (float)reg->get_num().int_num;
    }
}


void add_global_int(std::string& name, int initval) {
    global_consts.insert({name, initval});
}

bool is_global_scope() {
    return scope_stack.size() == 1;
}

std::shared_ptr<Ir::MemObject> get_mem(std::string &name) {
    auto it = mem_stack.rbegin();
    auto scope = *it;
    while (scope != nullptr && scope->find(name) == scope->cend() ) {
        it++;
        if (it == mem_stack.rend()) {
            break;
        } else {
            scope = *it;
        }
    }
    std::shared_ptr<Ir::MemObject> ret(new Ir::MemObject(Ir::Int, -1));
    if (it == mem_stack.rend()) {
        return ret;
    } else {
        ret = scope->find(name)->second;
    }
    return ret;
}


std::shared_ptr<Ir::Reg> get_reg(const string &name) {
    auto it = scope_stack.rbegin();
    auto scope = *it;
    while (scope != nullptr && scope->find(name) == scope->cend() ) {
        it++;
        if (it == scope_stack.rend()) {
            break;
        } else {
            scope = *it;
        }
    }
    std::shared_ptr<Ir::Reg> ret;
    if (it == scope_stack.rend()) {
        assert(false);
    } else {
        ret = scope->find(name)->second;
    }
    return ret;
}

void add_global_float(std::string& name, float initval) {
    global_floats.insert({name, initval});
}

