//
// Created by shliba on 2023/6/21.
//

#ifndef COMPILER_SYMBOL_TABLE_H
#define COMPILER_SYMBOL_TABLE_H
#include "unordered_map"
#include "unordered_set"
#include "list"
#include "string"

#include "../optimization/ir.h"
#include "../utils/utils.h"
#include "symbol_table.h"


using std::string;



typedef std::unordered_map<std::string, std::shared_ptr<Ir::Reg>> RegMap;

typedef std::unordered_map<std::string, std::shared_ptr<Ir::MemObject>> MemObjectMap;

typedef std::unordered_set<std::string> GlobalRegMap;

typedef std::unordered_map<std::string, Ir::Function *> FuncMap;

static Ir::CompileUnit* table_cu = nullptr;

static std::list<std::shared_ptr<RegMap>> scope_stack;

static std::list<std::shared_ptr<MemObjectMap>> mem_stack;

std::shared_ptr<Ir::MemObject> get_mem(std::string &name);

static FuncMap functions;

static std::unordered_map<std::string, int> global_consts;

static std::unordered_map<std::string, float > global_floats;

static std::unordered_map<std::string, std::vector<Ir::RegConst>> global_arrays;

std::vector<Ir::RegConst> & get_global_array(std::string &name);

void add_global_array(const std::string &name, std::vector<Ir::RegConst>& array, Ir::ScalaType type);

void add_func(const string& name, Ir::Function *);

Ir::Function* get_func(const std::string& name);

void create_scope();

void delete_scope();

void add_new_reg(const std::string& name, const std::shared_ptr<Ir::Reg> & reg);

void add_global( const std::shared_ptr<Ir::Type> & t);

void add_new_mem(const std::string& name, const std::shared_ptr<Ir::MemObject> & vector);

void add_global_mem(const std::string& name, const std::shared_ptr<Ir::MemObject> & vector);


std::shared_ptr<Ir::Reg> get_reg(const std::string& name);

int get_localarray();

void add_global_float(std::string& name, float initval);

int get_const_int(const std::string& name);

float get_const_float(const std::string& name);

void add_global_int(std::string& name, int initval);

bool is_global_scope();

void set_cu(Ir::CompileUnit *);

#endif // COMPILER_SYMBOL_TABLE_H
