//
// Created by 11067 on 2023/6/27.
//

#include "ABlock.h"
#include "instructions/AInstrPseudo.h"
#include <utility>

namespace armv7 {
    void ABlock::toAssemblyString(std::ostream &out) {
        if (aligned) {
            out << "\t.align 4" << std::endl;
        }
        out << label << ":" << std::endl;
        for (auto instr: *this) {
            out << '\t' << instr->toAssemblyString() << std::endl;
        }
        if (trailingLTORG && ajTable) {
            panic("jtable with ltorg");
        }
        if (ajTable) {
            ajTable->toAssemblyString(out);
        }
        if (trailingLTORG) {
            out << "\t.ltorg" << std::endl;
//            out << "\t.space " << trailingLTORG * 4 << std::endl;
        }
//        if (!myLiteralPool.empty()) {
//            out << label << "LP" << ":" << std::endl;
//            for (auto item: myLiteralPool) {
//                out << "\t.word " << item << std::endl;
//            }
//        }
//        whenDebug{
//            out << "\t@liveIn: ";
//            for (auto &reg: liveIn) {
//                out << aReg2String(reg) << ", ";
//            }
//            out <<"\n\t@liveOut: ";
//            for (auto &reg: liveOut) {
//                out << aReg2String(reg) << ", ";
//            }
//            out <<"\n\t@def: ";
//            for (auto &reg: def) {
//                out << aReg2String(reg) << ", ";
//            }
//            out <<"\n\t@use: ";
//            for (auto &reg: use) {
//                out << aReg2String(reg) << ", ";
//            }
//            out << '\n';
//        }
        whenDebug {
            out << "\t@prec: ";
            for (auto &bb: prec) {
                out << bb->label << ", ";
            }
            out << "\n\t@succ: ";
            for (auto &bb: succ) {
                out << bb->label << ", ";
            }
            out << "\n\t@succB: ";
            if (succBranch) {
                out << succBranch->label << ' ';
            }
            out << "failB: ";
            if (failBranch) {
                out << failBranch->label;
            }
            out << '\n';
        };
    }

    ABlock::ABlock(std::string label) : label(std::move(label)) {}

    size_t ABlock::getInstrCntNoComment() {
        size_t ret = 0;
        for (auto instr: *this) {
            if (dynamic_cast<AInstrComment*>(instr) == nullptr) {
                ret++;
            }
        }
        return ret;
    }

    void AJTable::toAssemblyString(std::ostream& out) const {
        out << label << ":\n";
        out << "\tadd pc, pc, " << aReg2String(switchReg) << ", LSL #2" << '\n';
//        out << "\tldr pc, [pc, " <<aReg2String(switchReg) << ", LSL #2]" << '\n';
//        out << "\tb " << exitBranch->label << '\n';
        out << "\t.zero 4\n";
        int slot = 0;
        for (auto [idx, b]: to) {
            auto offset = (idx - slot) * 4;
            if (offset) {
                out << "\t.zero ";
                out << std::to_string(offset);
                out << '\n';
            }
//            out << "\t.word ";
            out << "\tb ";
            out << b->label;
            out << '\n';
            slot = idx + 1;
        }
    }
} // armv7