//
// Created by 11067 on 2023/6/29.
//

#ifndef COMPILER_ACODEGEN_H
#define COMPILER_ACODEGEN_H

#include "../../optimization/ir.h"
#include "ACompileUnit.h"
#include <unordered_map>
#include <optional>

namespace armv7 {

    struct IndexHint{
        AReg base{};
        std::variant<int, AReg> index{}; // 如果是reg需要LSL 2 才是偏移
    };

    class ACodeGen {
        Ir::CompileUnit *compileUnit;


        AFunction *curAFunction{};
        Ir::Function *irFunction{};
        Ir::CallInstr *curCallInstr{};

        ABlock *curABlock{};
        Ir::BasicBlock *irBlock{};
        std::unordered_map<Ir::Function *, AFunction *> funcMap{};
        std::unordered_map<Ir::BasicBlock *, ABlock *> bbMap{};
        std::unordered_map<Ir::Type *, AData *> globVarMap{}; //记录IR中全局值在后端分配的AData
        std::unordered_map<std::string, AData *> globVecNameMap{};
        int blockCnt = 1;

        std::unordered_map<Ir::Reg *, AReg> irReg2AReg{};
        std::unordered_map<Ir::MemObject *, AReg> irMemObject2AReg{};
        std::unordered_map<Ir::Vector *, int> irStackVector2offset{}; //局部数组相对栈底的偏移
        std::unordered_map<Ir::MemObject *, int> irStackVar2offset{}; //局部变量相对栈底的偏移，mem2reg前会生成这种代码
        std::unordered_map<Ir::BasicBlock *, ABlock *> irBB2ABB{}; //IR基本块到后端IR的基本块
        std::unordered_map<Ir::Vector *, AReg> irParamVector2BaseReg{};//从参数传如的数组对应的基地址指针所在寄存器
        std::unordered_map<Ir::Reg *, Ir::Reg *> irAliasCvt{}; //处理不存在的bool类型
        std::unordered_map<Ir::MemObject *, IndexHint> indexHints{}; // 最后一次地址计算可以并入取地址
        std::vector<AInstrLS*> loadStoreMemObj{};
        VRegAllocator vRegAllocator{};

        std::unordered_map<AReg, AReg> savedRegMap{};

        bool divOpt{false}; //除常数优化
        bool mergeCmpBr{false}; //合并比较和跳转（激进的gvngcm可能造成错误）
        bool mergeBlock{false}; //合并基本块
        bool mergeCondExec{false}; //合并条件执行
        bool peepHole{false}; // 窥孔优化
        bool jtable{false}; //跳转表

        void genFunction();

        void genData();

        void genEachFunction();

        void genBlock();

        void resolveParam();

        void genProtectCtx();

        void genRecoverCtx();

        void genCallInstr();

        void placeHolderReplace();

        void arrayIndexLastOffset();


        VReg genLoadImm32(int num);

        VReg genLoadImm32F(float num);

        void genLoadImm32R(const AReg &aReg, int num);

        void genLoadImm32RF(const AReg &aReg, float num);

        void genOptimizedDiv(const AReg& dest, const AReg& left, int imm);
        void genOptimizedMod(const AReg& dest, const AReg& left, int imm);

        VReg genReg(Ir::Reg *reg);

        AReg tryReplacePlaceHolder(AReg original);

        Operand2 tryReplacePlaceHolderOp2(Operand2 operand2);

        void destructPhi();

        void replaceJumpTarget(ABlock *inWhich, ABlock *old, ABlock *nnew);

        void replacePhiBlock(ABlock *inWhich, ABlock *old, ABlock *nnew);

        std::optional<int> isHoldingConstantInt(AReg &aReg);

        std::optional<float> isHoldingConstantFloat(AReg &aReg);
    public:
        ACompileUnit *mCompileUnit = new ACompileUnit{};

        explicit ACodeGen(Ir::CompileUnit *_compileUnit);

        ACodeGen(Ir::CompileUnit *_compileUnit, const std::unordered_set<std::string>& optEnabled);

        void run();

    };

} // armv7

#endif //COMPILER_ACODEGEN_H
