//
// Created by 11067 on 2023/6/29.
//

#include <sstream>
#include "ACompileUnit.h"


namespace armv7 {
    std::string ACompileUnit::toAssemblyString() {
        std::stringstream out;
        out << "\t.arch armv7ve\n"; // 虚拟化扩展，支持sdiv指令
        out << "\t.syntax unified\n";
        out << "\t.fpu neon\n"; //使用neon （可选vfpv3-d16）

        out << "\t.arm\n"; //使用32位arm指令
        out << "\t.data\n"; //开始数据段
//        info("reached here");
        for (const auto &d: data) {
            d->toAssemblyString(out);
        }
//        for (const auto &j: jtables) {
//            j->toAssemblyString(out);
//        }
        out << "\t.bss\n"; // 开始初始化为0的数据段
        for (const auto &d: bss) {
            d->toAssemblyString(out);
        }
        out << "\t.text\n";
        out << "\t.global main" << std::endl;
        out << "\t.type main, %function" << std::endl;
        if (genAtriBuiltin) {
            out << supportFunctions[0];
            out << supportFunctions[1];
        }
        for (const auto &func: *this) {
            func->toAssemblyString(out);
        }
        return out.str();

    }

    ACompileUnit::ACompileUnit() {
        atriBuiltinName = {
                "_atri_memset",
                "_atri_memcpy"
        };
        // 这里可以查到系统调用号
        // https://github.com/torvalds/linux/blob/master/arch/x86/entry/syscalls/syscall_32.tbl
        // MMAP_FLAGS = MAP_PRIVATE | MAP_ANONYMOUS | MAP_STACK
        // MMAP_PROT = PROT_READ | PROT_WRITE
        // 给系统调用传参的方法
        // https://gist.github.com/yamnikov-oleg/454f48c3c45b735631f2
        // 如果不产生栈溢出，其实可以不给子进程分配栈？thu去年就是这样干的
        // _atri_memcpy(dest,src,len) len以字节为单位，要求4byte对齐,
        // _atri_memset(dest, len) len以字节为单位，要求4byte对齐
        supportFunctions = {
                R"(.align 4
_atri_memset:
    mov r2, #0
    mov r3, #0
    add r1, r1, r0

    cmp r0, r1
    bge _atri_memset_2

    tst r0, #7
    streq r2, [r0], #4

    cmp r0, r1
    bge _atri_memset_2

    tst r1, #7
    streq r2, [r1, #-4]!

    cmp r0, r1
    bge _atri_memset_2
_atri_memset_1:
    strd r2, r3, [r0], #8
    cmp r0, r1
    blt _atri_memset_1
_atri_memset_2:
    bx lr
)",
                R"(.align 4
_atri_memcpy:
    add r12, r2, r0
    cmp r0, r12
    bge _atri_memcpy8
    tst r2, #7
    beq _atri_memcpy_1
    add r3, r2, r1
    ldr r2, [r3, #-4]
    str r2, [r12, #-4]!
    cmp r0, r12
    bge _atri_memcpy8
_atri_memcpy_1:
    ldrd r2, r3, [r1], #8
    strd r2, r3, [r0], #8
    cmp r0, r12
    blt _atri_memcpy_1
_atri_memcpy8:
    bx lr
)",
                R"(     SYS_clone = 120
    SYS_old_mmap = 90
    SYS_mmap_pgoff = 192
    SYS_wait = 284
    SYS_exit = 1
    CLONE_VM = 256
    SIGCHLD = 17

    MMAP_FLAGS = 131106
    MMAP_PROT = 3
    STACK_SIZE = 4096

    P_ALL = 0
    WEXITED = 4
_atri_ssfork_raw: @share stack
	vmov s0, r4
	vmov s1, r5
	vmov s2, r6
	vmov s3, r7
    mov r0, #(CLONE_VM | SIGCHLD)
    mov r1, sp
    mov r2, #0
    mov r3, #0
    mov r4, #0
    mov r5, #0
    mov r6, #0
    mov r7, #SYS_clone
    swi #0
    vmov r4, s0
    vmov r5, s1
    vmov r6, s2
    vmov r7, s3
    bx lr


_atri_join_raw:
    vmov s0, r4
	vmov s1, r5
	vmov s2, r6
	vmov s3, r7

    mov r0, #P_ALL
    mov r1, #0
    mov r2, #0
    mov r3, #WEXITED
    mov r4, #0
    mov r5, #0
    mov r6, #0
    mov r7, #SYS_wait
    swi #0

    vmov r4, s0
    vmov r5, s1
    vmov r6, s2
    vmov r7, s3

_atri_exit:
    mov r7, #SYS_exit
    swi #0

_atri_memcpy:
    add r2, r0, r2
_atri_memcpy_L1:
    ldrd r3, r12, [r1]
    add r1, r1, #8
    strd r3, r12, [r0]
    add r0, r0, #8
    cmp r0, r2
    blt _atri_memcpy_L1
_atri_memcpy_L2:
    bx lr
)"
        };
        funcTemplate = {
                R"(_atri_fork_raw_%s:
	vmov s0, r4
	vmov s1, r5
	vmov s2, r6
	vmov s3, r7

    @mmap stack
    mov r0, 0 @addr
    mov r1, #STACK_SIZE @length
    mov r2, #MMAP_PROT @prot
    mov r3, #MMAP_FLAGS @flags
    mov r4, -1 @fd
    mov r5, 0
    mov r6, 0
    mov r7, #SYS_mmap_pgoff
    swi #0

    add r1, SP, %s
    mov r4, r1 @ save child stack bottom
    mov r2, %s
    bl _atri_memcpy

    mov r0, #(CLONE_VM | SIGCHLD)
    mov r1, r4 @ child stack bottom
    mov r2, #0
    mov r3, #0
    mov r4, #0
    mov r6, #0
    mov r7, #SYS_clone
    swi #0
    vmov r4, s0
    vmov r5, s1
    vmov r6, s2
    vmov r7, s3
    bx lr
)"
        };
    }
} // armv7