//
// Created by 11067 on 2023/6/28.
//

#include "AData.h"

#include <utility>
#include <iostream>
#include "../../utils/bitUtils.h"
namespace armv7 {


    void AData::toAssemblyString(std::ostream &out) const {
        out << label << ':' << std::endl;
        if (auto pIntArr = std::get_if<std::vector<IntOrZero>>(&data)) {
            for (auto &val: *pIntArr) {
                if(auto zeros = std::get_if<int>(&val)) {
                    unsigned long long s = *zeros;
                    if (s * 4 > INT32_MAX) {
                        auto r = s % 4;
                        s /= 4;
                        out << "\t.zero " << s * 4 << std::endl;
                        out << "\t.zero " << s * 4 << std::endl;
                        out << "\t.zero " << s * 4 << std::endl;
                        out << "\t.zero " << s * 4 << std::endl;
                        out << "\t.zero " << r * 4 << std::endl;
                    } else {
                        out << "\t.zero " << s * 4 << std::endl;
                    }

                } else {
                    auto &arr = std::get<std::vector<int>>(val);
                    for (auto elem: arr) {
                        out << "\t.word " << toUI32(elem) << std::endl;
                    }
                }
            }
        } else {
            auto &floatArr = std::get<std::vector<FloatOrZero>>(data);
            for (auto &val: floatArr) {
                if(auto zeros = std::get_if<int>(&val)) {
                    unsigned long long s = *zeros;
                    if (s * 4 > INT32_MAX) {
                        auto r = s % 4;
                        s /= 4;
                        out << "\t.zero " << s * 4 << std::endl;
                        out << "\t.zero " << s * 4 << std::endl;
                        out << "\t.zero " << s * 4 << std::endl;
                        out << "\t.zero " << s * 4 << std::endl;
                        out << "\t.zero " << r * 4 << std::endl;
                    } else {
                        out << "\t.zero " << s * 4 << std::endl;
                    }
                } else {
                    auto &arr = std::get<std::vector<float>>(val);
                    for (auto elem: arr) {
                        out << "\t.word " << toUI32(elem) << std::endl;
                    }
                }
            }
        }
    }



    AData * makeADataFloat(const std::string& label, const std::vector<float>& array) {
        FloatOrZero arr = std::vector<float>{array};
        return new AData{label, std::vector<FloatOrZero>{arr}};
    }

    AData * makeADataInt(const std::string& label, const std::vector<int>& array) {
        IntOrZero arr = std::vector<int>{array};
        return new AData{label,std::vector<IntOrZero>{arr}};
    }


    AData *makeADataFloatZeroed(const std::string &label, int s) {
        FloatOrZero arr = s;
        return new AData{label, std::vector<FloatOrZero>{arr}};
    }

    AData *makeADataIntZeroed(const std::string &label, int s) {
        IntOrZero arr = s;
        return new AData{label, std::vector<IntOrZero>{arr}};
    }



} // armv7