//
// Created by 11067 on 2023/6/27.
//

#include "AFunction.h"
#include "instructions/AInstrB.h"

#include <utility>

namespace armv7 {

    void AFunction::toAssemblyString(std::ostream &out) {
        if (!isLib) {
            out << label << ":" << std::endl;
            for (auto bb: *this) {
                bb->toAssemblyString(out);
            }
        }
    }

    AFunction::AFunction(std::string label) : label(std::move(label)) {}

    bool AFunction::checkPrecSucc() {
        for (auto pABlock: *this) {
            for (auto prec: pABlock->prec) {
                if (!prec->succ.count(pABlock)) {
                    return false;
                }
            }
            for (auto succ: pABlock->succ) {
                if (!succ->prec.count(pABlock)) {
                    return false;
                }
            }
        }
        return true;
    }

    void AFunction::calcPrecSucc() {
        for (auto bb: *this) {
            bb->prec.clear();
            bb->succ.clear();
            bb->succBranch = nullptr;
            bb->failBranch = nullptr;
            auto lastInstr = bb->getLast();
            auto last2Instr = lastInstr->getPrev(); // may null
            auto flag = false;
            if (auto b = dynamic_cast<AInstrB *>(last2Instr)) {
                if (!b->isCall()) {
                    flag = true;
                    const auto succ = std::get<ABlock *>(b->target);
                    bb->succ.insert(succ);
                    bb->succBranch = succ;
                }
            }
            if (auto b = dynamic_cast<AInstrB *>(lastInstr)) {
                const auto succ = std::get<ABlock *>(b->target);
                bb->succ.insert(succ);
                if (flag) {
                    if (bb->succBranch == succ) {
                        last2Instr->removeMe();
                        delete last2Instr;
//                        bb->succBranch = succ;
                    } else {
                        bb->failBranch = succ;
                    }
                } else {
                    bb->succBranch = succ;
                }
            }
        }
        for (auto bb: *this) {
            for (auto &succ: bb->succ) {
                succ->prec.insert(bb);
            }
        }
    }
} // armv7