//
// Created by 11067 on 2023/6/26.
//

#ifndef COMPILER_AINSTR_H
#define COMPILER_AINSTR_H


#include "../../utils/LinkedList.h"
#include "aarch.h"
#include "instructions/common.h"
#include <vector>

namespace armv7 {

    /**
     * @brief ARMv7指令
     */
    struct AInstr: atri::LinkedNode<AInstr> {
        Cond cond{Cond::None}; // 大部分指令都可以条件执行
        explicit AInstr(Cond cond);
        AInstr() = default;
        virtual std::vector<AReg> getDef() = 0;
        virtual std::vector<AReg> getUse() = 0;
        /**
         * @brief 分配寄存器
         * @param vReg 被分配的虚拟寄存器
         * @param pReg 虚拟寄存器被分配到的物理寄存器
         */
        virtual void allocate(const VReg &vReg, const AReg &pReg) = 0;
        /**
         * 汇编指令字符串
         * @return 汇编指令字符串形式
         */
        virtual std::string toAssemblyString() = 0;
        /**
         * @brief 尝试为指令设定条件执行
         * @param cond 条件
         * @return 是否允许条件执行
         */
        virtual bool setCond(Cond _cond);
        /**
         * @brief 是否允许条件执行
         * @return 是否允许条件执行
         */
        virtual bool permitCond();
        virtual ~AInstr() = default;
    };

} // armv7

#endif //COMPILER_AINSTR_H
