//
// Created by 11067 on 2023/6/25.
//

#include "aarch.h"

namespace armv7 {

    static Reg regs[]{
            Reg{0, "r0", RegAttribute::CallerSaved},
            Reg{1, "r1", RegAttribute::CallerSaved},
            Reg{2, "r2", RegAttribute::CallerSaved},
            Reg{3, "r3", RegAttribute::CallerSaved},
            Reg{4, "r4", RegAttribute::CalleeSaved},
            Reg{5, "r5", RegAttribute::CalleeSaved},
            Reg{6, "r6", RegAttribute::CalleeSaved},
            Reg{7, "r7", RegAttribute::CalleeSaved},
            Reg{8, "r8", RegAttribute::CalleeSaved},
            Reg{9, "r9", RegAttribute::CalleeSaved},
            Reg{10, "r10", RegAttribute::CalleeSaved},
            Reg{11, "r11", RegAttribute::CalleeSaved},
            Reg{12, "r12", RegAttribute::CallerSaved},//ip
            Reg{13, "SP", RegAttribute::NA},
            Reg{14, "LR", RegAttribute::NA},
            Reg{15, "PC", RegAttribute::NA},
            Reg{16, "s0", RegAttribute::CallerSaved},
            Reg{17, "s1", RegAttribute::CallerSaved},
            Reg{18, "s2", RegAttribute::CallerSaved},
            Reg{19, "s3", RegAttribute::CallerSaved},
            Reg{20, "s4", RegAttribute::CallerSaved},
            Reg{21, "s5", RegAttribute::CallerSaved},
            Reg{22, "s6", RegAttribute::CallerSaved},
            Reg{23, "s7", RegAttribute::CallerSaved},
            Reg{24, "s8", RegAttribute::CallerSaved},
            Reg{25, "s9", RegAttribute::CallerSaved},
            Reg{26, "s10", RegAttribute::CallerSaved},
            Reg{27, "s11", RegAttribute::CallerSaved},
            Reg{28, "s12", RegAttribute::CallerSaved},
            Reg{29, "s13", RegAttribute::CallerSaved},
            Reg{30, "s14", RegAttribute::CallerSaved},
            Reg{31, "s15", RegAttribute::CallerSaved},
            Reg{32, "s16", RegAttribute::CalleeSaved},
            Reg{33, "s17", RegAttribute::CalleeSaved},
            Reg{34, "s18", RegAttribute::CalleeSaved},
            Reg{35, "s19", RegAttribute::CalleeSaved},
            Reg{36, "s20", RegAttribute::CalleeSaved},
            Reg{37, "s21", RegAttribute::CalleeSaved},
            Reg{38, "s22", RegAttribute::CalleeSaved},
            Reg{39, "s23", RegAttribute::CalleeSaved},
            Reg{40, "s24", RegAttribute::CalleeSaved},
            Reg{41, "s25", RegAttribute::CalleeSaved},
            Reg{42, "s26", RegAttribute::CalleeSaved},
            Reg{43, "s27", RegAttribute::CalleeSaved},
            Reg{44, "s28", RegAttribute::CalleeSaved},
            Reg{45, "s29", RegAttribute::CalleeSaved},
            Reg{46, "s30", RegAttribute::CalleeSaved},
            Reg{47, "s31", RegAttribute::CalleeSaved},
    };
    static SIMDReg simdRegs[]{
        SIMDReg{0, "d0"},
        SIMDReg{1, "d1"},
        SIMDReg{2, "d2"},
        SIMDReg{3, "d3"},
        SIMDReg{4, "d4"},
        SIMDReg{5, "d5"},
        SIMDReg{6, "d6"},
        SIMDReg{7, "d7"},
        SIMDReg{8, "d8"},
        SIMDReg{9, "d9"},
        SIMDReg{10, "d10"},
        SIMDReg{11, "d11"},
        SIMDReg{12, "d12"},
        SIMDReg{13, "d13"},
        SIMDReg{14, "d14"},
        SIMDReg{15, "d15"},
        SIMDReg{16, "d16"},
        SIMDReg{17, "d17"},
        SIMDReg{18, "d18"},
        SIMDReg{19, "d19"},
        SIMDReg{20, "d20"},
        SIMDReg{21, "d21"},
        SIMDReg{22, "d22"},
        SIMDReg{23, "d23"},
        SIMDReg{24, "d24"},
        SIMDReg{25, "d25"},
        SIMDReg{26, "d26"},
        SIMDReg{27, "d27"},
        SIMDReg{28, "d28"},
        SIMDReg{29, "d29"},
        SIMDReg{30, "d30"},
        SIMDReg{31, "d31"},

    };
    bool VReg::operator==(const VReg &rhs) const {
        return id == rhs.id;
    }

    bool VReg::operator!=(const VReg &rhs) const {
        return !(rhs == *this);
    }

    std::string VReg::toAssemblyString() const {
        if (id > 0) {
            if (type == Type::Float) {
                return std::string{"fvir"} + std::to_string(id);
            } else {
                return std::string{"vir"} + std::to_string(id);
            }
        } else if (id == 0) {
            return std::string{"vir_null"};
        } else {
            return std::string{"placeHolder"} + std::to_string(-id);
        }
    }

    bool VReg::isPlaceHolder() const {
        return id <= -1;
    }

    bool VReg::nullReg() const {
        return id == 0;
    }

    Reg &get_reg_by_id(int i) {
        return regs[i];
    }

    Reg &get_freg_by_id(int i) {
        return regs[i + 16];
    }

    std::vector<Reg> get_saved_reg() {
        std::vector<Reg> ret{};
        for (int i = 0; i < 16; ++i) {
            if (regs[i].regAttribute == RegAttribute::CalleeSaved) {
                ret.push_back(regs[i]);
            }
        }
        return ret;
    }

    std::vector<Reg> get_saved_freg() {
        std::vector<Reg> ret{};
        for (int i = 16; i < 48; ++i) {
            if (regs[i].regAttribute == RegAttribute::CalleeSaved) {
                ret.push_back(regs[i]);
            }
        }
        return ret;
    }

    std::vector<AReg> get_saved_reg_wrapped() {
        std::vector<AReg> ret{};
        for (int i = 0; i < 16; ++i) {
            if (regs[i].regAttribute == RegAttribute::CalleeSaved) {
                ret.emplace_back(regs[i]);
            }
        }
        return ret;
    }

    std::vector<AReg> get_saved_freg_wrapped() {
        std::vector<AReg> ret{};
        for (int i = 16; i < 48; ++i) {
            if (regs[i].regAttribute == RegAttribute::CalleeSaved) {
                ret.emplace_back(regs[i]);
            }
        }
        return ret;
    }

    Reg &get_sp() {
        return regs[13];
    }

    Reg &get_lr() {
        return regs[14];
    }

    Reg &get_ip() {
        return regs[12];
    }

    int fruid(const AReg &aReg) {
        constexpr int pRegCnt = sizeof regs / sizeof(Reg);
        if (auto vReg = std::get_if<VReg>(&aReg)) {
            return vReg->id - 1 + pRegCnt; // 0是不分配的，VReg(id=0)表示空
        } else {
            auto &reg = std::get<Reg>(aReg);;
            return reg.id;
        }
    }


    Reg &get_pc() {
        return regs[15];
    }

    std::vector<Reg> get_temp_reg() {
        std::vector<Reg> ret{};
        for (int i = 0; i < 4; ++i) {
            ret.push_back(regs[i]);
        }
        ret.push_back(regs[12]);
        for (int i = 0; i < 16; ++i) {
            ret.push_back(regs[i + 16]);
        }
        return ret;
    }

    SIMDReg &get_simd_reg_by_id(int i) {
        return simdRegs[i];
    }

    SIMDReg &to_alias(Reg freg) {
        if ((freg.id - 16) & 1) {
            panic("invalid simd reg, must be 2n,2n+1");
        }
        return get_simd_reg_by_id((freg.id - 16) / 2);
    }
}

bool armv7::Reg::operator==(const armv7::Reg &rhs) const {
    return id == rhs.id;
}

bool armv7::Reg::operator!=(const armv7::Reg &rhs) const {
    return !(*this == rhs);
}

std::string armv7::Reg::toAssemblyString() const {
    if (id < 48) {
        return std::string{abiName};
    } else {
        return std::string{"vir"} + std::to_string(id - 48);
    }
}
