//
// Created by 11067 on 2023/6/26.
//

#include "AInstrALU.h"
#include "../../../utils/logger.h"

using namespace std::literals::string_literals;
namespace armv7 {
    std::vector<AReg> AInstrALU::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrALU::getUse() {
        std::vector<AReg> ret{rn};
        if (std::holds_alternative<Operand2Reg>(operand2)) {
            auto &operand2Reg = std::get<Operand2Reg>(operand2);
            ret.push_back(operand2Reg.base);
            if (std::holds_alternative<AReg>(operand2Reg.shift)) {
                ret.push_back(std::get<AReg>(operand2Reg.shift));
            }
        }
        return ret;
    }

    void AInstrALU::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd,vReg)) {
            rd = pReg;
        }
        if (matchRegAlloc(rn, vReg)) {
            rn = pReg;
        }
        if (std::holds_alternative<Operand2Reg>(operand2)) {
            auto &operand2Reg = std::get<Operand2Reg>(operand2);
            if (matchRegAlloc(operand2Reg.base, vReg)) {
                operand2Reg.base = pReg;
            }
            if (std::holds_alternative<AReg>(operand2Reg.shift)) {
                auto &shift = std::get<AReg>(operand2Reg.shift);
                if (matchRegAlloc(shift, vReg)) {
                    shift = pReg;
                }
            }
        }

    }

    std::string AInstrALU::toAssemblyString() {
        const static char *ops[]{"add", "sub", "rsb", "and", "eor", "orr", "bic"};
        std::string op = ops[static_cast<int>(type)];
        if (setFlag) {
            op += 's';
        }
        if (cond != Cond::None) {
            op += cond2StringView(cond);
        }
        op += ' ';
        op += aReg2String(rd) + ", ";
        op += aReg2String(rn) + ", ";
        op += operand2ToString(operand2);
        return op;
    }

    AInstrALU::AInstrALU(AInstrALU::Type type, Cond cond, bool setFlag, const AReg &rd, const AReg &rn,
                         const Operand2 &operand2) : AInstr(cond), type(type), setFlag(setFlag), rd(rd), rn(rn),
                                                     operand2(operand2) {}

    AInstrALU *makeAInstrALURR(AInstrALU::Type arithmType, AReg rd, AReg rm, AReg rn) {
        return new AInstrALU{
                arithmType,
                Cond::None,
                false,
                rd,
                rm,
                Operand2Reg{
                        rn,
                        ShiftType::None,
                        0
                }
        };
    }

    AInstrCMP *makeAInstrCMPRR(AReg rm, AReg rn) {
        return new AInstrCMP{
                false,
                rm,
                Operand2Reg{
                        rn,
                        ShiftType::None,
                        0
                },
                Cond::None
        };
    }

    AInstrALU *makeAInstrALURImm(AInstrALU::Type arithmType, AReg rd, AReg rn, int imm8m) {
        if (!isImm8m(imm8m)) {
            warn(imm8m, " is not imm8m.");
        }
        return new AInstrALU{
                arithmType,
                Cond::None,
                false,
                rd,
                rn,
                imm8m
        };
    }

    std::vector<AReg> AInstrCMP::getDef() {
        return {};
    }

    std::vector<AReg> AInstrCMP::getUse() {
        std::vector<AReg> ret{rm};
        if (std::holds_alternative<Operand2Reg>(operand2)) {
            auto operand2Reg = std::get<Operand2Reg>(operand2);
            ret.push_back(operand2Reg.base);
            if (std::holds_alternative<AReg>(operand2Reg.shift)) {
                ret.push_back(std::get<AReg>(operand2Reg.shift));
            }
        }
        return ret;
    }

    void AInstrCMP::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rm,vReg)) {
            rm = pReg;
        }
        if (std::holds_alternative<Operand2Reg>(operand2)) {
            auto &operand2Reg = std::get<Operand2Reg>(operand2);
            if (matchRegAlloc(operand2Reg.base, vReg)) {
                operand2Reg.base = pReg;
            }
            if (std::holds_alternative<AReg>(operand2Reg.shift)) {
                auto &shift = std::get<AReg>(operand2Reg.shift);
                if (matchRegAlloc(shift, vReg)) {
                    shift = pReg;
                }
            }
        }
    }

    std::string AInstrCMP::toAssemblyString() {
        auto ret = "cmp"s;
        ret += cond2StringView(cond);
        ret += ' ';
        return  ret  + aReg2String(rm) + ", " + operand2ToString(operand2);
    }

    AInstrCMP::AInstrCMP(bool neg, const AReg &rm, const Operand2 &operand2, Cond _cond) : AInstr{_cond}, neg(neg), rm(rm), operand2(operand2){}



    AInstrCMP::AInstrCMP(bool neg, const AReg &rm, const Operand2 &operand2) : neg(neg), rm(rm), operand2(operand2) {}
} // armv7