//
// Created by 11067 on 2023/6/26.
//

#ifndef COMPILER_AINSTRALU_H
#define COMPILER_AINSTRALU_H

#include "../aarch.h"
#include "common.h"
#include "../AInstr.h"

namespace armv7 {
    /**
     * @brief ARM数据处理指令
     * ARMv7没有明确分出I型、R型指令，区别主要在operand2的格式
     */
    struct AInstrALU: AInstr {
        enum class Type{
            Add,
            Sub,
            Rsb,
            And,
            Eor,// xor
            Orr,// or
            Bic,// and not
        };

        Type type;
        bool setFlag; // 是否影响标志寄存器
        AReg rd;
        AReg rn;
        Operand2 operand2;

        AInstrALU(Type type, Cond cond, bool setFlag, const AReg &rd, const AReg &rn, const Operand2 &operand2);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

    };
    struct AInstrCMP: AInstr{
        bool neg; //for cmn
        AReg  rm;
        Operand2 operand2;

        AInstrCMP(bool neg, const AReg &rm, const Operand2 &operand2, Cond _cond);

        AInstrCMP(bool neg, const AReg &rm, const Operand2 &operand2);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

    };
    AInstrCMP *makeAInstrCMPRR(AReg rm, AReg rn);
    AInstrALU *makeAInstrALURR(AInstrALU::Type arithmType, AReg rd, AReg rm, AReg rn);
    AInstrALU *makeAInstrALURImm(AInstrALU::Type arithmType, AReg rd, AReg rn, int imm8m);

} // armv7

#endif //COMPILER_AINSTRALU_H
