//
// Created by 11067 on 2023/6/27.
//

#include "AInstrB.h"

using namespace std::literals;
namespace armv7 {
    std::vector<AReg> AInstrB::getDef() {
//        if (linked) {
//            return {get_lr()};
//        } else {
//            return {};
//        }
        return {};
    }

    std::vector<AReg> AInstrB::getUse() {
        return {}; //may use ASPR for cond.
    }

    void AInstrB::allocate(const VReg &vReg, const AReg &pReg) {
        // nothing to do
    }

    std::string AInstrB::toAssemblyString() {
        std::string ret;
        if (linked) {
            ret += "bl";
        } else {
            ret += "b";
        }
        if (cond != Cond::None) {
            ret += cond2StringView(cond);
        }
        if (auto pFunc = std::get_if<AFunction *>(&target)) {
            ret += ' ';
            ret += (**pFunc).label;
        } else {
            auto pBlock = std::get<ABlock *>(target);
            ret += ' ';
            ret += pBlock->label;
        }
        return ret;
    }

    AInstrB::AInstrB(const std::variant<AFunction *, ABlock *> &target, bool linked, Cond _cond) : AInstr{_cond},
                                                                                                   target(target),
                                                                                                   linked(linked) {}

    bool AInstrB::isCall() const {
        return linked;
    }

    AInstrB *makeJump(ABlock *target) {
        return new AInstrB{
                target,
                false,
                Cond::None,
        };
    }

    AInstrB *makeBranch(Cond cond, ABlock *target) {
        return new AInstrB{
                target,
                false,
                cond
        };
    }

    AInstrCall *makeCall(AFunction *target, AInstrCall::ReturnValType returnValType, std::vector<AReg> &&paramRegs) {
        return new AInstrCall{
                target,
                Cond::None,
                returnValType,
                paramRegs
        };
    }

    AInstrBX *makeRet() {
        return new AInstrBX{
                get_lr(),
                Cond::None
        };
    }

    std::vector<AReg> AInstrBX::getDef() {
        return {};
    }

    std::vector<AReg> AInstrBX::getUse() {
        return {target};
    }

    void AInstrBX::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(target, vReg)) {
            target = pReg;
        }
    }

    std::string AInstrBX::toAssemblyString() {
        auto ret = "bx"s;
        ret += cond2StringView(cond);
        ret += ' ';
        return ret + aReg2String(target);
    }

    AInstrBX::AInstrBX(const AReg &target, Cond cond) : AInstr{cond}, target(target) {}

    std::vector<AReg> AInstrCall::getDef() {
        switch (returnValType) {
            case ReturnValType::None:
                return {};
            case ReturnValType::Int:
                return {get_reg_by_id(0)};
            case ReturnValType::Float:
                return {get_freg_by_id(0)};
        }
    }

    std::vector<AReg> AInstrCall::getUse() {
        return std::vector<AReg>{usedParamPassReg};
    }

    void AInstrCall::allocate(const VReg &vReg, const AReg &pReg) {

    }

    std::string AInstrCall::toAssemblyString() {
        std::string ret{"bl"};
        if (cond != Cond::None) {
            ret += cond2StringView(cond);
        }
        ret += ' ';
        ret += target->label;
        return ret;
    }

    AInstrCall::AInstrCall(AFunction *target, Cond _cond, AInstrCall::ReturnValType returnValType,
                           const std::vector<AReg> &usedParamPassReg) : AInstr{_cond}, target(target),
                                                                        returnValType(returnValType),
                                                                        usedParamPassReg(usedParamPassReg) {}


} // armv7