//
// Created by 11067 on 2023/6/27.
//

#ifndef COMPILER_AINSTRB_H
#define COMPILER_AINSTRB_H

#include <variant>
#include "../AFunction.h"
#include "common.h"
namespace armv7 {
    /**
     * @brief 跳转指令
     *
     * 范围是32MB
     * 其实可以不用管这个范围，汇编器会处理
     *
     * @cite Machine-level BL instructions have restricted ranges from the address of the current instruction.
     * However, you can use these instructions even if label is out of range. Often you do not know where the
     * linker places label. When necessary, the linker adds code to enable longer branches. The added code is
     * called a veneer.
     */
    struct AInstrB: AInstr {
        std::variant<AFunction*,ABlock*> target;
        bool linked; //是否链接，若链接为bl指令
        AInstrB(const std::variant<AFunction*, ABlock*> &target, bool linked, Cond _cond);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

        bool isCall() const;

    };

    struct AInstrCall: AInstr {
        AFunction *target;
        enum class ReturnValType{
            None,
            Int,
            Float,
        };
        ReturnValType returnValType;
        std::vector<AReg> usedParamPassReg{};

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

        AInstrCall(AFunction *target, Cond _cond, ReturnValType returnValType,
                   const std::vector<AReg> &usedParamPassReg);

    };
    AInstrB *makeJump(ABlock *target);
    AInstrB *makeBranch(Cond cond, ABlock *target);
    AInstrCall * makeCall(AFunction *target, AInstrCall::ReturnValType returnValType, std::vector<AReg> &&paramRegs);
    struct AInstrBX: AInstr {
        AReg target;
        AInstrBX(const AReg &target, Cond cond);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

    };
    AInstrBX *makeRet();
} // armv7

#endif //COMPILER_AINSTRB_H
