//
// Created by 11067 on 2023/6/28.
//

#include "AInstrLS.h"
#include "../../../utils/logger.h"
#include <cstdlib>

using namespace std::literals;
namespace armv7 {
    std::vector<AReg> AInstrLaHigh::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrLaHigh::getUse() {
        return {};
    }

    void AInstrLaHigh::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
    }

    std::string AInstrLaHigh::toAssemblyString() {
        auto ret = "movt"s;
        ret += cond2StringView(cond);
        ret += ' ';
        return ret + aReg2String(rd) + ", #:upper16:" + label->label;
    }

    AInstrLaHigh::AInstrLaHigh(const AReg &rd, const AData *label) : rd(rd), label(label) {}


    std::vector<AReg> AInstrLaLow::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrLaLow::getUse() {
        return {};
    }

    void AInstrLaLow::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
    }

    std::string AInstrLaLow::toAssemblyString() {
        auto ret = "movw"s;
        ret += cond2StringView(cond);
        ret += ' ';
        return ret + aReg2String(rd) + ", #:lower16:" + label->label;
    }

    AInstrLaLow::AInstrLaLow(AReg &rd, AData *label) : rd(rd), label(label) {}


    std::vector<AReg> AInstrPush::getDef() {
        return {};
    }

    std::vector<AReg> AInstrPush::getUse() {
        return std::vector<AReg>{reglist}; //copy constructor
    }

    void AInstrPush::allocate(const VReg &vReg, const AReg &pReg) {
        //不会生成push虚拟寄存器的代码
        warn("allocating push directly is ambiguous.");
    }

    std::string AInstrPush::toAssemblyString() {
        std::string ret{"push"};
        ret += cond2StringView(cond);
        ret += " {";
        for (auto &reg: reglist) {
            ret += aReg2String(reg);
            ret += ", ";
        }
        ret.pop_back();
        ret.pop_back();
        ret += '}';
        return ret;
    }

    AInstrPush::AInstrPush(const std::vector<AReg> &reglist, bool isFloat) : reglist(reglist), isFloat(isFloat) {}

    std::vector<AReg> AInstrPop::getDef() {
        return std::vector<AReg>{reglist};
    }

    std::vector<AReg> AInstrPop::getUse() {
        return {};
    }

    void AInstrPop::allocate(const VReg &vReg, const AReg &pReg) {
        warn("allocating pop directly is ambiguous.");
    }

    std::string AInstrPop::toAssemblyString() {
        std::string ret{"pop"};
        ret += cond2StringView(cond);
        ret += " {";
        for (auto &reg: reglist) {
            ret += aReg2String(reg);
            ret += ", ";
        }
        ret.pop_back();
        ret.pop_back();
        ret += '}';
        return ret;
    }

    AInstrPop::AInstrPop(const std::vector<AReg> &reglist, bool isFloat) : reglist(reglist), isFloat(isFloat) {}

    AInstrLS *makeLdRImm(const AReg &rd, const AReg &rn, int offset) {
        return new AInstrLS{
                false,
                Cond::None,
                AInstrLS::Type::Load,
                rd,
                rn,
                offset,
                VReg{},
                AInstrLS::SetType::None
        };
    }

    AInstrLS *makeLdRImmF(const AReg &rd, const AReg &rn, int offset) {
        return new AInstrLS{
                true,
                Cond::None,
                AInstrLS::Type::Load,
                rd,
                rn,
                offset,
                VReg{},
                AInstrLS::SetType::None
        };
    }

    AInstrLS *makeLdRR(const AReg &rd, const AReg &rn, const AReg &rm) {
        return new AInstrLS{
                false,
                Cond::None,
                AInstrLS::Type::Load,
                rd,
                rn,
                RegLS{
                        rm,
                        ShiftType::None,
                        0,
                        false,
                },
                VReg{},
                AInstrLS::SetType::None
        };
    }

    AInstrLS *makeLdRRF(const AReg &rd, const AReg &rn, const AReg &rm) {
        return new AInstrLS{
                true,
                Cond::None,
                AInstrLS::Type::Load,
                rd,
                rn,
                RegLS{
                        rm,
                        ShiftType::None,
                        0,
                        false,
                },
                VReg{},
                AInstrLS::SetType::None
        };
    }

    AInstrLS *makeStrRImm(const AReg &rd, const AReg &rn, int offset) {
        return new AInstrLS{
                false,
                Cond::None,
                AInstrLS::Type::Store,
                rd,
                rn,
                offset,
                VReg{},
                AInstrLS::SetType::None
        };
    }

    AInstrLS *makeStrRImmF(const AReg &rd, const AReg &rn, int offset) {
        return new AInstrLS{
                true,
                Cond::None,
                AInstrLS::Type::Store,
                rd,
                rn,
                offset,
                VReg{},
                AInstrLS::SetType::None
        };
    }

    AInstrLS *makeStrRR(const AReg &rd, const AReg &rn, const AReg &rm) {
        return new AInstrLS{
                false,
                Cond::None,
                AInstrLS::Type::Store,
                rd,
                rn,
                RegLS{
                        rm,
                        ShiftType::None,
                        0,
                        false
                },
                VReg{},
                AInstrLS::SetType::None
        };
    }

    AInstrLS *makeStrRRF(const AReg &rd, const AReg &rn, const AReg &rm) {
        return new AInstrLS{
                true,
                Cond::None,
                AInstrLS::Type::Store,
                rd,
                rn,
                RegLS{
                        rm,
                        ShiftType::None,
                        0,
                        false
                },
                VReg{},
                AInstrLS::SetType::None
        };
    }

    AInstrPush *makePush(const std::vector<AReg> &regList) {
        return new AInstrPush{
                regList,
                false
        };
    }

    AInstrPush *makePushF(std::vector<AReg> &regList) {
        return new AInstrPush{
                regList,
                true
        };
    }

    AInstrPop *makePop(const std::vector<AReg> &regList) {
        return new AInstrPop{
                regList,
                false,
        };
    }

    AInstrPop *makePopF(std::vector<AReg> &regList) {
        return new AInstrPop{
                regList,
                true,
        };
    }

    bool isSameRegLS(std::variant<int, RegLS> addr1, std::variant<int, RegLS> addr2) {
        if (std::holds_alternative<int>(addr1) && std::holds_alternative<int>(addr2)) {
            return std::get<int>(addr1) == std::get<int>(addr2);
        } else if (std::holds_alternative<RegLS>(addr1) && std::holds_alternative<RegLS>(addr2)) {
            auto reg1 = std::get<RegLS>(addr1);
            auto reg2 = std::get<RegLS>(addr2);
            return reg1 == reg2;
        } else {
            return false;
        }
    }

    bool isSameAddr(AInstrLS *ls1, AInstrLS *ls2) {
        return isSameRegLS(ls1->operand2, ls2->operand2) && ls1->rn == ls2->rn;
    }

    AInstrLS::AInstrLS(bool isFloat, Cond cond, AInstrLS::Type type, const AReg &rd, const AReg &rn,
                       const std::variant<int, RegLS> &operand2, const AReg &rd2, SetType setType) : AInstr{cond},
                                                                                                     isFloat(isFloat),
                                                                                                     type(type), setType(setType),
                                                                                                     rd(rd),
                                                                                                     rd2(rd2),
                                                                                                     rn(rn),
                                                                                                     operand2(operand2) {}

    std::vector<AReg> AInstrLS::getDef() {
        std::vector<AReg> ret{rd};
        if (!aRegNull(rd2)) {
            ret.push_back(rd2);
        }
        if (type == Type::Load) {
            return ret;
        } else {
            return {};
        }
    }

    std::vector<AReg> AInstrLS::getUse() {
        std::vector<AReg> ret{rn};
        if (auto pRegLS = std::get_if<RegLS>(&operand2)) {
            ret.push_back(pRegLS->rm);
        }
        if (type == Type::Store) {
            if (!aRegNull(rd2)) {
                ret.push_back(rd2);
            }
            ret.push_back(rd);
        }
        return ret;
    }

    void AInstrLS::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
        if (matchRegAlloc(rd2, vReg)) {
            rd2 = pReg;
        }
        if (matchRegAlloc(rn, vReg)) {
            rn = pReg;
        }
        if (auto pRegLS = std::get_if<RegLS>(&operand2)) {
            if (matchRegAlloc(pRegLS->rm, vReg)) {
                pRegLS->rm = pReg;
            }
        }

    }

    std::string AInstrLS::toAssemblyString() {
        std::string ret{};
        if (isFloat) {
            if (type == Type::Load) {
                ret += "vldr";
            } else {
                ret += "vstr";
            }
            if (cond != Cond::None) {
                ret += cond2StringView(cond);
            }
            if (!aRegNull(rd2)) {
                ret += ".64";
            } else {
                ret += ".32";
            }
        } else {
            if (type == Type::Load) {
                ret += "ldr";
            } else {
                ret += "str";
            }
            if (!aRegNull(rd2)) {
                ret += "d";
            }
            if (cond != Cond::None) {
                ret += cond2StringView(cond);
            }
        }
        ret += ' ';
        if (!aRegNull(rd2)) {
            if (isFloat) {
                ret += to_alias(std::get<Reg>(rd)).abiName;
            } else {
                ret += aReg2String(rd);
                ret += ", ";
                ret += aReg2String(rd2);
            }
        } else {
            ret += aReg2String(rd);
        }

        ret += ", [";
        ret += aReg2String(rn);
        if (auto offset = std::get_if<int>(&operand2)) {
            if (*offset != 0) {
                if ((isFloat && std::abs(*offset) > 1020) || std::abs(*offset) > 4095) {
                    warn("load store instruction offset immediate exceed: ", *offset);
                }
                ret += ", #";
                ret += std::to_string(*offset);
            }

        } else {
            auto pRegLS = std::get_if<RegLS>(&operand2);
            ret += ", ";
            if (pRegLS->minus) {
                ret += '-';
            }
            ret += aReg2String(pRegLS->rm);
            if (pRegLS->shiftType != ShiftType::None && pRegLS->shift != 0) {
                ret += ", ";
                ret += shiftType2CStr(pRegLS->shiftType);
                ret += " #";
                ret += std::to_string(pRegLS->shift);
            }
        }
        ret += ']';
        return ret;
    }

    std::vector<AReg> AInstrMovAddr32::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrMovAddr32::getUse() {
        return {};
    }

    void AInstrMovAddr32::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
    }

    std::string AInstrMovAddr32::toAssemblyString() {
        auto ret = "mov32 "s;
        ret += cond2StringView(cond);
        ret += ' ';
        return ret + aReg2String(rd) + ", " + label->label;
    }

    AInstrMovAddr32::AInstrMovAddr32(const AReg &rd, AData *label) : rd(rd), label(label) {}

    bool RegLS::operator==(const RegLS &rhs) const {
        return rm == rhs.rm &&
               shiftType == rhs.shiftType &&
               shift == rhs.shift &&
               minus == rhs.minus;
    }

    bool RegLS::operator!=(const RegLS &rhs) const {
        return !(rhs == *this);
    }
} // armv7