//
// Created by 11067 on 2023/6/26.
//

#ifndef COMPILER_AINSTRMUL_H
#define COMPILER_AINSTRMUL_H

#include "../aarch.h"
#include "../AInstr.h"
#include "common.h"
#include "AInstrALU.h"

namespace armv7 {
    /**
     * @brief 乘法指令
     * Rd := (Rm * Rs)[31:0]
     */
    struct AInstrMul : AInstr {
        AReg rd;
        AReg rm;
        AReg rs;
        bool setFlag;

        AInstrMul(const AReg &rd, const AReg &rm, const AReg &rs, bool setFlag, Cond cond);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    /**
     * @brief 乘加指令 update: 不仅包含乘加
     * Rd := (Rn + (Rm * Rs))[31:0]
     */
    struct AInstrMla : AInstr {
        AReg rd;
        AReg rm;
        AReg rs;
        AReg rn;
        bool setFlag;
        bool addFlag;  // 加法/减法
        bool signFlag; // 注意这不是是否有符号，这是是否取高32位

        AInstrMla(const AReg &rd, const AReg &rm, const AReg &rs, const AReg &rn, bool setFlag, bool addFlag,
                  bool signFlag, Cond cond);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    AInstrMla *makeSMMla(const AReg &rd, const AReg &r1, const AReg &r2, const AReg &acc);
    AInstrMla *makeMls(const AReg &rd, const AReg &r1, const AReg &r2, const AReg &acc);
    AInstrMla *makeMla(const AReg &rd, const AReg &r1, const AReg &r2, const AReg &acc);
    /**
     * @brief 除法指令
     * 文档说ARMv7-A除法是软实现的，但是貌似能直接这样用
     * Rd := Rn / Rm
     */
    struct AInstrSDiv : AInstr {
        AReg rd;
        AReg rm;
        AReg rn;
        bool setFlag;

        AInstrSDiv(const AReg &rd, const AReg &rn, const AReg &rm, bool setFlag, Cond cond);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    AInstrSDiv *makeSDiv(AReg rd, AReg rm, AReg rn);

    AInstrMul *makeMul(AReg rd, AReg rm, AReg rn);

    std::tuple<AInstrSDiv *, AInstrMul *, AInstrALU *> makeRem(AReg rd, AReg rm, AReg rn);

    /**
     * @brief 高位乘法，用于除常数优化
     * Rd := (Rm * Rs)[63:32]
     */
    struct AInstrSMMUL : AInstr {
        AReg rd;
        AReg rm;
        AReg rs;

        AInstrSMMUL(const AReg &rd, const AReg &rm, const AReg &rs, Cond cond);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

}
#endif //COMPILER_AINSTRMUL_H
