//
// Created by 11067 on 2023/7/5.
//

#include <sstream>
#include <utility>
#include "AInstrPseudo.h"
#include "common.h"


using namespace std::literals;
namespace armv7 {
    std::vector<AReg> AInstrParallelCopy::getDef() {
        return {};
    }

    std::vector<AReg> AInstrParallelCopy::getUse() {
        return {};
    }

    void AInstrParallelCopy::allocate(const VReg &vReg, const AReg &pReg) {

    }

    std::string AInstrParallelCopy::toAssemblyString() {
        std::string ret{"PCopy "};
        for (auto &[reg, reg1]: copy) {
            ret += "[" + aReg2String(reg) + "<-" + aReg2String(reg1) + "], ";
        }
        return ret;
    }

    std::vector<AReg> AInstrPhi::getDef() {
        return {};
    }

    std::vector<AReg> AInstrPhi::getUse() {
        return {};
    }

    void AInstrPhi::allocate(const VReg &vReg, const AReg &pReg) {

    }

    std::string AInstrPhi::toAssemblyString() {
        std::string ret{"aPhi "};
        for (auto &[reg, bb]: pairs) {
            ret += "[" + aReg2String(reg) + ',' + bb->label + "] ";
        }
        return ret;
    }

    std::vector<AReg> AInstrSpAdd::getDef() {
        return {};
    }

    std::vector<AReg> AInstrSpAdd::getUse() {
        return {};
    }

    void AInstrSpAdd::allocate(const VReg &vReg, const AReg &pReg) {

    }

    std::string AInstrSpAdd::toAssemblyString() {
        std::string ret{};
        std::string op = offset > 0 ? "add" : "sub";
        unsigned int imm = abs(offset);
        unsigned int mask = 0xff;
        for (int i = 0; i < 4; ++i) {
            if ((imm & mask) != 0) {
                ret += op + " sp, sp, #" + std::to_string(imm & mask);
                ret += "\n\t";
            }
            mask <<= 8;
        }
        return ret;
    }

    AInstrSpAdd::AInstrSpAdd(int offset) : offset(offset) {}

    std::vector<AReg> AInstrAddRImmPse::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrAddRImmPse::getUse() {
        return {rm};
    }

    void AInstrAddRImmPse::allocate(const VReg &vReg, const AReg &pReg) {
        //warn("pseudo instruction should not be allocated here.");
        if (matchRegAlloc(rm, vReg)) {
            rm = pReg;
        }
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
    }


    std::string AInstrAddRImmPse::toAssemblyString() {
        return std::string("addrimm ") +
               aReg2String(rd) + ", " +
               aReg2String(rm) + ", " +
               std::to_string(offset);
    }

    AInstrAddRImmPse::AInstrAddRImmPse(const AReg &rd, const AReg &rm, int offset) : rd(rd), rm(rm), offset(offset) {
    }

    AInstrAddRImmPse::~AInstrAddRImmPse() = default;

    std::vector<AReg> AInstrLiveInfoHint::getDef() {
        return std::vector{def};
    }

    std::vector<AReg> AInstrLiveInfoHint::getUse() {
        return std::vector{use};
    }

    void AInstrLiveInfoHint::allocate(const VReg &vReg, const AReg &pReg) {
        //nothing to do
    }

    std::string AInstrLiveInfoHint::toAssemblyString() {
        std::stringstream out{};
        out << "@def: ";
        for (auto &reg: def) {
            out << reg << ", ";
        }
        out << "use: ";
        for (auto &reg: use) {
            out << reg << ", ";
        }
        return out.str();
    }

    AInstrLiveInfoHint::AInstrLiveInfoHint(const std::vector<AReg> &use, const std::vector<AReg> &def) : use(use),
                                                                                                         def(def) {}

    std::vector<AReg> AInstrComment::getDef() {
        return {};
    }

    std::vector<AReg> AInstrComment::getUse() {
        return {};
    }

    void AInstrComment::allocate(const VReg &vReg, const AReg &pReg) {

    }

    std::string AInstrComment::toAssemblyString() {
        return "@"s + comment;
    }

    AInstrComment::AInstrComment(std::string comment) : comment(std::move(comment)) {}
} // armv7