//
// Created by 11067 on 2023/7/5.
//

#ifndef COMPILER_AINSTRPSEUDO_H
#define COMPILER_AINSTRPSEUDO_H

#include "../aarch.h"
#include "../AInstr.h"
#include "../ABlock.h"
#include <utility>
#include <vector>


namespace armv7 {

    /**
     * @brief 并行复制指令
     *
     * 用于在拆解Phi时临时使用
     */
    struct AInstrParallelCopy : AInstr {
        std::vector<std::pair<AReg, AReg>> copy;

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    struct AInstrPhi : AInstr {
        std::vector<std::pair<AReg, ABlock *>> pairs;
        AReg dest;

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    /**
     * @brief add sp,sp,offset 用于改变栈指针
     */
    struct AInstrSpAdd : AInstr {
        int offset;

        explicit AInstrSpAdd(int offset);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        /**
         * @brief 因为在栈指针准备好之前，局部变量（VReg）无法使用，该指令会被输出
         * 为1-4条Add。
         * @return 指令
         */
        std::string toAssemblyString() override;
    };

    struct AInstrAddRImmPse : AInstr {
        AReg rd;
        AReg rm;
        int offset;

        AInstrAddRImmPse(const AReg &rd, const AReg &rm, int offset);


        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

        ~AInstrAddRImmPse() override;
    };

    struct AInstrLiveInfoHint : AInstr {
        std::vector<AReg> use;
        std::vector<AReg> def;

        AInstrLiveInfoHint(const std::vector<AReg> &use, const std::vector<AReg> &def);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

    };

    struct AInstrComment : AInstr {
        std::string comment;

        explicit AInstrComment(std::string comment);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };


} // armv7

#endif //COMPILER_AINSTRPSEUDO_H
