//
// Created by 11067 on 2023/6/28.
//

#include "AInstrVFP.h"

namespace armv7 {
    std::vector<AReg> AInstrVMRS::getDef() {
        return {};
    }

    std::vector<AReg> AInstrVMRS::getUse() {
        return {};
    }

    void AInstrVMRS::allocate(const VReg &vReg, const AReg &pReg) {
        //nothing to do
    }

    std::string AInstrVMRS::toAssemblyString() {
        std::string ret{"vmrs"};
        ret += cond2StringView(cond);
        return ret + " APSR_nzcv, fpscr";
    }

    AInstrVMov *makeVMovS2S(const AReg &rd, const AReg &rs) {
        return new AInstrVMov{Cond::None, rd, rs, AInstrVMov::Type::S2S};
    }

    AInstrVMov::AInstrVMov(Cond cond, const AReg &rd, const AReg &rm, const Type type) : AInstr{cond}, rd(rd), type(type),
                                                                                         rm(rm) {}

    std::vector<AReg> AInstrVMov::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrVMov::getUse() {
        return {rm};
    }

    void AInstrVMov::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
        if (matchRegAlloc(rm, vReg)) {
            rm = pReg;
        }
    }

    std::string AInstrVMov::toAssemblyString() {
        std::string ret{"vmov"};
        if (cond != Cond::None) {
            ret += cond2StringView(cond);
        }
        switch (type) {
            case Type::S2R:
                break;
            case Type::R2S:
                break;
            case Type::S2S:
                ret += ".f32";
                break;
        }
        ret += ' ';
        ret += aReg2String(rd);
        ret += ", ";
        ret += aReg2String(rm);
        return ret;
    }

    bool AInstrVMov::canMerge() const {
        return cond == Cond::None && type == Type::S2S;
    }

    AInstrFloatBinary::AInstrFloatBinary(AInstrFloatBinary::Type type, Cond cond, const AReg &rd, const AReg &rm,
                                         const AReg &rn) : AInstr{cond}, type(type), rd(rd), rm(rm), rn(rn) {}

    std::vector<AReg> AInstrFloatBinary::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrFloatBinary::getUse() {
        return {rm, rn};
    }

    void AInstrFloatBinary::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
        if (matchRegAlloc(rn, vReg)) {
            rn = pReg;
        }
        if (matchRegAlloc(rm, vReg)) {
            rm = pReg;
        }
    }

    std::string AInstrFloatBinary::toAssemblyString() {
        std::string ret{};
        switch (type) {
            case Type::Vadd:
                ret += "vadd";
                break;
            case Type::Vsub:
                ret += "vsub";
                break;
            case Type::Vdiv:
                ret += "vdiv";
                break;
            case Type::Vmul:
                ret += "vmul";
                break;
        }
        if (cond != Cond::None) {
            ret += cond2StringView(cond);
        }
        ret += ".f32 ";
        ret += aReg2String(rd);
        ret += ", ";
        ret += aReg2String(rm);
        ret += ", ";
        ret += aReg2String(rn);
        return ret;
    }

    AInstrFloatUnary::AInstrFloatUnary(AInstrFloatUnary::Type type, Cond cond, const AReg &rd, const AReg &rm) : AInstr{
            cond}, type(
            type), rd(rd), rm(rm) {}

    std::vector<AReg> AInstrFloatUnary::getDef() {
        return {rd};
    }

    std::vector<AReg> AInstrFloatUnary::getUse() {
        return {rm};
    }

    void AInstrFloatUnary::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rd, vReg)) {
            rd = pReg;
        }
        if (matchRegAlloc(rm, vReg)) {
            rm = pReg;
        }
    }

    std::string AInstrFloatUnary::toAssemblyString() {
        std::string ret;
        switch (type) {
            case Type::Vabs:
                ret += "vabs";
                break;
            case Type::Vneg:
                ret += "vneg";
                break;
            case Type::Vsqrt:
                ret += "vsqrt";
                break;
            case Type::CvtI2F:
                ret += "vcvt";
                break;
            case Type::CvtF2I:
                ret += "vcvt";
                break;
        }
        if (cond != Cond::None) {
            ret += cond2StringView(cond);
        }
        if (type == Type::CvtF2I) {
            ret += ".s32.f32 ";
        } else if (type == Type::CvtI2F) {
            ret += ".f32.s32 ";
        } else {
            ret += ".f32 ";
        }
        ret += aReg2String(rd);
        ret += ", ";
        ret += aReg2String(rm);
        return ret;
    }


    AInstrFloatBinary *makeFloatBinary(AInstrFloatBinary::Type type, AReg rd, AReg rm, AReg rn) {
        return new AInstrFloatBinary{
                type,
                Cond::None,
                rd,
                rm,
                rn
        };
    }

    AInstrFloatUnary *makeFloatUnary(AInstrFloatUnary::Type type, AReg rd, AReg rm) {
        return new AInstrFloatUnary{
                type,
                Cond::None,
                rd,
                rm
        };
    }

    AInstrFloatCmp *makeAInstrFloatCmp(AReg &rm, AReg &rn) {
        return new AInstrFloatCmp{
                Cond::None,
                rm,
                rn
        };
    }

    std::vector<AReg> AInstrFloatCmp::getDef() {
        return {};
    }

    std::vector<AReg> AInstrFloatCmp::getUse() {
        std::vector<AReg> ret{rm};
        if (auto pReg = std::get_if<AReg>(&rn)) {
            ret.push_back(*pReg);
        }
        return ret;
    }

    void AInstrFloatCmp::allocate(const VReg &vReg, const AReg &pReg) {
        if (matchRegAlloc(rm, vReg)) {
            rm = pReg;
        }
        if (auto ptrReg = std::get_if<AReg>(&rn)) {
            if (matchRegAlloc(*ptrReg, vReg)) {
                rn = pReg;
            }
        }
    }

    std::string AInstrFloatCmp::toAssemblyString() {
        std::string ret{"vcmp"};
        if (cond != Cond::None) {
            ret += cond2StringView(cond);
        }
        ret += ".f32 ";
        ret += aReg2String(rm);
        ret += ", ";
        if (auto ptrReg = std::get_if<AReg>(&rn)) {
            ret += aReg2String(*ptrReg);
        } else {
            auto imm = std::get<int>(rn);
            if (imm != 0) {
                error("only #0 can be used as rn for vcmp");
            }
            ret += "#0";
        }
        return ret;
    }

    AInstrFloatCmp::AInstrFloatCmp(Cond cond, const AReg &rm, const std::variant<AReg, int> &rn) : AInstr{cond}, rm(rm),
                                                                                                   rn(rn) {}
} // armv7