//
// Created by 11067 on 2023/6/28.
//
// 浮点相关指令，使用VFP
// 可能可以使用NEON优化，但是有NEON不完全符合IEEE754的说法(subnormal float被当作0)。

#ifndef COMPILER_AINSTRVFP_H
#define COMPILER_AINSTRVFP_H

#include "../aarch.h"
#include "../AInstr.h"
#include "common.h"

namespace armv7 {
    struct AInstrFloatBinary: AInstr{
        enum class Type{
            Vadd,
            Vsub,
            Vdiv,
            Vmul,
        };
        Type type;
        AReg rd;
        AReg rm;
        AReg rn;

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

        AInstrFloatBinary(Type type, Cond cond, const AReg &rd, const AReg &rm, const AReg &rn);
    };
    AInstrFloatBinary *makeFloatBinary(AInstrFloatBinary::Type type, AReg rd, AReg rm, AReg rn);
    struct AInstrFloatUnary: AInstr {
        enum class Type{
            Vabs,
            Vneg,
            Vsqrt,
            CvtI2F,
            CvtF2I,
        };
        Type type;
        AReg rd;
        AReg rm;

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;

        AInstrFloatUnary(Type type, Cond cond, const AReg &rd, const AReg &rm);
    };
    AInstrFloatUnary *makeFloatUnary(AInstrFloatUnary::Type type, AReg rd, AReg rm);
    struct AInstrFloatCmp: AInstr {
        AReg rm;
        std::variant<AReg, int> rn; //立即数只能是0
        AInstrFloatCmp(Cond cond, const AReg &rm, const std::variant<AReg, int> &rn);

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };
    AInstrFloatCmp *makeAInstrFloatCmp(AReg &rm, AReg &rn);
    /**
     * @brief 将浮点比较结果传送到状态寄存器
     */
    struct  AInstrVMRS: AInstr {
        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
    };

    struct AInstrVMov: AInstr {
        enum class Type{
            S2R, //VFP to GPR
            R2S, //GPR to VFP
            S2S, //VFP to VFP
        };

        std::vector<AReg> getDef() override;

        std::vector<AReg> getUse() override;

        void allocate(const VReg &vReg, const AReg &pReg) override;

        std::string toAssemblyString() override;
        bool canMerge() const;

        AReg rd;
        Type type;

        AInstrVMov(Cond cond, const AReg &rd, const AReg &rm, Type type);

        AReg rm;
    };
    AInstrVMov *makeVMovS2S(const AReg &rd, const AReg &rs);
} // armv7

#endif //COMPILER_AINSTRVFP_H
