//
// Created by 11067 on 2023/6/26.
//

#ifndef COMPILER_COMMON_H
#define COMPILER_COMMON_H

#include <variant>
#include <cstring>
#include "../aarch.h"

namespace armv7 {
    enum class Cond {
        None,
        Eq,
        Ne,
        Ge,
        Lt,
        Gt,
        Le,
    };
    inline Cond revCond(Cond cond) {
        switch (cond) {
            case Cond::None:
                return Cond::None;
            case Cond::Eq:
                return Cond::Ne;
            case Cond::Ne:
                return Cond::Eq;
            case Cond::Ge:
                return Cond::Lt;
            case Cond::Lt:
                return Cond::Ge;
            case Cond::Gt:
                return Cond::Le;
            case Cond::Le:
                return Cond::Gt;
        }
    }
    inline std::string_view cond2StringView(Cond cond) {
        switch (cond) {
            case Cond::None:
                return "";
            case Cond::Eq:
                return "eq";
            case Cond::Ne:
                return "ne";
            case Cond::Ge:
                return "ge";
            case Cond::Lt:
                return "lt";
            case Cond::Gt:
                return "gt";
            case Cond::Le:
                return "le";
        }
    }
    enum class ShiftType {
        None,
        LSL,
        LSR,
        ASR,
        ROR,
    };
    inline const char* shiftType2CStr(ShiftType shiftType) {
        switch (shiftType) {
            case ShiftType::None:
                return "";
            case ShiftType::LSL:
                return "LSL";
            case ShiftType::LSR:
                return "LSR";
            case ShiftType::ASR:
                return "ASR";
            case ShiftType::ROR:
                return "ROR";
        }
    }


    /**
     * @brief Armv7的第二个操作数作为寄存器的情况，通常包含移位部分
     */
    struct Operand2Reg {
        AReg base;
        ShiftType shiftType;
        std::variant<AReg, int> shift;
    };
    typedef std::variant<int, Operand2Reg> Operand2;
    std::string operand2ToString(Operand2 operand2);
    /**
     * @brief 是否能以imm8m格式表达
     * @param x 立即数
     * @return 能否表达
     */
    inline bool isImm8m(int x) {
        unsigned int x2;
        std::memcpy(&x2, &x, sizeof x);
        unsigned int mask = 0xff000000;
        for (int i = 0; i < 16; ++i) {
            if ((x2 & (~mask)) == 0) {
                return true;
            }
            mask >>= 2;
        }
        return false;
    }
} // armv7



#endif //COMPILER_COMMON_H
