//
// Created by 11067 on 2023/7/17.
//

#include "AfterCodegenPass.h"
#include "../instructions/AInstrMov.h"
#include "../../../utils/bitUtils.h"

namespace armv7 {
    void AfterCodegenPass::run(ACompileUnit &compileUnit) {
        //insertLTORG(compileUnit);
        assemblePseudoInstruction(compileUnit);
    }

    void AfterCodegenPass::insertLTORG(ACompileUnit &compileUnit) {
        for (auto func: compileUnit) {

            for (auto bb: *func) {
                int BBLength = bb->getSize();
                int cnt = 0;
                for (auto instr: *bb) {
                    if (auto pseLdr = dynamic_cast<AInstrPseLdr *>(instr)) {
//                        auto cnst = pseLdr->getConst();
//                        auto poolOffset = bb->myLiteralPool.size();
//                        bb->myLiteralPool.push_back(cnst);
                        bb->trailingLTORG++;
                    }
                    cnt++;
                }
            }
        }
    }

    void AfterCodegenPass::assemblePseudoInstruction(ACompileUnit &compileUnit) {
        for (auto func: compileUnit) {
            for (auto bb: *func) {
                for (auto instr: bb->toVector()) {
                    if (auto mov32 = dynamic_cast<AInstrMov32 *>(instr)) {
                        int imm32 = mov32->imm32;
                        auto *movw = new AInstrMovLi{mov32->rd, getLo16(imm32)};
                        movw->setCond(mov32->cond);
                        mov32->insertBeforeMe(movw);
                        if (!isImm16(imm32)) {
                            auto *movt = new AInstrMovt{mov32->rd, getHi16(imm32)};
                            movt->setCond(mov32->cond);
                            mov32->insertBeforeMe(movt);
                        }
                        mov32->removeMe();
                        delete mov32;
                        continue;
                    }
                    if (auto mov32 = dynamic_cast<AInstrMovAddr32 *>(instr)) {
                        mov32->insertBeforeMe(new AInstrLaLow{mov32->rd, mov32->label});
                        mov32->insertBeforeMe(new AInstrLaHigh{mov32->rd, mov32->label});
                        mov32->removeMe();
                        delete mov32;
                        continue;
                    }
                }
            }
        }
    }
} // armv7