//
// Created by 11067 on 2023/7/7.
//

#include "BakaAllocator.h"
#include "../instructions/AInstrMov.h"
#include "../instructions/AInstrALU.h"

#include <algorithm>

namespace armv7 {
    void BakaAllocator::run(ACompileUnit &compileUnit) {
        for (auto func: compileUnit) {
            if (!func->isLib) {
                runOnFunction(func);
            }
        }
    }

    void BakaAllocator::runOnFunction(AFunction *function) {
        auto localOffset = (function->vRegAllocated + function->stackArrayAllocated) * 4;
        function->reserveLocal->offset = -localOffset;
        function->cleanLocal->offset = localOffset;
        if (localOffset > 1020) {
            warn("too many local, offset:", localOffset);
        }
        // +1 是lr
        auto localAddParamOffset = localOffset + (saved_freg_cnt + saved_freg_cnt + 1) * 4;
        Reg d = get_reg_by_id(0);
        Reg r1 = get_reg_by_id(1);
        Reg r2 = get_reg_by_id(2);
        Reg r3 = get_reg_by_id(3);
        Reg df = get_freg_by_id(0);
        Reg f1 = get_freg_by_id(1);
        Reg f2 = get_freg_by_id(2);
        //替换伪指令，ip被用作固定临时寄存器
        for (auto instr: function->loadLocalArrayInstructions) {
            instr->offset += function->vRegAllocated * 4;
            if (isImm8m(instr->offset)) {
                instr->insertBeforeMe(new AInstrALU{
                        AInstrALU::Type::Add,
                        Cond::None,
                        false,
                        instr->rd,
                        instr->rm, // should be sp
                        instr->offset
                });

            } else {
                instr->insertBeforeMe(new AInstrMov32{get_ip(), instr->offset});
                instr->insertBeforeMe(makeAInstrALURR(AInstrALU::Type::Add, instr->rd, instr->rm, get_ip()));
            }
            instr->removeMe();
            delete instr;
        }
        int ipOffset = 0;
        for (auto instr: function->loadParamInstructions) {
            auto imm = std::get<int>(instr->operand2);
            if ((instr->isFloat && imm - ipOffset > 1020) || imm - ipOffset > 4095) {
                ipOffset += 1020;
                instr->insertBeforeMe(makeAInstrALURImm(
                        AInstrALU::Type::Add,
                        AReg{get_ip()},
                        AReg{get_ip()},
                        1020));
            }
            instr->operand2 = imm - ipOffset;
        }
        for (auto bb: *function) {
            for (auto instr: bb->toVector()) {
                std::unordered_map<VReg, Reg> allocMap{};
                std::vector<AReg> uses{instr->getUse()};
                std::vector<AReg> defs{instr->getDef()};
                int cnt = 4;
                int fcnt = 16; // 0固定放返回值,不要从参数寄存器左手倒右手
                for (auto &reg: defs) {
                    if (auto vReg = std::get_if<VReg>(&reg)) {
                        if (vReg->type == VReg::Type::Int) {
                            allocMap.insert({*vReg, get_reg_by_id(cnt++)});
                        } else {
                            allocMap.insert({*vReg, get_freg_by_id(fcnt++)});
                        }
                    }
                }
                for (auto &reg: uses) {
                    if (auto vReg = std::get_if<VReg>(&reg)) {
                        if (vReg->type == VReg::Type::Int) {
                            allocMap.insert({*vReg, get_reg_by_id(cnt++)});
                        } else {
                            allocMap.insert({*vReg, get_freg_by_id(fcnt++)});
                        }
                    }
                }
                for (auto &[vReg, pReg]: allocMap) {
                    int offset = vReg.id * 4;
                    if (std::find(uses.begin(), uses.end(), AReg{vReg}) != uses.end()) {

                        if (vReg.type == VReg::Type::Int) {
                            if (offset > 4095) {
                                instr->insertBeforeMe(new AInstrMov32{get_ip(), offset});
                                instr->insertBeforeMe(makeLdRR(pReg, get_sp(), get_ip()));
                            } else {
                                instr->insertBeforeMe(makeLdRImm(pReg, get_sp(), vReg.id * 4));
                            }

                        } else {
                            if (offset > 1020) {
                                instr->insertBeforeMe(new AInstrMov32{get_ip(), offset});
                                instr->insertBeforeMe(makeLdRRF(pReg, get_sp(), get_ip()));
                            } else {
                                instr->insertBeforeMe(makeLdRImmF(pReg, get_sp(), vReg.id * 4));
                            }

                        }
                    }
                    if (std::find(defs.begin(), defs.end(), AReg{vReg}) != defs.end()) {
                        if (vReg.type == VReg::Type::Int) {
                            if (offset > 4095) {
                                auto *pInstrAfterThis = new AInstrMov32{get_ip(), offset};
                                instr->insertAfterMe(pInstrAfterThis);
                                pInstrAfterThis->insertAfterMe(makeStrRR(pReg, get_sp(), get_ip()));
                            } else {
                                instr->insertAfterMe(makeStrRImm(pReg, get_sp(), vReg.id * 4));
                            }

                        } else {
                            if (offset > 1020) {
                                auto *pInstrAfterThis = new AInstrMov32{get_ip(), offset};
                                instr->insertAfterMe(pInstrAfterThis);
                                pInstrAfterThis->insertAfterMe(makeStrRRF(pReg, get_sp(), get_ip()));
                            } else {
                                instr->insertAfterMe(makeStrRImmF(pReg, get_sp(), vReg.id * 4));
                            }
                        }
                    }
                }
                for (auto &[vReg, pReg]: allocMap) {
                    instr->allocate(vReg, pReg);
                }
            }
        }
    }
} // armv7