//
// Created by 11067 on 2023/7/10.
//

#ifndef COMPILER_COLORALLOCATOR_H
#define COMPILER_COLORALLOCATOR_H

#include "APass.h"
#include <unordered_map>
#include <vector>
#include <stack>
#include <deque>
#include "../../../utils/config.h"

using std::unordered_map;
using std::vector;
using std::unordered_set;
using std::stack;
using std::deque;

using Move = std::pair<armv7::AReg, armv7::AReg>;
template<>
struct std::hash<Move>
{
    std::size_t operator()(Move const& move) const noexcept
    {
        std::size_t h1 = std::hash<armv7::AReg>{}(move.first);
        std::size_t h2 = std::hash<armv7::AReg>{}(move.second);
        return h1 ^ (h2 << 1); // or use boost::hash_combine
    }
};

namespace armv7 {
    class ColorAllocatorExecutor {

        AFunction *function;
        vector<double> spillPriority{};
        vector <AReg> colors{};
        unordered_set <AReg> preColored{};
        unordered_set <AReg> initial{};
        vector <AReg> simplifyWorkList{};
        unordered_set <AReg> freezeWorkList{};
        unordered_set <AReg> spillWorkList{};
        unordered_set <AReg> spilledNodes{};
        unordered_set <AReg> coalescedNodes{};
        unordered_set <AReg> coloredNodes{};
        vector <AReg> selectStack{};
        unordered_set <AReg> selectSet{};
        unordered_set <Move> coalescesMoves{};
        unordered_set <Move> constrainedMoves{};
        unordered_set <Move> frozenMoves{};
        unordered_set <Move> workListMoves{};
        unordered_set <Move> activeMoves{};
        int spilledVarCnt{0};
        unordered_set <AReg> spillCreated{};
        unordered_set <std::pair<AReg, AReg >> adjSet{};
        vector <unordered_set<AReg >> adjList{};
        vector<int> degree{};
        vector <unordered_set<Move>> moveList{};
        vector <AReg> alias{};

        unordered_set<int> spiltedReg{}; //生命周期已被分割的Reg 的id

        int getAllRegCnt();

        void calculateSpillPriority();

        void clear();

        void build();

        void makeWorkList();

        void simplify();
        void coalesce();
        bool ok(const AReg& t, const AReg& r);
        bool conservative(const unordered_set<AReg>& nodes);
        void combine(const AReg& u, const AReg& v);
        void freeze();
        void selectSpill();
        void assignColors();
        void rewriteProgram();
        void finish();

        bool moveRelated(const AReg &reg);
        unordered_set<Move> nodeMoves(const AReg &reg);
        void decrementDegree(AReg &reg);
        unordered_set<AReg> adjacent(const AReg &reg);
        void enableMoves(unordered_set<AReg> &nodes);
        AReg getAlias(const AReg &reg);
        void addWorkList(const AReg &u);
        void freezeMoves(const AReg &u);
        void addEdge(const AReg &u, const AReg &v);
        VReg allocSplitVReg(VReg::Type type1);

        void fusionEnableMove(const AReg &reg);
    public:
        explicit ColorAllocatorExecutor(AFunction *function);

        void run();
    };

    /**
     * @brief 图着色寄存器分配
     *
     * 书上没有关于浮点数如何分配的内容，也许通过建冲突边（比如在浮点寄存器和整数虚拟寄存器之间）？
     */
    class ColorAllocator : public APass {
    public:
        void run(ACompileUnit &compileUnit) override;
    };


} // armv7

#endif //COMPILER_COLORALLOCATOR_H
