//
// Created by 11067 on 2023/8/7.
//

#include "JTable.h"
#include "../instructions/AInstrB.h"
#include "../instructions/AInstrALU.h"
#include <algorithm>
#include <unordered_set>
namespace armv7 {
    static int jTableCnt = 1;
    void JTable::run(ACompileUnit &compileUnit) {
        aCompileUnit = &compileUnit;
        for (auto func: compileUnit) {
            if (!func->isLib) {
                runOnFunction(func);
            }
        }
    }

    void JTable::runOnFunction(AFunction *function) {
        std::unordered_set<ABlock*> toRemove{};
        for (auto bb: function->toVector()) {
            if (toRemove.count(bb)) {
                continue;
            }
            auto lastInstr = bb->getLast();
            auto last2Instr = lastInstr->getPrev();
            if (last2Instr == nullptr) {
                continue;
            }
            auto last3Instr = last2Instr->getPrev();
            auto beq = dynamic_cast<AInstrB *>(last2Instr);
            if (beq == nullptr) {
                continue;
            }
            if (beq->isCall() || beq->cond != Cond::Eq) {
                continue;
            }
            auto cmp = dynamic_cast<AInstrCMP *>(last3Instr);
            if (cmp == nullptr) {
                continue;
            }
            if (!std::holds_alternative<int>(cmp->operand2) || cmp->neg) {
                continue;
            }
            std::vector<std::tuple<int, ABlock *, ABlock *>> jt{};
            int imm = std::get<int>(cmp->operand2);
            auto target = std::get<ABlock *>(beq->target);
            jt.emplace_back(imm, bb, target);
            AReg switchReg = cmp->rm;
            ABlock *bbSeq = bb->failBranch;
            int minImm = imm;
            int maxImm = imm;
            while (true) {
                if (bbSeq->getInstrCntNoComment() != 3) {
                    break;
                }
                auto lastInstr1 = bbSeq->getLast();
                auto last2Instr1 = lastInstr1->getPrev();
                if (last2Instr1 == nullptr) {
                    break;
                }
                auto last3Instr1 = last2Instr1->getPrev();
                auto beq1 = dynamic_cast<AInstrB *>(last2Instr1);
                if (beq1 == nullptr) {
                    break;
                }
                if (beq1->isCall() || beq1->cond != Cond::Eq) {
                    break;
                }
                auto cmp1 = dynamic_cast<AInstrCMP *>(last3Instr1);
                if (cmp1 == nullptr) {
                    break;
                }
                if (!std::holds_alternative<int>(cmp1->operand2) || cmp1->neg) {
                    break;
                }
                if (cmp1->rm != switchReg) {
                    break;
                }
                int imm1 = std::get<int>(cmp1->operand2);
                maxImm = std::max(maxImm, imm1);
                minImm = std::min(minImm, imm1);
                auto target1 = std::get<ABlock*>(beq1->target);
                jt.emplace_back(imm1, bbSeq, target1);
                bbSeq = bbSeq->failBranch;
            }
            if (jt.size() <= 3 || maxImm - minImm > 256) {
                continue;
            }
            ABlock *quitBranch = std::get<1>(jt.back())->failBranch;
            std::sort(jt.begin(), jt.end(),[](auto tuple1, auto tuple2) {
                return std::get<0>(tuple1) < std::get<0>(tuple2);
            });
            auto ajTable = new AJTable{};
            //aCompileUnit->jtables.emplace_back(ajTable);
            ajTable->switchReg = switchReg;
            ajTable->from = bb;
            ajTable->exitBranch = quitBranch;
            ajTable->label = ".JTABLE";
            ajTable->label += std::to_string(jTableCnt++);
            for (auto [i,b1,b2]: jt) {
                ajTable->to.emplace_back(i, b2);
                if (b1 != bb) {
                    toRemove.insert(b1);
                }
            }
            lastInstr->removeMe();
            last2Instr->removeMe();
            last3Instr->removeMe();
            delete lastInstr;
            delete last2Instr;
            delete last3Instr;
            bb->ajTable = ajTable;
        }
        for (auto bb: toRemove) {
            bb->removeMe(); //leak memory
        }
    }
} // armv7