//
// Created by 11067 on 2023/7/8.
//

#include "LiveInfo.h"

#include "../../../utils/config.h"
namespace armv7 {
    using atri::Bitset;


    void LiveInfo::run(ACompileUnit &compileUnit) {

    }

    void LiveInfo::runOnFunction(AFunction *function) {
        def.clear();
        use.clear();
        liveIn.clear();
        liveOut.clear();
        int2reg.clear();
        regsCnt = function->vRegAllocated + 48; //48个物理寄存器

        int2reg.resize(regsCnt, VReg{}); //VReg 占位
        for (int i = 0; i < 48; ++i) {
            int2reg[i] = get_reg_by_id(i);
        }
//        info(int2reg.size());
        for (auto bb: *function) {
            for (auto instr: *bb) {
//                info("instr here:", instr->toAssemblyString());
                for (auto &reg: instr->getDef()) {
                    if (auto vReg = std::get_if<VReg>(&reg)) {
//                        info(fruid(reg));
                        int2reg[fruid(reg)] = *vReg;
                    }
                }
                for (auto &reg: instr->getUse()) {
                    if (auto vReg = std::get_if<VReg>(&reg)) {
                        int2reg[fruid(reg)] = *vReg;
                    }
                }
            }
        }
        info("liveinfo build complete");
        for (auto bb: *function) {
            bb->liveOut.clear();
            bb->liveIn.clear();
            bb->use.clear();
            bb->def.clear();
            def.insert({bb, Bitset(regsCnt)});
            use.insert({bb, Bitset(regsCnt)});
            liveIn.insert({bb, Bitset(regsCnt)});
            liveOut.insert({bb, Bitset(regsCnt)});
            for (auto instr: *bb) {
                for (auto &reg: instr->getUse()) {
                    if (!def.at(bb)[fruid(reg)]) {
                        info(aReg2String(reg));
                        use.at(bb)[fruid(reg)] = true;
                    }
                }
                for (auto &reg: instr->getDef()) {
                    if (!use.at(bb)[fruid(reg)]) {
                        info(aReg2String(reg));
                        def.at(bb)[fruid(reg)] = true;
                    }
                }
            }
        }
        info("liveinfo block def use complete");
        if (function->getSize() > LiveInfoRecurThreshold) {
            info("too many blocks for recursive method, regress to iterating");
            changed = true;
            auto blocks = function->toVector();
            while (changed) {
                changed = false;
                for (int i = blocks.size() - 1; i >= 0 ; --i) {
                    auto nw = blocks[i];
                    calcuABlock(nw);
                }
            }
        } else {
            changed = true;
            while (changed) {
                changed = false;
                visited.clear();
                walkABlock(function->getFirst());
            }
        }
        info("liveinfo compute complete");
        //从bitset映射回一般的表达形式
        for (auto bb: *function) {
            Bitset &bitset = liveIn.at(bb);
            for (int i = 0; i < bitset.getBitLength(); ++i) {
                if (bitset[i]) {
                    bb->liveIn.push_back(int2reg[i]);
                }
            }
            Bitset &bitset1 = liveOut.at(bb);
            for (int i = 0; i < bitset1.getBitLength(); ++i) {
                if (bitset1[i]) {
                    bb->liveOut.push_back(int2reg[i]);
                }
            }
            Bitset &bitset2 = def.at(bb);
            for (int i = 0; i < bitset2.getBitLength(); ++i) {
                if (bitset2[i]) {
                    bb->def.push_back(int2reg[i]);
                }
            }
            Bitset &bitset3 = use.at(bb);
            for (int i = 0; i < bitset3.getBitLength(); ++i) {
                if (bitset3[i]) {
                    bb->use.push_back(int2reg[i]);
                }
            }
            info("liveinfo map back at Block: " , bb->label);
        }
        info("liveinfo map back complete");
    }

    void LiveInfo::walkABlock(ABlock *nw) {
        if (visited.count(nw)) {
            return;
        }
        visited.insert(nw);
        for (auto &item: nw->succ) {
            walkABlock(item);
        }
        calcuABlock(nw);

    }

    void LiveInfo::calcuABlock(ABlock *nw) {
        Bitset &old = liveOut.at(nw);
        Bitset nwLiveOut{static_cast<size_t>(regsCnt)};
        for (const auto &item: nw->succ) {
            nwLiveOut |= liveIn.at(item);
        }
        if (old != nwLiveOut) {
            changed = true;
        }
        liveOut.at(nw) = std::move(nwLiveOut);

        auto  &oldIn = liveIn.at(nw);
        Bitset nwLiveIn{ static_cast<size_t>(regsCnt)};
        Bitset t{ static_cast<size_t>(regsCnt)};

        nwLiveIn |= use.at(nw);
        t |= liveOut.at(nw);
        t.andNot(def.at(nw));
        nwLiveIn |= t;

        if (oldIn != nwLiveIn) {
            changed = true;
        }
        liveIn.at(nw) = std::move(nwLiveIn);
    }
} // armv7