//
// Created by 11067 on 2023/7/30.
//

#include "MergeBlock.h"
#include "../instructions/AInstrB.h"
#include "../../../utils/config.h"
#include <algorithm>

namespace armv7 {
    void MergeBlock::run(ACompileUnit &compileUnit) {
        for (auto func: compileUnit) {
            if (!func->isLib) {
                func->calcPrecSucc();
                mergeSinglePath(func);
                func->calcPrecSucc();
                if(!func->checkPrecSucc()){
                    panic("prec succ error");
                }
                mergeJump(func);
                if(!func->checkPrecSucc()){
                    panic("prec succ error");
                }
//                mergeSinglePath(func);
            }

        }
        info(compileUnit.toAssemblyString());
    }

    void MergeBlock::mergeSinglePath(AFunction *function) {
        auto blocks = function->toVector();
        std::unordered_set<ABlock *> toRemove{};
        for (auto bb: blocks) {
            if (toRemove.count(bb)) {
                continue;
            }
            while (true) {
                auto lastInstr = bb->getLast();
                auto last2Instr = lastInstr->getPrev(); // may null
                if (dynamic_cast<AInstrB *>(last2Instr) != nullptr) {
                    break;
                }
                if (dynamic_cast<AInstrBX *>(lastInstr) != nullptr) { // bx lr(ret)
                    break;
                }
                atriAssert(dynamic_cast<AInstrB *>(lastInstr))
                auto jump = dynamic_cast<AInstrB *>(lastInstr);
                atriAssert(jump->cond == Cond::None); // 无条件跳转
                atriAssert(!(jump->isCall()));
                auto target = std::get<ABlock *>(jump->target);
                // 只有这一个前驱
                if (target->prec.size() == 1 && *target->prec.begin() == bb) {
                    if (bb->getSize() + target->getSize() > 1024) { // 太长了，够不着常量池
                        break;
                    }
                    jump->removeMe();
                    delete jump;
                    for (auto item: target->toVector()) {
                        bb->addLast(item);
                    }
                    bb->succ = target->succ;
                    for (auto nxtBlockSucc: target->succ) {
                        nxtBlockSucc->prec.erase(target);
                        nxtBlockSucc->prec.insert(bb);
                    }
                    bb->succBranch = target->succBranch;
                    toRemove.insert(target);
                } else {
                    break;
                }
            }
        }

        for (auto item: toRemove) {
            item->removeMe();
            delete item;
        }

    }

    void MergeBlock::mergeJump(AFunction *function) {
        for (auto bb: *function) {
            auto lastInstr = bb->getLast();
            auto last2Instr = lastInstr->getPrev(); // may null
            auto flag = false;
            if (auto b = dynamic_cast<AInstrB *>(last2Instr)) {
                if (!b->isCall()) {
                    flag = true;
                    auto origin = std::get<ABlock *>(b->target);
                    auto dest = combineMultiJump(bb, origin);
                    if (origin != dest) {
                        b->target = dest;
                        bb->succ.erase(origin);
                        bb->succ.insert(dest);
                        origin->prec.erase(bb);
                        dest->prec.insert(bb);
                        bb->succBranch = dest;
                    }

                }
            }
            if (auto b = dynamic_cast<AInstrB *>(lastInstr)) {
                auto origin = std::get<ABlock *>(b->target);
                auto dest = combineMultiJump(bb, origin);
                if (origin != dest) {
                    b->target = dest;

                    bb->succ.erase(origin);
                    bb->succ.insert(dest);
                    origin->prec.erase(bb);
                    dest->prec.insert(bb);
                    if (flag) {
                        bb->failBranch = dest;
                    } else {
                        bb->succBranch = dest;
                    }
//                    if (bb->succBranch == origin) {
//                        bb->succBranch = dest;
//                    }
//                    if (bb->failBranch == origin) {
//                        bb->failBranch = dest;
//                    }

                }

            }

        }
        function->calcPrecSucc();
        //clean up
        // 没有前驱的块（不可达）应删除,除了第一个块
        auto workList = function->toVector();
        workList.erase(workList.begin());
        std::unordered_set<ABlock*> toRemove{};
        while (!workList.empty()) {
            auto bb = workList.back();
            workList.pop_back();
            if (!toRemove.count(bb) && bb->prec.empty()) {
                bb->removeMe();
                toRemove.insert(bb);
                for (auto succBlock: bb->succ) {
                    if (!toRemove.count(succBlock)) {
                        succBlock->prec.erase(bb);
                        workList.push_back(succBlock);
                    }
                }
            }
        }
        for (auto bb: toRemove) {
            delete bb;
        }
    }

    ABlock *MergeBlock::combineMultiJump(ABlock *srcBlock, ABlock *block) {
        auto ret = block;
        while (true) {
            if (ret->getInstrCntNoComment() != 1) {
                break;
            }
            auto instr = ret->getLast();
            if (auto nxtJmp = dynamic_cast<AInstrB *>(instr)) {
                atriAssert(nxtJmp->cond == Cond::None);
                atriAssert(!nxtJmp->isCall());
//                ret->prec.erase(srcBlock);
                ret = std::get<ABlock *>(nxtJmp->target);
            } else {
                break;
            }
        }
        return ret;
    }


} // armv7