//
// Created by 11067 on 2023/5/22.
//

#ifndef ATRI_BACKEND_MCBLOCK_H
#define ATRI_BACKEND_MCBLOCK_H

#include "MCFunction.h"
#include "MCCompileUnit.h"
#include "../../utils/LinkedList.h"
#include "instructions/MAssignInstr.h"
#include "instructions/MCalInstr.h"
#include "instructions/MCondJumpInstr.h"
#include "instructions/MFuncInstr.h"
#include "instructions/MPseudoInstr.h"
#include "instructions/MOptInstr.h"
#include "../../utils/bitUtils.h"
#include <string>
#include <memory>
#include <set>
#include <map>

namespace riscv {
struct MCFunction;
struct MCCompileUnit;
struct Mapping;

struct MCBlock : public atri::LinkedNode<MCBlock>, public atri::LinkedContainer<MRInstr> {
    std::string label;
    int loopDepth{};
    MCFunction *function;
    std::vector<MCBlock *> prev{}, succ{};
    std::set<Reg> activeUse, activeDef, activeIn, activeOut;
    std::unordered_map<std::string, Reg> globalName2Reg;
    MCBlock(MCFunction *_function, std::string name);
    void construct(Ir::BasicBlock *ir_block, Mapping *mapping);
    void push_back(MRInstr* instr);
    void push_back(const std::vector<MRInstr*>& instrs);
    void gen_asm(std::ostream &out);
    void gen_mul_opt(const Reg &op1, const Reg &op2, const Reg &op3);
    void clear_active();
    void clear();
};
} // namespace riscv

#endif // ATRI_BACKEND_MCBLOCK_H
