//
// Created by 11067 on 2023/5/22.
//

#ifndef ATRI_BACKEND_MCFUNCTION_H
#define ATRI_BACKEND_MCFUNCTION_H

#include "../../utils/LinkedList.h"
#include "MCBlock.h"
#include "MCCompileUnit.h"
#include "instructions/MPseudoInstr.h"
#include "MColorAllocator.h"
#include "MOptimizer.h"
#include <memory>
#include <algorithm>
#include <iterator>
#include <set>
namespace riscv {
struct MCCompileUnit;
struct MCBlock;
struct Reg;
class MColorAllocator;
class MOptimizer;

struct Mapping {
    std::unordered_map<int, Reg> operand_map;
    std::set<Reg> float_operand;
    std::vector<Ir::ScalaType> vRegType;
    MCBlock *curBlock{};
    std::unordered_map<int, int> irCvt{};
    std::unordered_map<int, Reg> reg_map;
    std::unordered_map<Ir::BasicBlock *, MCBlock *> irBB2MCBB;
    std::unordered_map<MCBlock *, Ir::BasicBlock *> mcBB2IRBB;
    std::set<Reg> float_reg_set;
    int reg_num;

    Mapping() :
        reg_num(PREG_NUM) {
    }
    Reg get_operand(Ir::Reg reg);
    Reg get_operand(const Ir::MemObject &memObject);
    Reg get_global_addr(const std::string& name);

    Reg gen_const_operand(int val);
    Reg gen_const_operand(double val);
    Reg gen_var_operand(bool isFloat = false);
};

struct MCFunction : public atri::LinkedNode<MCFunction>, public atri::LinkedContainer<MCBlock> {
    std::string label;
    bool is_lib{false};
    std::unordered_map<Ir::MemObject *, int> obj_stack_arr;
    std::unordered_map<Ir::Vector *, int> vec_stack_arr;
    std::unordered_map<Ir::Vector *, Reg> para_vec_base;
    std::unordered_map<Ir::MemObject *, Reg> para_mem_base;
    std::unordered_map<Reg, std::pair<Reg, int>> ls_opt_map;
    MCBlock *entry{};
    std::vector<MCBlock *> exitBlock{};
    MCCompileUnit *compileUnit;
    Mapping map{};
    int stackTop{};
    int stackBottom{0};
    int ra_stack_arr{0};
    MCFunction(MCCompileUnit *compileUnit, Ir::Function *ir_func);
    void gen_asm(std::ostream &out);
    void active_analyse();
    void destructPhi();
    void reg_alloc();
    void replaceJumpTarget(MCBlock *inWhich, MCBlock *old, MCBlock *nnew);
    void replacePhiBlock(MCBlock *inWhich, MCBlock *old, MCBlock *nnew);
    void foreOPT();
    void afterOPT();
    void clear();

    static std::set<Reg> set_union(const std::set<Reg> &setA, const std::set<Reg> &setB);
    static std::set<Reg> set_difference(const std::set<Reg> &setA, const std::set<Reg> &setB);
    static bool set_equal(const std::set<Reg> &setA, const std::set<Reg> &setB);

    template <typename T>
    bool is_var(const T &arg);
};
} // namespace riscv

#endif // ATRI_BACKEND_MCFUNCTION_H
