//
// Created by MSI-NB on 2023/7/13.
//

#ifndef COMPILER_MCOLORALLOCATOR_H
#define COMPILER_MCOLORALLOCATOR_H

#include "./MCCompileUnit.h"
#include "./MRInstr.h"
#include "./instructions/MCalInstr.h"
#include <unordered_map>
#include <unordered_set>
#include <algorithm>
#include <valarray>

using std::set;
using std::unordered_set;
using std::vector;
using MMove = std::pair<riscv::Reg, riscv::Reg>;
template <>
struct std::hash<MMove> {
    std::size_t operator()(MMove const &move) const noexcept {
        std::size_t h1 = std::hash<riscv::Reg>{}(move.first);
        std::size_t h2 = std::hash<riscv::Reg>{}(move.second);
        return h1 ^ (h2 << 1); // or use boost::hash_combine
    }
};

namespace riscv {
struct MCBlock;
struct MCFunction;
struct MCCompileUnit;
class MColorAllocator {
    MCFunction *function;
    unordered_set<Reg> initial{};
    vector<double> spillPriority{};
    vector<Reg> colors{};
    int run_time{0};
    unordered_set<Reg> preColored{};
    vector<Reg> simplifyWorkList{};
    unordered_set<Reg> freezeWorkList{};
    unordered_set<Reg> spillWorkList{};
    unordered_set<Reg> spilledNodes{};
    unordered_set<Reg> coalescedNodes{};
    unordered_set<Reg> coloredNodes{};
    vector<Reg> selectStack{};
    unordered_set<Reg> selectSet{};
    unordered_set<MMove> coalescesMoves{};
    unordered_set<MMove> constrainedMoves{};
    unordered_set<MMove> frozenMoves{};
//    unordered_set<MMove> workListMoves{};
    vector<MMove> workListMoves{};
    unordered_set<MMove> activeMoves{};
    int spilledVarCnt{0};
    unordered_set<Reg> spillCreated{};
    unordered_set<std::pair<Reg, Reg>> adjSet{};
    vector<unordered_set<Reg>> adjList{};
    vector<int> degree{};
    vector<unordered_set<MMove>> moveList{};
    vector<Reg> alias{};
    unordered_set<Reg> int_alloced{};
    unordered_set<Reg> float_alloced{};
    vector<Reg> getDef(MRInstr *instr);
    vector<Reg> getUse(MRInstr *instr);

    int getAllRegCnt();

    void calculateSpillPriority();

    void clear();

    void build();

    void makeWorkList();

    void simplify();
    void coalesce();
    void freeze();
    void selectSpill();
    void assignColors();
    void rewriteProgram();
    void finish();
    void addEdge(int from_id, Reg to);
    bool conservative(unordered_set<Reg> &nodes);
    void combine(Reg u, Reg v);
    static unordered_set<Reg> set_union(unordered_set<Reg> &a, unordered_set<Reg> &b);
    bool OK(Reg t, Reg r);
    bool moveRelated(const Reg &reg);
    unordered_set<MMove> nodeMoves(const Reg &reg);
    void decrementDegree(Reg &reg);
    unordered_set<Reg> adjacent(Reg &reg);
    void enableMoves(unordered_set<Reg> &nodes);
    Reg getAlias(const Reg &reg);
    void addWorkList(const Reg &u);
    void freezeMoves(const Reg &u);
    void allocReg(MRInstr *instr, Reg &vreg, Reg &reg);
    Reg allocSplitVReg(Ir::ScalaType type);
    void addWorkListMove(MMove move);

public:
    explicit MColorAllocator(MCFunction *_function);
    void run();
    unordered_set<Reg> get_int_alloc() {
        return int_alloced;
    }
    unordered_set<Reg> get_float_alloc() {
        return float_alloced;
    }
};
} // namespace riscv

#endif // COMPILER_MCOLORALLOCATOR_H
