//
// Created by MSI-NB on 2023/7/30.
//

#include "MOptimizer.h"
namespace riscv {
MOptimizer::MOptimizer(MCFunction *_function) { this->function = _function; }

void MOptimizer::foreOPT() {
  mergeBlock();
  dce();
  mergeInstr();
  forePeepHole();
  mergeInstr();
}

void MOptimizer::afterOPT() {
  afterPeepHole();
  dce();
  delEmptyBlk();
}

void MOptimizer::moveBlock() {
  auto blocks = function->toVector();
  for (auto bb : blocks) {
    auto instrList = bb->toVector();
    MCBlock *directSucc = nullptr;
    for (int i = instrList.size() - 1; i >= 0; i--) {
      auto instr = instrList[i];
      Instanceof(jump, GotoInstr *, instr) { directSucc = jump->target; }
      else {
        break;
      }
    }
    if (directSucc != nullptr) {
      directSucc->removeMe();
      bb->insertAfterMe(directSucc);
    }
  }
}

void MOptimizer::mergeBlock() {
  moveBlock();
  auto blocks = function->toVector();
  unordered_set<MCBlock *> removedBlks{};
  // 简单合并紧邻块
  for (auto bb : blocks) {
    if (removedBlks.count(bb)) {
      continue;
    }
    while (true) {
      auto lastInstr = bb->getLast();
      if (lastInstr == nullptr ||
          dynamic_cast<IfConInstr *>(lastInstr) != nullptr) {
        break;
      }
      if (lastInstr->getPrev() != nullptr &&
          dynamic_cast<IfConInstr *>(lastInstr->getPrev())) {
        break;
      }
      Instanceof(jump, GotoInstr *, lastInstr) {
        auto target = jump->target;
        if (bb->succ.size() == 1 && target->prev.size() == 1 &&
            target->prev[0] == bb) {
          jump->removeMe();
          delete jump;
          for (auto code : target->toVector()) {
            bb->addLast(code);
          }
          bb->succ = target->succ;
          for (auto sucBlock : target->succ) {
            auto it = std::remove(sucBlock->prev.begin(), sucBlock->prev.end(),
                                  target);
            sucBlock->prev.erase(it, sucBlock->prev.end());
            sucBlock->prev.push_back(bb);
          }
          if (std::find(function->exitBlock.begin(), function->exitBlock.end(),
                        target) != function->exitBlock.end()) {
            auto it = std::remove(function->exitBlock.begin(),
                                  function->exitBlock.end(), target);
            function->exitBlock.erase(it, function->exitBlock.end());
            function->exitBlock.push_back(bb);
          }
          removedBlks.insert(target);
        } else {
          break;
        }
      }
      else {
        break;
      }
    }
  }
  for (auto bb : removedBlks) {
    bb->removeMe();
    delete bb;
  }
  // 删除无用跳转
  auto after_blocks = function->toVector();
  for (int i = 0; i < after_blocks.size() - 1; i++) {
    auto bb = after_blocks[i];
    auto lastInstr = bb->getLast();
    if (lastInstr != nullptr) {
      Instanceof(jump, GotoInstr *, lastInstr) {
        if (jump->target == after_blocks[i + 1]) {
          jump->removeMe();
          delete jump;
        }
      }
    }
  }
}

void MOptimizer::dce() {
  function->active_analyse();
  for (auto block : *function) {
    unordered_set<Reg> live{};
    for (auto reg : block->activeOut) {
      live.insert(reg);
    }
    auto instrList = block->toVector();

    for (int i = instrList.size() - 1; i >= 0; i--) {
      auto &instr = instrList[i];
      auto def = getDef(instr);
      if (def.size() == 1 && is_pesudo(def[0]) && !live.count(def[0])) {
        instr->removeMe();
        delete instr;
      } else {
        for (auto reg : getUse(instr)) {
          live.insert(reg);
        }
      }
    }
  }
}

void MOptimizer::mergeInstr() {
  function->active_analyse();
  // 合并branch指令
  for (auto block : *function) {
    auto instrList = block->toVector();
    for (int i = instrList.size() - 1; i >= 0; i--) {
      auto &instr = instrList[i];
      // todo branch是否需要位扩展
      Instanceof(ifConInstr, IfConInstr *, instr) {
        if (i != 0) {
          auto target = ifConInstr->target;
          auto &preInstr = instrList[i - 1];
          bool replaced = false;
          if (preInstr->op1 == ifConInstr->op2 && !preInstr->op1.isConst &&
              !block->activeOut.count(preInstr->op1)) {
            Instanceof(sltInstr, SltInstr *, preInstr) {
              ifConInstr->insertAfterMe(
                  new BltInstr(target, preInstr->op2, preInstr->op3));
              replaced = true;
            }
            else Instanceof(sleInstr, SleInstr *, preInstr) {
              ifConInstr->insertAfterMe(
                  new BleInstr(target, preInstr->op2, preInstr->op3));
              replaced = true;
            }
            else Instanceof(sgtInstr, GreInstr *, preInstr) {
              ifConInstr->insertAfterMe(
                  new BgtInstr(target, preInstr->op2, preInstr->op3));
              replaced = true;
            }
            else Instanceof(sgeInstr, GeqInstr *, preInstr) {
              ifConInstr->insertAfterMe(
                  new BgeInstr(target, preInstr->op2, preInstr->op3));
              replaced = true;
            }
            else Instanceof(eqInstr, EqInstr *, preInstr) {
              ifConInstr->insertAfterMe(
                  new BeqInstr(target, preInstr->op2, preInstr->op3));
              replaced = true;
            }
            else Instanceof(neInstr, NeInstr *, preInstr) {
              ifConInstr->insertAfterMe(
                  new BneInstr(target, preInstr->op2, preInstr->op3));
              replaced = true;
            }
            if (replaced) {
              ifConInstr->removeMe();
              delete ifConInstr;
              preInstr->removeMe();
              delete preInstr;
              break;
            }
          }
        }
      }
    }
  }
  // 合并生成FMAdd、FMSub指令
  //  for (auto block : *function) {
  //    unordered_set<Reg> live{};
  //    for (auto reg : block->activeOut) {
  //      live.insert(reg);
  //    }
  //    auto instrList = block->toVector();
  //    for (int i = instrList.size() - 1; i >= 0; i--) {
  //      auto &instr = instrList[i];
  //      if (instr->op1.type == Ir::Float &&
  //          (dynamic_cast<AddInstr *>(instr) != nullptr ||
  //           dynamic_cast<SubInstr *>(instr) != nullptr)) {
  //        auto preInstr = instr->getPrev();
  //        // 跳过add/sub指令浮点赋值
  //        if (preInstr != nullptr &&
  //            dynamic_cast<Bit2FInstr *>(preInstr) != nullptr) {
  //          preInstr = preInstr->getNext()->getNext();
  //          i -= 2;
  //        }
  //        if (preInstr != nullptr && preInstr->op1.type == Ir::Float &&
  //            dynamic_cast<MulInstr *>(preInstr) != nullptr) {
  //          if (!live.count(preInstr->op1) && instr->op2 != instr->op3) {
  //            Reg op4;
  //            bool mergeable = false;
  //            bool first = true;
  //            if (instr->op2 == preInstr->op1) {
  //              op4 = instr->op3;
  //              mergeable = true;
  //            } else if (instr->op3 == preInstr->op1) {
  //              op4 = instr->op2;
  //              mergeable = true;
  //              first = false;
  //            }
  //            if (mergeable) {
  //              if (dynamic_cast<AddInstr *>(instr) != nullptr) {
  //                instr->insertAfterMe(new FMAddInstr(instr->op1,
  //                preInstr->op2,
  //                                                    preInstr->op3, op4));
  //              } else {
  //                if(first) {
  //                  instr->insertAfterMe(new FMSubInstr(instr->op1,
  //                  preInstr->op2,
  //                                                      preInstr->op3, op4));
  //                } else {
  //                  instr->insertAfterMe(new FNMSubInstr(instr->op1,
  //                  preInstr->op2,
  //                                                      preInstr->op3, op4));
  //                }
  //              }
  //              live.erase(instr->op1);
  //              live.insert(preInstr->op2);
  //              live.insert(preInstr->op3);
  //              live.insert(op4);
  //              instr->removeMe();
  //              preInstr->removeMe();
  //              delete instr;
  //              delete preInstr;
  //              i--;
  //              continue;
  //            }
  //          }
  //        }
  //      }
  //      for (auto reg : getDef(instr)) {
  //        live.erase(reg);
  //      }
  //      for (auto reg : getUse(instr)) {
  //        live.insert(reg);
  //      }
  //    }
  //  }
  // la addr + addi dest addr 0 ->  la  dest
  for (auto block : *function) {
    unordered_set<Reg> live{};
    for (auto reg : block->activeOut) {
      live.insert(reg);
    }
    auto instrList = block->toVector();
    for (int i = instrList.size() - 1; i >= 0; i--) {
      auto &instr = instrList[i];
      if (i != 0) {
        auto &preInstr = instrList[i - 1];
        if (dynamic_cast<Add64Instr *>(instr) != nullptr &&
            dynamic_cast<LoadLocalAddrInstr *>(preInstr) != nullptr &&
            !live.count(preInstr->op1) &&
            (!instr->op2.isConst && instr->op3.isConst &&
             instr->op3.int_num == 0) &&
            instr->op2 == preInstr->op1) {
          preInstr->op1 = instr->op1;
          instr->removeMe();
          delete instr;
          i--;
          continue;
        }
      }

      for (auto reg : getDef(instr)) {
        live.erase(reg);
      }
      for (auto reg : getUse(instr)) {
        live.insert(reg);
      }
    }
  }
}

void MOptimizer::forePeepHole() {
  for (auto block : *function) {
    auto instrList = block->toVector();
    for (int i = 0; i < instrList.size(); i++) {
      auto &instr = instrList[i];
      if (dynamic_cast<AddInstr *>(instr) != nullptr) {
        if ((instr->op1 == instr->op2 && instr->op3.isConst &&
             instr->op3.int_num == 0) ||
            (instr->op1 == instr->op3 && instr->op2.isConst &&
             instr->op2.int_num == 0)) {
          instr->removeMe();
          delete instr;
          continue;
        }
      } else if (i != instrList.size() - 1) {
        auto nextInstr = instrList[i + 1];
        if (dynamic_cast<MoveInstr *>(instr) != nullptr &&
            dynamic_cast<MoveInstr *>(nextInstr) != nullptr &&
            instr->op1 == nextInstr->op2 && nextInstr->op1 == instr->op2) {
          nextInstr->removeMe();
          delete nextInstr;
          i++;
          continue;
        } else if (dynamic_cast<CvtI2FInstr *>(instr) != nullptr &&
                   dynamic_cast<CvtF2IInstr *>(nextInstr) != nullptr &&
                   instr->op1 == nextInstr->op2 &&
                   nextInstr->op1 == instr->op2) {
          nextInstr->removeMe();
          delete nextInstr;
          i++;
          continue;
        } else if (dynamic_cast<CvtF2IInstr *>(instr) != nullptr &&
                   dynamic_cast<CvtI2FInstr *>(nextInstr) != nullptr &&
                   instr->op1 == nextInstr->op2 &&
                   nextInstr->op1 == instr->op2) {
          nextInstr->removeMe();
          delete nextInstr;
          i++;
          continue;
        } else if (auto sd = dynamic_cast<StoreInstr *>(instr)) {
          if (auto ld = dynamic_cast<LoadInstr *>(nextInstr)) {
            if (ld->op2 == sd->op3 && ld->op3.int_num == sd->offset) {
              if (ld->op1 != sd->op2) {
                nextInstr->insertAfterMe(new MoveInstr(ld->op1, sd->op2));
              }
              nextInstr->removeMe();
              delete nextInstr;
              i++;
            }
          }
        }
      }
    }
  }
}

void MOptimizer::afterPeepHole() {
  for (auto block : *function) {
    auto instrList = block->toVector();
    for (int i = 0; i < instrList.size(); i++) {
      auto &instr = instrList[i];
      // 删除无用指令
      if (dynamic_cast<MoveInstr *>(instr) != nullptr) {
        if (instr->op1 == instr->op2) {
          instr->removeMe();
          delete instr;
          continue;
        }
      } else if (dynamic_cast<AddInstr *>(instr) != nullptr) {
        if ((instr->op1 == instr->op2 && instr->op3.isConst &&
             instr->op3.int_num == 0) ||
            (instr->op1 == instr->op3 && instr->op2.isConst &&
             instr->op2.int_num == 0)) {
          instr->removeMe();
          delete instr;
          continue;
        }
      }
//      else if (i != instrList.size() - 1) {
//        auto afterInstr = instrList[i + 1];
//        if (dynamic_cast<MoveInstr *>(instr) != nullptr &&
//            dynamic_cast<MoveInstr *>(afterInstr) != nullptr &&
//            instr->op1 == afterInstr->op2 && instr->op2 == afterInstr->op1) {
//          afterInstr->removeMe();
//          delete afterInstr;
//          i++;
//          continue ;
//        }
//      }

      // 删除重复指令
      //      else if (i != instrList.size() - 1) {
      //        auto afterInstr = instrList[i + 1];
      //        if (!((dynamic_cast<BinaryInstr *>(instr) != nullptr &&
      //               !(instr->op2.isConst && instr->op3.isConst)) ||
      //              (dynamic_cast<FMInstr *>(instr) != nullptr) ||
      //              (dynamic_cast<CallInstr *>(instr)))) {
      //          std::stringstream ss1, ss2;
      //          instr->gen_asm(ss1);
      //          afterInstr->gen_asm(ss2);
      //          if (ss1.str() == ss2.str()) {
      //            instr->removeMe();
      //            delete instr;
      //            continue;
      //          }
      //        }
      //      }
    }
  }
}

void MOptimizer::delEmptyBlk() {
  auto blocks = function->toVector();
  unordered_set<MCBlock *> removedBlks{};
  for (auto block : blocks) {
    auto instrList = block->toVector();
    // 删除只有一条无条件跳转指令的基本块
    if (instrList.size() == 1) {
      Instanceof(jump, GotoInstr *, instrList[0]) {
        auto target = jump->target;
        auto it1 = std::remove(target->prev.begin(), target->prev.end(), block);
        target->prev.erase(it1, target->prev.end());
        for (auto preBlock : block->prev) {
          target->prev.push_back(preBlock);
          auto it =
              std::remove(preBlock->succ.begin(), preBlock->succ.end(), block);
          preBlock->succ.erase(it, preBlock->succ.end());
          preBlock->succ.push_back(target);
          // 前驱块无跳转直接后继
          if (block->getPrev() != nullptr && block->getPrev() == preBlock) {
            if (preBlock->getLast() == nullptr ||
                (preBlock->getLast() != nullptr &&
                 dynamic_cast<GotoInstr *>(preBlock->getLast()) == nullptr)) {
              if (block->getNext() != target) {
                preBlock->push_back(new GotoInstr(target));
              }
            }
          }
          // 前驱块有跳转
          if (preBlock->getLast() != nullptr) {
            if (auto jump1 = dynamic_cast<GotoInstr *>(preBlock->getLast())) {
              if (jump1->target == block) {
                jump1->target = target;
              }
              if (preBlock->getLast()->getPrev() != nullptr) {
                if (auto branch = dynamic_cast<BranchInstr *>(
                        preBlock->getLast()->getPrev())) {
                  if (branch->target == block) {
                    branch->target = target;
                  }
                } else if (auto ifCon = dynamic_cast<IfConInstr *>(
                               preBlock->getLast()->getPrev())) {
                  if (ifCon->target == block) {
                    ifCon->target = target;
                  }
                }
              }
            } else if (auto branch1 =
                           dynamic_cast<BranchInstr *>(preBlock->getLast())) {
              if (branch1->target == block) {
                branch1->target = target;
              }
            } else if (auto ifCon1 =
                           dynamic_cast<IfConInstr *>(preBlock->getLast())) {
              if (ifCon1->target == block) {
                ifCon1->target = target;
              }
            }
          }
        }
        jump->removeMe();
        delete jump;
        removedBlks.insert(block);
        continue;
      }
    }
  }
  for (auto bb : removedBlks) {
    bb->removeMe();
    delete bb;
  }
}

vector<Reg> MOptimizer::getDef(MRInstr *instr) {
  vector<Reg> ret;
  Instanceof(call_inst, CallInstr *, instr) { ret = call_inst->get_def(); }
  else Instanceof(fm_inst, FMInstr *, instr) {
    ret = fm_inst->get_def();
  }
  else {
    if (!instr->op1.isConst) {
      ret.push_back(instr->op1);
    }
  }
  return ret;
}

vector<Reg> MOptimizer::getUse(MRInstr *instr) {
  vector<Reg> ret;
  Instanceof(call_inst, CallInstr *, instr) { ret = call_inst->get_use(); }
  else Instanceof(fm_inst, FMInstr *, instr) {
    ret = fm_inst->get_use();
  }
  else {
    if (!instr->op2.isConst) {
      ret.push_back(instr->op2);
    }
    if (!instr->op3.isConst) {
      ret.push_back(instr->op3);
    }
  }
  return ret;
}

} // namespace riscv