//
// Created by 11067 on 2023/5/18.
//

#ifndef ATRI_BACKEND_ARCH_H
#define ATRI_BACKEND_ARCH_H

#include <ostream>
#include <string>
#include "../../optimization/ir.h"
namespace riscv {
enum class RegAttribute {
    CallerSaved,
    CalleeSaved,
    NA,
    VIR
};
/**
 * @brief 寄存器
 * 约定0-31是通用寄存器，32-63是浮点寄存器，>63是虚拟寄存器
 */

constexpr int ARG_REG_NUM = 8;
constexpr double SpillParam = 5.14;
constexpr int ARG_REGS[ARG_REG_NUM] = {10, 11, 12, 13, 14, 15, 16, 17};
constexpr int INT_SIZE = 4;
constexpr int PREG_NUM = 64;
constexpr int PREG_SP_NUM = 6;
constexpr int PREG_ALLOC_NUM = PREG_NUM - PREG_SP_NUM;
constexpr int FLOAT_REG_BEGIN = 32;

struct Reg {
    int id{-1}; // 常数均为-1
    Ir::ScalaType type{Ir::Void};
    const char *abiName{};
    RegAttribute regAttribute{RegAttribute::NA};
    bool isConst{true};
    int int_num{};
    bool is_64{false};
    double float_num{};

    bool operator==(const Reg &rhs) const;
    bool operator!=(const Reg &rhs) const;
    bool operator<(const Reg &rhs) const;
    bool operator>(const Reg &rhs) const;

    Reg() :
        id(-1), type(Ir::Void), abiName(nullptr), regAttribute(RegAttribute::NA), isConst(true), int_num(0){};
    Reg(int id, const char *abiName, RegAttribute regAttribute) :
        id(id), abiName(abiName), regAttribute(regAttribute), isConst(false), int_num(0) {
        is_64 = true;
        if (id >= 0 && id <= 31)
            type = Ir::Int;
        else if (id >= 32 && id <= 63)
            type = Ir::Float;
    }
    Reg(int id, bool isFloat);
    Reg(int num);
};
std::ostream &operator<<(std::ostream &os, const Reg &reg);

Reg &get_reg_by_id(int i);
/**
 * @brief 按id获取浮点寄存器
 * @param i f32寄存器id
 * @return 寄存器单例
 */
Reg &get_freg_by_id(int i);
Reg &get_sp();
Reg &get_ra();
Reg &get_temp();
Reg &get_zero();
Reg &get_ret(bool isFloat = false);
bool isImm12(int num);
std::vector<Reg> getSPReg();
std::vector<Reg> getTempReg();
std::vector<Reg> getAllocReg();
bool is_pesudo(Reg reg);

static int get_vec_size(const std::vector<int> &dims) {
    int s = 1;
    for (const auto &item : dims) {
        s *= item;
    }
    return s;
}

} // namespace riscv
template <>
struct std::hash<riscv::Reg> {
    std::size_t operator()(riscv::Reg const &reg) const noexcept {
        std::size_t h1 = std::hash<int>{}(reg.id);
        std::size_t h2 = std::hash<int>{}(static_cast<int>(reg.type));
        return h1 ^ (h2 << 1); // or use boost::hash_combine
    }
};
#endif // ATRI_BACKEND_ARCH_H
