//
// Created by MSI-NB on 2023/7/6.
//

#include "MAssignInstr.h"

#include <utility>
namespace riscv {
void LoadInstr::gen_asm(std::ostream &out) {
    if (!isImm12(op3.int_num)) {
        out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
        out << "add " << get_temp() << ", " << get_temp() << ", " << op2 << "\n\t";
        out << "lw " << op1 << ", " << 0 << "(" << get_temp() << ")\n";
    } else {
        out << "lw " << op1 << ", " << op3.int_num << "(" << op2 << ")\n";
    }
}

void Load64Instr::gen_asm(std::ostream &out) {
    if (!isImm12(op3.int_num)) {
        out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
        out << "add " << get_temp() << ", " << get_temp() << ", " << op2 << "\n\t";
        out << "ld " << op1 << ", " << 0 << "(" << get_temp() << ")\n";
    } else {
        out << "ld " << op1 << ", " << op3.int_num << "(" << op2 << ")\n";
    }
}

void StoreInstr::gen_asm(std::ostream &out) {
    if (!isImm12(offset)) {
        out << "li " << get_temp() << ", " << offset << "\n\t";
        out << "add " << get_temp() << ", " << get_temp() << ", " << op3 << "\n\t";
        out << "sw " << op2 << ", " << 0 << "(" << get_temp() << ")\n";
    } else {
        out << "sw " << op2 << ", " << offset << "(" << op3 << ")\n";
    }
}

void Store64Instr::gen_asm(std::ostream &out) {
    if (!isImm12(offset)) {
        out << "li " << get_temp() << ", " << offset << "\n\t";
        out << "add " << get_temp() << ", " << get_temp() << ", " << op3 << "\n\t";
        out << "sd " << op2 << ", " << 0 << "(" << get_temp() << ")\n";
    } else {
        out << "sd " << op2 << ", " << offset << "(" << op3 << ")\n";
    }
}

void MoveInstr::gen_asm(std::ostream &out) {
    if (op2.isConst) {
        out << "li " << op1 << ", " << op2.int_num << "\n";
    } else {
        if (op2.type == Ir::Float) {
            out << "fmv.s " << op1 << ", " << op2 << "\n";
        } else {
            out << "mv " << op1 << ", " << op2 << "\n";
        }
    }
}
void LoadLocalAddrInstr::gen_asm(std::ostream &out) {
    out << "la " << op1 << ", " << label << "\n";
}
void LoadFInstr::gen_asm(std::ostream &out) {
    if (!isImm12(op3.int_num)) {
        out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
        out << "add " << get_temp() << ", " << get_temp() << ", " << op2 << "\n\t";
        out << "flw " << op1 << ", " << 0 << "(" << get_temp() << ")\n\t";
    } else {
        out << "flw " << op1 << ", " << op3.int_num << "(" << op2 << ")\n";
    }
}
void StoreFInstr::gen_asm(std::ostream &out) {
    if (!isImm12(offset)) {
        out << "li " << get_temp() << ", " << offset << "\n\t";
        out << "add " << get_temp() << ", " << get_temp() << ", " << op3 << "\n\t";
        out << "fsw " << op2 << ", " << 0 << "(" << get_temp() << ")\n";
    } else {
        out << "fsw " << op2 << ", " << offset << "(" << op3 << ")\n";
    }
}
// fcvt.s.w	fa5,a5
void CvtI2FInstr::gen_asm(std::ostream &out) {
    out << "fcvt.s.w " << op1 << ", " << op2 << "\n";
}
// fcvt.w.s a5,fa5,rtz
void CvtF2IInstr::gen_asm(std::ostream &out) {
    out << "fcvt.w.s " << op1 << ", " << op2 << ", rtz\n";
}
void LoadStackInstr::gen_asm(std::ostream &out) {
    if (!op3.isConst) {
        panic("loadStackInstrError");
    }
    int padding = 4 - (func->stackTop + func->stackBottom) % 4;
    int offset = (func->stackTop + func->stackBottom + padding) * INT_SIZE - op3.int_num;
    if (!isImm12(offset)) {
        out << "li " << get_temp() << ", " << offset << "\n\t";
        out << "add " << get_temp() << ", " << get_temp() << ", " << get_sp() << "\n";
        if (op1.type == Ir::Int) {
            out << "lw " << op1 << ", " << 0 << "(" << get_temp() << ")\n";
        } else if (op1.type == Ir::Float) {
            out << "flw " << op1 << ", " << 0 << "(" << get_temp() << ")\n";
        }
    } else {
        if (op1.type == Ir::Int) {
            out << "lw " << op1 << ", " << offset << "(sp)\n";
        } else if (op1.type == Ir::Float) {
            out << "flw " << op1 << ", " << offset << "(sp)\n";
        }
    }
}
void StoreStackInstr::gen_asm(std::ostream &out) {
    //    if (!op3.isConst) {
    //        panic("storeStackInstrError");
    //    }
    //    int offset = (func->stackTop + func->stackBottom - 1) * INT_SIZE - op3.int_num;
    //    if (op1.type == Ir::Int) {
    //        out << "sw " << op2 << ", " << offset << "(sp)\n";
    //    } else if (op1.type == Ir::Float) {
    //        out << "slw " << op2 << ", " << offset << "(sp)\n";
    //    }
}

void PushStackInstr::gen_asm(std::ostream &out) {
    int padding = 4 - (func->stackBottom + func->stackTop) % 4;
    int size = (func->stackBottom + func->stackTop + padding) * INT_SIZE;
    if (!isImm12(-size)) {
        out << "li " << get_temp() << ", " << -size << "\n\t";
        out << "add sp, sp, " << get_temp() << "\n";
    } else {
        out << "addi sp, sp, " << -size << "\n";
    }
}
void PopStackInstr::gen_asm(std::ostream &out) {
    int padding = 4 - (func->stackBottom + func->stackTop) % 4;
    int size = (func->stackBottom + func->stackTop + padding) * INT_SIZE;
    if (!isImm12(size)) {
        out << "li " << get_temp() << ", " << size << "\n\t";
        out << "add sp, sp, " << get_temp() << "\n\t";
    } else {
        out << "addi sp, sp, " << size << "\n\t";
    }
//    out << "jr ra\n";
    out << "ret\n";
}

CvtI2FInstr::CvtI2FInstr(const Reg &_op1, const Reg &_op2) {
    op1 = _op1;
    op2 = _op2;
}
CvtF2IInstr::CvtF2IInstr(const Reg &_op1, const Reg &_op2) {
    op1 = _op1;
    op2 = _op2;
}

LoadInstr::LoadInstr(const Reg &_op1, const Reg &_op2) {
    op1 = _op1;
    op2 = _op2;
}

LoadInstr::LoadInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3) {
    op1 = _op1;
    op2 = _op2;
    op3 = _op3;
}

LoadFInstr::LoadFInstr(const Reg &_op1, const Reg &_op2) {
    op1 = _op1;
    op2 = _op2;
}

LoadFInstr::LoadFInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3) {
    op1 = _op1;
    op2 = _op2;
    op3 = _op3;
}

StoreInstr::StoreInstr(const Reg &_op1, const Reg &_op2) {
    op2 = _op1;
    op3 = _op2;
}

StoreInstr::StoreInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3) {
    op2 = _op1;
    op3 = _op2;
    if (_op3.isConst) {
        offset = _op3.int_num;
    }
}

StoreFInstr::StoreFInstr(const Reg &_op1, const Reg &_op2) {
    op2 = _op1;
    op3 = _op2;
}

StoreFInstr::StoreFInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3) {
    op2 = _op1;
    op3 = _op2;
    if (_op3.isConst) {
        offset = _op3.int_num;
    }
}
MoveInstr::MoveInstr(const Reg &_op1, const Reg &_op2) {
    op1 = _op1;
    op2 = _op2;
}

LoadStackInstr::LoadStackInstr(const Reg &_op1, const Reg &_op3, MCFunction *_func) {
    op1 = _op1;
    op3 = _op3;
    func = _func;
}

// StoreStackInstr::StoreStackInstr(const Reg &_op1, const Reg &_op3, MCFunction *_func) {
//     op2 = _op1;
//     op3 = _op3;
//     func = _func;
// }

PushStackInstr::PushStackInstr(MCFunction *_func) {
    func = _func;
}

PopStackInstr::PopStackInstr(MCFunction *_func) {
    func = _func;
}

LoadLocalAddrInstr::LoadLocalAddrInstr(const Reg &_op1, std::string _label) {
    op1 = _op1;
    label = std::move(_label);
}
} // namespace riscv
