//
// Created by MSI-NB on 2023/7/6.
//

#ifndef COMPILER_MASSIGNINSTR_H
#define COMPILER_MASSIGNINSTR_H

#include "../MRInstr.h"
#include "../MCFunction.h"

namespace riscv {
struct MCFunction;
struct CvtI2FInstr : MRInstr {
    CvtI2FInstr(const Reg &_op1, const Reg &_op2);
    void gen_asm(std::ostream &out) override;
};

struct CvtF2IInstr : MRInstr {
    CvtF2IInstr(const Reg &_op1, const Reg &_op2);
    void gen_asm(std::ostream &out) override;
};

struct LoadInstr : MRInstr {
    // op3表示偏移 为value型
    LoadInstr(const Reg &_op1, const Reg &_op2);
    LoadInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3);
    void gen_asm(std::ostream &out) override;
};

struct Load64Instr : LoadInstr {
    // op3表示偏移 为value型
    void gen_asm(std::ostream &out) override;
    Load64Instr(const Reg &_op1, const Reg &_op2, const Reg &_op3) :
        LoadInstr(_op1, _op2, _op3){
            op1.is_64 = true;
            op2.is_64 = true;
            op3.is_64 = true;
        };
};

struct StoreInstr : MRInstr {
    int offset{0};
    // op3表示偏移 为value型
    StoreInstr(const Reg &_op1, const Reg &_op2);
    StoreInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3);
    void gen_asm(std::ostream &out) override;
};

struct Store64Instr : StoreInstr {
    Store64Instr(const Reg &_op1, const Reg &_op2, const Reg &_op3) :
        StoreInstr(_op1, _op2, _op3) {
            op1.is_64 = true;
            op2.is_64 = true;
            op3.is_64 = true;
    }
    void gen_asm(std::ostream &out) override;
};

struct MoveInstr : MRInstr {
    bool is_spilled{false};
    MoveInstr(const Reg &_op1, const Reg &_op2);
    void gen_asm(std::ostream &out) override;
};

struct LoadFInstr : MRInstr {
    // op3表示偏移 为value型
    LoadFInstr(const Reg &op1, const Reg &op2);
    LoadFInstr(const Reg &op1, const Reg &op2, const Reg &op3);
    void gen_asm(std::ostream &out) override;
};

struct StoreFInstr : MRInstr {
    int offset{0};
    StoreFInstr(const Reg &op1, const Reg &op2);
    StoreFInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3);
    void gen_asm(std::ostream &out) override;
};

struct LoadStackInstr : MRInstr {
    MCFunction *func;
    LoadStackInstr(const Reg &_op1, const Reg &_op3, MCFunction *_func);
    void gen_asm(std::ostream &out) override;
};

struct StoreStackInstr : MRInstr {
//    MCFunction *func;
//    StoreStackInstr(const Reg &_op1, const Reg &_op3, MCFunction *_func);
    void gen_asm(std::ostream &out) override;
};

struct PushStackInstr : MRInstr {
    MCFunction *func;
    explicit PushStackInstr(MCFunction *_func);
    void gen_asm(std::ostream &out) override;
};

struct PopStackInstr : MRInstr {
    MCFunction *func;
    PopStackInstr(MCFunction *_func);
    void gen_asm(std::ostream &out) override;
};

struct LoadLocalAddrInstr : MRInstr {
    std::string label;
    LoadLocalAddrInstr(const Reg &_op1, std::string _label);
    void gen_asm(std::ostream &out) override;
};

} // namespace riscv

#endif // COMPILER_MASSIGNINSTR_H
