//
// Created by MSI-NB on 2023/7/6.
//

#include "MCalInstr.h"

#include <utility>
namespace riscv {
void BinaryInstr::gen_asm(std::ostream &out) { panic("why a binary here!"); }

void Add64Instr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << op2.int_num + op3.int_num << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(op3.int_num)) {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "add " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "addi " << op1 << ", " << op2 << ", " << op3.int_num << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (!isImm12(op2.int_num)) {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "add " << op1 << ", " << get_temp() << ", " << op3 << "\n";
    } else {
      out << "addi " << op1 << ", " << op3 << ", " << op2.int_num << "\n";
    }
  } else {
    if (op1.type == Ir::Int) {
      out << "add " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op1.type == Ir::Float) {
      out << "fadd.s " << op1 << ", " << op2 << ", " << op3 << "\n";
    }
  }
}

void AddInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << op2.int_num + op3.int_num << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(op3.int_num)) {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "addw " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "addiw " << op1 << ", " << op2 << ", " << op3.int_num << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (!isImm12(op2.int_num)) {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "addw " << op1 << ", " << get_temp() << ", " << op3 << "\n";
    } else {
      out << "addiw " << op1 << ", " << op3 << ", " << op2.int_num << "\n";
    }
  } else {
    if (op1.type == Ir::Int) {
      out << "addw " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op1.type == Ir::Float) {
      out << "fadd.s " << op1 << ", " << op2 << ", " << op3 << "\n";
    }
  }
}
void SubInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << op2.int_num - op3.int_num << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(op3.int_num)) {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "subw " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "addiw " << op1 << ", " << op2 << ", " << -op3.int_num << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
    out << "subw " << op1 << ", " << get_temp() << ", " << op3 << "\n";
  } else {
    if (op1.type == Ir::Int) {
      out << "subw " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op1.type == Ir::Float) {
      out << "fsub.s " << op1 << ", " << op2 << ", " << op3 << "\n";
    }
  }
}
void DivInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << op2.int_num / op3.int_num << "\n";
  } else if (!op2.isConst && op3.isConst) {
    out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
    out << "divw " << op1 << ", " << op2 << ", " << get_temp() << "\n";
  } else if (op2.isConst && !op3.isConst) {
    out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
    out << "divw " << op1 << ", " << get_temp() << ", " << op3 << "\n";
  } else {
    if (op1.type == Ir::Int) {
      out << "divw " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op1.type == Ir::Float) {
      out << "fdiv.s " << op1 << ", " << op2 << ", " << op3 << "\n";
    }
  }
}
void MulInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << op2.int_num * op3.int_num << "\n";
  } else if (!op2.isConst && op3.isConst) {
    out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
    out << "mulw " << op1 << ", " << get_temp() << ", " << op2 << "\n";
  } else if (op2.isConst && !op3.isConst) {
    out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
    out << "mulw " << op1 << ", " << get_temp() << ", " << op3 << "\n";
  } else {
    if (op1.type == Ir::Int) {
      out << "mulw " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op1.type == Ir::Float) {
      out << "fmul.s " << op1 << ", " << op2 << ", " << op3 << "\n";
    }
  }
}
void RemInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << op2.int_num % op3.int_num << "\n";
  } else if (!op2.isConst && op3.isConst) {
    out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
    out << "remw " << op1 << ", " << op2 << ", " << get_temp() << "\n";
  } else if (op2.isConst && !op3.isConst) {
    out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
    out << "remw " << op1 << ", " << get_temp() << ", " << op3 << "\n";
  } else {
    if (op1.type == Ir::Int) {
      out << "remw " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op1.type == Ir::Float) {
      panic("rem op is float");
      //            out<<"wrong rem\n";
      //            out << "remwf " << op1 << ", " << op2 << ", " << op3 <<
      //            "\n";
    }
  }
}

// todo sext.w 有无必要
void GreInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num > op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (op3.int_num == 0) {
      out << "sgt " << op1 << ", " << op2 << ", " << get_zero() << "\n";
    } else {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      //        out << "sext.w " << op2 << ", " << op2 << "\n";
      out << "sgt " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num >= 0x7fff) {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      //            out << "sext.w " << op3 << ", " << op3 << "\n";
      out << "slt " << op1 << ", " << op3 << ", " << get_temp() << "\n";
    } else {
      //            out << "sext.w " << op3 << ", " << op3 << "\n";
      out << "slti " << op1 << ", " << op3 << ", " << op2.int_num << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      //            out << "sext.w " << op2 << ", " << op2 << "\n";
      //            out << "sext.w " << op3 << ", " << op3 << "\n";
      out << "sgt " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op2.type == Ir::Float) {
      out << "fgt.s " << op1 << ", " << op2 << ", " << op3 << "\n";
    }
  }
}

void GeqInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num >= op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if(op3.int_num==1) {
      out << "sgt " << op1 << ", " << op2 << ", " << get_zero() << "\n";
    } else {
      out << "li " << get_temp() << ", " << op3.int_num - 1 << "\n\t";
      //        out << "sext.w " << op2 << ", " << op2 << "\n";
      out << "sgt " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    out << "addiw " << op1 << ", " << op3 << ", " << -op2.int_num << "\n\t";
    //        out << "sext.w " << op1 << ", " << op1 << "\n";
    out << "slti " << op1 << ", " << op1 << ", 1\n";
  } else {
    if (op2.type == Ir::Int) {
      out << "slt " << op1 << ", " << op2 << ", " << op3 << "\n\t";
      //            out << "sext.w " << op2 << ", " << op2 << "\n";
      //            out << "sext.w " << op3 << ", " << op3 << "\n";
      out << "xori " << op1 << ", " << op1 << ", 1\n";
    } else if (op2.type == Ir::Float) {
      out << "fge.s " << op1 << ", " << op2 << ", " << op3 << "\n";
    }
  }
}

void SltInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num < op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(op3.int_num)) {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      //            out << "sext.w " << op2 << ", " << op2 << "\n";
      out << "slt " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      //            out << "sext.w " << op2 << ", " << op2 << "\n";
      out << "slti " << op1 << ", " << op2 << ", " << op3.int_num << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "slt " << op1 << ", " << get_zero() << ", " << op3 << "\n";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      //        out << "sext.w " << op3 << ", " << op3 << "\n";
      out << "slt " << op1 << ", " << get_temp() << ", " << op3 << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      //            out << "sext.w " << op2 << ", " << op2 << "\n";
      //            out << "sext.w " << op3 << ", " << op3 << "\n";
      out << "slt " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op2.type == Ir::Float) {
      out << "flt.s " << op1 << ", " << op2 << ", " << op3 << "\n";
    }
  }
}

void SleInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num <= op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (op3.int_num == 0) {
      out << "sgt " << op1 << ", " << op2 << ", " << get_zero() << "\n\t";
    } else {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      //        out << "sext.w " << op2 << ", " << op2 << "\n";
      out << "sgt " << op1 << ", " << op2 << ", " << get_temp() << "\n\t";
    }
    out << "xori " << op1 << ", " << op1 << ", 1\n";
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "sgt " << op1 << ", " << get_zero() << ", " << op3 << "\n\t";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      //        out << "sext.w " << op3 << ", " << op3 << "\n";
      out << "sgt " << op1 << ", " << get_temp() << ", " << op3 << "\n\t";
    }
    out << "xori " << op1 << ", " << op1 << ", 1\n";
  } else {
    if (op2.type == Ir::Int) {
      //            out << "sext.w " << op2 << ", " << op2 << "\n";
      //            out << "sext.w " << op3 << ", " << op3 << "\n";
      out << "sgt " << op1 << ", " << op2 << ", " << op3 << "\n\t";
      out << "xori " << op1 << ", " << op1 << ", 1\n";
    } else if (op2.type == Ir::Float) {
      out << "fle.s " << op1 << ", " << op2 << ", " << op3 << "\n";
    }
  }
}

void ShlInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num << op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(op3.int_num)) {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "sllw " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "slliw " << op1 << ", " << op2 << ", " << op3.int_num << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "sllw " << op1 << ", " << get_zero() << ", " << op3 << "\n";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "sllw " << op1 << ", " << get_temp() << ", " << op3 << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      out << "sllw " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op2.type == Ir::Float) {
      panic("shift op is float");
    }
  }
}

void LShrInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num >> op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(op3.int_num)) {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "srl " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "srliw " << op1 << ", " << op2 << ", " << op3.int_num << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "srlw " << op1 << ", " << get_zero() << ", " << op3 << "\n";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "srlw " << op1 << ", " << get_temp() << ", " << op3 << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      if (op2 == op3) {
        out << "mv " << op1 << ", " << get_zero() << "\n";
      } else {
        out << "srlw " << op1 << ", " << op2 << ", " << op3 << "\n";
      }
    } else if (op2.type == Ir::Float) {
      panic("shift op is float");
    }
  }
}

void AShr64Instr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num >> op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(op3.int_num)) {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "sra " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "srai " << op1 << ", " << op2 << ", " << op3.int_num << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "sra " << op1 << ", " << get_zero() << ", " << op3 << "\n";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "sra " << op1 << ", " << get_temp() << ", " << op3 << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      if (op2 == op3) {
        out << "mv " << op1 << ", zero\n";
      } else {
        out << "sra " << op1 << ", " << op2 << ", " << op3 << "\n\t";
        out << "sext.w " << op1 << ", " << op1 << "\n";
      }
    } else if (op2.type == Ir::Float) {
      panic("shift op is float");
    }
  }
}

void AShrInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num >> op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(op3.int_num)) {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "sraw " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "sraiw " << op1 << ", " << op2 << ", " << op3.int_num << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "sraw " << op1 << ", " << get_zero() << ", " << op3 << "\n";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "sraw " << op1 << ", " << get_temp() << ", " << op3 << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      if (op2 == op3) {
        out << "mv " << op1 << ", zero\n";
      } else {
        out << "sraw " << op1 << ", " << op2 << ", " << op3 << "\n";
      }
    } else if (op2.type == Ir::Float) {
      panic("shift op is float");
    }
  }
}
// 按位
void AndInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num & op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(op3.int_num)) {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "and " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "andi " << op1 << ", " << op2 << ", " << op3.int_num << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (!isImm12(op2.int_num)) {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "and " << op1 << ", " << op3 << ", " << get_temp() << "\n";
    } else {
      out << "andi " << op1 << ", " << op3 << ", " << op2.int_num << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      out << "and " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op2.type == Ir::Float) {
      panic("shift op is float");
    }
  }
}

void OrInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num | op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(op3.int_num)) {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "or " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "ori " << op1 << ", " << op2 << ", " << op3.int_num << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (!isImm12(op2.int_num)) {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "or " << op1 << ", " << op3 << ", " << get_temp() << "\n";
    } else {
      out << "ori " << op1 << ", " << op3 << ", " << op2.int_num << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      out << "or " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op2.type == Ir::Float) {
      panic("shift op is float");
    }
  }
}

void XorInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num ^ op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(op3.int_num)) {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "xor " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "xori " << op1 << ", " << op2 << ", " << op3.int_num << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (!isImm12(op2.int_num)) {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "xor " << op1 << ", " << op3 << ", " << get_temp() << "\n";
    } else {
      out << "xori " << op1 << ", " << op3 << ", " << op2.int_num << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      out << "xor " << op1 << ", " << op2 << ", " << op3 << "\n";
    } else if (op2.type == Ir::Float) {
      panic("shift op is float");
    }
  }
}
void NeInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num != op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(-op3.int_num)) {
      out << "li " << get_temp() << ", " << -op3.int_num << "\n\t";
      out << "addw " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "addiw " << op1 << ", " << op2 << ", " << -op3.int_num << "\n";
    }
    //        out << "sext.w " << op1 << ", " << op1 << "\n";
    out << "sltu " << op1 << ", zero"
        << ", " << op1 << "\n";
  } else if (op2.isConst && !op3.isConst) {
    if (!isImm12(-op2.int_num)) {
      out << "li " << get_temp() << ", " << -op2.int_num << "\n\t";
      out << "addw " << op1 << ", " << op3 << ", " << get_temp() << "\n";
    } else {
      out << "addiw " << op1 << ", " << op3 << ", " << -op2.int_num << "\n";
    }
    //        out << "sext.w " << op1 << ", " << op1 << "\n";
    out << "sltu " << op1 << ", zero"
        << ", " << op1 << "\n";
  } else {
    if (op2.type == Ir::Int) {
      out << "subw " << op1 << ", " << op2 << ", " << op3 << "\n\t";
      //            out << "sext.w " << op1 << ", " << op1 << "\n";
      out << "snez " << op1 << ", " << op1 << "\n";
    } else if (op2.type == Ir::Float) {
      out << "feq.s " << op1 << ", " << op2 << ", " << op3 << "\n\t";
      out << "seqz " << op1 << ", " << op1 << "\n";
    }
  }
}
void EqInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    out << "li " << op1 << ", " << (op2.int_num == op3.int_num) << "\n";
  } else if (!op2.isConst && op3.isConst) {
    if (!isImm12(-op3.int_num)) {
      out << "li " << get_temp() << ", " << -op3.int_num << "\n\t";
      out << "addw " << op1 << ", " << op2 << ", " << get_temp() << "\n\t";
    } else {
      out << "addiw " << op1 << ", " << op2 << ", " << -op3.int_num << "\n\t";
    }
    //        out << "sext.w " << op1 << ", " << op1 << "\n";
    out << "sltiu " << op1 << ", " << op1 << ", 1\n";
  } else if (op2.isConst && !op3.isConst) {
    if (!isImm12(-op2.int_num)) {
      out << "li " << get_temp() << ", " << -op2.int_num << "\n\t";
      out << "addw " << op1 << ", " << op3 << ", " << get_temp() << "\n\t";
    } else {
      out << "addiw " << op1 << ", " << op3 << ", " << -op2.int_num << "\n\t";
    }
    //        out << "sext.w " << op1 << ", " << op1 << "\n";
    out << "sltiu " << op1 << ", " << op1 << ", 1\n";
  } else {
    if (op2.type == Ir::Int) {
      out << "subw " << op1 << ", " << op2 << ", " << op3 << "\n\t";
      //            out << "sext.w " << op1 << ", " << op1 << "\n";
      out << "seqz " << op1 << ", " << op1 << "\n";
    } else if (op2.type == Ir::Float) {
      out << "feq.s " << op1 << ", " << op2 << ", " << op3 << "\n";
    }
  }
}

void NegInstr::gen_asm(std::ostream &out) {
  if (op2.isConst) {
    out << "li " << op1 << ", " << -op2.int_num << "\n";
  } else {
    if (op2.type == Ir::Int) {
      out << "negw " << op1 << ", " << op2 << "\n";
    } else if (op2.type == Ir::Float) {
      out << "fneg.s " << op1 << ", " << op2 << "\n";
    }
  }
}

// 实现每类构造函数，单纯赋值
BinaryInstr::BinaryInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3) {
  op1 = _op1;
  op2 = _op2;
  op3 = _op3;
}

BinaryInstr::BinaryInstr(const Reg &_op1, const Reg &_op2) {
  op1 = _op1;
  op2 = _op2;
}

BinaryInstr::BinaryInstr() = default;

} // namespace riscv
