//
// Created by MSI-NB on 2023/7/6.
//

#ifndef COMPILER_MCALINSTR_H
#define COMPILER_MCALINSTR_H

#include "../MRInstr.h"

namespace riscv {
struct BinaryInstr : MRInstr {
  BinaryInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3);
  BinaryInstr(const Reg &_op1, const Reg &_op2);
  BinaryInstr();
  void gen_asm(std::ostream &out) override;
};

struct CompareInstr : BinaryInstr {
  CompareInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override{};
};

struct AddInstr : BinaryInstr {
  AddInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct Add64Instr : AddInstr {
  Add64Instr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : AddInstr(_op1, _op2, _op3) {
    op1.is_64 = true;
    op2.is_64 = true;
    op3.is_64 = true;
  };
  void gen_asm(std::ostream &out) override;
};

struct SubInstr : BinaryInstr {
  SubInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct DivInstr : BinaryInstr {
  DivInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct MulInstr : BinaryInstr {
  MulInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct RemInstr : BinaryInstr {
  RemInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct GreInstr : CompareInstr {
  GreInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : CompareInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct GeqInstr : CompareInstr {
  GeqInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : CompareInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct SltInstr : CompareInstr {
  SltInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : CompareInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct SleInstr : CompareInstr {
  SleInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : CompareInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct NeInstr : CompareInstr {
  NeInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : CompareInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct EqInstr : CompareInstr {
  EqInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : CompareInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct NegInstr : BinaryInstr {
  NegInstr(const Reg &_op1, const Reg &_op2) : BinaryInstr(_op1, _op2){};
  void gen_asm(std::ostream &out) override;
};

struct ShlInstr : BinaryInstr {
  ShlInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct AShr64Instr : BinaryInstr {
  AShr64Instr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct AShrInstr : BinaryInstr {
  AShrInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct LShrInstr : BinaryInstr {
  LShrInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct AndInstr : BinaryInstr {
  AndInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct OrInstr : BinaryInstr {
  OrInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct XorInstr : BinaryInstr {
  XorInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

} // namespace riscv
#endif // COMPILER_MCALINSTR_H
