//
// Created by MSI-NB on 2023/7/30.
//

#include "MOptInstr.h"
namespace riscv {
void BranchInstr::gen_asm(std::ostream &out) {
  panic("unexpected br instr here");
}

void FMInstr::gen_asm(std::ostream &out) { panic("unexpected fm instr here"); }

void FMAddInstr::gen_asm(std::ostream &out) {
  out << "fmadd.s " << op1 << ", " << op2 << ", " << op3 << ", " << op4 << "\n";
}

void FMSubInstr::gen_asm(std::ostream &out) {
  out << "fmsub.s " << op1 << ", " << op2 << ", " << op3 << ", " << op4 << "\n";
}

void FNMSubInstr::gen_asm(std::ostream &out) {
  out << "fnmsub.s " << op1 << ", " << op2 << ", " << op3 << ", " << op4
      << "\n";
}

void BeqInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    if (op2.int_num == op3.int_num) {
      out << "j " << target->label << "\n";
    }
  } else if (!op2.isConst && op3.isConst) {
    if (op3.int_num == 0) {
      out << "beq " << op2 << ", " << get_zero() << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "beq " << op2 << ", " << get_temp() << ", " << target->label
          << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "beq " << op3 << ", " << get_zero() << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "beq " << op3 << ", " << get_temp() << ", " << target->label
          << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      out << "beq " << op2 << ", " << op3 << ", " << target->label << "\n";
    } else if (op2.type == Ir::Float) {
      out << "feq.s " << get_temp() << ", " << op2 << ", " << op3 << "\n\t";
      out << "bnez " << get_temp() << ", " << target->label << "\n";
    }
  }
}

void BneInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    if (op2.int_num != op3.int_num) {
      out << "j " << target->label << "\n";
    }
  } else if (!op2.isConst && op3.isConst) {
    if (op3.int_num == 0) {
      out << "bne " << op2 << ", " << get_zero() << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "bne " << op2 << ", " << get_temp() << ", " << target->label
          << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "bne " << op3 << ", " << get_zero() << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "bne " << op3 << ", " << get_temp() << ", " << target->label
          << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      out << "bne " << op2 << ", " << op3 << ", " << target->label << "\n";
    } else if (op2.type == Ir::Float) {
      out << "feq.s " << get_temp() << ", " << op2 << ", " << op3 << "\n\t";
      out << "beqz " << get_temp() << ", " << target->label << "\n";
    }
  }
}

void BgeInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    if (op2.int_num >= op3.int_num) {
      out << "j " << target->label << "\n";
    }
  } else if (!op2.isConst && op3.isConst) {
    if (op3.int_num == 0) {
      out << "bge " << op2 << ", " << get_zero() << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "bge " << op2 << ", " << get_temp() << ", " << target->label
          << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "bge " << get_zero() << ", " << op3 << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "bge " << get_temp() << ", " << op3 << ", " << target->label
          << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      out << "bge " << op2 << ", " << op3 << ", " << target->label << "\n";
    } else if (op2.type == Ir::Float) {
      out << "fge.s " << get_temp() << ", " << op2 << ", " << op3 << "\n\t";
      out << "bnez " << get_temp() << ", " << target->label << "\n";
    }
  }
}

void BgtInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    if (op2.int_num > op3.int_num) {
      out << "j " << target->label << "\n";
    }
  } else if (!op2.isConst && op3.isConst) {
    if (op3.int_num == 0) {
      out << "blt " << get_zero() << ", " << op2 << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "blt " << get_temp() << ", " << op2 << ", " << target->label
          << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "blt " << op3 << ", " << get_zero() << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "blt " << op3 << ", " << get_temp() << ", " << target->label
          << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      out << "blt " << op3 << ", " << op2 << ", " << target->label << "\n";
    } else if (op2.type == Ir::Float) {
      out << "fgt.s " << get_temp() << ", " << op2 << ", " << op3 << "\n\t";
      out << "bnez " << get_temp() << ", " << target->label << "\n";
    }
  }
}

void BleInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    if (op2.int_num <= op3.int_num) {
      out << "j " << target->label << "\n";
    }
  } else if (!op2.isConst && op3.isConst) {
    if (op3.int_num == 0) {
      out << "bge " << get_zero() << ", " << op2 << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "bge " << get_temp() << ", " << op2 << ", " << target->label
          << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "bge " << op3 << ", " << get_zero() << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "bge " << op3 << ", " << get_temp() << ", " << target->label
          << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      out << "bge " << op3 << ", " << op2 << ", " << target->label << "\n";
    } else if (op2.type == Ir::Float) {
      out << "fle.s " << get_temp() << ", " << op2 << ", " << op3 << "\n\t";
      out << "bnez " << get_temp() << ", " << target->label << "\n";
    }
  }
}

void BltInstr::gen_asm(std::ostream &out) {
  if (op2.isConst && op3.isConst) {
    if (op2.int_num < op3.int_num) {
      out << "j " << target->label << "\n";
    }
  } else if (!op2.isConst && op3.isConst) {
    if (op3.int_num == 0) {
      out << "blt " << op2 << ", " << get_zero() << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op3.int_num << "\n\t";
      out << "blt " << op2 << ", " << get_temp() << ", " << target->label
          << "\n";
    }
  } else if (op2.isConst && !op3.isConst) {
    if (op2.int_num == 0) {
      out << "blt " << get_zero() << ", " << op3 << ", " << target->label
          << "\n";
    } else {
      out << "li " << get_temp() << ", " << op2.int_num << "\n\t";
      out << "blt " << get_temp() << ", " << op3 << ", " << target->label
          << "\n";
    }
  } else {
    if (op2.type == Ir::Int) {
      out << "blt " << op2 << ", " << op3 << ", " << target->label << "\n";
    } else if (op2.type == Ir::Float) {
      out << "flt.s " << get_temp() << ", " << op2 << ", " << op3 << "\n\t";
      out << "bnez " << get_temp() << ", " << target->label << "\n";
    }
  }
}

void Move64Instr::gen_asm(std::ostream &out) {
  out << "li " << op1 << ", " << val << "\n";
}

void MoveU32Instr::gen_asm(std::ostream &out) {
  out << "li " << op1 << ", " << val << "\n";
}
// void AddUnsignInstr::gen_asm(std::ostream &out) {
//     if(!op2.isConst&&!op3.isConst) {
//         out<<"add"
//     } else {
//         panic("addu para error");
//     }
// }
void MulHi64Instr::gen_asm(std::ostream &out) {
  out << "mul " << op1 << ", " << op2 << ", " << op3 << "\n\t";
  out << "srli " << op1 << ", " << op1 << ", 32\n";
}

void BICInstr::gen_asm(std::ostream &out) {
  if (!op2.isConst && op3.isConst) {
    int rev_num = ~op3.int_num;
    if (!isImm12(rev_num)) {
      out << "li " << get_temp() << ", " << rev_num << "\n\t";
      out << "and " << op1 << ", " << op2 << ", " << get_temp() << "\n";
    } else {
      out << "andi " << op1 << ", " << op2 << ", " << rev_num << "\n";
    }
  }
}

void Bit2FInstr::gen_asm(std::ostream &out) {
  out << "fmv.w.x " << op1 << ", " << op2 << "\n";
}

Move64Instr::Move64Instr(const Reg &_op1, long long _val) {
  op1 = _op1;
  val = _val;
}

MoveU32Instr::MoveU32Instr(const Reg &_op1, unsigned int _val) {
  op1 = _op1;
  val = _val;
}

Bit2FInstr::Bit2FInstr(const Reg &_op1, const Reg &_op2) {
  op1 = _op1;
  op2 = _op2;
}

BranchInstr::BranchInstr(MCBlock *_target, const Reg &_op2, const Reg &_op3) {
  op2 = _op2;
  op3 = _op3;
  target = _target;
}

FMInstr::FMInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3,
                 const Reg &_op4) {
  op1 = _op1;
  op2 = _op2;
  op3 = _op3;
  op4 = _op4;
}

std::vector<Reg> FMInstr::get_def() {
  std::vector<Reg> ret;
  ret.push_back(op1);
  return ret;
}

std::vector<Reg> FMInstr::get_use() {
  std::vector<Reg> ret;
  ret.push_back(op2);
  ret.push_back(op3);
  ret.push_back(op4);
  return ret;
}

} // namespace riscv