//
// Created by MSI-NB on 2023/7/30.
//

#ifndef COMPILER_MOPTINSTR_H
#define COMPILER_MOPTINSTR_H

#include "../MCBlock.h"
#include "../MRInstr.h"
#include "./MCalInstr.h"

namespace riscv {
struct MCBlock;
struct BinaryInstr;
struct Move64Instr : MRInstr {
  long long val;
  Move64Instr(const Reg &_op1, long long _val);
  void gen_asm(std::ostream &out) override;
};

struct MoveU32Instr : MRInstr {
  long long val;
  MoveU32Instr(const Reg &_op1, unsigned int _val);
  void gen_asm(std::ostream &out) override;
};

struct Bit2FInstr : MRInstr {
  Bit2FInstr(const Reg &_op1, const Reg &_op2);
  void gen_asm(std::ostream &out) override;
};

struct MulHi64Instr : BinaryInstr {
  MulHi64Instr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct BICInstr : BinaryInstr {
  BICInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3)
      : BinaryInstr(_op1, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct BranchInstr : MRInstr {
  MCBlock *target;
  BranchInstr(MCBlock *_target, const Reg &_op2, const Reg &_op3);
  void gen_asm(std::ostream &out) override;
};

struct BeqInstr : BranchInstr {
  BeqInstr(MCBlock *_target, const Reg &_op2, const Reg &_op3)
      : BranchInstr(_target, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct BneInstr : BranchInstr {
  BneInstr(MCBlock *_target, const Reg &_op2, const Reg &_op3)
      : BranchInstr(_target, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct BgeInstr : BranchInstr {
  BgeInstr(MCBlock *_target, const Reg &_op2, const Reg &_op3)
      : BranchInstr(_target, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct BleInstr : BranchInstr {
  BleInstr(MCBlock *_target, const Reg &_op2, const Reg &_op3)
      : BranchInstr(_target, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct BltInstr : BranchInstr {
  BltInstr(MCBlock *_target, const Reg &_op2, const Reg &_op3)
      : BranchInstr(_target, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct BgtInstr : BranchInstr {
  BgtInstr(MCBlock *_target, const Reg &_op2, const Reg &_op3)
      : BranchInstr(_target, _op2, _op3){};
  void gen_asm(std::ostream &out) override;
};

struct FMInstr : MRInstr {
  Reg op4;
  FMInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3, const Reg &_op4);
  void gen_asm(std::ostream &out) override;
  std::vector<Reg> get_def();
  std::vector<Reg> get_use();
};

struct FMAddInstr : FMInstr {
  FMAddInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3, const Reg &_op4)
      : FMInstr(_op1, _op2, _op3, _op4){};
  void gen_asm(std::ostream &out) override;
};

struct FMSubInstr : FMInstr {
  FMSubInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3, const Reg &_op4)
      : FMInstr(_op1, _op2, _op3, _op4){};
  void gen_asm(std::ostream &out) override;
};

// struct FNMAddInstr:FMInstr{
//   FNMAddInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3, const Reg
//   &_op4):FMInstr(_op1, _op2,_op3,_op4){}; void gen_asm(std::ostream &out)
//   override;
// };
//
 struct FNMSubInstr:FMInstr{
   FNMSubInstr(const Reg &_op1, const Reg &_op2, const Reg &_op3, const Reg
   &_op4):FMInstr(_op1, _op2,_op3,_op4){}; void gen_asm(std::ostream &out)
   override;
 };

} // namespace riscv

#endif // COMPILER_MOPTINSTR_H
