//
// Created by 11069 on 3/7/2023.
//

#include <algorithm>
#include "CFG.h"

namespace opt {
using namespace Ir;
CFG::CFG(Ir::Function * _func): func(_func) {}

void CFG::dom_level_compute(Ir::BasicBlock *root, int level) { // dfs
    root->dom_level = level;
    for(auto & bb:root->idom) {
        dom_level_compute(bb,level+1);
    }
}

static void useless_bb_elimination(Ir::Function *func) { // dfs find all accessible bb
    // mark accessible bb
    std::unordered_set<Ir::BasicBlock *> visited_bb;
    std::vector<Ir::BasicBlock *> s;
    s.push_back(func->get_bbs().begin()->get());
    while (s.size()) {
        auto bb = s.back();
        s.pop_back();
        if (visited_bb.count(bb)) {
            continue;
        }
        visited_bb.insert(bb);
        for (auto succ : bb->succ) {
            if (visited_bb.count(succ)) {
                continue;
            }
            s.push_back(succ);
        }
    }

    for (auto iter = func->get_bbs().begin(); iter != func->get_bbs().end();) {
        auto bb = iter->get();
        if (visited_bb.count(bb)) {
            iter++;
            continue;
        } // erase this bb
        for (auto &succ : bb->succ) {
            succ->prev.erase(bb);
            // remember modify the phi instr
            for (auto &inst : succ->get_instrs()) {
                Instanceof(phi, Ir::PhiInstr *, inst.get()) {
                    if (phi->get_bb_2_value().count(bb)) {
                        phi->get_bb_2_value().erase(bb);
                    }
                }
                else {
                    break;
                }
            }
        }
        for (auto &prev : bb->prev) {
            prev->succ.erase(bb);
            Instanceof(br,Ir::BranchInstr*,prev->get_instrs().back().get()) {
                if(br->get_bb_1() == bb) {
                    br->set_bb1(br->get_bb_2());
                    br->set_bb2(nullptr);
                    br->set_reg(nullptr);
                    br->remove_use_def();
                } else if(br->get_bb_2() == bb) {
                    br->set_bb2(nullptr);
                    br->set_reg(nullptr);
                    br->remove_use_def();
                }
            }
        }
        for (auto &inst : bb->get_instrs()) {
            inst->remove_use_def();
        }
        iter = func->get_bbs().erase(iter);
        // maintain the def place and use place
    }
}

void CFG::build() {
    if (func->is_lib()) {
        return ;
    }
    for (auto &bb : func->get_bbs()) {
        bb->succ.clear();
        bb->prev.clear();
    }

    for (auto bb_iter = func->get_bbs().begin(); bb_iter != func->get_bbs().end();) {
        auto bb = bb_iter->get();
        auto &ins = bb->get_instrs().back();
        Instanceof(branch, Ir::BranchInstr *, ins.get()) {
            bb->succ.insert(branch->get_bb_1());
            branch->get_bb_1()->prev.insert(bb);
            if (branch->get_bb_2()) { // jump
                bb->succ.insert(branch->get_bb_2());
                branch->get_bb_2()->prev.insert(bb);
            }
            bb_iter++;
        }
        else Instanceof(ret, Ir::Ret *, ins.get()) { // ret can not access the next bb
            bb_iter++;
        }
        else {         // access the next basicblock
            bb_iter++; // get next bb
            assert(bb_iter != func->get_bbs().end());
            bb->succ.insert(bb_iter->get());
            bb_iter->get()->prev.insert(bb);
        }
    }
}
int N;
unordered_map<Ir::BasicBlock*,int> dfnum;
unordered_map<int,Ir::BasicBlock*> vertex;
unordered_map<Ir::BasicBlock*,unordered_set<Ir::BasicBlock*>> bucket;
unordered_map<Ir::BasicBlock*,Ir::BasicBlock*> semi, parent,samedom,best,ancestor;

void dfs(BasicBlock* p,BasicBlock* n) {
    vector<BasicBlock*> stack;
    stack.push_back(p);
    stack.push_back(n);
    while(stack.size()) {
        auto n_bb = stack.back();
        stack.pop_back();
        auto p_bb = stack.back();
        stack.pop_back();
        if(dfnum.count(n_bb))
            continue ;
        dfnum[n_bb] = N;
        vertex[N] = n_bb;
        parent[n_bb] = p_bb;
        N++;
        for(auto &bb:n_bb->succ) {
           stack.push_back(n_bb);
           stack.push_back(bb);
        }
    }
}

BasicBlock* AncestorWithLowestSemi(BasicBlock* v) {
    BasicBlock* a = ancestor[v];
    if(ancestor[a]) {
        BasicBlock *b = AncestorWithLowestSemi(a);
        ancestor[v] = ancestor[a];
        if(dfnum[semi[b]] < dfnum[semi[best[v]]])
            best[v] = b;
    }
    return best[v];
}

void link(BasicBlock*p,BasicBlock* n) {
    ancestor[n] = p;
    best[n] = n;
}

void dom_calc(BasicBlock* root) {
    root->dom.insert(root);
    for(auto &succ:root->idom) {
        dom_calc(succ);
        root->dom.insert(succ->dom.begin(),succ->dom.end());
    }
}

void CFG::dom_build() {  // Lengauer-Tarjan
    if (func->is_lib()) {
        return ;
    }
    useless_bb_elimination(func);
    // init
    N = 0;
    vertex.clear();
    dfnum.clear();
    bucket.clear();
    semi.clear();
    parent.clear();
    samedom.clear();
    ancestor.clear();
    best.clear();
    for(auto &bb:func->get_bbs()) {
        bb->dom.clear();
        bb->dom_by.clear();
        bb->idom_by = nullptr;
        bb->idom.clear();
        bb->df.clear();
    }

    // dfs tree
    dfs(nullptr,func->get_first_bb());

    for(int i = N-1;i>=1;i--) {
        BasicBlock * n = vertex[i];
        BasicBlock * p = parent[n];
        BasicBlock * s = p;
        for(auto &v:n->prev) {
            BasicBlock* s_; // 寻找候选semi dom
            if(dfnum[v] <= dfnum[n]) {
                s_ = v;
            } else {
                s_ = semi[AncestorWithLowestSemi(v)];
            }
            if(dfnum[s_] < dfnum[s]) {
                s = s_;
            }
        }
        semi[n] = s;
        bucket[s].insert(n);
        link(p,n);
        for(auto &v:bucket[p]) {
            BasicBlock* y = AncestorWithLowestSemi(v);
            if (semi[y] == semi[v]) {
                v->idom_by = p;
                p->idom.insert(v);
            } else {
                samedom[v] = y;
            }
        }
        bucket[p].clear();
    }
    for(int i = 1;i <= N - 1;i++) {
        BasicBlock* n = vertex[i];
        if(samedom[n]) {
            n->idom_by = samedom[n]->idom_by;
            n->idom_by->idom.insert(n);
        }
    }

    // calc dom
    // dom = i U i.idom U (i.idom).idom
    dom_calc(func->get_first_bb());
    // calc df
    for(auto & bb:func->get_bbs()) {
        if(bb->prev.size() < 2)
            continue ;
        for(auto &prev:bb->prev) {
            auto runner = prev;
            while(!runner->dom.count(bb.get()) || runner == bb.get()) { // if runner == bb ,bb is a loop body
                runner->df.insert(bb.get());
                runner = runner->idom_by;
            }
        }
    }
    dom_level_compute(func->get_first_bb(),0);
}

void calc_loop_level(Ir::Loop *loop) {
    if(!loop) {
        return ;
    }
    if(loop->get_depth() != 0) {
        return ;
    }
    if(loop->get_parent()) {
        calc_loop_level(loop->get_parent());
        loop->set_depth(loop->get_parent()->get_depth() + 1);
    } else {
        loop->set_depth(1);
    }
}

void CFG::loop_analysis() {
    func->get_loops().clear();
    if(func->is_lib() || func->get_bbs().empty())
        return;
    vector<Ir::BasicBlock *> stack;
    for(auto &bb:func->get_bbs()) {
        bb->set_belong_to_loop(nullptr);
    }
    stack.push_back(func->get_bbs().front().get());
    while (!stack.empty()) {
        auto bb = stack.back();
        stack.pop_back();
        for(auto &dom_bb:bb->idom) {
            stack.push_back(dom_bb);
        }
        vector<Ir::BasicBlock *>back_edge_bbs;
        for(auto &pre:bb->prev) {
            if(bb->dom.count(pre)) { // find back edge
                back_edge_bbs.push_back(pre);
            }
        }
        if(!back_edge_bbs.empty()) {
            // 为什么不用while,如果有同一个header怎么办呢
            // 如果是 if do while的结构,do里面的才是header,所以应该不会共header?
            // 多个 back edge的话分不清continue和多个loop,当成一个loop 算了

            // BasicBlock *header, BasicBlock *exitingBb, int depth, Function *belongToFunc, Loop *parent
            Ir::BasicBlock* header = bb;
            Ir::BasicBlock *exitingBb = nullptr;
            int depth = 0;
            Ir::Function *belongtoFunc = bb->get_belongto_func();
            vector<Ir::BasicBlock *>stack_bbs;
            unordered_set<Ir::BasicBlock *>loop_bbs;
            stack_bbs.reserve(back_edge_bbs.size());
            for(auto &back_edge_bb:back_edge_bbs) {
                stack_bbs.push_back(back_edge_bb);
            }
            // 找到所有的bbs
            while(!stack_bbs.empty()) {
                auto stack_bb = stack_bbs.back();
                stack_bbs.pop_back();
                loop_bbs.insert(stack_bb);
                if(stack_bb != bb) {
                    for(auto prev:stack_bb->prev) {
                        if(!loop_bbs.count(prev)) {
                            stack_bbs.push_back(prev);
                        }
                    }
                }
            }
            // 找出口bbs
            for(auto &loop_bb:loop_bbs) {
                for(auto &next:loop_bb->succ) {
                    if(!loop_bbs.count(next)) {
                        exitingBb = next;
                        break ;
                    }
                }
            }
            // 新建loop
            auto* loop = new Ir::Loop(header,exitingBb,0, belongtoFunc, nullptr);
            loop->set_all_bbs(loop_bbs);
            belongtoFunc->get_loops().emplace_back(loop);
            for(auto &loop_bb:loop_bbs) {
                if(loop_bb->get_belong_to_loop() && !loop->get_parent() && loop_bb->get_belong_to_loop() != loop) {
                    loop->set_parent(loop_bb->get_belong_to_loop());
                    loop_bb->get_belong_to_loop()->children.insert(loop);
                }
                loop_bb->set_belong_to_loop(loop);
            }
        }
    }

    for(auto &loop:func->get_loops()) {
        calc_loop_level(loop.get());
    }

}

} // namespace opt