//
// Created by 11069 on 19/7/2023.
//

#include "opt.h"
using namespace Ir;
namespace opt {

void arrayssa_destruction(Function* func) {
    for(auto &bb:func->get_bbs()) {
        for(auto iter = bb->get_instrs().begin(); iter !=bb->get_instrs().end();) {
            auto &inst = *iter;
            Instanceof(mem_read,MemRead*,inst.get()) {
                mem_read->remove_use_def();
                if(mem_read->call_read) {
                    Instanceof(vector,MemObject*,mem_read->get_addr().get()) {
                        copy_propagation(func->use_place,mem_read->get_dst()->get_id(), static_pointer_cast<Vector>(mem_read->get_addr()));
                    }
                    iter = bb->get_instrs().erase(iter);
                } else {
                    // 转换成load或者loadGlobal
                    Instanceof(memObject,MemObject*,mem_read->get_addr().get()) {
                        auto load = new LoadInstr(static_pointer_cast<Reg>(mem_read->get_dst()), static_pointer_cast<MemObject>(mem_read->get_addr()));
                        iter->reset(load);
                        load->set_belongto_bb(bb.get());
                        load->add_use_def();
                    } else {
                        auto loadGlobal = new LoadGlobal(static_pointer_cast<Reg>(mem_read->get_dst()), static_pointer_cast<Reg>(mem_read->get_addr()));
                        iter->reset(loadGlobal);
                        loadGlobal->set_belongto_bb(bb.get());
                        loadGlobal->add_use_def();
                    }
                    iter++;
                }
            } else Instanceof(mem_write,MemWrite*,inst.get()) {
                mem_write->remove_use_def();
                if(mem_write->call_write) {
                    iter = bb->get_instrs().erase(iter);
                } else {
                    // 转换成store或者storeGlobal
                    Instanceof(memObject,MemObject*,mem_write->get_addr().get()) { // store
                        auto store = new StoreInstr(static_pointer_cast<Reg>(mem_write->get_val()),
                            static_pointer_cast<MemObject>(mem_write->get_addr()));
                        iter->reset(store);
                        store->set_belongto_bb(bb.get());
                        store->add_use_def();
                    } else {
                        auto storeGlobal = new StoreGlobal(static_pointer_cast<Reg>(mem_write->get_val()),
                           static_pointer_cast<Reg>(mem_write->get_addr()));
                        iter->reset(storeGlobal);
                        storeGlobal->set_belongto_bb(bb.get());
                        storeGlobal->add_use_def();
                    }
                    iter++;
                }
            } else Instanceof(phi,ArrayPhi*,inst.get()) {
                phi->remove_use_def();
                iter = bb->get_instrs().erase(iter);
            } else Instanceof(call,CallInstr*,inst.get()) {
                call->remove_use_def();
                if(call->get_dst() && call->is_void_flag()) {
                    call->get_belongto_bb()->get_belongto_func()->def_place.erase(call->get_dst()->get_id());
                    call->set_dst(nullptr);
                }
                call->used_global.clear();
                call->add_use_def();
                iter++;
            }
            else { // call 已经在mem read 被change use了,不需要再维护了,好中好
                iter++;
            }
        }
    }
}

void arrayssa_destruction(CompileUnit* compileUnit) {
    for(auto &func:compileUnit->get_functions()) {
        Factory::reset(func.get());
        arrayssa_destruction(func.get());
        func->set_bb_max_num(Factory::bb_num);
        func->set_reg_max_num(Factory::reg_num);
    }
    CompileUnit::in_array_ssa = false;
    info("off array ssa");
}


}