//
// Created by 11069 on 27/7/2023.
//

// 该pass 类似窥孔,
// 遍历所有的binary instr,记录其两个操作数
// 如果发现某条的binary instr的操作可以化简就化简

// 分为两个阶段
// 一个针对加减
// 一个针对乘除
// 加减优化b = a + imme c =  b + imme b = a + d c = b - d 等
// 乘除只针对 a = b * c d =a / c

#include <cmath>
#include "opt.h"

using namespace Ir;

namespace opt {

bool is_const_zero(const shared_ptr<Reg>& src) {
    return src->is_constant() && ((src->get_num().int_num == 0 && src->get_type() == Int ) || (abs(src->get_num().float_num - 0.0) <= 1e-6 &&src->get_type() == Float ));
}

void insert_add_map(unordered_map<shared_ptr<Reg>,int> & inst_map,const shared_ptr<Reg>& reg, int step) {
    if(reg->is_constant()) {
        shared_ptr<Reg> target = nullptr;
        int times = 0;
        for(auto &map_iter:inst_map) {
            if(map_iter.first->is_constant()) {
                target = map_iter.first;
                times = map_iter.second;
            }
        }
        if(!target) {
            inst_map[reg] += step;
        } else {
            RegConst new_const_reg ;
            if(reg->get_type() == Int) {
                new_const_reg.int_num = target->get_num().int_num * times + reg->get_num().int_num * step;
            } else {
                new_const_reg.float_num = target->get_num().float_num * times + reg->get_num().float_num * step;
            }
            auto new_reg = Factory::get_const_reg(reg->get_type(),new_const_reg);
            inst_map.erase(target);
            if(is_const_zero(new_reg))
                return;
            inst_map[new_reg] = 1;
        }
    } else {
        inst_map[reg] += step;
    }
}

void insert_mul_map(unordered_map<shared_ptr<Reg>,int> & inst_map,const shared_ptr<Reg>& reg, int step) {
    if(reg->is_constant()) {
        shared_ptr<Reg> target = nullptr;
        int times = 0;
        for(auto &map_iter:inst_map) {
            if(map_iter.first->is_constant()) {
                target = map_iter.first;
                times = map_iter.second;
            }
        }
        if(!target) {
            inst_map[reg] += step;
        } else {
            RegConst new_const_reg ;
            if(reg->get_type() == Int) {
                new_const_reg.int_num = pow(target->get_num().int_num,times) * pow(reg->get_num().int_num,step);
            } else {
                new_const_reg.float_num = pow(double(target->get_num().float_num),times*1.0) * pow(double(reg->get_num().float_num),step*1.0);
            }
            auto new_reg = Factory::get_const_reg(reg->get_type(),new_const_reg);
            inst_map.erase(target);
            inst_map[new_reg] = 1;
        }
    } else {
        inst_map[reg] += step;
    }
}

void merge_src1(unordered_map<shared_ptr<Reg>,unordered_map<shared_ptr<Reg>,int>> &total_map, unordered_map<shared_ptr<Reg>,int> &inst_map,const shared_ptr<Reg> &src1) {
    if(total_map.count(src1)) {
        inst_map = total_map[src1];
    } else {
        inst_map[src1] = 1;
    }
}

void merge_src2(unordered_map<shared_ptr<Reg>,unordered_map<shared_ptr<Reg>,int>> &total_map, unordered_map<shared_ptr<Reg>,int> &inst_map,const shared_ptr<Reg> &src2,BinaryType curType,BinaryType positive_type) {
    if(total_map.count(src2)) {
        for(auto &src:total_map.at(src2)) {
            if(positive_type == ADD) {
                insert_add_map(inst_map,src.first,(curType == positive_type ? 1 : -1)*src.second);
            } else {
                insert_mul_map(inst_map,src.first,(curType == positive_type ? 1 : -1)*src.second);
            }
        }
    } else {
        if(positive_type == ADD) {
            insert_add_map(inst_map,src2,(curType == positive_type ? 1 : -1));
        } else {
            insert_mul_map(inst_map,src2,(curType == positive_type ? 1 : -1));
        }
    }
}



void binary_opt(Function * func) {
    unordered_map<shared_ptr<Reg>,unordered_map<shared_ptr<Reg>,int>> add_sub_map;
    unordered_map<shared_ptr<Reg>,unordered_map<shared_ptr<Reg>,int>> mul_div_map;
    for(auto &bb:func->get_bbs()) {
        for(auto iter = bb->get_instrs().begin();iter != bb->get_instrs().end();iter++) {
            auto &inst = *iter;
            Instanceof(binary,BinaryInstr*,inst.get()) {
                unordered_map<shared_ptr<Reg>,int> inst_map;
                if((binary->get_binary_type()==ADD || binary->get_binary_type() == SUB) && binary->get_dst()->get_type() == Int) {
                    merge_src1(add_sub_map,inst_map,binary->get_src_1());
                    merge_src2(add_sub_map,inst_map,binary->get_src_2(),binary->get_binary_type(),ADD);
                    unordered_map<shared_ptr<Reg>,int>new_inst_map;
                    for(auto map_iter = inst_map.begin();map_iter != inst_map.end();) {
                        if(map_iter->second == 0) {
                            map_iter = inst_map.erase(map_iter);
                        } else {
                            map_iter++;
                        }
                    }
                    add_sub_map[binary->get_dst()] = inst_map;
                    // 对该指令化简
                    if(inst_map.size() == 0) {
                        RegConst tmp;
                        if(binary->get_dst()->get_type() == Int) {
                            tmp.int_num = 0;
                        } else {
                            tmp.float_num = 0.0;
                        }
                        copy_propagation(func->use_place,binary->get_dst()->get_id(),Factory::get_const_reg(binary->get_dst()->get_type(),tmp));
                    } else if(inst_map.size() == 1) {
                        auto single_map = inst_map.begin();
                        if (single_map->second == 1) {
                            copy_propagation(func->use_place,binary->get_dst()->get_id(),single_map->first);
                        } else if (single_map->second == -1){
                            binary->remove_use_def();
                            binary->set_src_1(Factory::get_zero_reg(binary->get_dst()->get_type()));
                            binary->set_src_2(single_map->first);
                            binary->setBinaryType(SUB);
                        }
                    } else if(inst_map.size() == 2) { // 存疑,感觉不一定能优化,人为延长了某些变量的寿命
                        bool valid = true;
                        bool one_neg = false;
                        int pos_times = 0;
                        for(auto &[reg,times]:inst_map) {
                            if(abs(times) != 1) {
                                valid = false;
                            }
                            if(one_neg && times == -1) {
                                valid = false;
                            }
                            if(times == -1) {
                                one_neg = true;
                            }
                            if(times == 1) {
                                pos_times += 1;
                            }
                        }
                        if(!valid) {
                            continue;
                        } else {
                            auto iter = inst_map.begin();
                            shared_ptr<Reg> new_src1 = iter->first;
                            iter++;
                            shared_ptr<Reg> new_src2 = iter->first;
                            binary->remove_use_def();
                            if (pos_times == 2) {
                                binary->set_src_1(new_src1);
                                binary->set_src_2(new_src2);
                                binary->setBinaryType(ADD);
                            } else {
                                if (iter->second == -1) {
                                    binary->set_src_1(new_src1);
                                    binary->set_src_2(new_src2);
                                    binary->setBinaryType(SUB);
                                } else {
                                    binary->set_src_2(new_src1);
                                    binary->set_src_1(new_src2);
                                    binary->setBinaryType(SUB);
                                }
                            }
                            binary->add_use_def();
                        }
                    }
                }
                else if(binary->get_binary_type()==MUL || binary->get_binary_type() == SDIV) {
                    continue ;
                    // 乘除法有精度损失,以下优化全部废除


                }
            }

        }
    }
}

void binary_opt(CompileUnit* compileUnit) {
    for(auto &func:compileUnit->get_functions()) {
        if(func->is_lib()) {
            continue;
        }
        Factory::reset(func.get());
        binary_opt(func.get());
        func->set_bb_max_num(Factory::bb_num);
        func->set_reg_max_num(Factory::reg_num);
    }
}

}
