//
// Created by shliba on 2023/8/7.
//

#include "opt.h"

namespace opt {
    std::unordered_map<string , Ir::BinaryType>  bit_map;
    bool analysis_and(Ir::Function * func) {
        if (func->is_void() || !(func->get_params().size() == 2) || !func->pure) {
            return false;
        }
        auto it = func->get_params().begin();
        auto first = it->get();
        it++;
        auto second = it->get();
        auto ret_bb = func->get_bbs().back().get();
        Instanceof(ret, Ir::Ret*, ret_bb->get_instrs().back().get()) {
            auto src = ret->get_src();
            auto def = func->def_place[src->get_id()];
            Instanceof(add, Ir::BinaryInstr*, def) {
                if (add->get_binary_type() == Ir::ADD && add->get_src_1()->get_id() == first->get_id()
                    && add->get_src_2()->get_id() == second->get_id()) {
                    auto name = func->get_name();
                    if (bit_map.count(name) == 0) {
                        bit_map.insert({name, Ir::ADD});
                    }
                    return true;
                }
            }
        }
        return false;
    }

    bool judge_negetive(const Ir::shared_ptr<Ir::Reg>& reg, Ir::Function* func) {
        if (reg->get_id() < 0) {
            return false;
        }
        auto instr = func->def_place[reg->get_id()];
        Instanceof(binary, Ir::BinaryInstr*, instr) {
            if (binary->get_binary_type() == Ir::SUB) {
                return is_const_zero(binary->get_src_1()) && binary->get_src_2()->get_num().int_num == 1;
            }
        }
        return false;
    }

    bool judge_and(const Ir::shared_ptr<Ir::Reg>& reg, Ir::Function* func) {
        if (reg->get_id() < 0) {
            return false;
        }
        auto instr = func->def_place[reg->get_id()];
        Instanceof(call, Ir::CallInstr*, instr) {
            if (bit_map.count(call->get_function()->get_name())) {
                auto x = bit_map.find(call->get_function()->get_name())->second;
                return x == Ir::ADD;
            }
        }
        return false;
    }

    bool judge_xor(const Ir::shared_ptr<Ir::Reg>& reg, Ir::Function* func) {
        if (reg->get_id() < 0) {
            return false;
        }
        auto instr = func->def_place[reg->get_id()];
        Instanceof(call, Ir::CallInstr*, instr) {
            if (bit_map.count(call->get_function()->get_name())) {
                auto x = bit_map.find(call->get_function()->get_name())->second;
                if (x == Ir::XOR) {
                    auto xor_instr = func->def_place[call->get_params()[0]->get_id()];
                    Instanceof(xor_call, Ir::CallInstr*, instr) {
                        if (bit_map.count(xor_call->get_function()->get_name())) {
                            auto t = bit_map.find(xor_call->get_function()->get_name())->second;
                            return t == Ir::XOR;
                        }
                    }
                }
            }
        }
        return false;
    }

    bool analysis_not(Ir::Function * func) {
        if (func->is_void() || !(func->get_params().size() == 1) || !func->pure) {
            return false;
        }
        auto it = func->get_params().begin();
        auto first = it->get();
        auto ret_bb = func->get_bbs().back().get();
        Instanceof(ret, Ir::Ret*, ret_bb->get_instrs().back().get()) {
            auto src = ret->get_src();
            auto def = func->def_place[src->get_id()];
            Instanceof(add, Ir::BinaryInstr*, def) {
                if (add->get_binary_type() == Ir::SUB && judge_negetive(add->get_src_1(), func)
                    && add->get_src_2()->get_id() == first->get_id()) {
                    auto name = func->get_name();
                    if (bit_map.count(name) == 0) {
                        bit_map.insert({name, Ir::SUB});
                    }
                    return true;
                }
            }
        }
        return false;
    }

    bool analysis_xor(Ir::Function * func) {
        if (func->is_void() || !(func->get_params().size() == 2) || !func->pure) {
            return false;
        }
        auto it = func->get_params().begin();
        auto first = it->get();
        it++;
        auto second = it->get();
        auto ret_bb = func->get_bbs().back().get();
        Instanceof(ret, Ir::Ret*, ret_bb->get_instrs().back().get()) {
            auto src = ret->get_src();
            auto def = func->def_place[src->get_id()];
            Instanceof(add, Ir::BinaryInstr*, def) {
                if (add->get_binary_type() == Ir::SUB && judge_and(add->get_src_2(), func)
                    ) {
                    auto name = func->get_name();
                    if (bit_map.count(name) == 0) {
                        bit_map.insert({name, Ir::XOR});
                    }
                    return true;
                }
            }
        }
        return false;
    }

    bool analysis_or(Ir::Function * func) {
        if (func->is_void() || !(func->get_params().size() == 2) || !func->pure) {
            return false;
        }
        auto it = func->get_params().begin();
        auto first = it->get();
        it++;
        auto second = it->get();
        auto ret_bb = func->get_bbs().back().get();
        Instanceof(ret, Ir::Ret*, ret_bb->get_instrs().back().get()) {
            auto src = ret->get_src();
            auto def = func->def_place[src->get_id()];
            Instanceof(call, Ir::CallInstr*, def) {
                if (call->get_params().size() == 2 &&
                    judge_xor(std::dynamic_pointer_cast<Ir::Reg>(call->get_params()[0]), func)){
                    auto name = func->get_name();
                    if (bit_map.count(name) == 0) {
                        bit_map.insert({name, Ir::OR});
                    }
                    return true;
                }
            }
        }
        return false;
    }


    void bitwise_opt(Ir::CompileUnit* cu) {
        for (auto &func: cu->get_functions()) {
             if (analysis_and(func.get()) || analysis_not(func.get()) ||
                analysis_or(func.get()) || analysis_xor(func.get())) {
                continue ;
             }
             Ir::Factory::reset(func.get());
             for (const auto &bb: func->get_bbs()) {
                auto it = bb->get_instrs().begin();
                while(it != bb->get_instrs().end()) {
                    auto instr = (*it).get();
                    Instanceof(call, Ir::CallInstr*, instr) {
                        if (analysis_and(call->get_function())) {
                            auto reg = Ir::Factory::get_reg(call->get_function()->get_type());
                            auto first = std::dynamic_pointer_cast<Ir::Reg>(call->get_params()[0]);
                            auto second = std::dynamic_pointer_cast<Ir::Reg>(call->get_params()[1]);
                            auto binary = Ir::Factory::get_binary(reg, first, second, Ir::ADD);
                            replace_all_use(call->get_dst(), binary->get_dst(), func.get());
                            call->remove_use_def();
                            binary->set_belongto_bb(bb.get());
                            binary->add_use_def();
                            *it = std::move(binary);
                        } else if (analysis_xor(call->get_function())){
                            auto reg = Ir::Factory::get_reg(call->get_function()->get_type());
                            auto first = std::dynamic_pointer_cast<Ir::Reg>(call->get_params()[0]);
                            auto second = std::dynamic_pointer_cast<Ir::Reg>(call->get_params()[1]);
                            auto binary = Ir::Factory::get_binary(reg, first, second, Ir::ADD);
                            auto ans = Ir::Factory::get_reg(call->get_function()->get_type());;
                            auto answer = Ir::Factory::get_binary(ans, Ir::Factory::get_zero_reg(reg->get_type()), reg, Ir::SUB);
                            replace_all_use(call->get_dst(),answer->get_dst(), func.get());
                            call->remove_use_def();
                            binary->set_belongto_bb(bb.get());
                            answer->set_belongto_bb(bb.get());
                            answer->add_use_def();
                            binary->add_use_def();
                            *it = std::move(answer);
                            it = bb->get_instrs().insert(it, std::move(binary));
                        } else if (analysis_or(call->get_function())) {
                            auto zero = Ir::Factory::get_zero_reg(call->get_dst()->get_type());
                            replace_all_use(call->get_dst(), zero, func.get());
                            call->remove_use_def();
                            it = bb->get_instrs().erase(it);
                        } else if (analysis_not(call->get_function())) {
                            auto reg = Ir::Factory::get_reg(call->get_function()->get_type());
                            Ir::RegConst neg;
                            neg.int_num = -1;
                            auto first = Ir::Factory::get_const_reg(Ir::Int, neg);
                            auto second = std::dynamic_pointer_cast<Ir::Reg>(call->get_params()[0]);
                            auto binary = Ir::Factory::get_binary(reg, first, second, Ir::SUB);
                            replace_all_use(call->get_dst(), binary->get_dst(), func.get());
                            call->remove_use_def();
                            binary->set_belongto_bb(bb.get());
                            binary->add_use_def();
                            *it = std::move(binary);
                        }
                    }
                    it++;
                }
             }
             func->set_bb_max_num(Ir::Factory::bb_num);
             func->set_reg_max_num(Ir::Factory::reg_num);
        }
    }
}