//
// Created by shliba on 2023/7/7.
//
#include "opt.h"

namespace opt {

shared_ptr<Ir::Reg> build_convert(Ir::ScalaType dst_type, shared_ptr<Ir::Reg>& src, Ir::BasicBlock* bb, list<unique_ptr<Ir::Instruction>>::iterator& it) {
    if (!src->is_constant()) {
        auto new_reg = Ir::Factory::get_reg(dst_type);
        auto convert = Ir::Factory::get_convert(new_reg, src);
        auto ptr = convert.get();
        ptr->set_belongto_bb(bb);
        it = bb->get_instrs().insert(it, std::move(convert));
        ptr->add_use_def();
        return new_reg;
    } else {
        if (dst_type == Ir::Int) {
            auto num = src->get_num();
            if (src->get_type() == Ir::Float) {
                num.int_num = (int) num.float_num;
            }
            src->set_num(num);
        } else if (dst_type == Ir::Float)  {
            auto num = src->get_num();
            float t = num.int_num;
            num.float_num = t;
            src->set_num(num);
        }
        src->set_type(dst_type);
        return src;
    }
}

void convert_pass(Ir::CompileUnit *compileUnit) {

    for (const auto &func : compileUnit->get_functions()){
        Ir::Factory::reset(func.get());
        for (auto &bb : func->get_bbs()) {
            auto it = bb->get_instrs().begin();
            auto end = bb->get_instrs().end();
            while (it != end) {
                Instanceof(binary, Ir::BinaryInstr*, (*it).get()) {
                    auto dst = binary->get_dst();
                    auto src1 = binary->get_src_1();
                    auto src2 = binary->get_src_2();
                    switch (binary->get_binary_type()) {
                    case Ir::ADD:
                    case Ir::SUB:
                    case Ir::SDIV:
                    case Ir::MUL:
                    case Ir::SREM: /* {
                        if (dst->get_type() != src1->get_type()) {

                            auto new_reg = build_convert(dst->get_type(), src1, bb.get(), it);
                            binary->change_use_name(src1->get_id(), new_reg);
                        }
                        if (dst->get_type() != src2->get_type()) {

                            auto new_reg = build_convert(dst->get_type(), src2, bb.get(), it);

                            binary->change_use_name(src2->get_id(), new_reg);
                        }
                        break ;
                    }*/
                    {
                        if (src1->get_type() != src2->get_type()) {
                            if (src1->get_type() > src2->get_type()) {
                                auto new_reg = build_convert(src1->get_type(), src2, bb.get(), it);
                                binary->change_use_name(src2, new_reg);
                                binary->get_dst()->set_type(src1->get_type());
                            } else {
                                auto new_reg = build_convert(src2->get_type(), src1, bb.get(), it);
                                binary->change_use_name(src1, new_reg);
                                binary->get_dst()->set_type(src2->get_type());
                            }
                        } else if (src1->get_type() != dst->get_type()) {
                            if (src1->get_type() > dst->get_type()){
                                dst->set_type(src1->get_type());
                            } else {
                                if (dst->get_type() != src1->get_type()) {

                                    auto new_reg = build_convert(dst->get_type(), src1, bb.get(), it);
                                    binary->change_use_name(src1, new_reg);
                                }
                                if (dst->get_type() != src2->get_type()) {

                                    auto new_reg = build_convert(dst->get_type(), src2, bb.get(), it);

                                    binary->change_use_name(src2, new_reg);
                                }
                            }
                        }
                        break ;
                    }
                    case Ir::EQ:
                    case Ir::NE:
                    case Ir::SGT:
                    case Ir::SGE:
                    case Ir::SLE:
                    case Ir::SLT:  {
                        if (src1->get_type() != src2->get_type()) {
                            if (src1->get_type() > src2->get_type()) {

                                auto new_reg = build_convert(src1->get_type(), src2, bb.get(), it);

                                binary->change_use_name(src2->get_id(), new_reg);
                            } else {

                                auto new_reg = build_convert(src2->get_type(), src1, bb.get(), it);


                                binary->change_use_name(src1->get_id(), new_reg);
                            }
                        }
                        break ;
                    }
                    default: {
                        assert(false);
                    }
                    }
                } else Instanceof(store, Ir::StoreInstr*, (*it).get()){
                    auto val = store->get_val();
                    auto addr = store->get_addr();
                    if ( val->get_type() != addr->get_type()) {
                        auto new_reg = build_convert(addr->get_type(), val, bb.get(), it);
                        store->change_use_name(store->get_val()->get_id(), new_reg);
                    }
                }  else Instanceof(call, Ir::CallInstr*, (*it).get()) {
                    int i = 0;
                    for (auto &param : call->get_function()->get_params()) {
                        Instanceof(reg, Ir::Reg*, param.get()) {
                            Instanceof(p_reg, Ir::Reg*, call->get_params()[i].get()) {
                                if (reg->get_type() != p_reg->get_type()) {

                                    shared_ptr<Ir::Reg> src = std::dynamic_pointer_cast<Ir::Reg>(call->get_params()[i]);

                                    auto new_reg = build_convert(reg->get_type(), src, bb.get(), it);

                                    auto id = src->get_id();
                                    call->get_params()[i] = new_reg;

                                    if (id >= 0) {
                                        call->get_belongto_bb()->get_belongto_func()->use_place[id].erase(call);
                                        if (new_reg->get_id() >= 0) {
                                            call->get_belongto_bb()->get_belongto_func()->use_place[new_reg->get_id()].insert(call);
                                        }
                                    }
                                    it++;
                                }
                            }
                        }
                        i++;
                    }
                }/*else Instanceof(load, Ir::LoadInstr*, (*it).get()) {
                    auto dst = load->get_dst();
                    auto addr = store->get_addr();
                    if (dst->get_type() != addr->get_type()) {
                        auto new_reg = Ir::Factory::get_reg(addr->get_type());
                        auto convert = Ir::Factory::get_convert(dst, new_reg);
                                                    auto ptr = convert.get();
                            ptr->set_belongto_bb(bb.get());
                            it = bb->get_instrs().insert(it, std::move(convert));
                            ptr->add_use_def();

                    }
                }*/

                it++;
            }
        }
        func->set_reg_max_num(Ir::Factory::reg_num);
        func->set_bb_max_num(Ir::Factory::bb_num);
    }

}
}