//
// Created by shliba on 2022/11/17.
//

#include "opt.h"
namespace opt {


void solve_phi(Ir::BasicBlock * old_bb, Ir::BasicBlock* new_bb) {
    for (auto &item : old_bb->succ) {
        for (const auto &instr : item->get_instrs()) {
            Instanceof(phi, Ir::PhiInstr*, instr.get()) {
                auto val =  phi->get_bb_2_value()[old_bb];
                phi->get_bb_2_value()[new_bb] = val;
                phi->get_bb_2_value().erase(old_bb);
            } else {
                break;
            }
        }
        item->prev.erase(old_bb);
        new_bb->succ.insert(item);
    }
    old_bb->succ.clear();
    old_bb->succ.insert(new_bb);
}

int get_func(Ir::Function* func) {
    int num;
    for (const auto &item: func->get_bbs()) {
        for (const auto &item: item->get_instrs()) {
            num++;
        }
    }
    return num;
}

void func_inline(Ir::CompileUnit* compUnit) {
    for (auto &func : compUnit->get_functions()) {

        auto bb_it = func->get_bbs().begin();
        Ir::Factory::reset(func.get());

        while (bb_it != func->get_bbs().end()) {
            auto instr_it = (*bb_it)->get_instrs().begin();
            while (instr_it != (*bb_it)->get_instrs().end()) {

                Instanceof(call, Ir::CallInstr*, (*instr_it).get()) {
                    auto target_func = call->get_function();
                    if (target_func->get_bbs().size() > 20
                        || ((*bb_it)->get_belong_to_loop() != nullptr && target_func->get_params().size() < 3
                           && get_func(target_func) > 20 && get_func(target_func)< 1000) || target_func->get_loops().size() > 1) {
                        instr_it++;
                        target_func->set_recur(true);
                        continue ;
                    }
                    if (!target_func->is_lib() && !target_func->is_recur()) {
                        auto exit_bb = Ir::Factory::get_noadd_new_bb((*bb_it)->get_belong_to_loop(), func.get());

                        solve_phi(bb_it->get(), exit_bb.get());

                        // splitB

                        instr_it++;
                        while (instr_it != (*bb_it)->get_instrs().end()) {
                            exit_bb->push_back(*instr_it);
                            instr_it = (*bb_it)->get_instrs().erase(instr_it);
                        }


                        std::unordered_map<Ir::BasicBlock*, shared_ptr<Ir::Reg>> values;
                        if (!target_func->is_void()) {
                            auto reg = call->get_dst();
                            auto phi = Ir::Factory::get_phi(reg);
                            phi->set_belongto_bb(exit_bb.get());
                            exit_bb->get_instrs().push_front(std::move(phi));
                        }
                        instr_it--;
                        //copy


                        std::unordered_map<int, std::shared_ptr<Ir::Type>> regs_map;
                        std::unordered_map<Ir::BasicBlock*, Ir::BasicBlock*> bbs_map;
                        auto it = target_func->get_params().begin();
                        for (int i = 0 ; i < call->get_params().size(); ++i, it++) {
                            regs_map.insert({(*it)->get_id(), call->get_params()[i]});
                        }
                        list<std::unique_ptr<Ir::BasicBlock>> new_bbs;
                        // copy_bbs
                        for (const auto &old_bb : target_func->get_bbs()) {
                            auto new_bb = Ir::Factory::get_noadd_new_bb(nullptr, func.get());
                            if (old_bb->get_belong_to_loop() == nullptr) {
                                new_bb->set_belong_to_loop((*bb_it)->get_belong_to_loop());
                            }

                            bbs_map.insert({old_bb.get(), new_bb.get()});
                            new_bbs.push_back(std::move(new_bb));
                        }




                        //copy_regs
                        for (const auto &old_bb : target_func->get_bbs()) {
                            for (const auto &old_instr : old_bb->get_instrs()) {
                                Instanceof(def_inst, Ir::DefInstr*, old_instr.get()) {
                                    auto dst = def_inst->def();
                                    if (dst == nullptr) {
                                        continue ;
                                    }
                                    auto new_dst = Ir::Factory::copy_def(dst);
                                    regs_map.insert({dst->get_id(), new_dst});
                                }

                            }
                        }

                        for (const auto &old_bb : target_func->get_bbs()) {
                            auto new_bb = bbs_map[old_bb.get()];
                            for (const auto &old_instr : old_bb->get_instrs()) {
                                auto new_instr = old_instr->clone(regs_map, bbs_map);
                                new_instr->set_belongto_bb(new_bb);
                                Instanceof(ret_instr, Ir::Ret*, new_instr) {
                                    if (!target_func->is_void()) {
                                        values.insert({new_bb, ret_instr->get_src()});
                                    }
                                    Ir::Factory::build_branch(exit_bb.get(), *new_bb);
                                    delete ret_instr;
                                    continue;
                                }
                                auto unique_instr = std::unique_ptr<Ir::Instruction>(new_instr);
                                new_bb->push_back(unique_instr);
                            }
                        }

                        //delete and combine
                        (*instr_it)->remove_use_def();
                        (*bb_it)->get_instrs().erase(instr_it);
                        if (!target_func->is_void()) {
                            Instanceof(phi, Ir::PhiInstr*, exit_bb->get_instrs().begin()->get()) {
                                for (const auto &item : values) {
                                    phi->get_bb_2_value().insert(item);
                                }
                                phi->add_use_def();
                            }
                        }
                        (*bb_it)->set_terminal(false);
                        Ir::Factory::build_branch((*new_bbs.begin()).get(), **bb_it);
                        bb_it++;
                        auto new_bb_it = new_bbs.begin();
                        while (new_bb_it != new_bbs.end()) {
                            bb_it = func->get_bbs().insert(bb_it, std::move(*new_bb_it));
                            new_bb_it = new_bbs.erase(new_bb_it);
                            bb_it++;
                        }
                        instr_it = exit_bb->get_instrs().begin();
                        bb_it = func->get_bbs().insert(bb_it, std::move(exit_bb));
                        continue;
                    }

                }
                instr_it++;
            }
            bb_it++;
        }
        func->set_reg_max_num(Ir::Factory::reg_num);
        func->set_bb_max_num(Ir::Factory::bb_num);
        func->cfg->build();
    }
    auto func_it = compUnit->get_functions().begin();
    while (func_it != compUnit->get_functions().end()) {
        auto func = func_it->get();
        if (func->get_name() == "reduce") {
            func_it++;
            continue ;
        }
        if (!func->is_recur() && func->get_name() != "main" && !func->is_lib()) {
            (*func_it)->get_bbs().clear();
            (*func_it)->use_place.clear();
            (*func_it)->get_bbs().clear();
            (*func_it)->def_place.clear();
            (*func_it)->get_loops().clear();
            func_it = compUnit->get_functions().erase(func_it);
        } else {
            func_it++;
        }
    }
    info("end func inline");


}
}