//
// Created by 11069 on 4/7/2023.
//

// 第一阶段 统计所有全局变量,及其初始值
// 第二阶段 遍历所有函数,记录只被main函数使用过的全局变量
// 第三阶段 对main函数中的全局变量进行局部化(loadGlobal 删除,然后拷贝传播初始值)
// 第四阶段 统计所有没有被修改过的全局变量
// 第五阶段 将所有函数中对没有修改过的全局变量进行常量传播

// 该pass 在memtoreg之前,不需要进行basicblock和def/use维护
// 该pass不一定能优化成功,但是一旦将全局变量局部化,能省不少访存,copy propagation秒了


#include "opt.h"

using namespace Ir;
using namespace std;
namespace opt {

void global_to_local(CompileUnit *compileUnit) {
    unordered_map<string, shared_ptr<Reg>> main_used_global;
    unordered_map<string, shared_ptr<Reg>> func_used_global;
    // init main_used_global
    Function *main_func;

    for (auto &func : compileUnit->get_functions()) { // get_function should have returned ptr
        string name = func->get_name();
        if (name == "main") {
            main_func = func.get();
            for (auto &bb : func->get_bbs()) {
                for (auto &inst : bb->get_instrs()) {
                    Instanceof(loadGlobal, LoadGlobal *, inst.get()) {
                        main_used_global[loadGlobal->get_addr()->get_global_name()] = loadGlobal->get_addr();
                    }
                    else Instanceof(storeGlobal, StoreGlobal *, inst.get()) {
                        main_used_global[storeGlobal->get_addr()->get_global_name()] = storeGlobal->get_addr();
                    }
                }
            }
        } else {
            if(func->is_lib())
                continue ;
            for (auto &bb : func->get_bbs()) {
                for (auto &inst : bb->get_instrs()) {
                    Instanceof(loadGlobal, LoadGlobal *, inst.get()) {
                            func_used_global[loadGlobal->get_addr()->get_global_name()] = loadGlobal->get_addr();
                    }
                    else Instanceof(storeGlobal, StoreGlobal *, inst.get()) {
                            func_used_global[storeGlobal->get_addr()->get_global_name()] = storeGlobal->get_addr();
                    }
                }
            }
        }
    }
    Factory::reset(main_func);
    unordered_map<string, shared_ptr<MemObject>> global_to_local;
    auto &head = main_func->get_bbs().front();
    for (auto &[name, var] : main_used_global) {
        if(func_used_global.count(name))
            continue ;
        auto tmp_instr = Factory::get_alloca(var->get_type());
        Instanceof(alloca_instr, AllocaInstr *, tmp_instr.get()) {
            shared_ptr<MemObject> local_var = alloca_instr->get_addr();
            shared_ptr<Reg> val_reg = Factory::get_const_reg(var->get_type(), var->get_num());
            auto* store_instr = new StoreInstr(val_reg, local_var);
            // add store
            head->push_front(store_instr);
            // add alloca
            head->push_front(std::move(tmp_instr));
            global_to_local[name] = local_var;
        }
    }
    main_func->set_reg_max_num(Factory::reg_num);
    main_func->set_bb_max_num(Factory::bb_num);

    Factory::reset(main_func);
    for (auto &bb : main_func->get_bbs()) {
        for (auto iter = bb->get_instrs().begin(); iter != bb->get_instrs().end();) {
            // 对于 loadGlobal,转换其为load local val
            // 对于storeGlobal,转换为 store local val
            auto inst = iter->get();
            Instanceof(loadGlobal, LoadGlobal *, inst) {
                if (global_to_local.count(loadGlobal->get_addr()->get_global_name())) {
                    unique_ptr<Instruction> load_instr = make_unique<LoadInstr>(loadGlobal->get_dst(), global_to_local[loadGlobal->get_addr()->get_global_name()]);
                    loadGlobal->remove_use_def();
                    load_instr->set_belongto_bb(bb.get());
                    load_instr->add_use_def();

                    iter = bb->get_instrs().erase(iter);
                    iter = bb->get_instrs().insert(iter, std::move(load_instr));
                    continue;
                }
            }
            else Instanceof(storeGlobal, StoreGlobal *, inst) {
                if (global_to_local.count(storeGlobal->get_addr()->get_global_name())) {
                    unique_ptr<Instruction> store_instr = make_unique<StoreInstr>(storeGlobal->get_val(), global_to_local[storeGlobal->get_addr()->get_global_name()]);
                    storeGlobal->remove_use_def();
                    store_instr->set_belongto_bb(bb.get());
                    store_instr->add_use_def();

                    iter = bb->get_instrs().erase(iter);
                    iter = bb->get_instrs().insert(iter, std::move(store_instr));

                    continue;
                }
            }
            iter++;
        }
    }

    main_func->set_reg_max_num(Factory::reg_num);
    main_func->set_bb_max_num(Factory::bb_num);
}

void global_to_const(CompileUnit *compileUnit) {
    unordered_map<string, shared_ptr<Reg>> used_global_var;
    unordered_set<string> modified_global_var;
    for (auto &func : compileUnit->get_functions()) {
        if(func->is_lib())
            continue ;
        for (auto &bb : func->get_bbs()) {
            for (auto &inst : bb->get_instrs()) {
                Instanceof(loadGlobal, LoadGlobal *, inst.get()) {
                    used_global_var[loadGlobal->get_addr()->get_global_name()] = loadGlobal->get_addr();
                }
                else Instanceof(storeGlobal, StoreGlobal *, inst.get()) {
                    modified_global_var.insert(storeGlobal->get_addr()->get_global_name());
                }
            }
        }
    }

    unordered_map<string, shared_ptr<Reg>> unmodified_global_var;
    for (auto &[name, reg] : used_global_var) {
        if (!modified_global_var.count(name)) {
            unmodified_global_var[name] = reg;
        }
    }

    for (auto &func : compileUnit->get_functions()) {
        if(func->is_lib())
            continue ;
        Factory::reset(func.get());
        for (auto &bb : func->get_bbs()) {
            for (auto iter = bb->get_instrs().begin();iter != bb->get_instrs().end();) {
                auto &inst = *iter;
                Instanceof(loadGlobal, LoadGlobal *, inst.get()) {
                    if (unmodified_global_var.count(loadGlobal->get_addr()->get_global_name())) {
                        shared_ptr<Reg> global_var = unmodified_global_var[loadGlobal->get_addr()->get_global_name()];
                        shared_ptr<Reg> const_reg = Factory::get_const_reg(global_var->get_type(), global_var->get_num());
                        copy_propagation(func->use_place, loadGlobal->get_dst()->get_id(), const_reg);
                        inst->remove_use_def();
                        iter = bb->get_instrs().erase(iter);
                    } else iter++;
                } else iter++;
            }
        }
        func->set_reg_max_num(Factory::reg_num);
        func->set_bb_max_num(Factory::bb_num);
    }
}

void global_var_pass(CompileUnit *compileUnit) {
    global_to_local(compileUnit);
    global_to_const(compileUnit);
    // 先global to local,array ssa可以解决常数传播问题,并回收多余的地址
}
} // namespace opt