//
// Created by 11069 on 5/7/2023.
//

// gvn global value numbering
// gcm global code motion
// 很知名的优化算法,不需要多说,终于可以让我来实现了吗 我雷诺其卡拉
#include "opt.h"
#include "tuple"
namespace std{
template<class T1> struct hash<vector<T1>>
{
    size_t operator() (vector<T1> const& vec) const {
        size_t seed =vec.size();
        for(auto& i: vec) {
            seed ^= hash<T1>()(i) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
        }
        return seed;
    }
};
}

namespace opt {
using namespace Ir;
using std::unordered_map;
using std::pair;
using std::shared_ptr;
using std::unique_ptr;
using std::string;
using std::tuple;
using std::vector;

static bool array_ssa_gcm_swtich = true;

struct pair_hash
{
    template <class T1, class T2>
    std::size_t operator() (const pair<T1, T2>& p) const
    {
        auto h1 = std::hash<T1>()(p.first);
        auto h2 = std::hash<T2>()(p.second);
        return h1 * 1221821 + h2*31;
    }
};

struct tuple_hash
{
    template <class T1, class T2, class T3>
    std::size_t operator() (const std::tuple<T1, T2, T3>& r) const
    {
        return std::hash<T1>()(std::get<0>(r)) * 1221821 + std::hash<T2>()(std::get<1>(r)) * 1926817 +
               std::hash<T3>()(std::get<2>(r)) * 31;
    }
};

    // data struct for gvn
    // Binary table
    // 如果两个都是常数,"1c" + const val为key
    // 两个都不是常数,id为 key
    // 一个常数一个不是常数 id  和  "c" + const val为key
    // key放入的时候记得交换,使其有序(对于mul , plus 等有交换律的,以及对于没有交换律,但是可以通过变符号实现交换的)
    unordered_map<tuple<BinaryType,string,string>,BinaryInstr *,tuple_hash> binary_table;
    // unordered_map<string, LoadGlobal *> loadGlobal_table;
    // 考虑到右值可能是常数,类似binary对key进行建立
    // 常数就是 value,reg是 i + id
    unordered_map<string,Convert *> convert_table;
    // loadArray
    unordered_map<pair<string,vector<string>>,LoadArray *,pair_hash> loadArray_table;
    // loadVector
    // vector是局部就是id,是全局就是
    unordered_map<pair<string,vector<string>>,LoadVector *,pair_hash> loadVector_table;
    // call
    unordered_map<pair<string,vector<string>>,CallInstr *, pair_hash> call_table;

    // mem read
    unordered_map<tuple<string,string,bool>,MemRead *,tuple_hash> mem_read_table;

    // phi
    unordered_map<pair<string,vector<string>>,PhiInstr *,pair_hash> phi_table; // loop unroll 可能产生相同的phi


    string getHash(Type * type) {
        string tmp = "1c";
        Instanceof(reg,Reg *,type) {
            if(reg->get_id() >= 0) {
                return std::to_string(reg->get_id());
            } else {
                if(reg->is_global()) {
                    return reg->get_global_name();
                } else {
                    bool is_int = reg->get_type() == Int;
                    if(is_int) {
                        int iv = reg->get_num().int_num;
                        tmp.append(std::to_string(iv));
                        return tmp;
                    } else {
                        double fv = reg->get_num().float_num;
                        tmp.append(std::to_string(fv));
                        return tmp;
                    }
                }
            }
        } Instanceof(vector,Vector*,type) {
            if(vector->get_id() >= 0) {
                return std::to_string(vector->get_id());
            } else {
                return vector->get_name();
            }
        } Instanceof(memObject,MemObject*,type) {
            return std::to_string(memObject->get_id());
        }
    }

    void binarySwap(BinaryInstr* binaryInstr,const shared_ptr<Reg>& src_1,const shared_ptr<Reg>& src_2) {
        auto tmp_reg = Factory::get_reg(src_1->get_type());
        binaryInstr->change_use_name(src_1->get_id(),tmp_reg);
        binaryInstr->change_use_name(src_2->get_id(),src_1);
        binaryInstr->change_use_name(tmp_reg->get_id(),src_2);
    }

    void orderBinary(BinaryInstr * binaryInstr,BinaryType binaryType) {
        auto src_1 = binaryInstr->get_src_1();
        auto src_2 = binaryInstr->get_src_2();
        bool two_reg = src_1->get_id() >= 0 && src_2->get_id() >= 0;
        bool mixture = (src_1->get_id() >= 0 && src_2->get_id() < 0) || (src_2->get_id() >= 0 && src_1->get_id() < 0);
        if(two_reg) {
            if(src_1->get_id() < src_2->get_id()) {
                binarySwap(binaryInstr,src_1,src_2);
                binaryInstr->setBinaryType(binaryType);
            }
        } else if(mixture) {
            if(src_1->get_id() < 0) {
                binarySwap(binaryInstr,src_1,src_2);
                binaryInstr->setBinaryType(binaryType);
            }
        } else {
            if((src_1->get_type() == Int && src_1->get_num().int_num < src_2->get_num().int_num) ||
                (src_1->get_type() == Float && src_1->get_num().float_num < src_2->get_num().float_num)) {
                binarySwap(binaryInstr,src_1,src_2);
                binaryInstr->setBinaryType(binaryType);
            }
        }
    }

    // 如果两个都是常数,"c" + const val为key
    // 两个都不是常数,id为 key
    // 一个常数一个不是常数 id  和  "c" + const val为key, const val 必须为正数
    // key放入的时候记得交换,使其有序(对于mul , plus 等有交换律的,以及对于没有交换律,但是可以通过变符号实现交换的)
    // 同时,简化binary inst,后期方便点)
    tuple<BinaryType,string,string> binary_hash(BinaryInstr* binaryInstr) {
        // const 改为正数,只对add sub生效

        switch (binaryInstr->get_binary_type()) {
        // 对于加法,只针对有一个负数的情况,所以先order,然后看最后一个是不是负数,如果是负数,修改第二个src,add改sub,然后交给第二个SUB了
        case ADD:
            orderBinary(binaryInstr, binaryInstr->get_binary_type());
            if(binaryInstr->get_src_2()->get_id() < 0) {
                if (binaryInstr->get_src_2()->get_type() == Int && binaryInstr->get_src_2()->get_num().int_num < 0) {
                    binaryInstr->get_src_2()->get_num().int_num = abs(binaryInstr->get_src_2()->get_num().int_num);
                    binaryInstr->setBinaryType(SUB);
                } else if (binaryInstr->get_src_2()->get_type() == Float && binaryInstr->get_src_2()->get_num().float_num < 0) {
                    binaryInstr->get_src_2()->get_num().float_num = abs(binaryInstr->get_src_2()->get_num().float_num);
                    binaryInstr->setBinaryType(SUB);
                }
            }
            break;
        case SUB:
            if(binaryInstr->get_src_2()->get_id() < 0) {
                if (binaryInstr->get_src_2()->get_type() == Int && binaryInstr->get_src_2()->get_num().int_num < 0) {
                    binaryInstr->get_src_2()->get_num().int_num = abs(binaryInstr->get_src_2()->get_num().int_num);
                    binaryInstr->setBinaryType(ADD);
                } else if (binaryInstr->get_src_2()->get_type() == Float && binaryInstr->get_src_2()->get_num().float_num < 0) {
                    binaryInstr->get_src_2()->get_num().float_num = abs(binaryInstr->get_src_2()->get_num().float_num);
                    binaryInstr->setBinaryType(ADD);
                }
            }
            break;
        default:
            break;
        }

        switch (binaryInstr->get_binary_type()) {
        case ADD:
        case MUL:
        case NE:
        case EQ:
            orderBinary(binaryInstr, binaryInstr->get_binary_type());
            break;
        case SLE:
            orderBinary(binaryInstr, SGE);
            break;
        case SLT:
            orderBinary(binaryInstr, SGT);
            break;
        case SGT:
            orderBinary(binaryInstr, SLT);
            break;
        case SGE:
            orderBinary(binaryInstr, SLE);
            break;
        default:
        break;
    }


    auto hash = tuple<BinaryType,string,string>(binaryInstr->get_binary_type(),
                      getHash(binaryInstr->get_src_1().get()),
                          getHash(binaryInstr->get_src_2().get()));
    return hash;
    }

    shared_ptr<Reg> calcBinary(BinaryInstr * binaryInstr) {
        shared_ptr<Reg> ret = nullptr;
        bool is_int = binaryInstr->get_dst()->get_type() == Int ||  binaryInstr->get_dst()->get_type() == Bool;
        switch (binaryInstr->get_binary_type()) {
        case ADD:
            if(is_int) {
                int val = binaryInstr->get_src_1()->get_num().int_num + binaryInstr->get_src_2()->get_num().int_num;
                RegConst reg;
                reg.int_num = val;
                ret = Factory::get_const_reg(Int,reg);
            } else {
                float val = binaryInstr->get_src_1()->get_num().float_num + binaryInstr->get_src_2()->get_num().float_num;
                RegConst reg;
                reg.float_num = val;
                ret = Factory::get_const_reg(Float,reg);
            }
            break;
        case SUB:
            if(is_int) {
                int val = binaryInstr->get_src_1()->get_num().int_num - binaryInstr->get_src_2()->get_num().int_num;
                RegConst reg;
                reg.int_num = val;
                ret = Factory::get_const_reg(Int,reg);
            } else {
                float val = binaryInstr->get_src_1()->get_num().float_num - binaryInstr->get_src_2()->get_num().float_num;
                RegConst reg;
                reg.float_num = val;
                ret = Factory::get_const_reg(Float,reg);
            }
            break;
        case MUL:
            if(is_int) {
                int val = binaryInstr->get_src_1()->get_num().int_num * binaryInstr->get_src_2()->get_num().int_num;
                RegConst reg;
                reg.int_num = val;
                ret = Factory::get_const_reg(Int,reg);
            } else {
                float val = binaryInstr->get_src_1()->get_num().float_num * binaryInstr->get_src_2()->get_num().float_num;
                RegConst reg;
                reg.float_num = val;
                ret = Factory::get_const_reg(Float,reg);
            }
            break;
        case SDIV:
            if(is_int) {
                int val = binaryInstr->get_src_1()->get_num().int_num / binaryInstr->get_src_2()->get_num().int_num;
                RegConst reg;
                reg.int_num = val;
                ret = Factory::get_const_reg(Int,reg);
            } else {
                float val = binaryInstr->get_src_1()->get_num().float_num / binaryInstr->get_src_2()->get_num().float_num;
                RegConst reg;
                reg.float_num = val;
                ret = Factory::get_const_reg(Float,reg);
            }
            break;
        case SREM:
            if(is_int) {
                int val = binaryInstr->get_src_1()->get_num().int_num % binaryInstr->get_src_2()->get_num().int_num;
                RegConst reg;
                reg.int_num = val;
                ret = Factory::get_const_reg(Int,reg);
            }
            break;
        case SLT:
            if(is_int) {
                int val = binaryInstr->get_src_1()->get_num().int_num < binaryInstr->get_src_2()->get_num().int_num ? 1:0;
                RegConst reg;
                reg.int_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            } else {
                float val = binaryInstr->get_src_1()->get_num().float_num < binaryInstr->get_src_2()->get_num().float_num ? 1:0;
                RegConst reg;
                reg.float_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            }
            break;
        case SLE:
            if(is_int) {
                int val = binaryInstr->get_src_1()->get_num().int_num <= binaryInstr->get_src_2()->get_num().int_num ? 1:0;
                RegConst reg;
                reg.int_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            } else {
                float val = binaryInstr->get_src_1()->get_num().float_num <= binaryInstr->get_src_2()->get_num().float_num ? 1:0;
                RegConst reg;
                reg.float_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            }
            break;
        case SGE:
            if(is_int) {
                int val = binaryInstr->get_src_1()->get_num().int_num >= binaryInstr->get_src_2()->get_num().int_num ? 1:0;
                RegConst reg;
                reg.int_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            } else {
                float val = binaryInstr->get_src_1()->get_num().float_num >= binaryInstr->get_src_2()->get_num().float_num ? 1:0;
                RegConst reg;
                reg.float_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            }
            break;
        case SGT:
            if(is_int) {
                int val = binaryInstr->get_src_1()->get_num().int_num > binaryInstr->get_src_2()->get_num().int_num ? 1:0;
                RegConst reg;
                reg.int_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            } else {
                float val = binaryInstr->get_src_1()->get_num().float_num > binaryInstr->get_src_2()->get_num().float_num ? 1:0;
                RegConst reg;
                reg.float_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            }
            break;
        case EQ:
            if(is_int) {
                int val = binaryInstr->get_src_1()->get_num().int_num == binaryInstr->get_src_2()->get_num().int_num ? 1:0;
                RegConst reg;
                reg.int_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            } else {
                float val = binaryInstr->get_src_1()->get_num().float_num == binaryInstr->get_src_2()->get_num().float_num ? 1:0;
                RegConst reg;
                reg.float_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            }
            break;
        case NE:
            if(is_int) {
                int val = binaryInstr->get_src_1()->get_num().int_num != binaryInstr->get_src_2()->get_num().int_num ? 1:0;
                RegConst reg;
                reg.int_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            } else {
                float val = binaryInstr->get_src_1()->get_num().float_num != binaryInstr->get_src_2()->get_num().float_num ? 1:0;
                RegConst reg;
                reg.float_num = val;
                ret = Factory::get_const_reg(Bool,reg);
            }
            break;
        default:
            break;
        }
        return ret;
    }

    pair<string,vector<string>> loadArray_hash(LoadArray * loadArray) {
        auto hash = pair<string,vector<string>>(getHash(loadArray->get_addr().get()),
                                                 vector<string>());
        for(auto &each:loadArray->get_indexs()) {
            hash.second.push_back(getHash(each.get()));
        }
        return hash;
    }
    pair<string,vector<string>> loadVector_hash(LoadVector* loadVector) {
        auto hash = pair<string,vector<string>>(getHash(loadVector->get_addr().get()),
                                                 vector<string>());
        for(auto &each:loadVector->get_indexs()) {
            hash.second.push_back(getHash(each.get()));
        }
        return hash;
    }

    pair<string,vector<string>> call_hash(CallInstr* call) {
        auto hash = pair<string,vector<string>>(call->get_function()->get_name(),
                                                 vector<string>());
        for(auto &each:call->get_params()) {
            hash.second.push_back(getHash(each.get()));
        }
        return hash;
    }

//    string loadGlobal_hash(LoadGlobal* loadGlobal) {
//        return loadGlobal->get_addr()->get_global_name();
//    }

    string convert_hash(Convert* convert) {
        return getHash(convert->get_src().get());
    }

    tuple<string,string,bool> mem_read_hash(MemRead* mem_read) {
        auto hash = tuple<string,string,bool>(getHash(mem_read->get_addr().get()),
                                                      getHash(mem_read->get_scope().get()),mem_read->call_read);
        return hash;
    }

    pair<string,vector<string>> phi_hash(PhiInstr* phi) {
        vector<string> value_hash;
        for(auto &[bb,value]:phi->get_bb_2_value()) {
            value_hash.push_back(getHash(value.get()));
        }

        auto hash = pair<string,vector<string>>(std::to_string(phi->get_belongto_bb()->get_id()), value_hash);
        return hash;
    }

    shared_ptr<Type> table_get(Instruction * inst) {
        if (!inst)
            return nullptr;
        Instanceof(binary,BinaryInstr*,inst) {
            // 如果俩都是常数,直接返回不就行了
            bool two_const = binary->get_src_1()->is_constant() && binary->get_src_2()->is_constant();
            if(two_const) {
                return calcBinary(binary);
            }
            if(binary->get_src_2()->is_constant() && binary->get_src_2()->get_type() == Int
                && binary->get_src_2()->get_num().int_num == 0
                && (binary->get_binary_type() == ADD || binary->get_binary_type() == SUB) )
                return binary->get_src_1();
            auto hash = binary_hash(binary);
            if(!binary_table.count(hash)) {
                binary_table[hash] = binary;
            }
            return binary_table[hash]->get_dst();

        } /*else Instanceof(loadGlobal,LoadGlobal*,inst) {
            auto hash = loadGlobal_hash(loadGlobal);
            if(!loadGlobal_table.count(hash)) {
                loadGlobal_table[hash] = loadGlobal;
            }
            return loadGlobal_table[hash]->get_dst();
        }*/ else Instanceof(convert, Convert*,inst) {
            if(convert->get_src()->is_constant()) { // convert copy propagation
                RegConst ret;
                switch (convert->get_src()->get_type()) {
                case Int:
                    if(convert->get_dst()->get_type() == Float) { // int -> float
                        ret.float_num = float(convert->get_src()->get_num().int_num);
                    } else { // int -> bool
                        ret.int_num = convert->get_src()->get_num().int_num != 0 ? 1 : 0;
                    }
                    break;
                case Float:
                    if(convert->get_dst()->get_type() == Int) {
                        ret.int_num = int(convert->get_src()->get_num().float_num);
                    } else { // float -> bool
                        ret.int_num = convert->get_src()->get_num().float_num != 0.0 ? 1 : 0;
                    }
                    break;
                case Bool:
                    if(convert->get_dst()->get_type() == Int) {
                        ret.int_num = convert->get_src()->get_num().int_num;
                    } else {
                        ret.float_num = float(convert->get_src()->get_num().int_num);
                    }
                    break;
                case Void: break;
                }
                return Factory::get_const_reg(convert->get_dst()->get_type(),ret);
            }
            auto hash = convert_hash(convert);
            if(!convert_table.count(hash)) {
                convert_table[hash] = convert;
            }
            return convert_table[hash]->get_dst();
        } else Instanceof(loadArray,LoadArray*,inst) {
            auto hash = loadArray_hash(loadArray);
            if(!loadArray_table.count(hash)) {
                loadArray_table[hash] = loadArray;
            }
            return loadArray_table[hash]->get_dst();
        } else Instanceof(loadVector,LoadVector*,inst) {
            auto hash = loadVector_hash(loadVector);
            if(!loadVector_table.count(hash)) {
                loadVector_table[hash] = loadVector;
            }
            return loadVector_table[hash]->get_dst();
        } else Instanceof(call,CallInstr*,inst) {
            if((call->get_function()->pure || (CompileUnit::in_array_ssa && call->get_function()->array_ssa_pure))) {
                auto hash = call_hash(call);
                if(!call_table.count(hash)) {
                    call_table[hash] = call;
                }
                return call_table[hash]->get_dst();
            } else {
                return call->get_dst();
            }
        } else Instanceof(mem_read,MemRead*,inst) {
            auto hash = mem_read_hash(mem_read);
            if(!mem_read_table.count(hash)) mem_read_table[hash] = mem_read;
            return mem_read_table[hash]->get_dst();
        } else Instanceof(phi,PhiInstr*,inst) {
            auto hash = phi_hash(phi);
            if(!phi_table.count(hash)) phi_table[hash] = phi;
            return phi_table[hash]->get_dst();
        } else Instanceof(definst,DefInstr*,inst) {
            return definst->def();
        } else Instanceof(mem_write,MemWrite*,inst) {
            return mem_write->get_new_scope();
        } else Instanceof(arrayPhi,ArrayPhi*,inst) {
            return arrayPhi->get_dst();
        } else return nullptr;

    }

    void dfs(Function * func,BasicBlock* root) {

        // 保存table 环境
        auto last_binary_table = binary_table;
        auto last_convert_table = convert_table;
        auto last_loadArray_table = loadArray_table;
        auto last_loadVector_table = loadVector_table;
        auto last_call_table = call_table;
        auto last_read_table  = mem_read_table;
        phi_table.clear();
        for(auto &inst:root->get_instrs()) {
            Instanceof(binary,BinaryInstr*,inst.get()) {
                // 既然在dst被修改做了复制传播,就不用对use进行vn了
                // 假设,需要做vn
                // 那么存在两个表达式,是等价的,但是复制传播没有让其等价
                // 设r1,r2等价,但是复制传播没有传播
                // 回到r1,r2的def place,我们可知gvn没让其def等价,这与gvn的效果相矛盾
                // 所以不需要对其的use进行vn,因为其一定在def的时候vn了
                bool out = false;
                if(Ir::CompileUnit::in_arm) {
                    switch (binary->get_binary_type()) {
                    case SLE:
                    case SLT:
                    case SGE:
                    case SGT:
                    case EQ:
                    case NE:
                        out = true;
                        break;
                    default:
                        break;
                    }
                }

                for(auto &use:binary->get_uses()) {
                    Instanceof(reg,Reg*,use.get()) {
                        if(reg->get_id() < 0 || func->is_param(reg))
                            continue ;
                        auto new_use = table_get(func->def_place[reg->get_id()]);
                        Instanceof(new_reg,Reg*,new_use.get()) {
                            if(new_reg->get_id()!=reg->get_id()) {
                                binary->change_use_name(reg->get_id(),std::static_pointer_cast<Reg>(new_use));
                            }
                        }
                    }
                }

                auto new_dst = table_get(binary);
                Instanceof(new_dst_reg,Reg *,new_dst.get()) {
                    if(new_dst_reg->get_id() != binary->get_dst()->get_id()) {
                        if (!out) {
                            copy_propagation(func->use_place, binary->get_dst()->get_id(), std::static_pointer_cast<Reg>(new_dst));
                        } else { // arm 架构只支持常量传播,不能复制传播
                            if(new_dst_reg->get_id() < 0) {
                                copy_propagation(func->use_place, binary->get_dst()->get_id(), std::static_pointer_cast<Reg>(new_dst));
                            }
                        }
                    }
                }
            } else Instanceof(convert, Convert*,inst.get()) {
                for(auto &use:convert->get_uses()) {
                    Instanceof(reg,Reg*,use.get()) {
                        if(reg->get_id() < 0 || func->is_param(reg))
                            continue ;
                        auto new_use = table_get(func->def_place[reg->get_id()]);
                        Instanceof(new_reg,Reg*,new_use.get()) {
                            if(new_reg->get_id()!=reg->get_id()) {
                                convert->change_use_name(reg->get_id(),std::static_pointer_cast<Reg>(new_use));
                            }
                        }
                    }
                }
                auto new_dst = table_get(convert);
                Instanceof(new_dst_reg,Reg *,new_dst.get()) {
                    if(new_dst_reg->get_id() != convert->get_dst()->get_id()) {
                        copy_propagation(func->use_place,convert->get_dst()->get_id(),std::static_pointer_cast<Reg>(new_dst));
                    }
                }
            } else Instanceof(loadArray,LoadArray*,inst.get()) {
                for(auto &use:loadArray->get_uses()) {
                    Instanceof(reg,Reg*,use.get()) {
                        if(reg->get_id() < 0 || func->is_param(reg))
                            continue ;
                        auto new_use = table_get(func->def_place[reg->get_id()]);
                        Instanceof(new_reg,Reg*,new_use.get()) {
                            if(new_reg->get_id()!=reg->get_id()) {
                                loadArray->change_use_name(reg->get_id(),std::static_pointer_cast<Reg>(new_use));
                            }
                        }
                    } else Instanceof(vector,Vector *,use.get()) {
                        if(vector->get_id() < 0 || func->is_param(vector))
                            continue ;
                        auto new_use = table_get(func->def_place[vector->get_id()]);
                        Instanceof(new_vec,Vector *,new_use.get()) {
                            if(new_vec->get_id()!=vector->get_id()) {
                                loadArray->change_use_name(vector->get_id(),std::static_pointer_cast<Vector>(new_use));
                            }
                        }
                    }
                }
                auto new_dst = table_get(loadArray);
                Instanceof(new_dst_reg,MemObject *,new_dst.get()) {
                    if(new_dst_reg->get_id() != loadArray->get_dst()->get_id()) {
                        copy_propagation(func->use_place,loadArray->get_dst()->get_id(),std::static_pointer_cast<MemObject>(new_dst));
                    }
                }
            }
            else Instanceof(loadGlobal,LoadGlobal*,inst.get()) {
                auto addr = loadGlobal->get_addr();
                if(addr->is_constant()) {
                    if(addr->get_type() == Int) {
                        auto new_reg = Factory::get_const_reg(Int,addr->get_num());
                        copy_propagation(func->use_place,loadGlobal->get_dst()->get_id(),new_reg);
                    } else {
                        auto new_reg = Factory::get_const_reg(Float,addr->get_num());
                        copy_propagation(func->use_place,loadGlobal->get_dst()->get_id(),new_reg);
                    }
                }
            }
            else Instanceof(loadVector,LoadVector*,inst.get()) {
                for(int i = 0;i<int(loadVector->get_indexs().size());i++) {
                    auto index = loadVector->get_indexs()[i];
                    auto new_index = table_get(func->def_place[index->get_id()]);
                    Instanceof(new_index_reg,Reg*,new_index.get()) {
                        if(new_index_reg->get_id() != index->get_id()) {
                            loadVector->change_use_name(index->get_id(),std::static_pointer_cast<Reg>(new_index));
                        }
                    }
                }
                auto new_dst = table_get(loadVector);
                Instanceof(new_dst_reg,Vector *,new_dst.get()) {
                    if(new_dst_reg->get_id() != loadVector->get_dst()->get_id()) {
                        copy_propagation(func->use_place,loadVector->get_dst()->get_id(),std::static_pointer_cast<Vector>(new_dst));
                    }
                }
            } else Instanceof(call,CallInstr *,inst.get()) {
                for(auto &use:call->get_uses()) {
                    Instanceof(reg,Reg*,use.get()) {
                        if(reg->get_id() < 0 || func->is_param(reg))
                            continue ;
                        auto new_use = table_get(func->def_place[reg->get_id()]);
                        Instanceof(new_reg,Reg*,new_use.get()) {
                            if(new_reg->get_id()!=reg->get_id()) {
                                call->change_use_name(reg->get_id(),std::static_pointer_cast<Reg>(new_use));
                            }
                        }
                    } else Instanceof(vector,Vector *,use.get()) {
                        if(vector->get_id() < 0 || func->is_param(vector))
                            continue ;
                        auto new_use = table_get(func->def_place[vector->get_id()]);
                        Instanceof(new_vec,Vector *,new_use.get()) {
                            if(new_vec->get_id()!=vector->get_id()) {
                                call->change_use_name(vector->get_id(),std::static_pointer_cast<Vector>(new_use));
                            }
                        }
                    } else Instanceof(mem,MemObject *,use.get()) {
                        if(mem->get_id() < 0 || func->is_param(mem))
                            continue ;
                        auto new_use = table_get(func->def_place[mem->get_id()]);
                        Instanceof(new_mem,MemObject *,new_use.get()) {
                            if(new_mem->get_id()!=mem->get_id()) {
                                call->change_use_name(mem->get_id(),std::static_pointer_cast<MemObject>(new_use));
                            }
                        }
                    }
                }
                if(!call->is_void_flag()) {

                    auto new_dst = table_get(call);
                    Instanceof(new_dst_reg,Reg *,new_dst.get()) {
                        if(new_dst_reg->get_id() != call->get_dst()->get_id()) {
                            copy_propagation(func->use_place,call->get_dst()->get_id(),std::static_pointer_cast<Reg>(new_dst));
                        }
                    }
                }
            } else Instanceof(branch,BranchInstr*,inst.get()) { // branch 的常数传播,不允许修改dom环境,会导致gcm出问题
                continue;
            } else Instanceof(phi,PhiInstr*,inst.get()) { // 所有value都相同,可以直接copy,单感觉可能性不大,主要针对都是常数的吧感觉
                shared_ptr<Reg> tmp = nullptr;
                bool flag = true;
                for(auto &[bb,value]:phi->get_bb_2_value()) {
                    if(!tmp) tmp = value;
                    if(tmp->get_id() != value->get_id()) {
                        flag = false;
                        break;
                    }
                    if(tmp->get_id() < 0) {
                        if (tmp->get_num().int_num != value->get_num().int_num) {
                            flag = false;
                            break;
                        }
                    }
                }
                if(flag) {
                    copy_propagation(func->use_place,phi->get_dst()->get_id(),tmp);
                }

                auto new_dst = table_get(phi);
                if(new_dst->get_id() != phi->get_dst()->get_id()) {
                    copy_propagation(func->use_place,phi->get_dst()->get_id(), static_pointer_cast<Reg>(new_dst));
                }

            } else Instanceof(mem_read,MemRead*,inst.get()) {
                int addr_id = mem_read->get_addr()->get_id();
                if(addr_id >= 0) {
                    auto new_addr = table_get(func->def_place[addr_id]);
                    Instanceof(memObject,MemObject*,new_addr.get()) {
                        if(memObject->get_id() != addr_id) {
                            mem_read->change_use_name(addr_id, static_pointer_cast<MemObject>(new_addr));
                        }
                    }
                }
                auto new_dst = table_get(mem_read);
                if(new_dst->get_id() != mem_read->get_dst()->get_id()) {
                    Instanceof(reg,Reg*,new_dst.get()) {
                        copy_propagation(func->use_place,mem_read->get_dst()->get_id(), static_pointer_cast<Reg>(new_dst));
                    } else Instanceof(vec,Vector*,new_dst.get()) {
                        copy_propagation(func->use_place,mem_read->get_dst()->get_id(), static_pointer_cast<Vector>(new_dst));
                    }

                }
            }
        }
        for(auto &child:root->idom) {
            dfs(func,child);
        }
        binary_table =  last_binary_table;
        convert_table = last_convert_table;
        loadArray_table = last_loadArray_table;
        loadVector_table = last_loadVector_table;
        call_table = last_call_table;
        mem_read_table = last_read_table;
    }

    void gvn(Function* func) {
        // 根据支配树来dfs gvn
        binary_table.clear();
        convert_table.clear();
        loadArray_table.clear();
        loadVector_table.clear();
        call_table.clear();
        mem_read_table.clear();
        BasicBlock* root = func->get_first_bb();
        dfs(func,root);
    }

    unordered_set<Instruction *> scheduled_inst;
    unordered_map<Instruction *,BasicBlock *> schedule_place; // schedule之后就马上放过去,别逼我求你 orz

    bool movable_call(CallInstr* inst) {
        if(!CompileUnit::in_array_ssa) {
            return inst->get_function()->pure;
        } else {
            // array ssa 下考虑motion call 带来的call 对全局变量影响的提前,提前到了不应该到的出口
            if(inst->get_function()->get_name() == "main" && inst->used_global.empty()) {
                return true;
            } else { // 判断是否对全局造成影响,但是把,如果造成影响,其前面的mem def就不会能motion
                return false;

            }
        }
    }

    void schedule_early(Function * func, BasicBlock * root, Instruction *inst) { // 对load,store不能gcm,todo array ssa
        if (!inst || scheduled_inst.count(inst)) {
            return;
        }
        scheduled_inst.insert(inst);
        Instanceof(defInstr, DefInstr *, inst) {
            Instanceof(call, CallInstr *, defInstr) {
                if (!movable_call(call)) {
                    schedule_place[inst] = inst->get_belongto_bb();
                    return;
                }
            }
            else Instanceof(phi, PhiInstr *, inst) {
                schedule_place[inst] = inst->get_belongto_bb();
                return;
            }
            else Instanceof(binary, BinaryInstr *, defInstr) {
                switch (binary->get_binary_type()) {
                case SLT:
                case SLE:
                case SGT:
                case SGE:
                case NE:
                case EQ:
                    schedule_place[inst] = inst->get_belongto_bb();
                    return;
                default:
                    break;
                }
            }
            else Instanceof(array_phi, ArrayPhi *, defInstr) {
                schedule_place[inst] = inst->get_belongto_bb();
                return;
            }
            else Instanceof(load, LoadInstr *, inst) {
                schedule_place[inst] = inst->get_belongto_bb();
                return;
            }
            else Instanceof(loadGlobal, LoadGlobal *, inst) {
                schedule_place[inst] = inst->get_belongto_bb();
                return;
            }
            BasicBlock *place = root;
            for (const auto &use : inst->get_uses()) {
                if (func->is_param(use.get())) {
                    continue;
                }
                int id = 0;
                Instanceof(reg, Reg *, use.get()) {
                    if (reg->get_id() < 0)
                        continue;
                    id = reg->get_id();
                }
                else Instanceof(vector, Vector *, use.get()) {
                    if (vector->get_id() < 0)
                        continue;
                    id = vector->get_id();
                }
                else Instanceof(memObject, MemObject *, use.get()) {
                    if (memObject->get_id() < 0)
                        continue;
                    id = memObject->get_id();
                }
                schedule_early(func, root, func->def_place[id]);
                BasicBlock *tmp_place =
                    func->def_place[id] ? func->def_place[id]->get_belongto_bb()
                                        : root;
                if (tmp_place->dom_level > place->dom_level) {
                    place = tmp_place;
                }
            }
            schedule_place[inst] = place;
        }
          else Instanceof(mem_write,MemWrite*,inst) { // 统统motion,全部motion,全场motion
            if(!array_ssa_gcm_swtich) return;
              if(mem_write->get_belongto_bb()->get_belongto_func()->get_name() != "main") {
                  // 函数调用可能存在数组别名问题,所以不能motion
                  schedule_place[inst] = inst->get_belongto_bb();
                  return;
              }
              BasicBlock *place = root;
              for(const auto& use:mem_write->get_uses()) {
                  if(func->is_param(use.get())){
                      continue ;
                  }
                  int id = 0;
                  Instanceof(reg,Reg *,use.get()) {
                      if(reg->get_id() < 0)
                          continue ;
                      id = reg->get_id();
                  } else Instanceof(vector,Vector*,use.get()) {
                      if(vector->get_id() < 0)
                          continue ;
                      id = vector->get_id();
                  } else Instanceof(memObject, MemObject*, use.get()) {
                      if(memObject->get_id() < 0)
                          continue ;
                      id = memObject->get_id();
                  }
                  schedule_early(func,root,func->def_place[id]);
                  BasicBlock *tmp_place = func->def_place[id] ? func->def_place[id]->get_belongto_bb() : root;
                  if(tmp_place->dom_level > place->dom_level) {
                      place = tmp_place;
                  }
              }
              schedule_place[inst] = place;
          } else Instanceof(mem_read,MemRead*,inst) { // 统统motion,全部motion,全场motion
              if(!array_ssa_gcm_swtich) return;
              BasicBlock *place = root;
              for(const auto& use:mem_read->get_uses()) {
                  if(func->is_param(use.get())){ // addr,val,read_before_write
                      continue ;
                  }
                  int id = 0;
                  Instanceof(reg,Reg *,use.get()) {
                      if(reg->get_id() < 0)
                          continue ;
                      id = reg->get_id();
                  } else Instanceof(vector,Vector*,use.get()) {
                      if(vector->get_id() < 0)
                          continue ;
                      id = vector->get_id();
                  } else Instanceof(memObject, MemObject*, use.get()) {
                      if(memObject->get_id() < 0)
                          continue ;
                      id = memObject->get_id();
                  }
                  schedule_early(func,root,func->def_place[id]);
                  BasicBlock *tmp_place = func->def_place[id] ? func->def_place[id]->get_belongto_bb() : root;
                  if(tmp_place->dom_level > place->dom_level) {
                      place = tmp_place;
                  }
              }
              schedule_place[inst] = place;
        }
        else {
            schedule_place[inst] = inst->get_belongto_bb();
        }
        if(schedule_place[inst] != inst->get_belongto_bb()) {
            for(auto tmp = inst->get_belongto_bb()->get_instrs().begin();tmp != inst->get_belongto_bb()->get_instrs().end();) {
                if(tmp->get() == inst) {
                    tmp->release();
                    inst->get_belongto_bb()->get_instrs().erase(tmp);
                    break;
                }
                tmp++;
            }
            schedule_place[inst]->insert_before_br(inst);
        }
    }

    BasicBlock * find_lca(BasicBlock * a,BasicBlock * b) { // 寻找最近公共子祖先,我也不懂,抄论文就完事了
        while(a != b) {
            if(a->dom_level > b->dom_level) {
                a = a->idom_by;
            } else if(a ->dom_level < b->dom_level) {
                b = b->idom_by;
            } else {
                a = a->idom_by;
                b = b->idom_by;
            }
        }
        return a;
    }

    void schedule_late(Function* func, Instruction* inst) {
        if(!inst || scheduled_inst.count(inst))
            return;
        scheduled_inst.insert(inst);
        Instanceof(definst,DefInstr *,inst) {
            // 这些指令不能提升,内存有毒),array ssa应该就行了吧大概,哎,还是太菜了
            Instanceof(phi, PhiInstr *, definst) { return; }
            else Instanceof(load, LoadInstr *, definst) {
                return;
            }
            else Instanceof(loadGlobal, LoadGlobal *, definst) {
                return;
            }
            else Instanceof(call, CallInstr *, definst) {
                if (!movable_call(
                        call)) // array ssa 解决了内存读写的时序问题,准许motion
                    return;
            }
            else Instanceof(binary, BinaryInstr *, definst) {
                switch (binary->get_binary_type()) {
                case SLT:
                case SLE:
                case SGT:
                case SGE:
                case NE:
                case EQ:
                    return;
                default:
                    break;
                }
            }
            else Instanceof(array_phi, ArrayPhi *, definst) {
                return;
            }
            auto def = definst->def();
            BasicBlock *place = nullptr;
            for (auto use : func->use_place[def->get_id()]) {
                BasicBlock *tmp = nullptr;
                Instanceof(phi, PhiInstr *, use) { // phi can not be motion
                    for (auto &[bb, val] : phi->get_bb_2_value()) {
                        if (val->get_id() == def->get_id()) {
                            tmp = (tmp == nullptr) ? bb : find_lca(bb, tmp);
                        }
                    }
                }
                else {
                    schedule_late(func, use);
                    tmp = use->get_belongto_bb();
                }
                if (tmp == nullptr) {
                    assert(false);
                }
                if (!place)
                    place = tmp;
                else
                    place = find_lca(place, tmp);
            }
            if (!place)
                place =
                    inst->get_belongto_bb(); // 没有use,是个废指令,那就别motion,等着删掉吧
            BasicBlock *final = place;
            if (place != schedule_place[inst]) {
                do {
                    place = place->idom_by;
                    if (place->get_loop_level() < final->get_loop_level()) {
                        final = place;
                    }
                } while (place != schedule_place[inst]);
            }
            schedule_place[inst] = final;
          } else Instanceof(mem_read,MemRead*,inst) {
            if(!array_ssa_gcm_swtich) return;
              if(mem_read->get_dst()->get_id() == 121) {
                  int mtp = 0;
              }
              auto dst = mem_read->get_dst();
              BasicBlock * place = nullptr;
              for(auto use: func->use_place[dst->get_id()]) {
                  BasicBlock * tmp = nullptr;
                  Instanceof(phi, PhiInstr *,use) { // phi can not be motion
                      for(auto &[bb,val] : phi->get_bb_2_value()) {
                          if(val->get_id() == dst->get_id()) {
                              tmp = (tmp == nullptr) ? bb: find_lca(bb,tmp);
                          }
                      }
                  } else Instanceof(array_phi,ArrayPhi*,use) { // 好像不太可能?,array phi的value主要是vector/reg,针对mem write,哎,写都写了,写一个吧
                      for(auto &[bb,val] : array_phi->get_bb_2_value()) {
                          if(val->get_id() == dst->get_id()) {
                              tmp = (tmp == nullptr) ? bb: find_lca(bb,tmp);
                          }
                      }
                      if(tmp == nullptr) tmp = inst->get_belongto_bb(); // 说明是array 的 read before write,那不能去array phi 所在的bb
                  } else {
                      schedule_late(func,use);
                      tmp = use->get_belongto_bb();
                  }
                  if(tmp == nullptr) {
                      assert(false);
                  }
                  if (!place) place = tmp;
                  else
                      place = find_lca(place,tmp);
              }
              if(!place) place = inst->get_belongto_bb(); // 没有use,是个废指令,那就别motion,等着删掉吧
              BasicBlock *final = place;
              if(place != schedule_place[inst]) {
                  do{
                      place = place->idom_by;
                      if (place->get_loop_level() < final->get_loop_level()) {
                          final = place;
                      }
                  } while(place != inst->get_belongto_bb());
              }
              schedule_place[inst] = final;
          } else Instanceof(mem_write,MemWrite*,inst) {
              if(!array_ssa_gcm_swtich) return;
              if(mem_write->get_belongto_bb()->get_belongto_func()->get_name() != "main") {
                  schedule_place[inst] = inst->get_belongto_bb();
                  return;
              }
              auto dst = mem_write->get_new_scope();
              BasicBlock * place = nullptr;
              for(auto use: func->use_place[dst->get_id()]) {
                  BasicBlock * tmp = nullptr;
                  Instanceof(phi, PhiInstr *,use) { // 可能性不大说实话,感觉可以emm删掉! oAo
                      assert(false);
                  } else Instanceof(array_phi,ArrayPhi*,use) { // 好像不太可能?,array phi的value主要是vector/reg,针对mem write,哎,写都写了,写一个吧
                      for(auto &[bb,val] : array_phi->get_bb_2_value()) {
                          if(val->get_id() == dst->get_id()) {
                              tmp = (tmp == nullptr) ? bb: find_lca(bb,tmp);
                          }
                      }
                  } else {
                      schedule_late(func,use);
                      tmp = use->get_belongto_bb();
                  }
                  if(tmp == nullptr) {
                      assert(false);
                  }
                  if (!place) place = tmp;
                  else
                      place = find_lca(place,tmp);
              }
              if(!place) place = inst->get_belongto_bb(); // 没有use,是个废指令,那就别motion,等着删掉吧
               BasicBlock *final = place;
               if(place != schedule_place[inst]) {
                  do{
                      place = place->idom_by;
                      if (place->get_loop_level() < final->get_loop_level()) {
                          final = place;
                      }
                  } while(place != schedule_place[inst]);
              }
              schedule_place[inst] = final;
          }
         else {
            schedule_place[inst] = inst->get_belongto_bb();
        }
        if(schedule_place[inst] != inst->get_belongto_bb()) {
            for(auto tmp = inst->get_belongto_bb()->get_instrs().begin();tmp != inst->get_belongto_bb()->get_instrs().end();) {
                if(tmp->get() == inst) {
                    tmp->release();
                    inst->get_belongto_bb()->get_instrs().erase(tmp);
                    break;
                }
                tmp++;
            }
            schedule_place[inst]->insert_after_phi(inst);
        }
    }


    void gcm(Function* func) {

        scheduled_inst.clear();
        vector<Instruction *> all_inst;
        for(auto &bb:func->get_bbs()) {
            for(auto &inst:bb->get_instrs()) {
                all_inst.push_back(inst.get());
            }
        }

        for(auto &inst:all_inst) {
            schedule_early(func,func->get_bbs().front().get(),inst);
        }

        scheduled_inst.clear();

        for(auto &inst:all_inst) {
            schedule_late(func,inst);
        }


    }

    void gvn_gcm(Function* func) {
        func->cfg->build();
        func->cfg->dom_build();
        func->cfg->loop_analysis();
        binary_table.clear();
        convert_table.clear();
        loadArray_table.clear();
        loadVector_table.clear();
        call_table.clear();
        mem_read_table.clear();
        gvn(func);
        gcm(func);
    }

    void gvn_gcm(CompileUnit* compileUnit) {
        for(auto &func: compileUnit->get_functions()) {
            if (func->is_lib()) {
                continue ;
            }
            Factory::reset(func.get());
            gvn_gcm(func.get());
            func->set_bb_max_num(Factory::bb_num);
            func->set_reg_max_num(Factory::reg_num);
        }
        info("end gvn gcm");
    }
}