//
// Created by 11069 on 5/7/2023.
//
// this pass used after pure function mark

// engineering a compiler 10.2
// 该pass思想:找到所有的critical inst,然后根据def和use反推所有有用的指令,没有用的指令全部删除
// 什么是critical inst?
// 输入输出算critical inst,这里也就是call lib function
// 函数 return
// mem store,该mem值的是外界mem,既非本函数栈帧的空间,比如全局变量,全局数组,数组函数参数
// 对于call 方法,如果该函数不是pure,则算effective inst
// Branch指令,Jump指令

# include "opt.h"
namespace opt {
using std::unordered_set;
using std::unordered_map;

using Ir::Instruction;
using Ir::CompileUnit;
using Ir::Function;
using Ir::Vector;

bool is_param(Function* func,std::shared_ptr<Vector> arg) {
    for(auto &type:func->get_params()) {
        Instanceof(para,Vector*,type.get()) {
            if(para->get_id() == arg->get_id()) {
                return true;
            }
        }
    }

    return false;
}




void inst_clean_up(CompileUnit* compileUnit) { // 指令删除很神奇,对于内存修改,因为删除不仅影响函数内部,还影响外部函数,比如a在内存写了一个1,b函数load出来然后输出了,所以内存写入都不能删除
    for(auto &func:compileUnit->get_functions()) {
        if(func->is_lib() || func->get_bbs().empty())
            continue ;
        unordered_map<int,int> mem_map;
        unordered_set<Instruction *> critical_inst;
        unordered_set<Instruction *> useful_inst;

        for(auto &bb:func->get_bbs()) {
            // 第一阶段,初始化critical_inst
            for(auto &inst:bb->get_instrs()) {
                Instanceof(ret,Ir::Ret*,inst.get()) { // critical
                    critical_inst.insert(inst.get());
                } else Instanceof(call,Ir::CallInstr*,inst.get()) { // not pure func and lib func are critical
                    if(!(call->get_function()->pure || (CompileUnit::in_array_ssa && call->get_function()->array_ssa_pure))) {
                        critical_inst.insert(inst.get());
                    }
                } else Instanceof(store,Ir::StoreInstr*,inst.get()) { // store is critical
                        critical_inst.insert(inst.get());
                } else Instanceof(storeGlobal,Ir::StoreGlobal *,inst.get()) { // critical
                    critical_inst.insert(inst.get());
                } else Instanceof(branch,Ir::BranchInstr *,inst.get()) {
                    critical_inst.insert(inst.get());
                } else Instanceof(mem_write,Ir::MemWrite*,inst.get()) {
                    critical_inst.insert(inst.get());
                }

            }
        }
        // 第二阶段 mark所有的有用的指令
        while (!critical_inst.empty()) {
            auto iter = critical_inst.begin();
            auto ins = *iter;
            critical_inst.erase(iter);
            useful_inst.insert(ins);
            // 把所有的use的def都放进去
            Instanceof(mem_write,Ir::MemWrite*,ins) {
                // mem write 对于read不放入有用指令,有用指令只是在gcm的时候用于限制mem write的motion位置的
                for(auto &use:mem_write->get_pure_uses()) {
                    int id = -1;
                    Instanceof(reg,Ir::Reg*,use.get()) {
                        id = reg->get_id();
                    } else Instanceof(memObject,Ir::MemObject *,use.get()) {
                        id = memObject->get_id();
                    } else Instanceof(Vector,Ir::Vector*,use.get()) {
                        id = Vector->get_id();
                    }
                    if(id >= 0) {
                        if(func->def_place.count(id)) {
                            auto def = func->def_place[id];
                            if(useful_inst.count(def) || !def)
                                continue ;
                            critical_inst.insert(def);
                            useful_inst.insert(ins);
                        }
                    }
                }
                continue; // todo
            }
            for(auto &use:ins->get_uses()) {
                int id = -1;
                Instanceof(reg,Ir::Reg*,use.get()) {
                    id = reg->get_id();
                } else Instanceof(memObject,Ir::MemObject *,use.get()) {
                    id = memObject->get_id();
                } else Instanceof(Vector,Ir::Vector*,use.get()) {
                    id = Vector->get_id();
                }
                if(id >= 0) {
                    if(func->def_place.count(id)) {
                        auto def = func->def_place[id];
                        Instanceof(mem_write,Ir::MemWrite*,def) {

                        }
                        if(useful_inst.count(def) || !def)
                            continue ;
                        critical_inst.insert(def);
                    }
                }
            }
        }

        // 第三阶段 删除所有没有mark的指令

        for(auto &bb:func->get_bbs()) {
            for(auto iter = bb->get_instrs().begin();iter != bb->get_instrs().end();) {
                auto &inst = *iter;
                if(!useful_inst.count(inst.get())) {
                    inst->remove_use_def();
                    iter = bb->get_instrs().erase(iter);
                } else {
                    iter++;
                }
            }
        }
    }
}
}