//
// Created by shliba on 2023/5/21.
//

#include "ir.h"
#include <cstring>
#include <utility>

namespace Ir {
std::string type_transe(ScalaType type) {
    switch (type) {
    case Ir::Int:
        return "i32";
    case Ir::Float:
        return "float";
    case Ir::Bool:
        return "i1";
    default:
        return "void";
    }
}
inline void replace_reg (std::shared_ptr<Ir::Reg> &reg, unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps) {
        if (!reg->is_constant() && !reg->is_global()) {
            reg = std::dynamic_pointer_cast<Ir::Reg>(reg_maps[reg->get_id()]);
        }
}

inline void replace_reg (std::shared_ptr<Ir::Type> &reg, unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps) {
        if (reg->get_id() < 0) {
            return ;
        }
        Instanceof(ret_reg, Ir::Reg*, reg.get()) {
            if (!ret_reg->is_constant() && !ret_reg->is_global()) {
                reg = reg_maps[reg->get_id()];
            }
        } else {
            reg = reg_maps[reg->get_id()];
        }
}

inline void replace_reg (std::shared_ptr<Ir::MemObject> &reg, unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps) {
        if (reg->get_id() < 0) {
            return ;
        }
        reg = std::dynamic_pointer_cast<Ir::MemObject>(reg_maps[reg->get_id()]);

}

inline void replace_reg (std::shared_ptr<Ir::Vector> &reg, unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps) {
        if (reg->get_id() < 0) {
            return ;
        }
        reg = std::dynamic_pointer_cast<Ir::Vector>(reg_maps[reg->get_id()]);

}


std::string vector_transe(Ir::Vector* array, int deep) {
    if (array->get_dims().size() == 1 && array->get_dims()[0] == 0) {
        return type_transe(array->get_type());
    }
    if (deep == 0 && array->get_dims()[0] == 0) {
        return vector_transe(array, deep + 1);
    }
    if (deep + 1 < array->get_dims().size() ) {
        return "[" + std::to_string(array->get_dims()[deep]) + " x " + vector_transe(array, deep + 1)  + "]";
    } else {
        return "[" +std::to_string(array->get_dims()[deep]) + " x " + type_transe(array->get_type())  + "]";
    }
}



int Factory::reg_num = 0;
int Factory::bb_num = 0;
int Factory::global_num = 0;
std::unordered_set<std::shared_ptr<Type>> BranchInstr::get_uses() const {
    if (bb2 == nullptr) {
        return {};
    } else {
        return {reg};
    }
}

BranchInstr::BranchInstr(BasicBlock *bb1) :
    bb1(bb1) {
}

BranchInstr::BranchInstr(const shared_ptr<Reg> &reg, BasicBlock *bb1, BasicBlock *bb2) :
    reg(reg), bb1(bb1), bb2(bb2) {
}

const shared_ptr<Reg> &BranchInstr::get_reg() const {
    return reg;
}

BasicBlock *BranchInstr::get_bb_1() const {
    return bb1;
}

BasicBlock *BranchInstr::get_bb_2() const {
    return bb2;
}
void BranchInstr::add_use_def() {
    if (bb2 == nullptr) {
        return ;
    }
    if(!reg || reg->get_id() < 0) {
        return;
    }
    get_belongto_bb()->get_belongto_func()->use_place[reg->get_id()].insert(this);
}
void BranchInstr::remove_use_def() {
    if(!reg || reg->get_id() < 0)
        return;
    get_belongto_bb()->get_belongto_func()->use_place[reg->get_id()].erase(this);
}
void BranchInstr::change_use_name(int src_id, shared_ptr<Reg> new_val) {
    if(!reg ||reg->get_id() < 0 || reg->get_id() != src_id)
        return;
    get_belongto_bb()->get_belongto_func()->use_place[src_id].erase(this);
    reg = new_val;
    if(reg->get_id() >= 0)
        get_belongto_bb()->get_belongto_func()->use_place[reg->get_id()].insert(this);
}

void BranchInstr::gen_ir(std::ostream &os) const {
    if (bb2 == nullptr) {
        os << "br label %BB" << bb1->get_id();
    } else {
        os << "br " << type_transe(reg->get_type()) << " " << *reg.get()  << ", label %BB" << bb1->get_id() << ", label %BB" << bb2->get_id() ;
    }
}
void BranchInstr::set_bb1(BasicBlock *bb1) {
    BranchInstr::bb1 = bb1;
}
void BranchInstr::set_bb2(BasicBlock *bb2) {
    BranchInstr::bb2 = bb2;
}
void BranchInstr::set_reg(const shared_ptr<Reg> &reg) {
    BranchInstr::reg = reg;
}
Instruction *BranchInstr::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_bb1 = map[bb1];
    if (bb2 == nullptr) {
        auto new_br = new BranchInstr(new_bb1);
        return new_br;
    } else {
        auto new_bb2 = map[bb2];
        auto new_cond = reg;
        replace_reg(new_cond, reg_maps);
        auto new_br = new BranchInstr(new_cond, new_bb1, new_bb2);
        return new_br;
    }
}

unordered_set<shared_ptr<Type>> BinaryInstr::get_uses() const {
    return {src1, src2};
}
shared_ptr<Type> BinaryInstr::def() const {
    return dst;
}

BinaryInstr::BinaryInstr(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &src1, const shared_ptr<Reg> &src2, BinaryType binaryType) :
    dst(dst), src1(src1), src2(src2), binaryType(binaryType) {
}

 shared_ptr<Reg> &BinaryInstr::get_dst() {
    return dst;
}

shared_ptr<Reg> &BinaryInstr::get_src_1()  {
    return src1;
}

shared_ptr<Reg> &BinaryInstr::get_src_2() {
    return src2;
}


BinaryType BinaryInstr::get_binary_type() const {
    return binaryType;
}
void BinaryInstr::add_use_def() {
    int dst_id = dst->get_id();
    get_belongto_bb()->get_belongto_func()->def_place[dst_id] = this;
    int use_id = src1->get_id();
    if(use_id >= 0) {
        get_belongto_bb()->get_belongto_func()->use_place[use_id].insert(this);
    }
    use_id = src2->get_id();
    if(use_id >= 0) {
        get_belongto_bb()->get_belongto_func()->use_place[use_id].insert(this);
    }
}

void BinaryInstr::remove_use_def() {
    get_belongto_bb()->get_belongto_func()->def_place.erase(this->dst->get_id());
    int use_id = src1->get_id();
    if(use_id >= 0) { // 常量不需要放入use,待后端寄存器分配的时候先分配变量,有剩的再分配给立即数,没剩的就spill out
        get_belongto_bb()->get_belongto_func()->use_place[use_id].erase(this);
    }
    use_id = src2->get_id();
    if(use_id >= 0) {
        get_belongto_bb()->get_belongto_func()->use_place[use_id].erase(this);
    }
}
void BinaryInstr::change_use_name(int src_id,shared_ptr<Reg> new_val) {
    if(src1->get_id() == src_id) {
        get_belongto_bb()->get_belongto_func()->use_place[src1->get_id()].erase(this);
        src1 = new_val;
        if(src1->get_id() >= 0)
            get_belongto_bb()->get_belongto_func()->use_place[src1->get_id()].insert(this);
    }
    if(src2->get_id() == src_id) {
        get_belongto_bb()->get_belongto_func()->use_place[src2->get_id()].erase(this);
        src2 = new_val;
        if(src2->get_id() >= 0)
            get_belongto_bb()->get_belongto_func()->use_place[src2->get_id()].insert(this);
    }

}

std::string binay_transe (Ir::BinaryType type, Ir::ScalaType scala_type) {
    switch (type) {
    case ADD: {
        if (scala_type < Ir::Float) {
            return "add";
        } else if (scala_type == Ir::Float) {
            return "fadd";
        } else {
            assert(false);
        }
    }
    case SUB: {
        if (scala_type < Ir::Float) {
            return "sub";
        } else if (scala_type == Ir::Float) {
            return "fsub";
        } else {
            assert(false);
        }
    }
    case SDIV: {
        if (scala_type < Ir::Float) {
            return "sdiv";
        } else if (scala_type == Ir::Float) {
            return "fdiv";
        } else {
            assert(false);
        };
    }
    case MUL: {
        if (scala_type < Ir::Float) {
            return "mul";
        } else if (scala_type == Ir::Float) {
            return "fmul";
        } else {
            assert(false);
        };
    }
    case SREM: {
        if (scala_type < Ir::Float) {
            return "srem";
        } else if (scala_type == Ir::Float) {
            return "frem";
        } else {
            assert(false);
        };
    }
    case EQ: {
        if (scala_type < Ir::Float) {
            return "icmp eq";
        } else if (scala_type == Ir::Float) {
            return "fcmp oeq";
        } else {
            assert(false);
        }
    }
    case NE: {
        if (scala_type < Ir::Float) {
            return "icmp ne";
        } else if (scala_type == Ir::Float) {
            return "fcmp one";
        } else {
            assert(false);
        };
    }
    case SGT: {
        if (scala_type < Ir::Float) {
            return "icmp sgt";
        } else if (scala_type == Ir::Float) {
            return "fcmp ogt";
        } else {
            assert(false);
        };
    }
    case SGE: {
        if (scala_type < Ir::Float) {
            return "icmp sge";
        } else if (scala_type == Ir::Float) {
            return "fcmp oge";
        } else {
            assert(false);
        };
    }
    case SLE: {
        if (scala_type < Ir::Float) {
            return "icmp sle";
        } else if (scala_type == Ir::Float) {
            return "fcmp ole";
        } else {
            assert(false);
        };
    }
    case SLT: {
        if (scala_type < Ir::Float) {
            return "icmp slt";
        } else if (scala_type == Ir::Float) {
            return "fcmp olt";
        } else {
            assert(false);
        };
    }
    case SHL: break;
    case AND: break;
    case OR: {
        if (scala_type < Ir::Float) {
            return "or";
        } else if (scala_type == Ir::Float) {
            return "or";
        } else {
            assert(false);
        }
    };
    case XOR: {
        if (scala_type < Ir::Float) {
            return "xor";
        } else if (scala_type == Ir::Float) {
            return "xor";
        } else {
            assert(false);
        }
    };
    case ASHR: break;
    case LSHR: break;
    }
    assert(false);
}


void BinaryInstr::gen_ir(std::ostream &os) const {
    os << "%" << dst->get_id() << " = " << binay_transe( binaryType, src1->get_type()) << " "
       << type_transe(src1->get_type()) << " "  << *src1.get() << ", " << *src2 ;
}
void BinaryInstr::set_src_1(const shared_ptr<Reg> &src1) {
    BinaryInstr::src1 = src1;
}
void BinaryInstr::set_src_2(const shared_ptr<Reg> &src2) {
    BinaryInstr::src2 = src2;
}
void BinaryInstr::setBinaryType(BinaryType binaryType) {
    BinaryInstr::binaryType = binaryType;
}
void BinaryInstr::change_use_name(shared_ptr<Reg> src, shared_ptr<Reg> new_val) {
    auto func = get_belongto_bb()->get_belongto_func();
    if(*src1 == *src) {
        if(src1->get_id() >= 0) {
            func->use_place[src1->get_id()].erase(this);
        }
        src1 = new_val;
        if(new_val->get_id() >= 0) {
            func->use_place[new_val->get_id()].insert(this);
        }
    }
    if(*src2 == *src) {
        if(src2->get_id() >= 0) {
            func->use_place[src2->get_id()].erase(this);
        }
        src2 = new_val;
        if(new_val->get_id() >= 0) {
            func->use_place[new_val->get_id()].insert(this);
        }
    }
}


Instruction *BinaryInstr::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_src1 = src1;
    auto new_src2 = src2;
    replace_reg(new_src1, reg_maps);
    replace_reg(new_src2, reg_maps);
    auto new_dst = std::dynamic_pointer_cast<Ir::Reg>(reg_maps[dst->get_id()]);
    auto new_binary = new BinaryInstr(new_dst, new_src1, new_src2, binaryType);
    return new_binary;
}
void BinaryInstr::set_dst(const shared_ptr<Reg> &dst) {
    BinaryInstr::dst = dst;
}

std::unordered_set<std::shared_ptr<Type>> AllocaInstr::get_uses() const {
    return {};
}
AllocaInstr::AllocaInstr(ScalaType type, int id) {
    addr = std::make_shared<MemObject>(type, id);
}
const shared_ptr<MemObject> &AllocaInstr::get_addr() const {
    return addr;
}
AllocaInstr::AllocaInstr(const shared_ptr<MemObject> &addr) :
    addr(addr) {
}
void AllocaInstr::remove_use_def() {
    get_belongto_bb()->get_belongto_func()->def_place.erase(addr->get_id());

}
void AllocaInstr::add_use_def() {
    get_belongto_bb()->get_belongto_func()->def_place[addr->get_id()] = this;
}


void AllocaInstr::gen_ir(std::ostream &os) const {
    Instanceof(array, Ir::Vector*, addr.get()) {
        os << "%" << addr->get_id() << " = alloca " << vector_transe(array, 0);
    } else {
        os << "%" << addr->get_id() << " = alloca " << type_transe(addr->get_type());
    }
}
shared_ptr<Type> AllocaInstr::def() const {
    return addr;
}
Instruction *AllocaInstr::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_dst = std::dynamic_pointer_cast<Ir::MemObject>(reg_maps[addr->get_id()]);
    auto new_alloca = new AllocaInstr(new_dst);
    return new_alloca;
}

std::unordered_set<std::shared_ptr<Type>> StoreInstr::get_uses() const {
    return {val,addr};
}

StoreInstr::StoreInstr(const shared_ptr<Reg> &val, const shared_ptr<MemObject> &addr) :
    val(val), addr(addr) {
}

const shared_ptr<Reg> &StoreInstr::get_val() const {
    return val;
}

const shared_ptr<MemObject> &StoreInstr::get_addr() const {
    return addr;
}
void StoreInstr::add_use_def() { // only use

    if(val->get_id() >= 0)
        get_belongto_bb()->get_belongto_func()->use_place[val->get_id()].insert(this);
    get_belongto_bb()->get_belongto_func()->use_place[addr->get_id()].insert(this);
}
void StoreInstr::remove_use_def() { // only use
    if(val->get_id() >= 0)
        get_belongto_bb()->get_belongto_func()->use_place[val->get_id()].erase(this);
    get_belongto_bb()->get_belongto_func()->use_place[addr->get_id()].erase(this);
}
void StoreInstr::change_use_name(int src_id, shared_ptr<Reg> new_val) { // only use
    if(val->get_id() == src_id) {
        auto &user = get_belongto_bb()->get_belongto_func()->use_place[val->get_id()];
        get_belongto_bb()->get_belongto_func()->use_place[val->get_id()].erase(this);
        val = new_val;
        if(val->get_id() >= 0)
            get_belongto_bb()->get_belongto_func()->use_place[val->get_id()].insert(this);
    }
}
void StoreInstr::change_use_name(int src_id, shared_ptr<MemObject> new_val) { // only use
    if(addr->get_id() == src_id) {
        get_belongto_bb()->get_belongto_func()->use_place[addr->get_id()].erase(this);
        addr = new_val;
        get_belongto_bb()->get_belongto_func()->use_place[addr->get_id()].insert(this);
    }
}
void StoreInstr::gen_ir(std::ostream &os) const {
    os << "store " << type_transe(val->get_type()) << " " << *val.get()
       << ", " << type_transe(addr->get_type()) + "* %" << addr->get_id();
}
void StoreInstr::set_val(const shared_ptr<Reg> &val) {
    StoreInstr::val = val;
}
Instruction *StoreInstr::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_addr = addr;
    auto new_val = val;
    replace_reg(new_addr, reg_maps);
    replace_reg(new_val, reg_maps);
    auto store = new StoreInstr(new_val, new_addr);
    return store;
}

std::unordered_set<std::shared_ptr<Type>> LoadInstr::get_uses() const {
    return {addr};
}

shared_ptr<Type> LoadInstr::def() const {
    return dst;
}

LoadInstr::LoadInstr(const shared_ptr<Reg> &dst, const shared_ptr<MemObject> &addr) :
    dst(dst), addr(addr) {
}

const shared_ptr<Reg> &LoadInstr::get_dst() const {
    return dst;
}

const shared_ptr<MemObject> &LoadInstr::get_addr() const {
    return addr;
}
void LoadInstr::add_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place[dst->get_id()] = this;
    func->use_place[addr->get_id()].insert(this);
}
void LoadInstr::remove_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place.erase(dst->get_id());
    func->use_place[addr->get_id()].erase(this);
}
void LoadInstr::change_use_name(int src_id, shared_ptr<MemObject> new_val) {
    if(addr->get_id() != src_id) {
        printf("Error happening on change LoadInstr\n");
        assert(false);
        return;
    }
    auto func = get_belongto_bb()->get_belongto_func();
    func->use_place[addr->get_id()].erase(this);
    addr = new_val;
    func->use_place[addr->get_id()].insert(this);
}
void LoadInstr::gen_ir(std::ostream &os) const {
    os << "%" << dst->get_id()  << " = load " << type_transe(dst->get_type())
       << ", " << type_transe(addr->get_type())  << "* %" << addr->get_id();
}
Instruction *LoadInstr::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_addr = addr;
    replace_reg(new_addr, reg_maps);
    auto new_dst = std::dynamic_pointer_cast<Ir::Reg>(reg_maps[dst->get_id()]);

    auto new_load = new LoadInstr(new_dst, new_addr);
    return new_load;
}

PhiInstr::PhiInstr(const shared_ptr<Reg> &dst,const unordered_map<BasicBlock*, shared_ptr<Reg>> &bb2value): dst(dst),bb2value(bb2value) {}

std::unordered_set<std::shared_ptr<Type>> PhiInstr::get_uses() const {
    unordered_set<shared_ptr<Type>> uses;
    for (const auto &item : bb2value) {
        uses.insert(item.second);
    }
    return uses;
}
const shared_ptr<Reg> &PhiInstr::get_dst() const {
    return dst;
}
unordered_map<BasicBlock *, shared_ptr<Reg>> &PhiInstr::get_bb_2_value() {
    return bb2value;
}
void PhiInstr::add_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place[dst->get_id()] = this;
    for(auto &iter:bb2value) {
        if(iter.second->get_id() >= 0) {
            func->use_place[iter.second->get_id()].insert(this);
        }
    }
}
void PhiInstr::remove_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place.erase(dst->get_id());
    for(auto &iter:bb2value) {
        if(iter.second->get_id() >= 0) {
            func->use_place[iter.second->get_id()].erase(this);
        }
    }
}
void PhiInstr::change_use_name(int src_id, shared_ptr<Reg> new_val) {
    auto func = get_belongto_bb()->get_belongto_func();
    if(src_id == 6161) {
      int tmp = 0;
    }
    for(auto &iter:bb2value) {
        if(iter.second->get_id() >= 0 && iter.second->get_id() == src_id) { // 应该修改basicblock位置的,因为new_val的def不一定在同一个bb,但通过def去找应该问题不大
            func->use_place[iter.second->get_id()].erase(this);
            iter.second = new_val;
            if(iter.second->get_id() >= 0)
            func->use_place[iter.second->get_id()].insert(this);
        }
    }
}
PhiInstr::PhiInstr(const shared_ptr<Reg> &dst) :
    dst(dst) {
}
void PhiInstr::gen_ir(std::ostream &os) const {
    os  << "%" << dst->get_id() << " = phi " << type_transe(dst->get_type()) << " ";
    for (const auto &item : bb2value) {
        if (item != *bb2value.begin()) {
            os << ", ";
        }
        os << "[ " << *(item.second) << ", %BB" << item.first->get_id() << " ]";
    }
}
shared_ptr<Type> PhiInstr::def() const {
    return dst;
}
Instruction *PhiInstr::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_dst = std::dynamic_pointer_cast<Ir::Reg>(reg_maps[dst->get_id()]);
    auto new_phi = new PhiInstr(new_dst);
    for (const auto &item : bb2value) {
        auto new_bb = map[item.first];
        auto new_reg= item.second;
        if (new_bb == nullptr) {
            int x = 10;
        }
        replace_reg(new_reg, reg_maps);
        new_phi->bb2value.insert({new_bb, new_reg});
    }
    return new_phi;
}

std::shared_ptr<Reg> Factory::get_reg(ScalaType type) {

    auto reg = std::make_shared<Reg>(type, reg_num++);
    return reg;
}

std::shared_ptr<MemObject> Factory::get_mem(ScalaType type) {
    return std::make_shared<MemObject>(type,reg_num++);
}

std::shared_ptr<Vector> Factory::get_vector(ScalaType type, const std::string& name,  bool constant) {
    return std::make_shared<Vector>(type, name, constant);
}


std::shared_ptr<Vector> Factory::get_vector(ScalaType type) {
    return std::make_shared<Vector>(type, reg_num++);
}

void Factory::reset(Function *func) {
    reg_num = func->get_reg_max_num();
    bb_num = func->get_bb_max_num();
}
std::shared_ptr<Reg> Factory::get_const_reg(ScalaType type, RegConst num) {
    auto reg = std::make_shared<Reg>(type, -1);
    reg->set_constant(true);
    reg->set_num(num);
    return reg;
}

std::shared_ptr<Reg> Factory::get_global_reg(ScalaType type, RegConst num, std::string& name) {
    auto reg = std::make_shared<Reg>(type, -1);
    reg->set_global(true);
    reg->set_num(num);
    reg->set_global_name(name);
    return reg;
}

Function* Factory::get_func(std::string &name, ScalaType type,  CompileUnit * cu) {
    auto func = std::make_unique<Function>(name, type);
    auto ptr = func.get();
    cu->get_functions().push_back(std::move(func));
    return ptr;
}

void Factory::build_alloca(ScalaType type, BasicBlock &bb) {
    std::unique_ptr<Instruction> alloca_instr(new AllocaInstr(type, reg_num++));
    bb.push_back(alloca_instr);
}
bool isCompare(BinaryType binaryType) {
    return binaryType == EQ || binaryType == NE || binaryType == SGE || binaryType == SGT || binaryType == SLE ||
           binaryType == SLT;
}


void Factory::build_alloca(const shared_ptr<Vector> &addr, BasicBlock &bb) {
    std::unique_ptr<Instruction> alloca_instr(new AllocaInstr(addr));
    bb.push_back(alloca_instr);
}

std::unique_ptr<Instruction> Factory::get_alloca(ScalaType type) {
    std::unique_ptr<Instruction> alloca_instr(new AllocaInstr(type, reg_num++));
    return alloca_instr;
}

std::unique_ptr<PhiInstr> Factory::get_phi(const shared_ptr<Reg> &dst) {
    std::unique_ptr<PhiInstr> phi_instr(new PhiInstr(dst));
    return phi_instr;
}

std::unique_ptr<BinaryInstr> Factory::get_binary(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &src1, const shared_ptr<Reg> &src2, BinaryType binaryType) {
    std::unique_ptr<BinaryInstr> binary_instr(new BinaryInstr(dst, src1, src2, binaryType));
    return binary_instr;
}

void Factory::build_binary(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &src1, const shared_ptr<Reg> &src2, BinaryType binaryType, BasicBlock &bb) {
    std::unique_ptr<Instruction> binary_instr(new BinaryInstr(dst, src1, src2, binaryType));
    bb.push_back(binary_instr);
}


void Factory::build_store( const shared_ptr<Reg> &val, const shared_ptr<MemObject> &addr, BasicBlock& bb) {
    std::unique_ptr<Instruction> store_instr(new StoreInstr(val, addr));
    bb.push_back(store_instr);
}
BasicBlock* Factory::get_new_bb(Loop * cur_loop, Function* func) {

    std::unique_ptr<BasicBlock> bb(new BasicBlock(bb_num++));
    bb->set_belongto_func(func);
    if (cur_loop != nullptr) {

        cur_loop->get_all_bbs().insert(bb.get());
        bb->set_belong_to_loop(cur_loop);
        bb->set_depth(cur_loop->get_depth());
    }
    func->get_bbs().push_back(std::move(bb));
    return func->get_bbs().back().get();
}

std::unique_ptr<BasicBlock> Factory::get_noadd_new_bb(Loop * cur_loop, Function* func) {

    std::unique_ptr<BasicBlock> bb(new BasicBlock(bb_num++));
    bb->set_belongto_func(func);
    if (cur_loop != nullptr) {

        cur_loop->get_all_bbs().insert(bb.get());
        bb->set_belong_to_loop(cur_loop);
        bb->set_depth(cur_loop->get_depth());
    }

    return bb;
}

void Factory::build_load(const shared_ptr<Reg> &dst,
                         const shared_ptr<MemObject> &addr, BasicBlock &bb) {
    std::unique_ptr<Instruction> load_instr(new LoadInstr(dst, addr));
    bb.push_back(load_instr);
}
void Factory::build_call(const shared_ptr<Ir::Reg> &dst, Function *function,
                           const std::vector<shared_ptr<Type>>&params, BasicBlock& bb) {
    std::unique_ptr<Instruction> call_instr(new CallInstr(dst, function, params));
    bb.push_back(call_instr);
}

void Factory::build_call(Function *function, const std::vector<shared_ptr<Type>>&params, BasicBlock& bb){
    std::unique_ptr<Instruction> call_instr(new CallInstr(function, params));
    bb.push_back(call_instr);
}

void Factory::build_load_array(const shared_ptr<Reg> &dst, const shared_ptr<Vector> &addr,
                                    const std::vector<shared_ptr<Reg>>& indexs, BasicBlock& bb) {
    auto mem = std::make_shared<MemObject>(dst->get_type(), reg_num++);;
    std::unique_ptr<Instruction> load_array(new LoadArray(mem, addr, indexs));
    std::unique_ptr<Instruction> load_instr(new LoadInstr(dst,mem));
    bb.push_back(load_array);
    bb.push_back(load_instr);
}

void Factory::build_store_array(const shared_ptr<Reg> &src, const shared_ptr<Vector> &addr,
                       const std::vector<shared_ptr<Reg>>& indexs, BasicBlock&bb) {
    auto mem = std::make_shared<MemObject>(addr->get_type(), reg_num++);;
    std::unique_ptr<Instruction> load_array(new LoadArray(mem, addr, indexs));
    std::unique_ptr<Instruction> store_instr(new StoreInstr(src, mem));
    bb.push_back(load_array);
    bb.push_back(store_instr);
}


void Factory::build_load_vector(const shared_ptr<Vector> &dst, const shared_ptr<Vector> &addr,
                  const std::vector<shared_ptr<Reg>>&indexs, BasicBlock& bb) {
    std::unique_ptr<Instruction> load_vector(new LoadVector(dst, addr, indexs));
    bb.push_back(load_vector);
}

std::unique_ptr<Instruction> Factory::get_convert(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &src) {
    std::unique_ptr<Instruction> convert(new Convert(dst, src));
    return convert;
}

std::shared_ptr<Ir::Reg> Factory::get_zero_reg(ScalaType type) {
    auto reg =std::make_shared<Ir::Reg>(type, -1);
    reg->set_constant(true);
    return reg;
}

void Factory::build_branch(BasicBlock *bb1, BasicBlock &bb) {
    if (bb.is_terminal()) {
        return ;
    }
    std::unique_ptr<Instruction> brach(new BranchInstr(bb1));
    bb.push_back(brach);
    bb.set_terminal(true);
}

void Factory::build_branch(const shared_ptr<Reg> &reg, BasicBlock *bb1, BasicBlock *bb2,BasicBlock &bb) {
    if (bb.is_terminal()) {
        return ;
    }
    std::unique_ptr<Instruction> brach(new BranchInstr(reg, bb1, bb2));
    bb.push_back(brach);
    bb.set_terminal(true);
}

Loop *Factory::get_loop(BasicBlock *header, BasicBlock *exitingBb, int depth, Function *belongToFunc, Loop *pa) {
    std::unique_ptr<Loop> loop(new Loop(header, exitingBb, depth, belongToFunc, pa));
    belongToFunc->get_loops().push_back(std::move(loop));
    return belongToFunc->get_loops().back().get();
}
void Factory::build_ret(BasicBlock &bb) {
    if (bb.is_terminal()) {
        return ;
    }
    bb.set_terminal(true);
    std::unique_ptr<Instruction> ret(new Ret());
    bb.push_back(ret);
}
void Factory::build_ret(const shared_ptr<Reg> &src, BasicBlock& bb) {
    if (bb.is_terminal()) {
        return ;
    }
    bb.set_terminal(true);
    std::unique_ptr<Instruction> ret(new Ret(src));
    bb.push_back(ret);
}
void Factory::build_load_global(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &addr,
                                BasicBlock &bb) {
    std::unique_ptr<Instruction> load_global(new LoadGlobal(dst, addr));
    bb.push_back(load_global);
}

std::shared_ptr<Ir::Type> Factory::copy_def(const shared_ptr<Type>& def) {

    Instanceof(reg, Ir::Reg*, def.get()) {
        return get_reg(def->get_type());
    }
    Instanceof(reg, Ir::Vector*, def.get()) {
        auto vec = get_vector(def->get_type());
        vec->set_dims(reg->get_dims());
        return vec;
    }
    Instanceof(reg, Ir::MemObject*, def.get()) {
        auto mem = get_mem(def->get_type());
        return mem;
    }
    assert(false);

}

void Factory::build_store_global(const shared_ptr<Reg> &val,
                                 const shared_ptr<Reg> &addr, BasicBlock &bb) {
    std::unique_ptr<Instruction> store_global(new StoreGlobal(val, addr));
    bb.push_back(store_global);
}


Reg::Reg(ScalaType type, int id) :
    Type(id, type) {
}

bool Reg::is_constant() const {
    return constant;
}
void Reg::set_constant(bool constant) {
    Reg::constant = constant;
}
bool Reg::is_param() const {
    return param;
}
void Reg::set_param(bool param) {
    Reg::param = param;
}
bool Reg::is_global() const {
    return global;
}
void Reg::set_global(bool global) {
    Reg::global = global;
}

RegConst &Reg::get_num() {
    return num;
}
void Reg::set_num(const RegConst &num) {
    Reg::num = num;
}
const std::string &Reg::get_global_name() const {
    return global_name;
}
void Reg::set_global_name(const std::string &globalName) {
    global_name = globalName;
}
std::ostream &operator<<(std::ostream &os, const Reg &reg) {
    if (reg.global){
        os << "@" << reg.global_name;

    } else if (reg.constant) {
        if (reg.get_type() == Ir::Int || reg.get_type() == Ir::Bool) {
            os << reg.num.int_num;
        } else {
            os << std::fixed <<std::setprecision(32) <<reg.num.float_num;
        }
    } else  {
        os << "%" << reg.get_id();
    }
    return os;
}

bool Reg::operator==(const Reg &rhs) const {
    if (this->is_global() || rhs.is_global()) {
        return this->global_name == rhs.global_name;
    }
    if (this->is_constant() != rhs.is_constant() || (this->get_type() != rhs.get_type())) {
        return false;
    }
    if (this->is_constant()) {
        if (this->get_type() == Ir::Float) {
            return this->num.float_num == rhs.num.float_num;
        } else {
            return this->num.int_num == rhs.num.int_num;
        }
    } else {
        return this->get_id() == rhs.get_id();
    }
}

bool Reg::operator!=(const Reg &rhs) const {
    return !(rhs == *this);
}

const std::string &Vector::get_name() const {
    return name;
}
std::vector<int> &Vector::get_dims() {
    return dims;
}
Vector::Vector(ScalaType type, std::string name, bool constant) :
    MemObject(type, -1), name(std::move(name)), constant(constant) {
}
Vector::Vector(ScalaType type, int id) :
    MemObject(type, id) {
}
std::ostream &operator<<(std::ostream &os, const Vector &vector) {
    if (vector.get_id() == -1) {
        os << "@" << vector.name;
    } else {
        os << "%" << vector.get_id();
    }
    return os;
}
bool Vector::is_param() const {
    return param;
}
void Vector::set_param(bool param) {
    Vector::param = param;
}
bool Vector::is_constant() const {
    return constant;
}
void Vector::set_dims(const std::vector<int> &dims) {
    Vector::dims = dims;
}
Function::Function(std::string name, ScalaType type) :
    name(std::move(name)), type(type) {
}
bool Function::is_recur() const {
    return recur;
}
void Function::set_recur(bool recur) {
    Function::recur = recur;
}
const std::string &Function::get_name() const {
    return name;
}
ScalaType Function::get_type() const {
    return type;
}
std::list<std::unique_ptr<BasicBlock>> &Function::get_bbs()  {
    return bbs;
}

std::list<std::shared_ptr<Type>> &Function::get_params() {
    return params;
}

bool Function::is_void() const {
    return type == Void;
}
std::list<std::unique_ptr<Loop>> &Function::get_loops() {
    return loops;
}
std::ostream &operator<<(std::ostream &os, const Function &function) {
    if (!function.lib) {
        os << "define ";
    } else {
        os << "declare ";
    }
    os << type_transe(function.get_type()) << " @" << function.get_name() << "(";
    auto params = function.params;
    auto len = params.size();
    auto it = params.begin();
    for (int i = 0; i < len; ++i) {
        Instanceof(addr, Ir::Vector*, (*it).get()) {
            os << vector_transe(addr, 1) << "* " << "%" << addr->get_id();
        } else Instanceof(reg, Ir::Reg*, (*it).get()) {
            os <<  type_transe((*it)->get_type()) << " " << *reg;
        }
        if (i != len - 1) {
            os << ", ";
        }
        it++;
    }
    os << ")";
    if (function.lib) {
        return os;
    }
    os << "{" << std::endl;
    for (auto &bb : function.bbs) {
        os << *bb.get() << std::endl;
    }
    os << "}" << std::endl;
    return os;
}
int Function::get_reg_max_num() const {
    return reg_max_num;
}
void Function::set_reg_max_num(int regMaxNum) {
    reg_max_num = regMaxNum;
}
int Function::get_bb_max_num() const {
    return bb_max_num;
}
void Function::set_bb_max_num(int bbMaxNum) {
    bb_max_num = bbMaxNum;
}
bool Function::is_lib() const {
    return lib;
}
void Function::set_lib(bool lib) {
    Function::lib = lib;
}
bool Function::is_param(Type *type) {
    Instanceof(reg,Reg *,type) {
        for(const auto& para:params) {
            Instanceof(reg_p,Reg *,para.get()) {
                if(reg_p->get_id() == reg->get_id()) {
                    return true;
                }
            }
        }
    } else Instanceof(memObject,MemObject*,type) {
        for(const auto& para:params) {
            Instanceof(memObject_p,MemObject*,para.get()) {
                if(memObject_p->get_id() == memObject->get_id()) {
                    return true;
                }
            }
        }
    } else Instanceof(vector,Vector *,type) {
        for(const auto& para:params) {
            Instanceof(vector_p,Vector *,para.get()) {
                if(vector_p->get_id() == vector->get_id()) {
                    return true;
                }
            }
        }
    }
    return false;
}
BasicBlock *Function::get_first_bb() const {
    return first_bb;
}
void Function::set_first_bb(BasicBlock *firstBb) {
    first_bb = firstBb;
}

MemObject::MemObject(ScalaType type, int id) :
    Type(id, type) {}


std::list<std::unique_ptr<Instruction>> &BasicBlock::get_instrs() {
    return instrs;
}

BasicBlock *Instruction::get_belongto_bb() const {
    return belongto_bb;
}

void Instruction::set_belongto_bb(BasicBlock *belongtoBb) {
    belongto_bb = belongtoBb;
}
std::ostream &operator<<(std::ostream &os, const Instruction &instruction) {
    os << "    ";
    instruction.gen_ir(os);
    return os;
}
void Instruction::gen_ir(std::ostream &os) const {
    os << "no imp";
}

void BasicBlock::push_back(std::unique_ptr<Instruction>& instr) {
    instr->set_belongto_bb(this);
    instr->add_use_def();
    instrs.push_back(std::move(instr));

}

void BasicBlock::push_front(Instruction* instr) {
    instr->set_belongto_bb(this);
    instr->add_use_def();
    instrs.emplace_front(instr);
}

void BasicBlock::push_front(std::unique_ptr<Instruction> instr) {
    instr->set_belongto_bb(this);
    instr->add_use_def();
    instrs.push_front(std::move(instr));
}


BasicBlock::BasicBlock(int id) :
    id(id) {
}

Function *BasicBlock::get_belongto_func() const {
    return belongto_func;
}

void BasicBlock::set_belongto_func(Function *belongtoFunc) {
    belongto_func = belongtoFunc;
}
Loop *BasicBlock::get_belong_to_loop() const {
    return belong_to_loop;
}
void BasicBlock::set_belong_to_loop(Loop *belongToLoop) {
    belong_to_loop = belongToLoop;
}
int BasicBlock::get_depth() const {
    return depth;
}
void BasicBlock::set_depth(int depth) {
    BasicBlock::depth = depth;
}
bool BasicBlock::is_terminal() const {
    return terminal;
}
void BasicBlock::set_terminal(bool terminal) {
    BasicBlock::terminal = terminal;
}
std::ostream &operator<<(std::ostream &os, const BasicBlock &block) {
    os << "BB" << block.id << ":" << std::endl;
    if (block.idom_by != nullptr) {
        os << "\t; idom: BB" << block.idom_by->get_id() << std::endl;
    }
    os << "\t; df: ";
    for (const auto &item : block.df) {
        os << "BB" << item->get_id() << ", ";
    }
    os << std::endl;
    os << "\t;depth: " << block.get_depth() << std::endl;
    for (auto &inst : block.instrs) {
        os << *inst.get() << std::endl;
    }
    return os;
}
int BasicBlock::get_id() const {
    return id;
}
void BasicBlock::set_id(int id) {
    BasicBlock::id = id;
}
void BasicBlock::insert_after_phi(Instruction *instr) {
    auto iter = instrs.begin();
    for(; iter != instrs.end();) {
        Instanceof(phi,PhiInstr*,iter->get()) {
            iter ++;
        } else Instanceof(array,ArrayPhi*,iter->get()) {
            iter ++;
        }
        else {
            break;
        }
    }
    instrs.emplace(iter,instr);

    instr->set_belongto_bb(this);
    instr->add_use_def();
}
int BasicBlock::get_loop_level() {
    return !belong_to_loop ? 0 : belong_to_loop->get_depth();
}
void BasicBlock::insert_before_br(Instruction *instr) {
    auto iter = instrs.begin();
    while(iter != instrs.end()) {
        Instanceof(br,BranchInstr*,iter->get()) {
            break;
        } else Instanceof(binary,BinaryInstr*,iter->get()) {
            if(binary->get_binary_type() == EQ || binary->get_binary_type() == NE || binary->get_binary_type() == SLT
                || binary->get_binary_type() == SLE || binary->get_binary_type() == SGE || binary->get_binary_type() == SGT)
                break;
            else {
                iter ++;
            }
        } else {
            iter++;
        }
    }
    instrs.emplace(iter,instr);
    instr->set_belongto_bb(this);
    instr->add_use_def();
}

unordered_set<shared_ptr<Type>> LoadArray::get_uses() const {
    unordered_set<shared_ptr<Type>> uses;
    for (auto &item : indexs) {
        uses.insert(item);
    }
    uses.insert(addr);
    return uses;
}


LoadArray::LoadArray(const shared_ptr<MemObject> &dst, const shared_ptr<Vector> &addr,
                     const std::vector<shared_ptr<Reg>>& _indexs) :dst(dst), addr(addr) {
    for (const auto &item : _indexs) {
        indexs.push_back(item);
    }
}

const shared_ptr<Vector> &LoadArray::get_addr() const {
    return addr;
}
std::vector<shared_ptr<Reg>> &LoadArray::get_indexs() {
    return indexs;
}
const shared_ptr<MemObject> &LoadArray::get_dst() const {
    return dst;
}
void LoadArray::gen_ir(std::ostream &os) const {
    os << "%" << dst->get_id() << " = ";
    os << "getelementptr " << vector_transe(addr.get(), 0) <<  ", "
        << vector_transe(addr.get(), 0)  << "* " << *addr.get() << ", " ;
    if (addr->get_dims().front() != 0) {
        os << "i32 0, ";
    }
    for (int i = 0; i < indexs.size(); ++i) {
        os << type_transe(indexs[i]->get_type()) << " " << *indexs[i];
        if (i + 1 != indexs.size()) {
            os << ", ";
        }
    }
}
void LoadArray::add_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place[dst->get_id()] = this;
    func->use_place[addr->get_id()].insert(this);
    for(auto &reg:indexs) {
        if(reg->get_id() < 0)
            continue ;
        func->use_place[reg->get_id()].insert(this);
    }
}
void LoadArray::remove_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place.erase(dst->get_id());
    func->use_place[addr->get_id()].erase(this);
    for(auto &reg:indexs) {
        if(reg->get_id() < 0)
            continue ;
        func->use_place[reg->get_id()].erase(this);
    }
}
void LoadArray::change_use_name(int src_id, shared_ptr<Reg> new_val) {
    auto func = get_belongto_bb()->get_belongto_func();
    for(int i = 0;i < indexs.size();i++) {
        auto &reg = indexs[i];
        if(reg->get_id() < 0 || reg->get_id() != src_id)
            continue ;
        func->use_place[reg->get_id()].erase(this);
        indexs[i] = new_val;
        if(new_val->get_id())
            func->use_place[new_val->get_id()].insert(this);
    }
}
void LoadArray::change_use_name(int src_id, shared_ptr<Vector> new_val) {
    if(addr->get_id() != src_id) {
        printf("Error happening on LoadArray maintain use");
        assert(false);
        return ;
    }
    auto func = get_belongto_bb()->get_belongto_func();
    func->use_place[addr->get_id()].erase(this);
    addr = new_val;
    func->use_place[addr->get_id()].insert(this);
}

shared_ptr<Type> LoadArray::def() const {
    return dst;
}

Instruction *LoadArray::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_index = indexs;
    if (addr->get_name() == "que") {
        int x = 10;
    }
    for (auto &item : new_index) {
        replace_reg(item, reg_maps);
    }
    auto new_addr = addr;
    replace_reg(new_addr, reg_maps);
    auto new_dst = std::dynamic_pointer_cast<Ir::MemObject>(reg_maps[dst->get_id()]);
    auto new_load = new LoadArray(new_dst, new_addr, new_index);
    return new_load;
}

LoadVector::LoadVector(const shared_ptr<Vector> &dst, const shared_ptr<Vector> &addr,
                       const std::vector<shared_ptr<Reg>>& indexs): dst(dst), addr(addr),
    indexs(indexs){
}

CallInstr::CallInstr(const shared_ptr<Ir::Reg> &dst, Function *function,const std::vector<shared_ptr<Type>>& params) :
    dst(dst), function(function), params(params), void_flag(false){
}

CallInstr::CallInstr( Function *function,const std::vector<shared_ptr<Type>>& params) :
    function(function), params(params), void_flag(true) {
}

unordered_set<shared_ptr<Type>> CallInstr::get_uses() const {
    unordered_set<shared_ptr<Type>> uses;
    for (const auto &item: params) {
        uses.insert(item);
    }
    for (const auto &item: used_global) {
        uses.insert(item);
    }
    return uses;
}

shared_ptr<Type> CallInstr::def() const {
    if (void_flag) {
        return nullptr;
    } else {
        return {dst};
    }
}
shared_ptr<Ir::Reg> &CallInstr::get_dst() {
    return dst;
}
Function *CallInstr::get_function() const {
    return function;
}
std::vector<shared_ptr<Type>> &CallInstr::get_params() {
    return params;
}
bool CallInstr::is_void_flag() const {
    return void_flag;
}
void CallInstr::gen_ir(std::ostream &os) const {
    if (dst) {
        os << *dst.get() << " = ";
    }
    os << "call "  << type_transe(function->get_type()) << " @" + function->get_name() << "(";
    auto len = params.size();
    for (int i = 0; i < len; ++i) {
        Instanceof(addr, Ir::Vector*, params[i].get()) {
            os << vector_transe(addr, 0) << "* " << "%" << addr->get_id();
        } else Instanceof(reg, Ir::Reg*, params[i].get()) {
            os <<  type_transe(params[i]->get_type()) << " " << *reg;
        }
        if (i != len - 1) {
            os << ", ";
        }
    }
    os << ")";

}

void CallInstr::add_use_def() {
    for(auto& type:params) {
        Instanceof(reg,Reg*,type.get()) {
            if(reg->get_id() >= 0) {
                get_belongto_bb()->get_belongto_func()->use_place[reg->get_id()].insert(this);
            }
        } else Instanceof(vector,Vector*,type.get()) {
            get_belongto_bb()->get_belongto_func()->use_place[vector->get_id()].insert(this);
        } else Instanceof(memobject,MemObject*,type.get()) {
            get_belongto_bb()->get_belongto_func()->use_place[memobject->get_id()].insert(this);
        }
    }
    for(auto& type:used_global) {
        Instanceof(reg,Reg*,type.get()) {
            if(reg->get_id() >= 0) {
                get_belongto_bb()->get_belongto_func()->use_place[reg->get_id()].insert(this);
            }
        } else Instanceof(vector,Vector*,type.get()) {
            get_belongto_bb()->get_belongto_func()->use_place[vector->get_id()].insert(this);
        } else Instanceof(memobject,MemObject*,type.get()) {
            get_belongto_bb()->get_belongto_func()->use_place[memobject->get_id()].insert(this);
        }
    }
    if(!dst)
        return;
    get_belongto_bb()->get_belongto_func()->def_place[dst->get_id()] = this;
}

void CallInstr::remove_use_def() {
    for(auto& type:params) {
        Instanceof(reg,Reg*,type.get()) {
            if(reg->get_id() >= 0) {
                get_belongto_bb()->get_belongto_func()->use_place[reg->get_id()].erase(this);
            }
        } else Instanceof(vector,Vector*,type.get()) {
            get_belongto_bb()->get_belongto_func()->use_place[vector->get_id()].erase(this);
        } else Instanceof(memobject,MemObject*,type.get()) {
            get_belongto_bb()->get_belongto_func()->use_place[memobject->get_id()].erase(this);
        }
    }
    if(void_flag)
        return;
    get_belongto_bb()->get_belongto_func()->def_place.erase(dst->get_id());
}
void CallInstr::change_use_name(int src_id,shared_ptr<Reg> new_val) {
    int id = -1;
    for(int i = 0;i<params.size();i++) {
        auto& type = params[i];
        Instanceof(reg,Reg*,type.get()) {
            id = reg->get_id();
        }
        if(id == src_id) {
            params[i] = new_val;
            get_belongto_bb()->get_belongto_func()->use_place[id].erase(this);
            if(new_val->get_id() >= 0) {
                get_belongto_bb()->get_belongto_func()->use_place[new_val->get_id()].insert(this);
            }
            id = -1;
        }
    }

    for(auto iter = used_global.begin();iter != used_global.end();) {
        auto type = iter;
        id = -1;
        Instanceof(reg,Reg*,type->get()) {
            id = reg->get_id();
        }else {
            iter++;
            continue;
        }
        if(id == src_id) {
            iter = used_global.erase(iter);
            iter = used_global.insert(iter,new_val);
            get_belongto_bb()->get_belongto_func()->use_place[id].erase(this);
            if(new_val->get_id() >= 0) {
                get_belongto_bb()->get_belongto_func()->use_place[new_val->get_id()].insert(this);
            }
        } else {
            iter++;
            continue;
        }
    }
}
void CallInstr::change_use_name(int src_id,shared_ptr<MemObject> new_val) {
    int id = -1;
    for(int i = 0;i<params.size();i++) {
        auto& type = params[i];
        Instanceof(memobject,MemObject*,type.get()) {
            id = memobject->get_id();
        }else continue ;
        if(id == src_id) {
            params[i] = new_val;
            get_belongto_bb()->get_belongto_func()->use_place[id].erase(this);
            if(new_val->get_id() >= 0)
                get_belongto_bb()->get_belongto_func()->use_place[new_val->get_id()].insert(this);
            id = -1;
        }
    }

    for(auto iter = used_global.begin();iter != used_global.end();) {
        auto type = iter;
        id = -1;
        Instanceof(reg,MemObject*,type->get()) {
            id = reg->get_id();
        }else {
            iter++;
            continue;
        }
        if(id == src_id) {
            iter = used_global.erase(iter);
            iter = used_global.insert(iter,new_val);
            get_belongto_bb()->get_belongto_func()->use_place[id].erase(this);
            if(new_val->get_id() >= 0) {
                get_belongto_bb()->get_belongto_func()->use_place[new_val->get_id()].insert(this);
            }
        } else {
            iter++;
            continue;
        }
    }
}
void CallInstr::change_use_name(int src_id,shared_ptr<Vector> new_val) {
    int id = -1;
    for(int i = 0;i<params.size();i++) {
        auto& type = params[i];
        Instanceof(vector,Vector*,type.get()) {
            id = vector->get_id();
        }
        if(id == src_id) {
            params[i] = new_val;
            get_belongto_bb()->get_belongto_func()->use_place[id].erase(this);
            if(new_val->get_id() >= 0)
                get_belongto_bb()->get_belongto_func()->use_place[new_val->get_id()].insert(this);
            id = -1;
        }
    }

    for(auto iter = used_global.begin();iter != used_global.end();) {
        auto type = iter;
        id = -1;
        Instanceof(reg,Vector*,type->get()) {
            id = reg->get_id();
        }else {
            iter++;
            continue;
        }
        if(id == src_id) {
            iter = used_global.erase(iter);
            iter = used_global.insert(iter,new_val);
            get_belongto_bb()->get_belongto_func()->use_place[id].erase(this);
            if(new_val->get_id() >= 0) {
                get_belongto_bb()->get_belongto_func()->use_place[new_val->get_id()].insert(this);
            }
        } else {
            iter++;
            continue;
        }
    }
}
void CallInstr::set_dst(const shared_ptr<Ir::Reg> &dst) {
    CallInstr::dst = dst;
}
Instruction *CallInstr::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_params = params;

    for (auto &item : new_params) {
        replace_reg(item, reg_maps);
    }
    if (!is_void_flag()) {
        auto new_dst = std::dynamic_pointer_cast<Ir::Reg>(reg_maps[dst->get_id()]);
        auto new_call = new CallInstr(new_dst, function, new_params);
        return new_call;
    } else {
        auto new_call = new CallInstr(function, new_params);
        new_call->set_void_flag(true);
        return new_call;
    }
}
void CallInstr::set_void_flag(bool voidFlag) {
    void_flag = voidFlag;
}

ScalaType Type::get_type() const {
    return type;
}

void Type::set_type(ScalaType type) {
    Type::type = type;
}


int Type::get_id() const {
    return id;
}
void Type::set_id(int id) {
    Type::id = id;
}
Type::Type(int id, ScalaType type) :
    id(id), type(type) {
}

unordered_set<shared_ptr<Type>> LoadVector::get_uses() const {
    unordered_set<shared_ptr<Type>> uses;
    for (auto &item : indexs) {
        uses.insert(item);
    }
    uses.insert(addr);
    return uses;
}
const shared_ptr<Vector> &LoadVector::get_dst() const {
    return dst;
}
const shared_ptr<Vector> &LoadVector::get_addr() const {
    return addr;
}
std::vector<shared_ptr<Reg>> &LoadVector::get_indexs() {
    return indexs;
}
void LoadVector::gen_ir(std::ostream &os) const {
    os <<  *dst.get() << " = ";
    os << "getelementptr " << vector_transe(addr.get(), 0) <<  ", "
       << vector_transe(addr.get(), 0)  << "* " << *addr.get() << ", " ;
    if (addr->get_dims().front() != 0) {
        os << "i32 0";
           os << ", ";

    }
    for (int i = 0; i < indexs.size(); ++i) {
        os << type_transe(indexs[i]->get_type()) << " " << *indexs[i];

            os << ", ";

    };
    os << "i32 0";
}
void LoadVector::add_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place[dst->get_id()] = this;
    func->use_place[addr->get_id()].insert(this);
    for(auto &reg:indexs) {
        if(reg->get_id() < 0)
            continue ;
        func->use_place[reg->get_id()].insert(this);
    }

}
void LoadVector::remove_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place.erase(dst->get_id());
    func->use_place[addr->get_id()].erase(this);
    for(auto &reg:indexs) {
        if(reg->get_id() < 0)
            continue ;
        func->use_place[reg->get_id()].erase(this);
    }
}
void LoadVector::change_use_name(int src_id, shared_ptr<Reg> new_val) {
    auto func = get_belongto_bb()->get_belongto_func();
    for(int i = 0;i < indexs.size();i++) {
        auto &reg = indexs[i];
        if(reg->get_id() < 0 || reg->get_id() != src_id)
            continue ;
        func->use_place[reg->get_id()].erase(this);
        indexs[i] = new_val;
        if(new_val->get_id() >= 0)
            func->use_place[new_val->get_id()].insert(this);
    }
}
void LoadVector::change_use_name(int src_id, shared_ptr<Vector> new_val) {
    if(addr->get_id() != src_id) {
        printf("Error happening on LoadVector maintain use");
        assert(false);
        return ;
    }
    auto func = get_belongto_bb()->get_belongto_func();
    func->use_place[addr->get_id()].erase(this);
    addr = new_val;
    func->use_place[addr->get_id()].insert(this);
}
shared_ptr<Type> LoadVector::def() const {
    return dst;
}
Instruction *LoadVector::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_index = indexs;
    for (auto &item : new_index) {
        replace_reg(item, reg_maps);
    }
    auto new_addr = addr;
    replace_reg(new_addr, reg_maps);
    auto new_dst = std::dynamic_pointer_cast<Ir::Vector>(reg_maps[dst->get_id()]);
    auto new_load = new LoadVector(new_dst, new_addr, new_index);
    return new_load;
}

unordered_map<std::string, std::vector<int>> &CompileUnit::get_global_int_arrays() {
    return global_int_arrays;
}

 unordered_map<std::string, std::vector<float>> &CompileUnit::get_global_float_arrays() {
    return global_float_arrays;
}

std::list<shared_ptr<Type>> &CompileUnit::get_globals() {
    return globals;

}

inline void printfloat_16(const float fnum, std::ostream &os){
    uint8_t* c = (uint8_t*)&fnum;   //指针指向低位字节的地址，所以最先传给c的是低位字节
    os << "0x";
    char s[20] = {0};
    for(auto i = 7;i >= 0;i--) {
        char t[20];
        sprintf(t, "%02x", c[i]); // 先获得的是低位字节，所以应该把低位字节放在最后面
        strcat(s, t);
    }
    s[19] = 0;
    os << s;
}


void const_transe(RegConst c, ScalaType type, std::ostream &os) {
    if (type == Ir::Int || type == Ir::Bool) {
        os << c.int_num;
    }  else {

        os << std::fixed << std::setprecision(30)<<c.float_num;
    }
}

std::string CompileUnit::print_array(Vector *vector, int deep, int& num) const{
    if (vector->get_type() == Ir::Float) {
        auto& array = global_float_arrays.find(vector->get_name())->second;
        if (array.size() == 0) {
            return vector_transe(vector, 0) + " zeroinitializer";
        } else {
            auto string = std::string("");
            string +=  vector_transe(vector, deep) + " ";
            string += "[";
            if (deep + 1 == vector->get_dims().size()) {
                for (int i = 0; i < vector->get_dims()[deep]; ++i) {
                    string += "float ";
                    string += std::to_string(array[num + i]);
                    if (i + 1 != vector->get_dims()[deep]) {
                        string += ", ";
                    }
                }
                num += vector->get_dims()[deep];
            } else {

                for (int i = 0; i < vector->get_dims()[deep]; ++i) {
                    string += print_array(vector, deep + 1, num);
                    if (i + 1 != vector->get_dims()[deep]) {
                        string += ", ";
                    }
                }

            }
            string += "]";
            return string ;
        }
    } else {
        auto& array = global_int_arrays.find(vector->get_name())->second;
        if (array.size() == 0) {
            return vector_transe(vector, 0) + " zeroinitializer";
        } else {
            auto string = std::string("");
            string +=  vector_transe(vector, deep) + " ";
            string += "[";
            if (deep + 1 == vector->get_dims().size()) {
                for (int i = 0; i < vector->get_dims()[deep]; ++i) {
                    string += "i32 ";
                    string += std::to_string(array[num + i]);
                    if (i + 1 != vector->get_dims()[deep]) {
                        string += ", ";
                    }
                }
                num += vector->get_dims()[deep];
            } else {

                for (int i = 0; i < vector->get_dims()[deep]; ++i) {
                    string += print_array(vector, deep + 1, num);
                    if (i + 1 != vector->get_dims()[deep]) {
                        string += ", ";
                    }
                }

            }
            string += "]";
            return string;
        }
    }
}


std::ostream &operator<<(std::ostream &os, const CompileUnit &unit) {

    for (const auto &item : unit.globals) {
        auto type = item.get();
        Instanceof(reg, Reg*, type) {

            os << "@" << reg->get_global_name() << " = global "  << type_transe(reg->get_type()) << " ";
            const_transe(reg->get_num(), reg->get_type(), os);
        } else Instanceof(vector, Vector*, type) {
            int num = 0;
            os << "@" << vector->get_name() << " = global" << " " << unit.print_array(vector, 0, num);
        }
        os << std::endl;
    }
    for(auto &item : unit.functions) {
        os << *item.get() << std::endl;
    }
    return os;
}
std::list<std::unique_ptr<Function>> &CompileUnit::get_functions() {
    return functions;
}

bool CompileUnit::in_array_ssa = false;
bool CompileUnit::in_arm = false;

void CompileUnit::reset_num() {
    for (auto &func : functions) {
        Ir::Factory::reg_num = 0;
        Ir::Factory::bb_num = 0;
        for (const auto &item : func->get_params()) {
            item->set_id(Ir::Factory::reg_num++);
        }
        for (auto& bb: func->get_bbs()) {
            bb->set_id(Ir::Factory::bb_num++);
            for ( auto &instr : bb->get_instrs()) {
                Instanceof(def, DefInstr*, instr.get()) {
                    if (def->def()) {

                        def->def()->set_id(Ir::Factory::reg_num++);
                    }
                }
            }
        }
    }
}

//StoreArray::StoreArray(const shared_ptr<Reg> &src, const shared_ptr<Vector> &addr,
//                       const std::vector<shared_ptr<Reg>>& index):src(src), addr(addr), indexs(index) {}
//
//unordered_set<shared_ptr<Reg>> StoreArray::get_uses() const {
//    unordered_set<shared_ptr<Reg>> uses;
//    for (const auto &item : indexs) {
//        uses.insert(item);
//    }
//    return uses;
//}
//const shared_ptr<Reg> &StoreArray::get_src() const {
//    return src;
//}
//const shared_ptr<Vector> &StoreArray::get_addr() const {
//    return addr;
//}
//std::vector<shared_ptr<Reg>> &StoreArray::get_indexs(){
//    return indexs;
//}

BasicBlock *Loop::get_header() const {
    return header;
}
BasicBlock *Loop::get_exiting_bb() const {
    return exiting_bb;
}
int Loop::get_depth() const {
    return depth;
}
Function *Loop::get_belong_to_func() const {
    return belong_to_func;
}
BasicBlock *Loop::get_preheader() const {
    return preheader;
}
Loop *Loop::get_parent() const {
    return parent;
}
unordered_set<BasicBlock *> &Loop::get_all_bbs() {
    return all_bbs;
}
//unordered_set<BasicBlock *> &Loop::get_exit_bbs() {
//    return exit_bbs;
//}
//unordered_set<BasicBlock *> &Loop::get_back_bbs() {
//    return back_bbs;
//}
Loop::Loop(BasicBlock *header, BasicBlock *exitingBb, int depth, Function *belongToFunc, Loop *parent) :
    header(header), exiting_bb(exitingBb), depth(depth), belong_to_func(belongToFunc), parent(parent) {
    if (parent != nullptr) {
        parent->children.insert(this);
    }
    if(!header->get_belong_to_loop()) {
        header->set_belong_to_loop(this);
    } else {
        auto parent_loop = header->get_belong_to_loop();
        parent_loop->children.insert(this);
        header->set_belong_to_loop(this);
        this->set_parent(parent_loop);
    }
    header->set_depth(depth);
    all_bbs.insert(header);
}
void Loop::set_parent(Loop *parent) {
    Loop::parent = parent;
}
void Loop::set_depth(int depth) {
    Loop::depth = depth;
}
void Loop::set_all_bbs(const unordered_set<BasicBlock *> &allBbs) {
    all_bbs = allBbs;
}
unordered_set<shared_ptr<Type>> Ret::get_uses() const {
    if (void_flag) {
        return {};
    } else {
        return {src};
    }
}
Ret::Ret() :void_flag(true){
}
Ret::Ret(const shared_ptr<Reg> &src) :
    void_flag(false), src(src){
}
bool Ret::is_void_flag() const {
    return void_flag;
}
const shared_ptr<Reg> &Ret::get_src() const {
    return src;
}
void Ret::add_use_def() { // only use
    if(void_flag || src->get_id() < 0) {
        return;
    }
    auto func = get_belongto_bb()->get_belongto_func();
    func->use_place[src->get_id()].insert(this);
}
void Ret::remove_use_def() { // only use
    if(void_flag && src->get_id() < 0)
        return;
    auto func = get_belongto_bb()->get_belongto_func();
    func->use_place[src->get_id()].erase(this);
}
void Ret::change_use_name(int src_id, shared_ptr<Reg> new_val) { // only use
    if(void_flag || src_id != src->get_id()) {
        printf("Error trying to change a void ret");
        assert(false);
        return;
    }
    auto func = get_belongto_bb()->get_belongto_func();
    func->use_place[src->get_id()].erase(this);
    src = new_val;
    if(src->get_id() >= 0)
        func->use_place[src->get_id()].insert(this);
}

void Ret::gen_ir(std::ostream &os) const {
    if (void_flag) {
        os << "ret void";
    } else {
        os << "ret " << type_transe(src->get_type()) << " " << *src.get();
    }
}
Instruction *Ret::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    if (is_void_flag()) {
        auto new_ret = new Ret();
        return new_ret;
    } else {
        auto new_src = src;
        replace_reg(new_src, reg_maps);
        auto new_ret = new Ret(new_src);
        return new_ret;
    }
}
LoadGlobal::LoadGlobal(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &addr) :
    dst(dst), addr(addr) {
}
unordered_set<shared_ptr<Type>> LoadGlobal::get_uses() const {
    return {};
}
shared_ptr<Type> LoadGlobal::def() const {
    return dst;
}
const shared_ptr<Reg> &LoadGlobal::get_dst() const {
    return dst;
}
const shared_ptr<Reg> &LoadGlobal::get_addr() const {
    return addr;
}
void LoadGlobal::gen_ir(std::ostream &os) const {
    os << "%" << dst->get_id()  << " = load " << type_transe(dst->get_type())
       << ", " << type_transe(addr->get_type())  << "* "  << *addr.get();
}
void LoadGlobal::add_use_def() { // only def
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place[dst->get_id()] = this;
}

void LoadGlobal::remove_use_def() { // only def
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place.erase(dst->get_id());
}

Instruction *LoadGlobal::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_addr = addr;
    replace_reg(new_addr, reg_maps);
    auto new_dst = std::dynamic_pointer_cast<Ir::Reg>(reg_maps[dst->get_id()]);
    auto new_load = new LoadGlobal(new_dst, new_addr);
    return new_load;
}

StoreGlobal::StoreGlobal(const shared_ptr<Reg> &val, const shared_ptr<Reg> &addr) :
    val(val), addr(addr) {
}
const shared_ptr<Reg> &StoreGlobal::get_val() const {
    return val;
}
const shared_ptr<Reg> &StoreGlobal::get_addr() const {
    return addr;
}
unordered_set<shared_ptr<Type>> StoreGlobal::get_uses() const {
    return {val};
}
void StoreGlobal::add_use_def() {
    if(val->get_id() < 0)
        return;
    get_belongto_bb()->get_belongto_func()->use_place[val->get_id()].insert(this);
}
void StoreGlobal::remove_use_def() {
    if(val->get_id() < 0)
        return;
    get_belongto_bb()->get_belongto_func()->use_place[val->get_id()].erase(this);
}
void StoreGlobal::change_use_name(int src_id, shared_ptr<Reg> new_val) {
    if(val->get_id() == src_id) {
        get_belongto_bb()->get_belongto_func()->use_place[val->get_id()].erase(this);
        val = new_val;
        if(val->get_id() >= 0)
            get_belongto_bb()->get_belongto_func()->use_place[val->get_id()].insert(this);
    }
}
void StoreGlobal::gen_ir(std::ostream &os) const {
    os << "store " << type_transe(val->get_type()) << " " << *val.get()
       << ", " << type_transe(addr->get_type()) + "* " << *addr.get();
}
Instruction *StoreGlobal::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_addr = addr;
    auto new_val = val;
    replace_reg(new_addr, reg_maps);
    replace_reg(new_val, reg_maps);
    auto store = new StoreGlobal(new_val, new_addr);
    return store;
}
Convert::Convert(const shared_ptr<Reg> &dst, const shared_ptr<Reg> &src) :
    dst(dst), src(src) {
}
const shared_ptr<Reg> &Convert::get_dst() const {
    return dst;
}
const shared_ptr<Reg> &Convert::get_src() const {
    return src;
}
unordered_set<shared_ptr<Type>> Convert::get_uses() const {
    return {src};
}
shared_ptr<Type> Convert::def() const {
    return {dst};
}
void Convert::add_use_def() {
    if(dst->get_id() >= 0)
        get_belongto_bb()->get_belongto_func()->def_place[dst->get_id()] = this;
    if(src->get_id() >= 0) {
        auto& users = get_belongto_bb()->get_belongto_func()->use_place[src->get_id()];
        users.insert(this);

    }
}

void Convert::remove_use_def() {
    if(dst->get_id() >= 0)
        get_belongto_bb()->get_belongto_func()->def_place.erase(dst->get_id());
    if(src->get_id() >= 0)
        get_belongto_bb()->get_belongto_func()->use_place[src->get_id()].erase(this);
}
void Convert::change_use_name(int src_id,shared_ptr<Reg> new_val) {
    if(src->get_id() == src_id) {
        get_belongto_bb()->get_belongto_func()->use_place[src->get_id()].erase(this);
        src = new_val;
        if(new_val->get_id() >= 0) {
            get_belongto_bb()->get_belongto_func()->use_place[new_val->get_id()].insert(this);
        }
    }
}
void Convert::gen_ir(std::ostream &os) const {
    os << *dst.get() << " = ";
    if (dst->get_type() == Ir::Int && src->get_type() == Ir::Bool) {
        os << "zext ";
    } else if (dst->get_type() == Ir::Float) {
        os << "sitofp ";
    } else if (src->get_type() == Ir::Float){
        os << "fptosi ";
    } else {
        os << "trunc ";
    }
    os << type_transe(src->get_type()) << " " << *src.get()
       << " to " << type_transe(dst->get_type());
}
Instruction *Convert::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, std::unordered_map<Ir::BasicBlock *, Ir::BasicBlock *>& map) {
    auto new_src = src;
    replace_reg(new_src, reg_maps);
    auto new_dst = std::dynamic_pointer_cast<Ir::Reg>(reg_maps[dst->get_id()]);
    auto new_move = new Convert(new_dst, new_src);
    return new_move;
}
const shared_ptr<Type> &ArrayPhi::get_dst() const {
    return dst;
}
unordered_map<BasicBlock *, shared_ptr<Type>> &ArrayPhi::get_bb_2_value() {
    return bb2value;
}
ArrayPhi::ArrayPhi(const shared_ptr<Type>& dst,const shared_ptr<Type>& scope):dst(dst),scope(scope) {}
unordered_set<shared_ptr<Type>> ArrayPhi::get_uses() const {
    unordered_set<shared_ptr<Type>> ret;
    for(auto &[bb,val]:bb2value) {
        ret.insert(val);
    }
    ret.insert(read_before_write.begin(),read_before_write.end());
    return ret;
}
void ArrayPhi::add_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place[dst->get_id()] = this;
    for(auto &[bb,val]:bb2value) {
        if(val->get_id() >= 0)
        func->use_place[val->get_id()].insert(this);
    }
    for(auto &use:read_before_write) {
        if(use->get_id() >= 0)
        func->use_place[use->get_id()].insert(this);
    }
}
void ArrayPhi::remove_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place.erase(dst->get_id());
    for(auto &[bb,val]:bb2value) {
        if(val->get_id() >= 0)
        func->use_place[val->get_id()].erase(this);
    }
    for(auto &use:read_before_write) {
        if(use->get_id() >= 0)
        func->use_place[use->get_id()].erase(this);
    }
}
void ArrayPhi::change_use_name(int src_id, shared_ptr<Reg> new_val) {
    auto func = get_belongto_bb()->get_belongto_func();
    for(auto &iter:bb2value) {
        if(iter.second->get_id() >= 0 && iter.second->get_id() == src_id) { // 应该修改basicblock位置的,因为new_val的def不一定在同一个bb,但通过def去找应该问题不大
        func->use_place[iter.second->get_id()].erase(this);
        iter.second = new_val;
        if(iter.second->get_id() >= 0)
                func->use_place[iter.second->get_id()].insert(this);
        }
    }

    for(auto iter = read_before_write.begin();iter != read_before_write.end();) {
        auto use = iter->get();
        if(use->get_id() == src_id) {
        func->use_place[use->get_id()].erase(this);
          if(new_val->get_id() < 0) {
                  iter = read_before_write.erase(iter);
                  break;
          } else {
                  iter = read_before_write.erase(iter);
                  iter = read_before_write.insert(iter,new_val);
                  func->use_place[new_val->get_id()].insert(this);
                  break;
          }
        }
        iter++;
    }
}
void ArrayPhi::change_use_name(int src_id, shared_ptr<Vector> new_val) {
    auto func = get_belongto_bb()->get_belongto_func();
    for(auto &iter:bb2value) {
        if(iter.second->get_id() >= 0 && iter.second->get_id() == src_id) { // 应该修改basicblock位置的,因为new_val的def不一定在同一个bb,但通过def去找应该问题不大
        func->use_place[iter.second->get_id()].erase(this);
        iter.second = new_val;
        if(iter.second->get_id() >= 0)
                func->use_place[iter.second->get_id()].insert(this);
        }
    }

    for(auto iter = read_before_write.begin();iter != read_before_write.end();) {
        auto use = iter->get();
        if(use->get_id() == src_id) {
        func->use_place[use->get_id()].erase(this);
        if(new_val->get_id() < 0) {
                iter = read_before_write.erase(iter);
                break;
        } else {
                iter = read_before_write.erase(iter);
                iter = read_before_write.insert(iter,new_val);
                func->use_place[new_val->get_id()].insert(this);
                break;
        }
        }
        iter++;
    }
}
const shared_ptr<Type> &ArrayPhi::get_scope() const {
    return scope;
}
shared_ptr<Type> ArrayPhi::def() const {
    return dst;
}
void ArrayPhi::gen_ir(std::ostream &os) const {
    os  << "%" << dst->get_id() << " = array phi " << type_transe(dst->get_type()) << " ";
    for (const auto &item : bb2value) {
        if (item != *bb2value.begin()) {
        os << ", ";
        }
        os << "[ " << item.second->get_id() << ", %BB" << item.first->get_id() << " ]";
    }
}
Instruction *ArrayPhi::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, unordered_map<Ir::BasicBlock *, Ir::BasicBlock *> &map) {
    assert(false);
    return nullptr;
}
MemWrite::MemWrite(const shared_ptr<Type> &addr, const shared_ptr<Reg> &val, const shared_ptr<Type> &scope, const shared_ptr<Type> new_scope,bool call_write) :
    addr(addr), val(val), scope(scope), call_write(call_write),new_scope(new_scope) {
}
const shared_ptr<Type> &MemWrite::get_addr() const {
    return addr;
}
const shared_ptr<Reg> &MemWrite::get_val() const {
    return val;
}
const shared_ptr<Type> &MemWrite::get_scope() const {
    return scope;
}
void MemWrite::add_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    for(auto &use:get_uses()) {
        if(use->get_id()>=0) {
            func->use_place[use->get_id()].insert(this);
        }
    }
    func->def_place[new_scope->get_id()] = this;
}
void MemWrite::remove_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    for(auto &use:get_uses()) {
        if(use->get_id()>=0) {
            func->use_place[use->get_id()].erase(this);
        }
    }
}
void MemWrite::change_use_name(int src_id, shared_ptr<Reg> new_val) {
    auto func = get_belongto_bb()->get_belongto_func();
    if(val->get_id() == src_id && val->get_id() >= 0) {
        func->use_place[src_id].erase(this);
        val = new_val;
        if(new_val->get_id() >= 0)
            func->use_place[new_val->get_id()].insert(this);
    } else if(addr->get_id() == src_id && addr->get_id() >= 0){
        Instanceof(reg_addr,Reg*,addr.get()) {
            if(reg_addr->get_id() == src_id) {
                func->use_place[src_id].erase(this);
                addr = new_val;
                if (new_val->get_id() >= 0)
                    func->use_place[new_val->get_id()].insert(this);
            }
        }
        Instanceof(reg_scope,Reg *,scope.get()) {
            if(reg_scope->get_id() == src_id) {
                func->use_place[src_id].erase(this);
                scope = new_val;
                if (new_val->get_id() >= 0)
                    func->use_place[new_val->get_id()].insert(this);
            }
        }
    } else {
        for(auto iter = read_before_write.begin();iter != read_before_write.end();) {
            auto use = iter->get();
            if(use->get_id() == src_id) {
                func->use_place[use->get_id()].erase(this);
                if(new_val->get_id() < 0) {
                    iter = read_before_write.erase(iter);
                    break;
                } else {
                    iter = read_before_write.erase(iter);
                    iter = read_before_write.insert(iter,new_val);
                    func->use_place[new_val->get_id()].insert(this);
                    break;
                }
            }
            iter++;
        }
    }
}
void MemWrite::change_use_name(int src_id, shared_ptr<Vector> new_val) { // scope
    auto func = get_belongto_bb()->get_belongto_func();
    Instanceof(reg_scope,Vector*,scope.get()) {
        if(reg_scope->get_id() == src_id) {
            func->use_place[src_id].erase(this);
            scope = new_val;
            if (new_val->get_id() >= 0)
                func->use_place[new_val->get_id()].insert(this);
            return;
        }
    }
    for(auto iter = read_before_write.begin();iter != read_before_write.end();) {
        auto use = iter->get();
        if(use->get_id() == src_id) {
            func->use_place[use->get_id()].erase(this);
            if(new_val->get_id() < 0) {
                iter = read_before_write.erase(iter);
                break;
            } else {
                iter = read_before_write.erase(iter);
                iter = read_before_write.insert(iter,new_val);
                func->use_place[new_val->get_id()].insert(this);
                break;
            }
        }
        iter++;
    }
}

void MemWrite::change_use_name(int src_id, shared_ptr<MemObject> new_val) { // addr
    auto func = get_belongto_bb()->get_belongto_func();
    Instanceof(reg_addr,MemObject*,addr.get()) {
        if(reg_addr->get_id() == src_id) {
            func->use_place[src_id].erase(this);
            addr = new_val;
            if (new_val->get_id() >= 0)
                func->use_place[new_val->get_id()].insert(this);
            return;
        }
    }
    Instanceof(reg_scope,MemObject*,scope.get()) {
        assert(false); // 感觉不太可能说实话
        if(reg_scope->get_id() == src_id) {
            func->use_place[src_id].erase(this);
            scope = new_val;
            if (new_val->get_id() >= 0)
                func->use_place[new_val->get_id()].insert(this);
        }
    }
}

unordered_set<shared_ptr<Type>> MemWrite::get_uses() const {
    unordered_set<shared_ptr<Type>> ret;
    ret.insert(read_before_write.begin(),read_before_write.end());
    ret.insert({addr, val, scope});
    return ret;
}
const shared_ptr<Type> &MemWrite::get_new_scope() const {
    return new_scope;
}
unordered_set<shared_ptr<Type>> MemWrite::get_pure_uses() const {
    return {addr,val,scope};
}
void MemWrite::gen_ir(std::ostream &os) const {
    Instanceof(reg,Reg*,scope.get()) {
        os << " "
           << "%" << new_scope->get_id() << " = "
           << "memwrite "
           << " " << *val
           << ", "
           << "* %" << addr->get_id() << " scope " << *reg << " call_write: " << call_write;
    } else {
        Instanceof(vec,Vector*,scope.get()) {
            if(!vec->get_name().empty()) {
                os << " "
                   << "%" << new_scope->get_id() << " = "
                   << "memwrite "
                   << " " << *val
                   << ", "
                   << "* %" << addr->get_id() << " scope " << vec->get_name() << " call_write: " << call_write;
                return;
            }
        }
        os << " "
           << "%" << new_scope->get_id() << " = "
           << "memwrite "
           << " " << *val
           << ", "
           << "* %" << addr->get_id() << " scope " << scope->get_id() << " call_write: " << call_write;
    }
}
Instruction *MemWrite::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, unordered_map<Ir::BasicBlock *, Ir::BasicBlock *> &map) {
    assert(false);
    return nullptr;
}

MemRead::MemRead(const shared_ptr<Type> &dst, const shared_ptr<Type> &scope, const shared_ptr<Type> &addr, bool call_read) :
    dst(dst), scope(scope), addr(addr), call_read(call_read) {
}
const shared_ptr<Type> &MemRead::get_dst() const {
    return dst;
}
const shared_ptr<Type> &MemRead::get_scope() const {
    return scope;
}
const shared_ptr<Type> &MemRead::get_addr() const {
    return addr;
}
unordered_set<shared_ptr<Type>> MemRead::get_uses() const {
    return {addr,scope};
}
void MemRead::add_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place[dst->get_id()] = this;
    for(auto &use:get_uses()) {
        if(use->get_id()>=0) {
            func->use_place[use->get_id()].insert(this);
        }
    }
}
void MemRead::remove_use_def() {
    auto func = get_belongto_bb()->get_belongto_func();
    func->def_place.erase(dst->get_id());
    for(auto &use:get_uses()) {
        if(use->get_id()>=0) {
            func->use_place[use->get_id()].erase(this);
        }
    }
}
void MemRead::change_use_name(int src_id, shared_ptr<Reg> new_val) {
    auto func = get_belongto_bb()->get_belongto_func();
    Instanceof(reg_addr,Reg*,addr.get()) {
        if(reg_addr->get_id() == src_id && src_id >=0) {
            func->use_place[src_id].erase(this);
            addr = new_val;
            if(new_val->get_id() >= 0)
                func->use_place[new_val->get_id()].insert(this);
        }
    }

    Instanceof(reg_scope,Reg*,scope.get()) {
        if(reg_scope->get_id() == src_id && src_id >=0) {
            func->use_place[src_id].erase(this);
            scope = new_val;
            if(new_val->get_id() >= 0)
                func->use_place[new_val->get_id()].insert(this);
        }
    }
}
void MemRead::change_use_name(int src_id, shared_ptr<Vector> new_val) { // scope
    auto func = get_belongto_bb()->get_belongto_func();
    Instanceof(reg_scope,Vector*,scope.get()) {
        if(reg_scope->get_id() == src_id) {
            func->use_place[src_id].erase(this);
            scope = new_val;
            if (new_val->get_id() >= 0)
                func->use_place[new_val->get_id()].insert(this);
        }
    }
    Instanceof(reg_addr,Vector*,addr.get()) {
        if(reg_addr->get_id() == src_id) {
            func->use_place[src_id].erase(this);
            addr = new_val;
            if (new_val->get_id() >= 0)
                func->use_place[new_val->get_id()].insert(this);
        }
    }
}
void MemRead::change_use_name(int src_id, shared_ptr<MemObject> new_val) {
    auto func = get_belongto_bb()->get_belongto_func();
    Instanceof(reg_addr,MemObject*,addr.get()) {
        if(reg_addr->get_id() == src_id) {
            func->use_place[src_id].erase(this);
            addr = new_val;
            if (new_val->get_id() >= 0)
                func->use_place[new_val->get_id()].insert(this);
        }
    }
    Instanceof(reg_scope,MemObject*,scope.get()) {
        if(reg_scope->get_id() == src_id) {
            func->use_place[src_id].erase(this);
            scope = new_val;
            if (new_val->get_id() >= 0)
                func->use_place[new_val->get_id()].insert(this);
        }
    }
}
void MemRead::gen_ir(std::ostream &os) const {
    Instanceof(reg,Reg*,addr.get()) {
        os << "%" << dst->get_id() << " = " << "memread * @" << *(reg) << " scope "<< "%" << scope->get_id() << " call_read " << call_read;
    } else {
        os << "%" << dst->get_id() << " = " << "memread * @" << addr->get_id() << " scope "<< "%" << scope->get_id() << " call_read " << call_read;
    }

}
Instruction *MemRead::clone(unordered_map<int, std::shared_ptr<Ir::Type>> &reg_maps, unordered_map<Ir::BasicBlock *, Ir::BasicBlock *> &map) {
    assert(false);
    return nullptr;
}

}

