//
// Created by 11069 on 10/7/2023.
//

// loop interchange 解决cache miss问题
// 如果 for j for i a[i][j] 这种,i j就可以调换
// but 这个循环必须是最内层的俩循环
// 且外层循环的header就是内存循环的pre header,并且,这个pre header只能有比较和跳转语句,不能有其他语句
// 并且 内存循环的出口就是外层循环的back edge
// 另外,内层循环不能有对i,j的其他use语句(除了phi和比较和update之外)
// 内层循环不能有continue和break

#include "opt.h"

namespace opt {
    using namespace Ir;

struct NestedLoopInfo {
        // inner
        BasicBlock * inner_header;
        BasicBlock * inner_exit_prev;
        BasicBlock * inner_exit;
        PhiInstr * inner_phi;
        BinaryInstr * inner_upd;
        BinaryInstr * inner_cond;
        BranchInstr * inner_br;
        // outer
        BasicBlock * outer_header;
        BasicBlock * outer_exit_prev;
        BasicBlock * outer_exit;
        PhiInstr * outer_phi;
        BinaryInstr * outer_upd;
        BinaryInstr * outer_cond;
        BranchInstr * outer_br;
};
typedef NestedLoopInfo NestedLoopInfo;

NestedLoopInfo info;

    std::unique_ptr<Instruction>& find_phi(PhiInstr* target) {
        for(auto &inst:target->get_belongto_bb()->get_instrs()) {
            if(inst.get() == target) return inst;
        }
    }

    void loop_interchange(Loop* inner,Loop* outer) {
        auto func = inner->get_belong_to_func();
        cout<<"begin interchange:"<<inner->get_header()->get_id()<<endl;
        auto reg_i = info.outer_phi->get_dst();
        auto reg_j = info.inner_phi->get_dst();
        // 删除use def
        info.outer_phi->remove_use_def();
        info.inner_phi->remove_use_def();

        // 修改phi的incoming bb
        // outer_phi 变成内层phi
        BasicBlock* into_loop = nullptr; // outer phi这个东西吧有个incoming来自循环外面,这个要给到inner phi
        unordered_map<BasicBlock *, shared_ptr<Reg>> new_incoming;
        for(auto &[bb,value]:info.outer_phi->get_bb_2_value()) {
            if(bb == info.outer_exit_prev) {
                new_incoming[info.inner_exit_prev] = value;
            } else {
                into_loop = bb;
                new_incoming[info.outer_header] = value;
            }
        }
        info.outer_phi->get_bb_2_value() = new_incoming;

        for(auto &[bb,value]:info.inner_phi->get_bb_2_value()) {
            if(bb == info.inner_exit_prev) {
                new_incoming[info.outer_exit_prev] = value;
            } else {
                into_loop = bb;
                new_incoming[into_loop] = value;
            }
        }
        info.outer_phi->get_bb_2_value() = new_incoming;

        // 现在两个phi都变成了对方的样子,但是bb改了,value没改,这里修改判断条件,修改update语句,修改跳转语句即可

        auto new_inner_upd = new BinaryInstr(Factory::get_reg(Int),info.outer_upd->get_src_1(),// 新建upd,该updade的两个src,一个来自phi,一个是常数,直接将outer的复制进来就行了,但是输出要改
                                 info.outer_upd->get_src_2(),info.outer_upd->get_binary_type());
        auto new_inner_cond = new BinaryInstr(Factory::get_reg(Bool),info.outer_cond->get_src_1() // 新的cond,一个乃是upd,一个是常数,先复制,下面再改
                                              ,info.outer_cond->get_src_2(),info.outer_cond->get_binary_type());
        new_inner_cond->change_use_name(info.inner_upd->get_dst()->get_id(),new_inner_upd->get_dst());
        info.inner_br->change_use_name(info.inner_br->get_reg()->get_id(),new_inner_cond->get_dst()); // 新的br和旧的br一样,保护cfg,但是条件改成新的cond的结果
        // 同上
        auto new_outer_upd = new BinaryInstr(Factory::get_reg(Int),info.inner_upd->get_src_1(),
                                             info.inner_upd->get_src_2(),info.inner_upd->get_binary_type());
        auto new_outer_cond = new BinaryInstr(Factory::get_reg(Bool),info.inner_cond->get_src_1()
                                              ,info.inner_cond->get_src_2(),info.inner_cond->get_binary_type());
        new_outer_cond->change_use_name(info.outer_upd->get_dst()->get_id(),new_outer_upd->get_dst());
        info.outer_br->change_use_name(info.outer_br->get_reg()->get_id(),new_outer_cond->get_dst());

        // insert insts 因为br没改,所以倒着吧指令放进去,之前的指令因为dst没有use而被删掉
        info.outer_exit_prev->insert_before_br(new_outer_upd);
        info.outer_exit_prev->insert_before_br(new_outer_cond);
        new_outer_upd->add_use_def();
        new_outer_cond->add_use_def();
        info.inner_exit_prev->insert_before_br(new_inner_upd);
        info.inner_exit_prev->insert_before_br(new_inner_cond);
        new_inner_upd->add_use_def();
        new_inner_cond->add_use_def();
        // phi 交换
        auto &origin_out_phi = find_phi(info.outer_phi);
        auto &origin_in_phi = find_phi(info.inner_phi);
        BasicBlock * tmp = info.outer_phi->get_belongto_bb();
        info.outer_phi->set_belongto_bb(info.inner_phi->get_belongto_bb());
        info.inner_phi->set_belongto_bb(tmp);
        origin_out_phi.reset(info.inner_phi);
        origin_in_phi.reset(info.outer_phi);
        origin_out_phi->change_use_name(info.inner_upd->get_dst()->get_id(),new_outer_upd->get_dst());
        origin_in_phi->change_use_name(info.outer_upd->get_dst()->get_id(),new_inner_upd->get_dst());
        // 修改phi的不在loop里面的incoming的基础值 ,也既 循环变量的初始值
        {
            auto new_out_phi = info.inner_phi;
            auto new_in_phi = info.outer_phi;
            shared_ptr<Reg> init_outer = nullptr;
            shared_ptr<Reg> init_inner = nullptr;
            for(auto &[bb,value]:new_out_phi->get_bb_2_value()) {
                if(bb == into_loop) {
                    init_outer = value; // 这个不一定是常量
                    break;
                }
            }
            for(auto &[bb,value]:new_in_phi->get_bb_2_value()) {
                if(bb == info.outer_header) {
                    init_inner = value; // 这个一定是个常量
                    break;
                }
            }
            new_out_phi->get_bb_2_value()[into_loop] = init_inner;
            new_in_phi->get_bb_2_value()[info.outer_header] = init_outer;
        }
        origin_in_phi->add_use_def();
        origin_out_phi->add_use_def();
    }

    bool find_inst(list<unique_ptr<Instruction>> &insts,Instruction* target) {
        bool ret = false;
        for(auto &inst:insts) {
            ret |= inst.get() == target;
        }
        return ret;
    }

    bool find_use_valid(Function* func,int id, Instruction* upd) {
        for(auto &use:func->use_place[id]) {
            Instanceof(binary,BinaryInstr *,use) {
                if(binary != upd) return false;
            } else Instanceof(loadArray,LoadArray*,use) {
                continue ;
            } else Instanceof(loadVector,LoadVector*,use) {
                continue;
            } else return false;
        }
        return true;
    }

    bool nested_loop_valid(Loop* inner,Loop* outer) {
        // init inner
        auto func = inner->get_belong_to_func();
        info.inner_header = inner->get_header();
        info.inner_exit = inner->get_exiting_bb();

        BasicBlock* exit_prev = nullptr;
        for(auto &bb:inner->get_all_bbs()) {
            for(auto &next:bb->succ) {
                if(!inner->get_all_bbs().count(next)) {
                        exit_prev = bb;
                        break;
                }
            }
        }
        info.inner_exit_prev = exit_prev;
        Instanceof(tmp_br ,BranchInstr *, exit_prev->get_instrs().back().get())
            info.inner_br = tmp_br;
        Instanceof(tmp_cond,BinaryInstr *,func->def_place[info.inner_br->get_reg()->get_id()])
            info.inner_cond = tmp_cond;
        auto reg_i = info.inner_cond->get_src_1();
        Instanceof(tmp_upd,BinaryInstr *,func->def_place[reg_i->get_id()])
            info.inner_upd = tmp_upd;
        reg_i = info.inner_upd->get_src_1();
        if(is_loop_const(reg_i.get(),func,inner)) {
            reg_i = info.inner_upd->get_src_2();
        }
        Instanceof(tmp_phi,PhiInstr *,func->def_place[reg_i->get_id()]) {
            info.inner_phi = tmp_phi;
        }

        // init outer
        info.outer_header = outer->get_header();
        info.outer_exit = outer->get_exiting_bb();

        exit_prev = nullptr;
        for(auto &bb:outer->get_all_bbs()) {
            for(auto &next:bb->succ) {
                if(!outer->get_all_bbs().count(next)) {
                        exit_prev = bb;
                        break;
                }
            }
        }
        info.outer_exit_prev = exit_prev;
        Instanceof(tmp_br ,BranchInstr *, exit_prev->get_instrs().back().get())
            info.outer_br = tmp_br;
        Instanceof(tmp_cond,BinaryInstr *,func->def_place[info.inner_br->get_reg()->get_id()])
            info.outer_cond = tmp_cond;
        reg_i = info.outer_cond->get_src_1();
        Instanceof(tmp_upd,BinaryInstr *,func->def_place[reg_i->get_id()])
            info.outer_upd = tmp_upd;
        reg_i = info.outer_upd->get_src_1();
        if(is_loop_const(reg_i.get(),func,inner)) {
            reg_i = info.outer_upd->get_src_2();
        }
        Instanceof(tmp_phi,PhiInstr *,func->def_place[reg_i->get_id()]) {
            info.outer_phi = tmp_phi;
        }
//        load j0 0
//          br j0 label1 label2
//              label1: outer header
//                  phi j1 j0 j2
//                  load i0 0
//                  br i0 label3 label4 // 这里有一个 比较省略了,所以应该是2条指令
//                      label3 inner header/exit_prev
//                          phi i1 i0 i2
//                          ... = j1 i1
//                          i2 = i1 + 1
//                          branch i2 label3 label4
//               label4: outer exit prev
//                  j2 = j1 + 1
//                  branch j2 label1 label2
//           label2
        // check validation
        if(inner->get_header()->idom_by != outer->get_header()) return false;
        if(info.outer_exit_prev != info.inner_exit) return false;
        if(info.outer_header->get_instrs().size() > 3) return false; // i的phi,j的cond,j的branch,没有j的初始化
        if(info.outer_header->get_instrs().front().get() != info.outer_phi) return false;
        if(info.outer_exit_prev->get_instrs().size() != 3) return false; // update cond br
        if(!find_inst(info.outer_exit_prev->get_instrs(),info.outer_upd)) return false;
        if(!find_inst(info.outer_exit_prev->get_instrs(),info.outer_phi)) return false;
        if(!find_inst(info.outer_exit_prev->get_instrs(),info.outer_cond)) return false;

        // check a i j a j i
        bool i_j = false;
        bool j_i = false;
        reg_i = info.outer_phi->get_dst();
        auto reg_j = info.inner_phi->get_dst();
        // i,j只能用于update和地址计算
        if (!find_use_valid(func,reg_i->get_id(),info.outer_upd)) {
            return false;
        }
        if(!find_use_valid(func,reg_j->get_id(),info.inner_upd)) {
            return false;
        }
        for(auto &bb:inner->get_all_bbs()) {
            for(auto &inst:bb->get_instrs()) {
                // 因为是为了解决cache miss,如果在很后面就没有意义了,前两个index才有意义
                Instanceof(loadArray,LoadArray*,inst.get()) {
                    if(loadArray->get_indexs().size() >= 2) {
                        if(loadArray->get_indexs()[0]->get_id() == reg_i->get_id()
                              && loadArray->get_indexs()[1]->get_id() == reg_j->get_id())
                            i_j = true;
                        else if(loadArray->get_indexs()[0]->get_id() == reg_j->get_id()
                            && loadArray->get_indexs()[1]->get_id() == reg_i->get_id())
                            j_i = true;
                    }
                } else Instanceof(loadVector,LoadVector*,inst.get()) {
                    if(loadVector->get_indexs().size() >= 2) {
                        if(loadVector->get_indexs().size() >= 2) {
                            if(loadVector->get_indexs()[0]->get_id() == reg_i->get_id()
                                && loadVector->get_indexs()[1]->get_id() == reg_j->get_id())
                                i_j = true;
                            else if(loadVector->get_indexs()[0]->get_id() == reg_j->get_id()
                                     && loadVector->get_indexs()[1]->get_id() == reg_i->get_id())
                                j_i = true;
                        }
                    }
                }
            }
        }
        // i 放进里面减少cache miss,那肯定是 j i的顺序才对
        return !i_j && j_i;
    }

    bool is_valid(Loop * loop) { // one back edge  one exit i appear only in phi  and update dst
        Function* func = loop->get_belong_to_func();
        // one way in one way out
        BasicBlock *exit = nullptr;
        BasicBlock *into_loop = nullptr;
        for(auto &bb:loop->get_all_bbs()) {
            for(auto &next:bb->succ) {
                if(!loop->get_all_bbs().count(next)) {
                    if(exit)
                        return false;
                    exit = next;
                }
            }
            for(auto &prev:bb->prev) {
                if(!loop->get_all_bbs().count(prev)) {
                    if(into_loop)
                        return false;
                    into_loop = prev;
                }
            }
        }
        if(!into_loop || !exit) {
            return false;
        }
        // no break no continue

        BasicBlock * back_edge = nullptr;
        BasicBlock * exit_prev = nullptr;
        for(auto &bb:exit->prev) {
            if(loop->get_all_bbs().count(bb)) {
                if(exit_prev)
                    return false;
                exit_prev = bb;
            }
        }

        for(auto &bb:loop->get_header()->prev) {
            if(loop->get_all_bbs().count(bb)) {
                if(back_edge)
                    return false;
                back_edge = bb;
            }
        }

        if(back_edge != exit_prev ) return false;

        // update check
        Instruction * inst = exit_prev->get_instrs().back().get();
        Instanceof(branch,BranchInstr *,inst) {
            if(branch->get_reg()->get_id() < 0) {
                return false;
            }
            Instanceof(cond,BinaryInstr*,branch->get_belongto_bb()->get_belongto_func()->def_place[branch->get_reg()->get_id()]) {
                auto reg_i = cond->get_src_1();
                auto reg_c = cond->get_src_2();
                auto func = cond->get_belongto_bb()->get_belongto_func();
                if(is_loop_const(reg_i.get(),func,loop) == is_loop_const(reg_c.get(),func,loop)) {
                    return false;
                }
                if(is_loop_const(reg_i.get(),func,loop) && !is_loop_const(reg_c.get(),func,loop)) {
                    if (!swap_cond(cond)) {
                        return false;
                    }
                    std::swap(reg_i,reg_c);
                }
                if(func->is_param(reg_i.get())) {
                    return false;
                }
                Instanceof(update,BinaryInstr *,func->def_place[reg_i->get_id()]) {
                    if(update->get_dst()->get_id() != Int) return false;
                    if(func->use_place[update->get_dst()->get_id()].size() > 2) return false;
                    reg_i = update->get_src_1();
                    reg_c = update->get_src_2();
                    if(func->is_param(reg_i.get()) || func->is_param(reg_c.get())) return false;
                    if(is_loop_const(reg_i.get(),func,loop)) {
                        std::swap(reg_i,reg_c);
                    }
                    if(!is_loop_const(reg_c.get(),func,loop)) {
                        return false;
                    }
                    Instanceof(phi,PhiInstr *,func->def_place[reg_i->get_id()]) {
                        if(phi->get_bb_2_value().size() != 2) return false;
                        if(phi->get_belongto_bb() != loop->get_header()) return false;
                        return true;
                    } else {
                        return false;
                    }
                }
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    void loop_interchange(Function* func) {
        func->cfg->build();
        func->cfg->dom_build();
        func->cfg->loop_analysis();
        for (auto &item:func->get_loops()) {
            auto loop = item.get();
            // 必须是最内层的循环,且该循环有一个外层循环
            if (!loop->get_parent())
                continue ;
            if (!loop->get_parent()->get_parent())
                continue ;
            if(!is_valid(loop) || !is_valid(loop->get_parent()) || !nested_loop_valid(loop,loop->get_parent())) {
                continue ;
            }

            loop_interchange(loop,loop->get_parent());

        }
    }

    void loop_interchange(CompileUnit* compileUnit) {
        for(auto &func : compileUnit->get_functions()) {
            Factory::reset(func.get());
            loop_interchange(func.get());
            func->set_bb_max_num(Factory::bb_num);
            func->set_reg_max_num(Factory::reg_num);
        }
    }
}