//
// Created by 11069 on 9/8/2023.
//


#include "opt.h"
using namespace Ir;
namespace opt {

  void loop_invariant_store_motion(Loop* loop) {
    Function* func = loop->get_belong_to_func();

    // 单出口单入口
    BasicBlock *exit = nullptr;
    BasicBlock *into_loop = nullptr;
    for(auto &bb:loop->get_all_bbs()) {
      for(auto &next:bb->succ) {
        if(!loop->get_all_bbs().count(next)) {
          if(exit)
            return ;
          exit = next;
        }
      }
      for(auto &prev:bb->prev) {
        if(!loop->get_all_bbs().count(prev)) {
          if(into_loop)
            return ;
          into_loop = prev;
        }
      }
    }
    if(!into_loop || !exit) {
      return ;
    }

    // 无break和continue
    BasicBlock * back_edge = nullptr;
    BasicBlock * exit_prev = nullptr;
    for(auto &bb:exit->prev) {
      if(loop->get_all_bbs().count(bb)) {
        if(exit_prev)
          return ;
        exit_prev = bb;
      }
    }

    for(auto &bb:loop->get_header()->prev) {
      if(loop->get_all_bbs().count(bb)) {
        if(back_edge)
          return ;
        back_edge = bb;
      }
    }

    for(auto iter = loop->get_header()->get_instrs().begin();iter != loop->get_header()->get_instrs().end();) {
      Instanceof(array_phi,ArrayPhi*,iter->get()) {
        bool erased = false;
        if(array_phi->get_bb_2_value().size() != 2) {
          iter++;
          continue ;
        }
        Instanceof(vector,Reg*,array_phi->get_dst().get()) {
          iter++;
          continue ;
        }
        if(!array_phi->get_bb_2_value().count(into_loop)  || !array_phi->get_bb_2_value().count(exit_prev))
          continue ;
        auto base_scope = array_phi->get_bb_2_value()[exit_prev];
        while(true) {
          auto inst = func->def_place[base_scope->get_id()];
          Instanceof(mem_write,MemWrite*,inst) {
            // store val is const
            if(!mem_write->get_val()->is_constant()) {
              break;
            }
            // addr is const
            Instanceof(loadArray,LoadArray*,func->def_place[mem_write->get_addr()->get_id()]) {
              if(func->def_place.count(loadArray->get_addr()->get_id())) break;
              if(loadArray->get_indexs().size() != 1) break;
              if(!loadArray->get_indexs()[0]->is_constant()) break;
            } else break;
            base_scope = mem_write->get_scope();
          } else Instanceof(array_phi_j,ArrayPhi*,inst) {
            if(array_phi_j != array_phi) {
              break;
            }
            copy_propagation(func->use_place,array_phi->get_dst()->get_id(), static_pointer_cast<Vector>(array_phi->get_bb_2_value()[into_loop]));
            (*iter)->remove_use_def();
            iter = loop->get_header()->get_instrs().erase(iter);
            erased = true;
            break;
          }
        }
        if(!erased) {
          iter++;
          continue ;
        }
      } else break;
    }

  }

  void loop_invariant_store_motion(Function* func) {
    func->cfg->build();
    func->cfg->dom_build();
    func->cfg->loop_analysis();
    for(auto &loop:func->get_loops()) {
      if(loop->children.size())
        continue ;

      loop_invariant_store_motion(loop.get());
    }


  }

  void loop_invariant_store_motion(CompileUnit* compileUnit) {
    for(auto &func:compileUnit->get_functions()) {
      if(func->is_lib())
        continue ;
      if(func->get_name() != "main")
        continue ;
      loop_invariant_store_motion(func.get());
      break;
    }
  }
}