//
// Created by 11069 on 11/7/2023.
//
// 该 pass 先于loop_unroll_dual 完成
// 该 pass 彻底展开循环
// 需要条件
// 循环判断的开始,结束,步数都为常数
// 无break,continue,因为是while循环,break和continue会导致步数不确定,比如continue在while循环不会让iter改变,步数就增多了,break会让大量被复制的basicblock浪费,严重的cache miss
// 入口出口唯一
#include "opt.h"
namespace opt {
using namespace Ir;
// 默认值 10 * 1000
static int MAX_INST = 3100;
struct CompleteUnrollInfo {
    BinaryType op;
    int begin;
    int end;
    int step;
    int times;
    int total_times;
    BinaryInstr * upd;
    BinaryInstr * cond;
    PhiInstr * phi;
    BranchInstr * br;
    BasicBlock * into_loop;
    BasicBlock * header;
    BasicBlock * back_edge;
    BasicBlock * exit_prev;
    BasicBlock * exit;
};

typedef CompleteUnrollInfo CompleteUnrollInfo;

int calc_times(BinaryType op,int begin,int step,int end) {
    switch (op) {
    case SLT:
        return end-begin;
    case SLE:
        return end-begin+1;
    case SGT:
        return begin-end;
    case SGE:
        return begin-end + 1;
    case NE:
        return abs(begin-end);
    case EQ: // 看不懂
        return false;
    default:
        return false;
    }
}

static CompleteUnrollInfo get_loop_unroll_info(Loop* loop) {
    CompleteUnrollInfo info;
    // 单出口单入口
    BasicBlock *exit = nullptr;
    BasicBlock *into_loop = nullptr;
    Function * func = loop->get_belong_to_func();
    for(auto &bb:loop->get_all_bbs()) {
        for(auto &next:bb->succ) {
            if(!loop->get_all_bbs().count(next)) {
                if(exit)
                    assert(false);
                exit = next;
            }
        }
        for(auto &prev:bb->prev) {
            if(!loop->get_all_bbs().count(prev)) {
                if(into_loop)
                    assert(false);
                into_loop = prev;
            }
        }
    }
    BasicBlock * exit_prev = nullptr;
    for(auto &bb:exit->prev) {
        if(loop->get_all_bbs().count(bb)) {
            if(exit_prev)
                assert(false);
            exit_prev = bb;
        }
    }

    int cnt = 0;
    for(auto &bb:loop->get_all_bbs()) {
        cnt += bb->get_instrs().size();
        cnt-=1;
    }

    BasicBlock * back_edge = nullptr;

    for(auto &bb:loop->get_header()->prev) {
        if(loop->get_all_bbs().count(bb)) {
            back_edge = bb;
        }
    }

    info.exit_prev = exit_prev;
    info.exit = exit;
    info.header = loop->get_header();
    info.back_edge = back_edge;
    info.into_loop = into_loop;

    Instanceof(br,BranchInstr *,exit_prev->get_instrs().back().get()) {
        info.br = br;
        Instanceof(cond,BinaryInstr*,func->def_place[br->get_reg()->get_id()]) {
            // 经过gvn了,src2肯定是常数,不是常数就return false
            info.end = cond->get_src_2()->get_num().int_num;
            info.cond = cond;
            info.op = cond->get_binary_type();
            Instanceof(upd,BinaryInstr*,func->def_place[cond->get_src_1()->get_id()]) {
                // upd 必须是1步的加减法,且必须合理
                info.step = upd->get_binary_type() == ADD ? 1: -1;
                info.upd = upd;
                Instanceof(phi,PhiInstr*,func->def_place[upd->get_src_1()->get_id()]) {
                    for(auto &[bb,value]:phi->get_bb_2_value()) {
                        if(!loop->get_all_bbs().count(bb)) {
                            info.begin = value->get_num().int_num;
                        }
                    }
                    info.phi = phi;
                }
            } else assert(false);
        } else assert(false);
    } else assert(false);

    info.total_times = calc_times(info.op,info.begin,info.step,info.end);
    if(cnt * info.total_times > MAX_INST) {
        int max_step = MAX_INST / cnt;
        while(max_step > 1) {
            if(info.total_times % max_step == 0) {
                info.times = max_step;
                break ;
            } else {
                max_step--;
            }
        }
        if(max_step == 1) {
            info.times = 1;
        }
    } else {
        info.times = info.total_times;
    }

    return info;
}

static bool is_valid_cmp(BinaryInstr * cond) { // 先兼容所有的吧,后面看实际展开的实现难度
    switch (cond->get_binary_type()) {
    case SLT:
    case SLE:
    case SGT:
    case SGE:
    case NE:
        return true;
    case EQ:
    default:
        return false;
    }
}

bool step_validation_check(BinaryType op,int begin,int step,int end) {
    switch (op) {
    case SLT:
        return begin < end && step > 0;
    case SLE:
        return begin < end && step > 0;
    case SGT:
        return begin > end && step < 0;
    case SGE:
        return begin > end && step < 0;
    case NE:
        return (begin > end && step < 0) || (begin < end && step > 0);
    case EQ: // 看不懂
        return false;
    default:
        return false;
    }
}

static bool unroll_valid(Loop* loop) {

    Function* func = loop->get_belong_to_func();
    if (func->get_bbs().size() > 21) {
    return false;
    }
    // 单出口单入口
    BasicBlock *exit = nullptr;
    BasicBlock *into_loop = nullptr;
    for(auto &bb:loop->get_all_bbs()) {
        for(auto &next:bb->succ) {
            if(!loop->get_all_bbs().count(next)) {
                if(exit)
                    return false;
                exit = next;
            }
        }
        for(auto &prev:bb->prev) {
            if(!loop->get_all_bbs().count(prev)) {
                if(into_loop)
                    return false;
                into_loop = prev;
            }
        }
    }
    if(!into_loop || !exit) {
        return false;
    }

    // 无break和continue
    BasicBlock * back_edge = nullptr;
    BasicBlock * exit_prev = nullptr;
    for(auto &bb:exit->prev) {
        if(loop->get_all_bbs().count(bb)) {
            if(exit_prev)
                return false;
            exit_prev = bb;
        }
    }

    for(auto &bb:loop->get_header()->prev) {
        if(loop->get_all_bbs().count(bb)) {
            if(back_edge)
                return false;
            back_edge = bb;
        }
    }

    if(back_edge != exit_prev ) return false;
    // 指令不能太多,防止cache miss 过大
    int cost = 0;
    for(auto &bb:loop->get_all_bbs()) {
        for(auto &inst:bb->get_instrs()) {
            if(cost >= 80) return false;
            Instanceof(call,CallInstr *,inst.get()) {
                if(call->get_function()->is_lib()) {
                    cost += 20;
                } else {
                    cost += 10;
                    for(auto &func_bb:call->get_function()->get_bbs()) {
                        if(cost >= 80) return false;
                        cost += func_bb->get_instrs().size();
                    }
                }
            } else {
                cost += 1;
            }
        }
    }
    // 获取开始,步数,终止条件
    BinaryType op = SLT;
    int end = 0;
    int step = 0;
    int begin = 0;
    Instanceof(br,BranchInstr *,exit_prev->get_instrs().back().get()) {
        if(!br->get_bb_2()) return false;
        Instanceof(cond,BinaryInstr*,func->def_place[br->get_reg()->get_id()]) {
            if(!is_valid_cmp(cond))
                return false;
            // 经过gvn了,src2肯定是常数,不是常数就return false
            if(cond->get_src_2()->get_id() >= 0 || cond->get_src_2()->get_type() != Int) return false;
            end = cond->get_src_2()->get_num().int_num;
            op = cond->get_binary_type();
            Instanceof(upd,BinaryInstr*,func->def_place[cond->get_src_1()->get_id()]) {
                // upd 必须是1步的加减法,且必须合理
                if(upd->get_binary_type() != ADD && upd->get_binary_type() != SUB) return false;
                if(upd->get_src_2()->get_id() >= 0 || upd->get_src_2()->get_type() != Int
                    || upd->get_src_2()->get_num().int_num != 1) return false;
                step = upd->get_binary_type() == ADD ? 1: -1;
                Instanceof(phi,PhiInstr*,func->def_place[upd->get_src_1()->get_id()]) {
                    if(phi->get_bb_2_value().size() > 2) return false;
                    if(phi->get_belongto_bb() != loop->get_header()) return false;
                    for(auto &[bb,value]:phi->get_bb_2_value()) {
                        if(!loop->get_all_bbs().count(bb)) {
                            if(value->get_id() >= 0) return false;
                            if(value->get_type() != Int) return false;
                            begin = value->get_num().int_num;
                        }
                    }
                } else return false;
            } else return false;
        } else return false;
    } else return false;

    // check begin step end validation

    return step_validation_check(op,begin,step,end);

}

static unordered_map<BasicBlock *,BasicBlock *> loop0_to_loop1_bb_map[2];
static unordered_map<shared_ptr<Type>,shared_ptr<Type>> loop0_to_loop1_dst_map[2];
static int cur;
static shared_ptr<Type> find_dst(shared_ptr<Type> old_dst) {
    if(loop0_to_loop1_dst_map[cur].count(old_dst))
        return loop0_to_loop1_dst_map[cur][old_dst];
    return old_dst;
}

static BasicBlock * find_bb(BasicBlock * old_bb) {
    if(loop0_to_loop1_bb_map[cur].count(old_bb))
        return loop0_to_loop1_bb_map[cur][old_bb];
    return old_bb;
}

void inst_copy(CompleteUnrollInfo info,BasicBlock * old_bb,BasicBlock * new_bb) {
    for(auto &inst:old_bb->get_instrs()) {
        unique_ptr<Instruction> new_inst = nullptr;
        Instanceof(binary,BinaryInstr*,inst.get()) {
            new_inst = make_unique<BinaryInstr>(static_pointer_cast<Reg>(find_dst(binary->get_dst()))
             ,static_pointer_cast<Reg>(find_dst(binary->get_src_1())),static_pointer_cast<Reg>(find_dst(binary->get_src_2())),binary->get_binary_type());
        } else Instanceof(call,CallInstr*,inst.get()) {
            auto *args = new vector<shared_ptr<Type>>;
            args->reserve(call->get_params().size());
            for(auto &arg:call->get_params()) {
                args->push_back(find_dst(arg));
            }
            if(call->is_void_flag()) {
                new_inst = make_unique<CallInstr>(call->get_function(),*args);
            } else {
                new_inst = make_unique<CallInstr>(static_pointer_cast<Reg>(find_dst(call->get_dst())),call->get_function(),*args);
            }

        } else Instanceof(br,BranchInstr*,inst.get()) {
            if(old_bb == info.exit_prev) {
                new_inst = make_unique<BranchInstr>(static_pointer_cast<Reg>(find_dst(br->get_reg())),loop0_to_loop1_bb_map[cur][info.header],info.exit);
            } else {
                new_inst = make_unique<BranchInstr>(static_pointer_cast<Reg>(find_dst(br->get_reg())), find_bb(br->get_bb_1()), find_bb(br->get_bb_2()));
            }

        } else Instanceof(store,StoreInstr*,inst.get()) {
            new_inst = make_unique<StoreInstr>(static_pointer_cast<Reg>(find_dst(store->get_val())), static_pointer_cast<MemObject>(find_dst(store->get_addr())));
        } else Instanceof(storeGlobal,StoreGlobal*,inst.get()) {
            new_inst = make_unique<StoreGlobal>(static_pointer_cast<Reg>(find_dst(storeGlobal->get_val())), static_pointer_cast<Reg>(find_dst(storeGlobal->get_addr())));
        } else Instanceof(loadVector,LoadVector*,inst.get()) {
            auto *indexs = new vector<shared_ptr<Reg>>;
            indexs->reserve(loadVector->get_indexs().size());
            for(auto &index:loadVector->get_indexs()) {
                indexs->push_back(static_pointer_cast<Reg>(find_dst(index)));
            }
            new_inst = make_unique<LoadVector>(static_pointer_cast<Vector>(find_dst(loadVector->get_dst())),
                static_pointer_cast<Vector>(find_dst(loadVector->get_addr())), *indexs);
        } else Instanceof(loadArray,LoadArray*,inst.get()) {
            auto *indexs = new vector<shared_ptr<Reg>>;
            indexs->reserve(loadArray->get_indexs().size());
            for(auto &index:loadArray->get_indexs()) {
                indexs->push_back(static_pointer_cast<Reg>(find_dst(index)));
            }
            new_inst = make_unique<LoadArray>(static_pointer_cast<MemObject>(find_dst(loadArray->get_dst())),
                                           static_pointer_cast<Vector>(find_dst(loadArray->get_addr())), *indexs);
        } else Instanceof(ret,Ret*,inst.get()) {
            if(ret->is_void_flag()) {
                new_inst = make_unique<Ret>();
            } else {
                new_inst = make_unique<Ret>(static_pointer_cast<Reg>(find_dst(ret->get_src())));
            }
        } else Instanceof(convert,Convert*,inst.get()) {
            new_inst = make_unique<Convert>(static_pointer_cast<Reg>(find_dst(convert->get_dst())), static_pointer_cast<Reg>(find_dst(convert->get_src())));
        } else Instanceof(load,LoadInstr*,inst.get()) {
            new_inst = make_unique<LoadInstr>(static_pointer_cast<Reg>(find_dst(load->get_dst())), static_pointer_cast<MemObject>(find_dst(load->get_addr())));
        } else Instanceof(loadGlobal,LoadGlobal*,inst.get()) {
            new_inst = make_unique<LoadGlobal>(static_pointer_cast<Reg>(find_dst(loadGlobal->get_dst())), static_pointer_cast<Reg>(find_dst(loadGlobal->get_addr())));
        } else Instanceof(phi,PhiInstr*,inst.get()) { // 只有header的phi会有外界的传入,这里修改一下就行了
            auto *bb2value = new unordered_map<BasicBlock*, shared_ptr<Reg>>;
            if(old_bb == info.header) { // 对于 header 的phi,来自外界的不copy,来自back edge copy 但是是指向loop 0
                    (*bb2value)[loop0_to_loop1_bb_map[cur^1][info.exit_prev]] = static_pointer_cast<Reg>(loop0_to_loop1_dst_map[cur^1][phi->get_bb_2_value()[info.exit_prev]]);
            } else {
                for(auto &[bb,value]:phi->get_bb_2_value()) {
                    (*bb2value)[find_bb(bb)] = static_pointer_cast<Reg>(find_dst(value));
                }
            }
            new_inst = make_unique<PhiInstr>(static_pointer_cast<Reg>(find_dst(phi->get_dst())),*bb2value);
        } else assert(false);
        new_bb->push_back(new_inst);
    }
}

void loop_unroll_complete(Function* func,Loop* loop, CompleteUnrollInfo info) {
    // basicblock和指令复制,需要新建不少的bb和inst,需要建立映射的map
    loop0_to_loop1_bb_map[0].clear();
    loop0_to_loop1_bb_map[1].clear();
    loop0_to_loop1_dst_map[0].clear();
    loop0_to_loop1_dst_map[1].clear();
    // 基本块复制和指令复制的时候,对于header的phi和exit_prev的branch都要特别小心
    // 首先原loop0的upd的phi,可以删除back edge 的 value
    // 对于exit_prev的branch,true target变成了下一个复制的header,false不变,这个可以在新建下一个basicblock 的时候修改,copy的时候不变
    // 全部unroll 完毕后,修改最后一个loop的exit_prev,这个prev的branch改成jump到exit
    // 对于loop到exit的phi,需要修改,记得记号,放在循环外面做

    // 初始化 map
    cur = 0;
    for (auto &bb : loop->get_all_bbs()) {
        loop0_to_loop1_bb_map[cur][bb] = bb;
        for (auto &inst : bb->get_instrs()) {
            Instanceof(def, DefInstr *, inst.get()) {
                loop0_to_loop1_dst_map[cur][def->def()] = def->def();
            }
            else Instanceof(phi, PhiInstr *, inst.get()) {
                loop0_to_loop1_dst_map[cur][phi->get_dst()] = phi->get_dst();
            }
            else Instanceof(loadArray, LoadArray *, inst.get()) {
                loop0_to_loop1_dst_map[cur][loadArray->get_dst()] = loadArray->get_dst();
            }
            else Instanceof(loadVector, LoadVector *, inst.get()) {
                loop0_to_loop1_dst_map[cur][loadVector->get_dst()] = loadVector->get_dst();
            }
        }
    }

    // 处理header的cfg关系
    info.header->prev.erase(info.exit_prev);

    for (int i = 1; i < info.times; i++) {
        cur ^= 1;
        // 新建 bb
        for (auto &bb : loop->get_all_bbs()) {
            loop0_to_loop1_bb_map[cur][bb] = Factory::get_new_bb(nullptr, func);
            loop0_to_loop1_bb_map[cur][bb]->set_belong_to_loop(loop);
            // 新建 var
            for (auto &inst : bb->get_instrs()) {
                Instanceof(phi, PhiInstr *, inst.get()) {
                    loop0_to_loop1_dst_map[cur][phi->get_dst()] = Factory::get_reg(phi->get_dst()->get_type());
                }
                else Instanceof(loadArray, LoadArray *, inst.get()) {
                    loop0_to_loop1_dst_map[cur][loadArray->get_dst()] = Factory::get_mem(loadArray->get_dst()->get_type());
                }
                else Instanceof(loadVector, LoadVector *, inst.get()) {
                    auto dst = Factory::get_vector(loadVector->get_dst()->get_type());
                    dst->get_dims().push_back(0);
                    loop0_to_loop1_dst_map[cur][loadVector->get_dst()] = dst;
                }
                else Instanceof(def, DefInstr *, inst.get()) {
                    if (def->def())
                        loop0_to_loop1_dst_map[cur][def->def()] = Factory::get_reg(def->def()->get_type());
                }
            }
        }

        // 重新建立basicblock的连接关系
        for (auto &bb : loop->get_all_bbs()) {
            if (bb != info.header)
                for (auto &prev : bb->prev) {
                    loop0_to_loop1_bb_map[cur][bb]->prev.insert(loop0_to_loop1_bb_map[cur][prev]);
                }
            if (bb != info.exit_prev)
                for (auto &succ : bb->succ) {
                    loop0_to_loop1_bb_map[cur][bb]->succ.insert(loop0_to_loop1_bb_map[cur][succ]);
                }
        }

        // loop1 header 建立

        loop0_to_loop1_bb_map[cur][info.header]->prev.insert(loop0_to_loop1_bb_map[cur ^ 1][info.exit_prev]);

        // loop0 exit prev 建立

        loop0_to_loop1_bb_map[cur ^ 1][info.exit_prev]->succ.insert(loop0_to_loop1_bb_map[cur][info.header]);

        // 修改loop0 exit prev 的 br 指令

        Instanceof(br, BranchInstr *, loop0_to_loop1_bb_map[cur ^ 1][info.exit_prev]->get_instrs().back().get()) {
            br->remove_use_def();
            br->set_bb2(nullptr);
            br->set_bb1(loop0_to_loop1_bb_map[cur][info.header]);
            br->set_reg(nullptr);
            loop0_to_loop1_bb_map[cur ^ 1][info.exit_prev]->succ.clear();
            loop0_to_loop1_bb_map[cur ^ 1][info.exit_prev]->succ.insert(loop0_to_loop1_bb_map[cur][info.header]);
            loop0_to_loop1_bb_map[cur][info.header]->prev.insert(loop0_to_loop1_bb_map[cur ^ 1][info.exit_prev]);
        }

        // 指令复制

        for (auto &bb : loop->get_all_bbs()) {
            inst_copy(info, bb, loop0_to_loop1_bb_map[cur][bb]);
        }
    }

    if (info.times == info.total_times) {
        // 处理header phi指令
        for (auto &inst : info.header->get_instrs()) {
            Instanceof(header_phi, PhiInstr *, inst.get()) {
                if (header_phi->get_bb_2_value().count(info.exit_prev)) {
                    header_phi->remove_use_def();
                    header_phi->get_bb_2_value().erase(info.exit_prev);
                    header_phi->add_use_def();
                } else
                    break;
            }
        }
        // 最后处理loop1的exit_prev的br,将其改为jump实现跳转
        Instanceof(br, BranchInstr *, loop0_to_loop1_bb_map[cur][info.exit_prev]->get_instrs().back().get()) {
            br->remove_use_def();
            br->set_bb1(info.exit);
            br->set_bb2(nullptr);
            br->set_reg(nullptr);
            br->remove_use_def();
            loop0_to_loop1_bb_map[cur][info.exit_prev]->succ.clear();
        }
    } else { // 重建exit prev 到 header 的路径
        // phi
        // branch
        // cfg

        // 处理header phi指令
        for (auto &inst : info.header->get_instrs()) {
            Instanceof(header_phi, PhiInstr *, inst.get()) {
                if (header_phi->get_bb_2_value().count(info.exit_prev)) {
                    header_phi->remove_use_def();
                    header_phi->get_bb_2_value()[loop0_to_loop1_bb_map[cur][info.exit_prev]] = static_pointer_cast<Reg>(find_dst(header_phi->get_bb_2_value()[info.exit_prev]));
                    header_phi->get_bb_2_value().erase(info.exit_prev);
                    header_phi->add_use_def();
                }
            }
            else break;
        }

        auto &last_br_ptr = loop0_to_loop1_bb_map[cur][info.exit_prev]->get_instrs().back();
        Instanceof(last_br, BranchInstr *, last_br_ptr.get()) {
            last_br->remove_use_def();
            last_br->set_reg(static_pointer_cast<Reg>(find_dst(info.cond->get_dst())));
            last_br->set_bb1(info.header);
            last_br->set_bb2(info.exit);
            last_br->add_use_def();
        }

        loop0_to_loop1_bb_map[cur][info.exit_prev]->succ.insert(info.header);
        info.header->prev.insert(loop0_to_loop1_bb_map[cur][info.exit_prev]);
    }

    loop0_to_loop1_bb_map[cur][info.exit_prev]->succ.insert(info.exit);
    info.exit->prev.erase(info.exit_prev);
    info.exit->prev.insert(loop0_to_loop1_bb_map[cur][info.exit_prev]);
    for (auto &inst_exit : info.exit->get_instrs()) {
        Instanceof(phi_exit, PhiInstr *, inst_exit.get()) {
            if (phi_exit->get_bb_2_value().count(info.exit_prev)) {
                phi_exit->remove_use_def();
                phi_exit->get_bb_2_value()[loop0_to_loop1_bb_map[cur][info.exit_prev]] = static_pointer_cast<Reg>(find_dst(phi_exit->get_bb_2_value()[info.exit_prev]));
                phi_exit->get_bb_2_value().erase(info.exit_prev);
                phi_exit->add_use_def();
            }
        }
        else break;
    }

    // 另外,将所有在循环里面定义,循环外面使用的reg经过reg map 一下
    for (auto &bb : loop->get_all_bbs()) {
        // loop0_to_loop1_bb_map[cur][bb] = Factory::get_new_bb(loop,func);
        //  新建 var
        for (auto &inst : bb->get_instrs()) {
            Instanceof(def, DefInstr *, inst.get()) {
                if (!def->def())
                    continue;
                unordered_set<Instruction *> target_inst; // use place modify bug fix
                for (auto &use : func->use_place[def->def()->get_id()]) {
                    if (use->get_belongto_bb()->get_belong_to_loop() != loop) {
                        target_inst.insert(use);
                    }
                }
                for (auto &use : target_inst) {
                    use->change_use_name(def->def()->get_id(), static_pointer_cast<Reg>(loop0_to_loop1_dst_map[cur][def->def()]));
                }
            }
        }
    }
}

void loop_unroll_complete(Function* func) {
    if(func->is_lib() || func->get_bbs().empty())
        return;
    func->cfg->build();
    func->cfg->dom_build();
    func->cfg->loop_analysis();
    for(auto &loop:func->get_loops()) {
        if(!loop->children.empty())
            continue ;
        if(!unroll_valid(loop.get())) {
            continue ;
        }
        auto info = get_loop_unroll_info(loop.get());
        if(info.times == 1)
            return;
        loop_unroll_complete(func,loop.get(),info);
    }
}

void loop_unroll_complete(CompileUnit* compileUnit) {
    for(auto &func : compileUnit->get_functions()) {
        Factory::reset(func.get());
        loop_unroll_complete(func.get());
        func->set_bb_max_num(Factory::bb_num);
        func->set_reg_max_num(Factory::reg_num);
    }
    info("end loop unroll complete");
}
}