//
// Created by 11069 on 11/7/2023.
//
// 该 pass 先于loop_unroll_complete 完成
// 该 pass 将步数变为两倍
// 需要条件:
// 无break,continue
// 入口出口唯一
// pre header 的 最后一条指令如果是 jump,那么循环终止条件,循环开始条件,循环步长必须已知,交给loop unroll complete解决应该就行了
// 如果是branch,那么才能使用loop unroll dual

#include "opt.h"
using namespace Ir;

namespace opt{
struct DualUnrollInfo{
    shared_ptr<Reg> step = nullptr;
    shared_ptr<Reg> end = nullptr;
    BinaryInstr * upd = nullptr;
    BinaryInstr * cond = nullptr;
    PhiInstr * phi = nullptr;
    BranchInstr * br = nullptr;
    BasicBlock * into_loop = nullptr;
    BasicBlock * header = nullptr;
    BasicBlock * exit_prev = nullptr;
    BasicBlock * exit = nullptr;
    BasicBlock * exit_tmp = nullptr;
    shared_ptr<Reg> new_end = nullptr;
    shared_ptr<Reg> begin = nullptr;
};
static unordered_map<BasicBlock *,BasicBlock *> loop0_to_loop1_bb_map[2];
static unordered_map<shared_ptr<Type>,shared_ptr<Type>> loop0_to_loop1_dst_map[2];
static int cur;

static shared_ptr<Type> find_dst(shared_ptr<Type> old_dst) {
    if(loop0_to_loop1_dst_map[cur].count(old_dst))
        return loop0_to_loop1_dst_map[cur][old_dst];
    return old_dst;
}

static BasicBlock * find_bb(BasicBlock * old_bb) {
    if(loop0_to_loop1_bb_map[cur].count(old_bb))
        return loop0_to_loop1_bb_map[cur][old_bb];
    return old_bb;
}

void inst_copy(DualUnrollInfo info,BasicBlock * old_bb,BasicBlock * new_bb) {
    for(auto &inst:old_bb->get_instrs()) {
        unique_ptr<Instruction> new_inst = nullptr;
        Instanceof(binary,BinaryInstr*,inst.get()) {
            if(binary == info.cond && cur == 1) {
                new_inst = make_unique<BinaryInstr>(static_pointer_cast<Reg>(find_dst(binary->get_dst()))
                ,static_pointer_cast<Reg>(find_dst(binary->get_src_1())),info.new_end,binary->get_binary_type());
            } else {
                new_inst = make_unique<BinaryInstr>(static_pointer_cast<Reg>(find_dst(binary->get_dst())),
                static_pointer_cast<Reg>(find_dst(binary->get_src_1())), static_pointer_cast<Reg>(find_dst(binary->get_src_2())), binary->get_binary_type());
            }
        } else Instanceof(call,CallInstr*,inst.get()) {
            auto *args = new vector<shared_ptr<Type>>;
            args->reserve(call->get_params().size());
            for(auto &arg:call->get_params()) {
                args->push_back(find_dst(arg));
            }
            if(call->is_void_flag()) {
                new_inst = make_unique<CallInstr>(call->get_function(),*args);
            } else {
                new_inst = make_unique<CallInstr>(static_pointer_cast<Reg>(find_dst(call->get_dst())),call->get_function(),*args);
            }

        } else Instanceof(br,BranchInstr*,inst.get()) {
            if(old_bb == info.exit_prev) {
                if(cur == 0) {
                    new_inst = make_unique<BranchInstr>(loop0_to_loop1_bb_map[1][info.header]);
                } else {
                    new_inst = make_unique<BranchInstr>(
                        static_pointer_cast<Reg>(find_dst(br->get_reg())),
                            loop0_to_loop1_bb_map[0][info.header], info.exit_tmp);
                }
            } else new_inst = make_unique<BranchInstr>(static_pointer_cast<Reg>(find_dst(br->get_reg())), find_bb(br->get_bb_1()), find_bb(br->get_bb_2()));
        } else Instanceof(store,StoreInstr*,inst.get()) {
            new_inst = make_unique<StoreInstr>(static_pointer_cast<Reg>(find_dst(store->get_val())), static_pointer_cast<MemObject>(find_dst(store->get_addr())));
        } else Instanceof(storeGlobal,StoreGlobal*,inst.get()) {
            new_inst = make_unique<StoreGlobal>(static_pointer_cast<Reg>(find_dst(storeGlobal->get_val())), static_pointer_cast<Reg>(find_dst(storeGlobal->get_addr())));
        } else Instanceof(loadVector,LoadVector*,inst.get()) {
            auto *indexs = new vector<shared_ptr<Reg>>;
            indexs->reserve(loadVector->get_indexs().size());
            for(auto &index:loadVector->get_indexs()) {
                indexs->push_back(static_pointer_cast<Reg>(find_dst(index)));
            }
            new_inst = make_unique<LoadVector>(static_pointer_cast<Vector>(find_dst(loadVector->get_dst())),
                                           static_pointer_cast<Vector>(find_dst(loadVector->get_addr())), *indexs);
        } else Instanceof(loadArray,LoadArray*,inst.get()) {
            auto *indexs = new vector<shared_ptr<Reg>>;
            indexs->reserve(loadArray->get_indexs().size());
            for(auto &index:loadArray->get_indexs()) {
                indexs->push_back(static_pointer_cast<Reg>(find_dst(index)));
            }
            new_inst = make_unique<LoadArray>(static_pointer_cast<MemObject>(find_dst(loadArray->get_dst())),
                                          static_pointer_cast<Vector>(find_dst(loadArray->get_addr())), *indexs);
        } else Instanceof(ret,Ret*,inst.get()) {
            if(ret->is_void_flag()) {
                new_inst = make_unique<Ret>();
            } else {
                new_inst = make_unique<Ret>(static_pointer_cast<Reg>(find_dst(ret->get_src())));
            }
        } else Instanceof(convert,Convert*,inst.get()) {
            new_inst = make_unique<Convert>(static_pointer_cast<Reg>(find_dst(convert->get_dst())), static_pointer_cast<Reg>(find_dst(convert->get_src())));
        } else Instanceof(load,LoadInstr*,inst.get()) {
            new_inst = make_unique<LoadInstr>(static_pointer_cast<Reg>(find_dst(load->get_dst())), static_pointer_cast<MemObject>(find_dst(load->get_addr())));
        } else Instanceof(loadGlobal,LoadGlobal*,inst.get()) {
            new_inst = make_unique<LoadGlobal>(static_pointer_cast<Reg>(find_dst(loadGlobal->get_dst())), static_pointer_cast<Reg>(find_dst(loadGlobal->get_addr())));
        } else Instanceof(phi,PhiInstr*,inst.get()) { // 只有header的phi会有外界的传入,这里修改一下就行了
            auto *bb2value = new unordered_map<BasicBlock*, shared_ptr<Reg>>;
            if(old_bb == info.header) { // 对于 header 的phi,来自外界的不copy,来自back edge copy 但是是指向loop 0
                if(cur == 0) {
                    (*bb2value)[info.into_loop] = phi->get_bb_2_value()[info.into_loop];
                    (*bb2value)[loop0_to_loop1_bb_map[1][info.exit_prev]] = static_pointer_cast<Reg>(
                        loop0_to_loop1_dst_map[1][phi->get_bb_2_value()[info.exit_prev]]);
                } else {
                    (*bb2value)[loop0_to_loop1_bb_map[0][info.exit_prev]] = static_pointer_cast<Reg>(
                        loop0_to_loop1_dst_map[0][phi->get_bb_2_value()[info.exit_prev]]);
                }
            } else {
                for(auto &[bb,value]:phi->get_bb_2_value()) {
                    (*bb2value)[find_bb(bb)] = static_pointer_cast<Reg>(find_dst(value));
                }
            }
            new_inst = make_unique<PhiInstr>(static_pointer_cast<Reg>(find_dst(phi->get_dst())),*bb2value);
        } else assert(false);
        new_bb->push_back(new_inst);
    }
}

static void loop_unroll_dual(Function * func,Loop* loop,DualUnrollInfo info) {
    loop0_to_loop1_bb_map[0].clear();
    loop0_to_loop1_bb_map[1].clear();
    loop0_to_loop1_dst_map[0].clear();
    loop0_to_loop1_dst_map[1].clear();
    // 新建two step 的 loop bb 和 reg dst
    cur = 0;
    for(auto &bb:loop->get_all_bbs()) {
        loop0_to_loop1_bb_map[cur][bb] = Factory::get_new_bb(nullptr,func);
        // 新建 var
        for(auto &inst:bb->get_instrs()) {
            Instanceof(phi,PhiInstr *,inst.get()) {
                loop0_to_loop1_dst_map[cur][phi->get_dst()] = Factory::get_reg(phi->get_dst()->get_type());
            } else Instanceof(loadArray,LoadArray *,inst.get()) {
                loop0_to_loop1_dst_map[cur][loadArray->get_dst()] = Factory::get_mem(loadArray->get_dst()->get_type());
            } else Instanceof(loadVector,LoadVector *,inst.get()) {
                auto dst = Factory::get_vector(loadVector->get_dst()->get_type());
                dst->get_dims().push_back(0);
                loop0_to_loop1_dst_map[cur][loadVector->get_dst()] = dst;
            } else Instanceof(def,DefInstr *,inst.get()) {
                if(def->def())
                    loop0_to_loop1_dst_map[cur][def->def()] = Factory::get_reg(def->def()->get_type());
            }
        }
    }
    cur = 1;
    for(auto &bb:loop->get_all_bbs()) {
        loop0_to_loop1_bb_map[cur][bb] = Factory::get_new_bb(nullptr,func);
        // 新建 var
        for(auto &inst:bb->get_instrs()) {
            Instanceof(phi,PhiInstr *,inst.get()) {
                loop0_to_loop1_dst_map[cur][phi->get_dst()] = Factory::get_reg(phi->get_dst()->get_type());
            } else Instanceof(loadArray,LoadArray *,inst.get()) {
                loop0_to_loop1_dst_map[cur][loadArray->get_dst()] = Factory::get_mem(loadArray->get_dst()->get_type());
            } else Instanceof(loadVector,LoadVector *,inst.get()) {
                auto dst = Factory::get_vector(loadVector->get_dst()->get_type());
                dst->get_dims().push_back(0);
                loop0_to_loop1_dst_map[cur][loadVector->get_dst()] = dst;
            } else Instanceof(def,DefInstr *,inst.get()) {
                if(def->def())
                loop0_to_loop1_dst_map[cur][def->def()] = Factory::get_reg(def->def()->get_type());
            }
        }
    }

    BasicBlock * exit_tmp = Factory::get_new_bb(nullptr,func);
    info.exit_tmp = exit_tmp;
    BasicBlock * pre_header_tmp = exit_tmp;
    auto* new_inst = new BinaryInstr(Factory::get_reg(info.end->get_type()),info.end,info.step,info.upd->get_binary_type() == ADD?SUB:ADD);
    info.new_end = new_inst->get_dst();
    { // insert new_inst into into_loop
        auto iter = info.into_loop->get_instrs().end();
        iter--; // br
        iter = info.into_loop->get_instrs().emplace(iter, new_inst);
        new_inst->set_belongto_bb(info.into_loop);
        new_inst->add_use_def();
        iter = info.into_loop->get_instrs().end();
    }
    // 对 pre header 进行改变,跳转到pre header tmp
    auto origin_into_loop_br_ptr = (--info.into_loop->get_instrs().end())->get();
    origin_into_loop_br_ptr->remove_use_def();
    auto origin_into_loop_br = dynamic_cast<BranchInstr*>(origin_into_loop_br_ptr);
    assert(origin_into_loop_br);
    auto new_into_loop_cnd = new BinaryInstr(Factory::get_reg(Bool),
                             info.phi->get_bb_2_value()[info.into_loop],new_inst->get_dst(),info.cond->get_binary_type());
    origin_into_loop_br->remove_use_def();
    origin_into_loop_br->set_reg(new_into_loop_cnd->get_dst());
    origin_into_loop_br->set_bb1(loop0_to_loop1_bb_map[0][info.header]);
    origin_into_loop_br->set_bb2(pre_header_tmp);
    origin_into_loop_br->add_use_def();
    auto iter = info.into_loop->get_instrs().end();
    iter--; // br
    iter = info.into_loop->get_instrs().emplace(iter, new_into_loop_cnd);
    new_into_loop_cnd->set_belongto_bb(info.into_loop);
    new_into_loop_cnd->add_use_def();

    shared_ptr<Reg> pre_header_tmp_cnd_reg = nullptr;
    unordered_map<shared_ptr<Reg>,shared_ptr<Reg>> loop1_step1_dst_map;
    // pre header tmp应该有 phi指令,其通过copy exit 的 phi实现的
    // 如果 exit 没有 phi 指令,那就先不对 pre header tmp的phi指令进行创建,待等下修改header时创建

    for (auto &exit_inst : info.exit->get_instrs()) {
        Instanceof(exit_phi, PhiInstr *, exit_inst.get()) {
            auto pre_header_tmp_phi = new PhiInstr(Factory::get_reg(exit_phi->get_dst()->get_type()));
            if (exit_phi->get_bb_2_value().count(info.into_loop)) { // into loop -> pre header
                exit_phi->remove_use_def();
                pre_header_tmp_phi->get_bb_2_value()[info.into_loop] = exit_phi->get_bb_2_value()[info.into_loop];
                exit_phi->get_bb_2_value()[pre_header_tmp] = pre_header_tmp_phi->get_dst();
                exit_phi->get_bb_2_value().erase(info.into_loop);
                exit_phi->add_use_def();
            }
            if (exit_phi->get_bb_2_value().count(info.exit_prev)) { // copy exit prev
                pre_header_tmp_phi->get_bb_2_value()[loop0_to_loop1_bb_map[1][info.exit_prev]] = static_pointer_cast<Reg>(find_dst(exit_phi->get_bb_2_value()[info.exit_prev]));
                loop1_step1_dst_map[static_pointer_cast<Reg>(find_dst(exit_phi->get_bb_2_value()[info.exit_prev]))] = pre_header_tmp_phi->get_dst();
            }
            pre_header_tmp->insert_after_phi(pre_header_tmp_phi);
        }
        else break;
    }


    // into loop 不去header了,成功跳转应该是 loop1 step0 header,pre header tmp 成功才去 header
    info.into_loop->succ.erase(info.header);
    info.into_loop->succ.insert(loop0_to_loop1_bb_map[0][info.header]);
    loop0_to_loop1_bb_map[0][info.header]->prev.insert(info.into_loop);
    info.header->prev.erase(info.into_loop);
    info.header->prev.insert(pre_header_tmp);
    pre_header_tmp->succ.insert(info.header);

    // into loop 不去exit了,失败跳转应该是 pre header tmp,pre header tmp 失败才取 exit
    info.into_loop->succ.erase(info.exit);
    info.into_loop->succ.insert(pre_header_tmp);
    pre_header_tmp->prev.insert(info.into_loop);
    pre_header_tmp->succ.insert(info.exit);
    info.exit->prev.erase(info.into_loop);
    info.exit->prev.insert(pre_header_tmp);

    // loop1 step0 bb relation rebuild and inst copy
    cur = 0;
    // 首先是 loop1 step0 的header 和 exit prev
    loop0_to_loop1_bb_map[cur][info.header]->prev.insert(info.into_loop);
    loop0_to_loop1_bb_map[cur][info.header]->prev.insert(loop0_to_loop1_bb_map[cur^1][info.exit_prev]);
    loop0_to_loop1_bb_map[cur][info.exit_prev]->succ.insert(loop0_to_loop1_bb_map[cur^1][info.header]);

    for(auto &loop_0_bb:loop->get_all_bbs()) {
        if(loop_0_bb != info.header) {
            // rebuild prev
            for(auto &prev:loop_0_bb->prev) {
                loop0_to_loop1_bb_map[cur][loop_0_bb]->prev.insert(loop0_to_loop1_bb_map[cur][prev]);
            }
        }
        if(loop_0_bb != info.exit_prev) {
            // rebuild succ
            for(auto &succ:loop_0_bb->succ) {
                loop0_to_loop1_bb_map[cur][loop_0_bb]->succ.insert(loop0_to_loop1_bb_map[cur][succ]);
            }
        }
    }

    // 指令copy
    for(auto &bb:loop->get_all_bbs()) {
        inst_copy(info,bb,loop0_to_loop1_bb_map[cur][bb]);
    }

    // loop1 step1 bb relation rebuild and inst copy

    cur = 1;
    // 首先是 loop1 step0 的header 和 exit prev
    loop0_to_loop1_bb_map[cur][info.header]->prev.insert(loop0_to_loop1_bb_map[cur^1][info.exit_prev]);
    loop0_to_loop1_bb_map[cur][info.exit_prev]->succ.insert(loop0_to_loop1_bb_map[cur^1][info.header]);
    loop0_to_loop1_bb_map[cur][info.exit_prev]->succ.insert(exit_tmp);

    for(auto &loop_0_bb:loop->get_all_bbs()) {
        if(loop_0_bb != info.header) {
            // rebuild prev
            for(auto &prev:loop_0_bb->prev) {
                loop0_to_loop1_bb_map[cur][loop_0_bb]->prev.insert(loop0_to_loop1_bb_map[cur][prev]);
            }
        }
        if(loop_0_bb != info.exit_prev) {
            // rebuild succ
            for(auto &succ:loop_0_bb->succ) {
                loop0_to_loop1_bb_map[cur][loop_0_bb]->succ.insert(loop0_to_loop1_bb_map[cur][succ]);
            }
        }
    }

    // 指令copy
    for(auto &bb:loop->get_all_bbs()) {
        inst_copy(info,bb,loop0_to_loop1_bb_map[cur][bb]);
    }

    //loop0 modify

    // 修改 header 的 phi,只有一个来源,就是exit prev
    for(auto &inst:info.header->get_instrs()) {
        Instanceof(phi,PhiInstr*,inst.get()) {
            phi->remove_use_def();
            auto exit_prev_reg = phi->get_bb_2_value()[info.exit_prev];
            auto loop1_reg = static_pointer_cast<Reg>(find_dst(exit_prev_reg));
            if(loop1_step1_dst_map.count(loop1_reg)) { // phi已经存在了
                phi->get_bb_2_value()[exit_tmp] = loop1_step1_dst_map.at(loop1_reg);
            } else { // phi不存在
                auto new_phi = new PhiInstr(Factory::get_reg(phi->get_dst()->get_type()));
                new_phi->get_bb_2_value()[loop0_to_loop1_bb_map[1][info.exit_prev]] = loop1_reg;
                new_phi->get_bb_2_value()[info.into_loop] = phi->get_bb_2_value()[info.into_loop];
                exit_tmp->insert_after_phi(new_phi);
                phi->get_bb_2_value()[exit_tmp] = new_phi->get_dst();
            }
            phi->get_bb_2_value().erase(info.into_loop);
            phi->get_bb_2_value().erase(info.exit_prev);
            phi->add_use_def();
        } else break;
    }

    // 修改exit prev的br
    info.br->remove_use_def();
    info.br->set_reg(nullptr);
    info.br->set_bb2(nullptr);
    info.br->set_bb1(info.exit);

    // 这里不需要修改循环里面定义,循环外面使用的变量了,因为exit 有 phi 指令,不会直接访问 循环bb的reg
    // 修改 exit_tmp 的 cnd
    for (auto &inst:exit_tmp->get_instrs()) {
        Instanceof(phi,PhiInstr*,inst.get()) {
            if(phi->get_bb_2_value().count(loop0_to_loop1_bb_map[1][info.exit_prev])
                && phi->get_bb_2_value().at(loop0_to_loop1_bb_map[1][info.exit_prev]) == static_pointer_cast<Reg>(loop0_to_loop1_dst_map[1][info.upd->get_dst()])) {
                pre_header_tmp_cnd_reg = phi->get_dst();
            }
        }
    }

    assert(pre_header_tmp_cnd_reg);
    auto pre_header_tmp_cnd = new BinaryInstr(Factory::get_reg(Bool),
                                              pre_header_tmp_cnd_reg,info.end,info.cond->get_binary_type());
    auto pre_header_tmp_br = new BranchInstr(pre_header_tmp_cnd->get_dst(),info.header,info.exit);
    pre_header_tmp->insert_after_phi(pre_header_tmp_br);
    pre_header_tmp->insert_after_phi(pre_header_tmp_cnd);

}

static DualUnrollInfo get_loop_unroll_info(Loop* loop) {
    DualUnrollInfo info;
    // 单出口单入口
    BasicBlock *exit = nullptr;
    BasicBlock *into_loop = nullptr;
    Function * func = loop->get_belong_to_func();
    for(auto &bb:loop->get_all_bbs()) {
        for(auto &next:bb->succ) {
            if(!loop->get_all_bbs().count(next)) {
                if(exit)
                    assert(false);
                exit = next;
            }
        }
        for(auto &prev:bb->prev) {
            if(!loop->get_all_bbs().count(prev)) {
                if(into_loop)
                    assert(false);
                into_loop = prev;
            }
        }
    }
    BasicBlock * exit_prev = nullptr;
    for(auto &bb:exit->prev) {
        if(loop->get_all_bbs().count(bb)) {
            if(exit_prev)
                assert(false);
            exit_prev = bb;
        }
    }

    BasicBlock * back_edge = nullptr;

    for(auto &bb:loop->get_header()->prev) {
        if(loop->get_all_bbs().count(bb)) {
            back_edge = bb;
        }
    }

    auto &into_loop_br_ptr = into_loop->get_instrs().back();
    Instanceof(into_loop_br,BranchInstr*,into_loop_br_ptr.get()) {
        if(into_loop_br->get_reg()== nullptr) {
            assert(false);
        }
    }

    info.exit_prev = exit_prev;
    info.exit = exit;
    info.header = loop->get_header();
    info.into_loop = into_loop;

    Instanceof(br,BranchInstr *,exit_prev->get_instrs().back().get()) {
        info.br = br;
        Instanceof(cond,BinaryInstr*,func->def_place[br->get_reg()->get_id()]) {
            info.cond = cond;
            info.end = cond->get_src_2();
            Instanceof(upd,BinaryInstr*,func->def_place[cond->get_src_1()->get_id()]) {
                // upd 的src为步长,如果upd为 +,判断条件- step,如果upd 为 -,判断条件 +step
                info.step = upd->get_src_2();
                info.upd = upd;
                Instanceof(phi,PhiInstr*,func->def_place[upd->get_src_1()->get_id()]) {
                    info.phi = phi;
                    info.begin = phi->get_bb_2_value()[info.into_loop];
                }
            } else assert(false);
        } else assert(false);
    } else assert(false);

    return info;
}

static bool is_valid_cmp(BinaryInstr * cond) { // 先兼容所有的吧,后面看实际展开的实现难度
    switch (cond->get_binary_type()) {
    case SLT:
    case SLE:
    case SGT:
    case SGE:
        return true;
    case NE:
    case EQ:
    default:
        return false;
    }
}

bool is_loop_const(Reg * reg,Function* func, Loop * loop) {
    if(reg->get_id() < 0) {
        return true;
    }
    if(func->is_param(reg)) {
            return true;
    }
    auto reg_bb = func->def_place[reg->get_id()]->get_belongto_bb();
    return !loop->get_all_bbs().count(reg_bb);
}

bool swap_cond(BinaryInstr * cond) {
    auto reg_i = cond->get_src_1();
    auto reg_c = cond->get_src_2();
    switch (cond->get_binary_type()) {
    case SLT:
        cond->setBinaryType(SGT);
        break;
    case SLE:
        cond->setBinaryType(SGE);
        break;
    case SGT:
        cond->setBinaryType(SLT);
        break;
    case SGE:
        cond->setBinaryType(SLE);
        break;
    case NE:
    case EQ:
        break;
    default:
        return false;
    }
    shared_ptr<Reg> tmp_reg = Factory::get_reg(Int);
    cond->change_use_name(reg_i->get_id(),tmp_reg);
    cond->change_use_name(reg_c->get_id(),reg_i);
    cond->change_use_name(tmp_reg->get_id(),reg_c);
    return true;
}

void swap_upd(BinaryInstr * upd) {
    auto reg_i = upd->get_src_1();
    auto reg_c = upd->get_src_2();
    switch (upd->get_binary_type()) {
    case ADD:
        upd->setBinaryType(ADD);
        break;
    case SUB:
        upd->setBinaryType(SUB);
        break;
    default:
        return;
    }
    shared_ptr<Reg> tmp_reg = Factory::get_reg(Int);
    upd->change_use_name(reg_i->get_id(),tmp_reg);
    upd->change_use_name(reg_c->get_id(),reg_i);
    upd->change_use_name(tmp_reg->get_id(),reg_c);
}

static bool unroll_valid(Loop* loop) {
    Function* func = loop->get_belong_to_func();
    // one way in one way out
    BasicBlock *exit = nullptr;
    BasicBlock *into_loop = nullptr;
    if(loop->get_all_bbs().size() >= 4) return false;
    for(auto &bb:loop->get_all_bbs()) {
        for(auto &next:bb->succ) {
            if(!loop->get_all_bbs().count(next)) {
                if(exit)
                    return false;
                exit = next;
            }
        }
        for(auto &prev:bb->prev) {
            if(!loop->get_all_bbs().count(prev)) {
                if(into_loop)
                    return false;
                into_loop = prev;
            }
        }
    }
    if(!into_loop || !exit) {
        return false;
    }
    // no break no continue

    BasicBlock * back_edge = nullptr;
    BasicBlock * exit_prev = nullptr;
    for(auto &bb:exit->prev) {
        if(loop->get_all_bbs().count(bb)) {
            if(exit_prev)
                return false;
            exit_prev = bb;
        }
    }

    for(auto &bb:loop->get_header()->prev) {
        if(loop->get_all_bbs().count(bb)) {
            if(back_edge)
                return false;
            back_edge = bb;
        }
    }

    if(back_edge != exit_prev ) return false;
    auto &into_loop_br_ptr = into_loop->get_instrs().back();
    Instanceof(into_loop_br,BranchInstr*,into_loop_br_ptr.get()) {
        if(into_loop_br->get_reg()== nullptr) {
            return false;
        } else if(into_loop_br->get_bb_2() != exit || into_loop_br->get_bb_1() != loop->get_header()) {
            return false;
        }
    }

    // 指令不能太多,防止cache miss 过大

    int cost = 0;
    for(auto &bb:loop->get_all_bbs()) {
        for(auto &inst:bb->get_instrs()) {
            if(cost >= 40) return false;
            Instanceof(call,CallInstr *,inst.get()) {
                if(call->get_function()->is_lib()) {
                    cost += 20;
                } else {
                    cost += 10;
                    for(auto &func_bb:call->get_function()->get_bbs()) {
                      if(cost >= 40) return false;
                      cost += func_bb->get_instrs().size();
                    }
                }
            } else {
                cost += 1;
            }
        }
    }
    Reg* step,*end;

    Instanceof(br,BranchInstr *,exit_prev->get_instrs().back().get()) {
        if(!br->get_bb_2()) return false;
        Instanceof(cond,BinaryInstr*,func->def_place[br->get_reg()->get_id()]) {
            if(!is_valid_cmp(cond))
                return false;
            // src2 必须是 loop const
            auto reg_i = cond->get_src_1();
            auto reg_c = cond->get_src_2();
            if(is_loop_const(reg_i.get(),func,loop) == is_loop_const(reg_c.get(),func,loop)) {
                return false;
            }
            if(is_loop_const(reg_i.get(),func,loop) && !is_loop_const(reg_c.get(),func,loop)) {
                if (!swap_cond(cond)) {
                    return false;
                }
                std::swap(reg_i,reg_c);
            }
            end = cond->get_src_2().get();
            Instanceof(upd,BinaryInstr*,func->def_place[cond->get_src_1()->get_id()]) {
                // 必须是加减法
                if(upd->get_binary_type()!= ADD && upd->get_binary_type() != SUB) return false;
                // 必须有一个loop const 且该const是src2
                auto src1 = upd->get_src_1();
                auto src2 = upd->get_src_2();
                if(!is_loop_const(src2.get(),func,loop)) {
                    if(!is_loop_const(src1.get(),func,loop)) return false;
                    // 交换操作数
                    if(upd->get_binary_type() == SUB) return false;// swap 只针对add
                    swap_upd(upd);
                    std::swap(src1,src2);
                }
                step = src2.get();
                Instanceof(phi,PhiInstr*,func->def_place[upd->get_src_1()->get_id()]) {
                    if(phi->get_bb_2_value().size() > 2) return false;
                    if(phi->get_belongto_bb() != loop->get_header()) return false;
                } else return false;
            } else return false;
        } else return false;
    } else return false;

    if(step->get_id() == end->get_id() ) {
        if(step->get_id() < 0) { // 都是常数那就要判断一下是否在范围里面了
            if(step->get_type() == Int) {
                return step->get_num().int_num < 1<< 30 && end->get_num().int_num < 1<< 30 && step->get_num().int_num > -2147483647 && end->get_num().int_num > -2147483647;
            } else {
                return step->get_num().float_num < float(1<< 30) && end->get_num().float_num < float(1<< 30) && step->get_num().float_num > float(-2147483647) && end->get_num().float_num > float(-2147483647);
            }
        } else {
            return false;
        }
    }

    return true;
}

void loop_unroll_dual(Function* func) {
    func->cfg->build();
    func->cfg->dom_build();
    func->cfg->loop_analysis();
    for(auto &loop:func->get_loops()) {
        if(!loop->children.empty())
            continue ;
        if(!unroll_valid(loop.get())) {
            continue ;
        }
        auto info = get_loop_unroll_info(loop.get());
        loop_unroll_dual(func,loop.get(),info);
    }
}

void loop_unroll_dual(CompileUnit* compileUnit) {
    for(auto &func:compileUnit->get_functions()) {
        if(func->is_lib() || func->get_bbs().empty())
            continue ;
        Factory::reset(func.get());
        loop_unroll_dual(func.get());
        func->set_bb_max_num(Factory::bb_num);
        func->set_reg_max_num(Factory::reg_num);
    }
    info("end loop unroll dual");
}
}
