//
// Created by 11069 on 2/7/2023.
//

// done before inst_clean_up
// mark pure function to detect critical inst
// engineering a compiler 10.2
#include "opt.h"

using namespace Ir;

namespace opt {


std::unordered_set<string> ret_used;

void mark_pure_func(CompileUnit *compileUnit, Function *func) { // true function to set the attribute pure of the function
    func->visited = true;
    // 读/写 全局变量
    // 往非该函数栈空间写入数据不是全局变量(主要是数组参数)
    std::unordered_set<int> dirtyId; // 地址是非外部空间的
    // init dirtyId
    std::unordered_set<int> globalId; // 地址是非外部空间的
    for (auto &arg : func->get_params()) {
        Instanceof(vector, Ir::MemObject *, arg.get()) {
            dirtyId.insert(vector->get_id()); // 这个id是外部空间,任何的读都要被统计进去,如果该空间被store,则不是pure function
        }
    }



    for (auto &bb : func->get_bbs()) {
        for (auto &inst : bb->get_instrs()) {
            Instanceof(loadGlobal, Ir::LoadGlobal *, inst.get()) {
                func->pure = false;
                func->array_ssa_pure = false;
            }
            else Instanceof(loadArray, Ir::LoadArray *, inst.get()) {
                if (dirtyId.count(loadArray->get_addr()->get_id())) {
                    dirtyId.insert(loadArray->get_dst()->get_id());
                }

                if(loadArray->get_addr()->get_id() < 0) {
                    dirtyId.insert(loadArray->get_dst()->get_id());
                    globalId.insert(loadArray->get_dst()->get_id());
                }

                if(globalId.count(loadArray->get_addr()->get_id())) {
                    globalId.insert(loadArray->get_dst()->get_id());
                }
            }
            else Instanceof(store, Ir::StoreInstr *, inst.get()) {
                if (dirtyId.count(store->get_addr()->get_id())) {
                    func->pure = false;
                }
                if(globalId.count(store->get_addr()->get_id())) {
                    func->array_ssa_pure = false;
                }
            }
            else Instanceof(call, Ir::CallInstr *, inst.get()) {
                if (call->get_function()->get_name() == "swap") {
                    int x = 10;
                }
                if (!call->is_void_flag() && !func->use_place[call->def()->get_id()].empty()) {
                    ret_used.insert(call->get_function()->get_name());
                }
                if (call->get_function() == func)
                    continue;
                if (!call->get_function()->visited)
                    mark_pure_func(compileUnit, call->get_function());
                func->pure &= call->get_function()->pure;
                func->array_ssa_pure &= call->get_function()->array_ssa_pure;
            } else Instanceof(load,Ir::LoadInstr *,inst.get()) {
                if(dirtyId.count(load->get_addr()->get_id())) {
                    func->pure = false;
                }

                if(globalId.count(load->get_addr()->get_id())) {
                    func->array_ssa_pure = false;
                }
            } else Instanceof(store_global,Ir::StoreGlobal*,inst.get()) {
                func->pure = false;
                func->array_ssa_pure = false;
            } else Instanceof(loadVector,LoadVector*,inst.get()) {
                if (dirtyId.count(loadVector->get_addr()->get_id())) {
                    dirtyId.insert(loadVector->get_dst()->get_id());
                }

                if(loadVector->get_addr()->get_id() < 0) {
                    dirtyId.insert(loadVector->get_dst()->get_id());
                    globalId.insert(loadVector->get_dst()->get_id());
                }

                if(globalId.count(loadVector->get_addr()->get_id())) {
                    globalId.insert(loadVector->get_dst()->get_id());
                }
            }
        }
    }
}

void mark_pure_func(CompileUnit *compileUnit) {
    for (auto &func : compileUnit->get_functions()) {
        if (func->get_bbs().size() || !func->is_lib()) {
            func->pure = true;
            func->visited = false; // init func
            func->array_ssa_pure = true;
        } else {
            func->visited = true;
            func->pure = false;
            func->array_ssa_pure = false;
        }
    }

    for (auto &func : compileUnit->get_functions()) {
        if (func->get_name() == "main") {
            mark_pure_func(compileUnit, func.get());
            break;
        }
    }

    for (auto iter = compileUnit->get_functions().begin(); iter != compileUnit->get_functions().end();) {
        if (!iter->get()->visited) {
            iter = compileUnit->get_functions().erase(iter);
        } else {
            iter++;
        }
    }
}

void ret_opt(Ir::CompileUnit* compUnit) {
    ret_used.insert("main");
    for (auto &func : compUnit->get_functions()) {
        if (!func->is_lib() && !ret_used.count(func->get_name()) && !func->is_void()) {
            for (auto &bb : func->get_bbs()) {
              for (const auto &instr : bb->get_instrs()) {
                    Instanceof(ret, Ir::Ret*, instr.get()) {
                      ret->change_use_name(ret->get_src()->get_id(), Ir::Factory::get_zero_reg(ret->get_src()->get_type()));
                    }
                }
            }
        }
    }
}

}