//
// Created by shliba on 2023/7/13.
//
#include "opt.h"
namespace opt {
void replace_all_use(std::unordered_set<Ir::Instruction*>& users,
                    int src_id, std::shared_ptr<Ir::Reg>& new_reg) {

    while (!users.empty()) {


        auto first = *users.begin();
        first->change_use_name(src_id, new_reg);
    }
}

void mem2reg(Ir::Function* func) {
    if (func->is_lib()) {
        return ;
    }

    func->cfg->build();


    func->cfg->dom_build();


    // first stage, phi-function insertion

    std::unordered_map<Ir::MemObject*, std::unordered_set<Ir::BasicBlock*>> defs;
    std::unordered_map<Ir::MemObject*, Ir::BasicBlock*> alloca2bb;
    std::unordered_map<Ir::PhiInstr*, Ir::MemObject*> phi2instr;
    std::unordered_map<Ir::BasicBlock*, std::unordered_map<Ir::MemObject*, std::shared_ptr<Ir::Reg>>> reachingDefs;
    for (const auto &bb : func->get_bbs()) {
        for (const auto &instr : bb->get_instrs()) {
            Instanceof(res, Ir::AllocaInstr *, instr.get()) {
                Instanceof(array, Ir::Vector *, res->get_addr().get()) {
                    continue ;
                }
                defs[res->get_addr().get()] = {};
                alloca2bb[res->get_addr().get()] = bb.get();

            }
            Instanceof(res, Ir::StoreInstr *, instr.get()) {
                auto addr = res->get_addr();
                if (alloca2bb.count(addr.get())) {
                    defs[addr.get()].insert(bb.get());
                }
            }
        }
    }

    // first stage, phi-function insertion
    for (auto &pair : alloca2bb) {
        auto v = pair.first;

        std::unordered_set<Ir::BasicBlock *> F;
        std::unordered_set<Ir::BasicBlock *> W;
        for (auto &item : defs[v]) {
            W.insert(item);
        }
        while (!W.empty()) {
            auto X = W.begin();
            bool dirty = false;
            for (auto &Y : (*X)->df) {
                if (!F.count(Y)) {
                    F.insert(Y);
                    auto dst = Ir::Factory::get_reg(v->get_type());
                    auto phi = Ir::Factory::get_phi(dst);
                    phi->set_belongto_bb(Y);
                    phi2instr[phi.get()] = v;
                    Y->get_instrs().push_front(std::move(phi));
                    if (!W.count(Y)) {
                        W.insert(Y);
                        if (Y == *X) {
                            dirty = true;
                        }
                    }
                }
            }
            if (!dirty) {
                W.erase(X);
            }
        }
    }


        //Variable renaming
        std::list<Ir::BasicBlock*> query;
        auto start = func->get_bbs().begin()->get();
        query.push_back(start);

        while (!query.empty()) {
            auto cur_bb = *query.begin();

            auto it = cur_bb->get_instrs().begin();
            auto &bb_defs = reachingDefs[cur_bb];
            while (it != cur_bb->get_instrs().end()) {

                auto instr = it->get();
                Instanceof(phi, Ir::PhiInstr*, instr) {
                    if (!phi2instr.count(phi)) {
                        it++;
                        continue;
                    }
                    bb_defs[phi2instr[phi]] = phi->get_dst();
                    //
                } else Instanceof(alloca, Ir::AllocaInstr*, instr) {
                    if (alloca2bb.count(alloca->get_addr().get())) {
                        it = cur_bb->get_instrs().erase(it);
                        continue;
                    }
                } else Instanceof(store, Ir::StoreInstr*, instr) {
                    auto addr = store->get_addr();
                    if (alloca2bb.count(addr.get())) {
                        bb_defs[addr.get()] = store->get_val();
                        instr->remove_use_def();
                        it = cur_bb->get_instrs().erase(it);
                        continue;
                    }
                } else Instanceof(load, Ir::LoadInstr*, instr) {
                    auto addr = load->get_addr().get();
                    if (alloca2bb.count(addr)) {
                        auto t_bb = cur_bb;
                        while (t_bb != nullptr && !reachingDefs[t_bb].count(addr)) {
                            t_bb = t_bb->idom_by;
                        }
                        std::shared_ptr<Ir::Reg> new_reg;
                        if (t_bb == nullptr) {
                            new_reg = Ir::Factory::get_zero_reg(addr->get_type());
                        } else {
                            new_reg = reachingDefs[t_bb][addr];
                        }

                        auto &users = func->use_place[load->get_dst()->get_id()];
                        replace_all_use(users, load->get_dst()->get_id(), new_reg);
                        it = cur_bb->get_instrs().erase(it);
                        continue;
                    }
                }
                it++;
            }
            for (auto &bb : cur_bb->succ) {

                for (const auto &instr : bb->get_instrs()) {
                    Instanceof(res ,Ir::PhiInstr*, instr.get()) {
                        if (!phi2instr.count(res)) {
                            continue;
                        }
                        auto addr = phi2instr[res];
                        auto& values = res->get_bb_2_value();
                        auto t_bb = cur_bb;
                        while (t_bb != nullptr && !reachingDefs[t_bb].count(addr)) {
                            t_bb = t_bb->idom_by;
                        }
                        std::shared_ptr<Ir::Reg> new_reg;
                        if (t_bb == nullptr) {
                            new_reg = Ir::Factory::get_zero_reg(addr->get_type());
                        } else {
                            new_reg = reachingDefs[t_bb][addr];
                        }
                        res->remove_use_def();
                        values[cur_bb] = new_reg;
//                        if (values.count(cur_bb)) {
//                            auto old  = values.find(cur_bb)->second;
//                            func->use_place[old->get_id()].erase(res);
//                            values[cur_bb] = new_reg;
//                            func->use_place[new_reg->get_id()].insert(res);
//                        } else {
//                            values[cur_bb] = new_reg;
//                            func->use_place[new_reg->get_id()].insert(res);
//                        }
                        res->add_use_def();
                    } else {
                        break;
                    }
                }
            }
            for (auto &item : cur_bb->idom) {
                if (cur_bb != item) {
                    query.push_back(item);
                }
            }
            query.pop_front();
        }

}

void mem2reg(Ir::CompileUnit *compileUnit) {

    for (auto &func : compileUnit->get_functions()) {
        if(func->is_lib() || func->get_bbs().empty())
            continue;
        Ir::Factory::reset(func.get());
        mem2reg(func.get());
        func->set_bb_max_num(Ir::Factory::bb_num);
        func->set_reg_max_num(Ir::Factory::reg_num);
    }
}
}
