//
// Created by 11069 on 19/7/2023.
//

#include "opt.h"

using namespace Ir;
namespace opt {

bool closelyConnect(Function* func,MemWrite* def,MemWrite* use) {
    return def->get_belongto_bb() == use->get_belongto_bb() ||
           (def->get_belongto_bb()->succ.size() == 1
            && def->get_belongto_bb()->succ.count(use->get_belongto_bb())); // 必须是同一个bb
}

void mem_const_propagation(Function* func,CompileUnit* comp) {
    unordered_set<Instruction*> useless_inst;
    for(auto &bb:func->get_bbs()) {
        for(auto iter = bb->get_instrs().begin();iter != bb->get_instrs().end();iter++) {
            auto inst = iter->get();
            Instanceof(mem_read,MemRead*,inst) {
                if(mem_read->call_read) {
                    continue ;
                }
                // 如果 write 的 addr 等于 read addr ,替换掉
                if(func->def_place.count(mem_read->get_scope()->get_id())) {
                    Instanceof(mem_write,MemWrite*,func->def_place[mem_read->get_scope()->get_id()]) {
                        if(mem_write->call_write)
                            continue ;
                        if(mem_write->get_addr() == mem_read->get_addr() && !mem_write->call_write) {
                            copy_propagation(func->use_place,mem_read->get_dst()->get_id(),mem_write->get_val());
                            mem_read->remove_use_def();
                            useless_inst.insert(mem_read);
                            //iter = bb->get_instrs().erase(iter);
                        }
                    }
                }

                //看能否常数传播
                Instanceof(vecotr,Vector*,mem_read->get_scope().get()) {
                    //addr是否是const看看,lol
                    auto def = func->def_place[mem_read->get_addr()->get_id()];
                    Instanceof(loadArray,LoadArray*,def) {
                        if( !((loadArray->get_addr()->is_constant())
                              || (func->get_name() == "main" && loadArray->get_addr() == mem_read->get_scope()))
                                )
                            continue;
                        auto is_const = true;
                        vector<int> indexs;
                        for(auto &index:loadArray->get_indexs()) {
                            is_const &= index->is_constant();
                            indexs.push_back(index->get_num().int_num);
                        }
                        if(is_const) {// index是常数,可以直接取值咯
                            auto dimension = vecotr->get_dims();
                            int base = 1,index = 0;
                            for(int i = dimension.size()-1;i >= 0;i--) {
                                index += base * indexs[i];
                                base *= dimension[i];
                            }
                            if(vecotr->get_type() == Int) {
                                auto value_vector = comp->get_global_int_arrays()[vecotr->get_name()];
                                int value = 0;
                                if(value_vector.size() > index) {
                                    value = value_vector[index];
                                }
                                RegConst tmp;
                                tmp.int_num=value;
                                auto reg_int = Factory::get_const_reg(Int,tmp);
                                copy_propagation(func->use_place,mem_read->get_dst()->get_id(),reg_int);
                            } else if (vecotr->get_type() == Float) {
                                auto value_vector = comp->get_global_float_arrays()[vecotr->get_name()];
                                float value = 0;
                                if(value_vector.size() > index) {
                                    value = value_vector[index];
                                }
                                RegConst tmp;
                                tmp.float_num=value;
                                auto reg_float = Factory::get_const_reg(Float,tmp);
                                copy_propagation(func->use_place,mem_read->get_dst()->get_id(),reg_float);
                            } else
                                assert(false);
                            mem_read->remove_use_def();
                            useless_inst.insert(mem_read);
                            //iter = bb->get_instrs().erase(iter);
                            continue ;
                        }
                    }
                }



            }
            else Instanceof(mem_write,MemWrite*,inst) {
                if(mem_write->call_write) {
                    continue;
                }
                if(func->def_place.count(mem_write->get_scope()->get_id())) { // 两次写的都是同一个地址
                    Instanceof(mem_write_def,MemWrite*,func->def_place[mem_write->get_scope()->get_id()]) {
                        if(mem_write_def->call_write)
                            continue ;
                        if(mem_write_def->get_addr() == mem_write->get_addr() && mem_write->read_before_write.empty() && closelyConnect(func,mem_write_def,mem_write)) {
                            mem_write_def->remove_use_def();
                            useless_inst.insert(mem_write_def);
                            Instanceof(reg,Reg*,mem_write_def->get_scope().get()) {
                                mem_write->change_use_name(mem_write_def->get_new_scope()->get_id(), static_pointer_cast<Reg>(mem_write_def->get_scope()));
                            }
                            else Instanceof(vector,Vector*,mem_write_def->get_scope().get()) {
                                mem_write->change_use_name(mem_write_def->get_new_scope()->get_id(), static_pointer_cast<Vector>(mem_write_def->get_scope()));
                            }
                            else Instanceof(mem,MemObject*,mem_write_def->get_scope().get()) {
                                mem_write->change_use_name(mem_write_def->get_new_scope()->get_id(), static_pointer_cast<MemObject>(mem_write_def->get_scope()));
                            }
                        }
                    }
                }
                // 从自己取值然后存给自己
                if(func->def_place.count(mem_write->get_val()->get_id())) {
                    Instanceof(mem_read_for_write,MemRead*,func->def_place[mem_write->get_val()->get_id()]) {
                        if(mem_read_for_write->call_read)
                            continue ;
                        if(mem_read_for_write->get_addr() == mem_write->get_addr() && mem_write->read_before_write.count(mem_read_for_write->get_dst())) {
                            mem_write->remove_use_def();
                            useless_inst.insert(mem_write);
                        }
                    }
                }
            }
        }
    }

    for(auto &bb:func->get_bbs()) {
        for(auto iter = bb->get_instrs().begin();iter != bb->get_instrs().end();) {
            auto inst = iter->get();
            if(useless_inst.count(inst)) {
                inst->remove_use_def();
                iter = bb->get_instrs().erase(iter);
            } else iter++;
        }
    }
}

void mem_const_propagation(CompileUnit* compileUnit) {
    for(auto &func:compileUnit->get_functions()) {
        Factory::reset(func.get());
        mem_const_propagation(func.get(),compileUnit);
        func->set_reg_max_num(Factory::reg_num);
        func->set_bb_max_num(Factory::bb_num);
    }
}
}