//
// Created by 11069 on 8/8/2023.
//

#include "opt.h"
using namespace Ir;
namespace opt {

static Function* memset_func;

static shared_ptr<Reg> end;
static shared_ptr<Reg> begin;

// _atri_memset(dest, len) len以字节为单位，要求4byte对齐
bool memset_insert_valid(Loop* loop) {
  Function* func = loop->get_belong_to_func();
  if(loop->get_all_bbs().size() != 1) return false;
  // 单出口单入口
  BasicBlock *exit = nullptr;
  BasicBlock *into_loop = nullptr;
  for(auto &bb:loop->get_all_bbs()) {
    for(auto &next:bb->succ) {
      if(!loop->get_all_bbs().count(next)) {
        if(exit)
          return false;
        exit = next;
      }
    }
    for(auto &prev:bb->prev) {
      if(!loop->get_all_bbs().count(prev)) {
        if(into_loop)
          return false;
        into_loop = prev;
      }
    }
  }
  if(!into_loop || !exit) {
    return false;
  }

  // 无break和continue
  BasicBlock * back_edge = nullptr;
  BasicBlock * exit_prev = nullptr;
  for(auto &bb:exit->prev) {
    if(loop->get_all_bbs().count(bb)) {
      if(exit_prev)
        return false;
      exit_prev = bb;
    }
  }

  for(auto &bb:loop->get_header()->prev) {
    if(loop->get_all_bbs().count(bb)) {
      if(back_edge)
        return false;
      back_edge = bb;
    }
  }

  if(back_edge != exit_prev ) return false;
  if(exit_prev != loop->get_header()) return false;
  if(loop->get_header()->get_instrs().size() != 6) return false; // phi gep store/mem set upd cnd br

  Instanceof(br,BranchInstr*,exit_prev->get_instrs().back().get()) {
    if(!br->get_bb_2()) return false;
    Instanceof(cnd,BinaryInstr*,func->def_place[br->get_reg()->get_id()]) {
      if(cnd->get_binary_type() != SLT) return false;
      if(is_loop_const(cnd->get_src_1().get(),func,loop) == is_loop_const(cnd->get_src_2().get(),func,loop)) return false;
      if(!is_loop_const(cnd->get_src_2().get(),func,loop))
        swap_cond(cnd);
      if(cnd->get_src_2()->get_type() != Int) return false;
      end = cnd->get_src_2();
      Instanceof(upd,BinaryInstr*,func->def_place[cnd->get_src_1()->get_id()]) {
        if(upd->get_binary_type() != ADD) return false;
        if(upd->get_src_2()->is_constant() == upd->get_src_1()->is_constant()) return false;
        if(!upd->get_src_2()->is_constant())
          swap_upd(cnd);
        if(upd->get_src_2()->get_type() != Int || upd->get_src_2()->get_num().int_num != 1) return false;
      }
    }
  }

  auto iter = exit_prev->get_instrs().begin();
  Instanceof(phi,PhiInstr*,iter->get()) {
    if(phi->get_bb_2_value().size() != 2) return false;
      begin = phi->get_bb_2_value()[into_loop];
      if(!is_loop_const(begin.get(),func,loop)) return false;
      iter ++;
      Instanceof(loadArray,LoadArray*,iter->get()) {
        iter++;
        Instanceof(store,StoreInstr*,iter->get()) {
          if(!store->get_val()->is_constant()) return false;
          if(store->get_val()->get_type() != Int) return false;
          if(store->get_val()->get_num().int_num != 0) return false;
          if(store->get_addr() != loadArray->get_dst()) return false;
        }

      } else return false;
  } else return false;

  return true;
}

void memset_insert(Loop* loop) {
  BasicBlock *exit = nullptr;
  BasicBlock *into_loop = nullptr;
  for(auto &bb:loop->get_all_bbs()) {
      for(auto &next:bb->succ) {
        if(!loop->get_all_bbs().count(next)) {
          if(exit)
          assert(false);
          exit = next;
        }
      }
      for(auto &prev:bb->prev) {
        if(!loop->get_all_bbs().count(prev)) {
          if(into_loop)
          assert(false);
          into_loop = prev;
        }
      }
  }


  auto bb = Factory::get_new_bb(nullptr,loop->get_belong_to_func());

  auto iter = loop->get_header()->get_instrs().begin();
  iter++;
  Instanceof(loadArray,LoadArray*,iter->get()) {
        auto new_dst = Factory::get_vector(loadArray->get_dst()->get_type());
        for(int i = 0;i<loadArray->get_addr()->get_dims().size()-loadArray->get_indexs().size()+1;i++) {
          new_dst->get_dims().push_back(loadArray->get_addr()->get_dims()[i]);
        }
        if(new_dst->get_dims().size() == 0) {
          new_dst->get_dims().push_back(loadArray->get_addr()->get_dims()[0]);
        }
        vector<shared_ptr<Reg>> new_index;
        for(int i = 0;i<loadArray->get_indexs().size()-1;i++) {
          new_index.push_back(loadArray->get_indexs()[i]);
        }
//        new_index.push_back(begin);
        auto new_loadvector = new LoadVector(new_dst,loadArray->get_addr(),new_index);
        auto new_len_binary = new BinaryInstr(Factory::get_reg(Int),end,begin,SUB);
        RegConst times;
        times.int_num = 4;
        auto new_mul_binary = new BinaryInstr(Factory::get_reg(Int),new_len_binary->get_dst(),Factory::get_const_reg(Int,times),MUL);
        vector<shared_ptr<Type>> params;
        params.push_back(new_loadvector->get_dst());
        params.push_back(new_mul_binary->get_dst());
//        params.push_back(Factory::get_zero_reg(Int));
        auto call_memset = new CallInstr(memset_func,params);

        bb->push_front(new BranchInstr(exit));
        bb->push_front(call_memset);
        bb->push_front(new_mul_binary);
        bb->push_front(new_len_binary);
        bb->push_front(new_loadvector);

        loop->get_header()->prev.clear();
        loop->get_header()->succ.clear();

        auto &into_loop_br_ptr = into_loop->get_instrs().back();
        Instanceof(into_loop_br,BranchInstr*,into_loop_br_ptr.get()) {
          if(into_loop_br->get_bb_1() == loop->get_header()) {
            into_loop_br->set_bb1(bb);
            into_loop->succ.erase(loop->get_header());
            into_loop->succ.insert(bb);
            bb->prev.insert(into_loop);
          }
        }
        exit->prev.erase(loop->get_header());
        exit->prev.insert(bb);
        bb->succ.insert(exit);
        for(auto &inst:loop->get_header()->get_instrs()) {
          inst->remove_use_def();
        }
  } else assert(false);
}

// _atri_memset(dest, len)
void memset_insert(Function* func) {
  func->cfg->build();
  func->cfg->dom_build();
  func->cfg->loop_analysis();
  begin = nullptr;
  end = nullptr;
  for(auto &loop:func->get_loops()) {
    if(loop->children.size())
      continue ;
    if(!memset_insert_valid(loop.get()))
      continue ;
    memset_insert(loop.get());
  }

}

void memset_insert(CompileUnit* compileUnit) {

  memset_func = nullptr;

  for(auto &func: compileUnit->get_functions()) {
    if(func->get_name() == memset_func_name) {
      memset_func = func.get();
    }
  }

  if(!memset_func) return;

  for(auto &func:compileUnit->get_functions()) {
    if(func->is_lib())
      continue ;
    Factory::reset(func.get());
    memset_insert(func.get());
    func->set_reg_max_num(Factory::reg_num);
    func->set_bb_max_num(Factory::bb_num);
  }
}
}


