//
// Created by 11067 on 2023/7/9.
//

#ifndef COMPILER_BITSET_H
#define COMPILER_BITSET_H

#include <cstddef>
#include <vector>


namespace atri {

    class Bitset {
        using FieldType = unsigned long long;
        size_t bitLength;
        FieldType *field;

        constexpr static size_t bitWidth() {
            return sizeof(FieldType) * 8;
        }

    public:
        size_t getBitLength() const;

        explicit Bitset(const size_t &bitLength);

        Bitset(const Bitset &rhs);

        Bitset &operator=(const Bitset &rhs);

        Bitset &operator=(Bitset &&rhs) noexcept;

        Bitset(Bitset &&rhs) noexcept;

        ~Bitset();

        Bitset &operator&=(const Bitset &rhs);

        Bitset &operator|=(const Bitset &rhs);

        Bitset &andNot(const Bitset &rhs);

        Bitset operator~();

        bool operator[](size_t idx) const;

        std::vector<int> bitView() const;


        class reference{
            size_t pos;
            FieldType *field;
        public:
            reference & operator=(bool rhs);
            explicit operator bool();
            reference(size_t pos, FieldType *field);
        };
        reference operator[](size_t idx);

        inline friend bool operator==(Bitset& lhs, Bitset& rhs) {
            if (lhs.bitLength != rhs.bitLength) {
                return false;
            }
            size_t wordLength = lhs.bitLength / bitWidth() + 1;
            for (size_t i = 0; i < wordLength; ++i) {
                if (lhs.field[i] != rhs.field[i]) {
                    return false;
                }
            }
            return true;
        }
        inline friend bool operator!=(Bitset& lhs, Bitset& rhs) {
            return !(lhs == rhs);
        }
    };

    inline Bitset operator&(Bitset& lhs, Bitset& rhs) {
        Bitset ret{lhs};
        ret &= rhs;
        return ret;
    }

    inline Bitset operator|(Bitset& lhs, Bitset& rhs) {
        Bitset ret{lhs};
        ret |= rhs;
        return ret;
    }

} // atri

#endif //COMPILER_BITSET_H
