//
// Created by 11067 on 2023/7/2.
//

#ifndef COMPILER_LOGGER_H
#define COMPILER_LOGGER_H

#include <ostream>
#include <iostream>
#include <any>
#include <ctime>

namespace atriLog {
    enum class LogLevel {
        Info,
        Warn,
        Error,
        Off,
    };
    constexpr std::string_view logType[]{"[INFO@", "[WARNING@", "[ERROR@"};
    constexpr std::string_view logColor[]{"\x1b[0m", "\x1b[33m", "\x1b[91m"};
    constexpr LogLevel level = LogLevel::Off;
    constexpr bool withTime = true;
    constexpr bool withColor = true;

    template<typename T1, typename... T2>
    void _atriLog1(const T1 &x, const T2 &... xs) {
        std::cerr << x;
        if constexpr (sizeof...(xs) > 0) {
            _atriLog1(xs...);
        }
    }

    template<typename T1, typename... T2>
    void _atriLog(LogLevel logLevel, const char *file, int line, const T1 &x, const T2 &... xs) {
        if constexpr (level != LogLevel::Off) {
            if (logLevel < level) {
                return;
            }
            if constexpr (withColor) {
                std::cerr << logColor[static_cast<int>(logLevel)];
            }
            std::cerr << logType[static_cast<int>(logLevel)];
            std::cerr << file << ':' << line;
            if constexpr (withTime) {
                std::time_t result = std::time(nullptr);
                std::cerr << '@' << result;
            }
            std::cerr << ']';
            if constexpr (withColor) {
                std::cerr << logColor[0];
            }
            std::cerr << x;
            if constexpr (sizeof...(xs) > 0) {
                _atriLog1(xs...);
            }
            std::cerr << std::endl;
        }
    }


#define info(...) atriLog::_atriLog(atriLog::LogLevel::Info, __FILE__, __LINE__, __VA_ARGS__)
#define warn(...) atriLog::_atriLog(atriLog::LogLevel::Warn, __FILE__, __LINE__, __VA_ARGS__)
#define error(...) atriLog::_atriLog(atriLog::LogLevel::Error, __FILE__, __LINE__, __VA_ARGS__)
#define whenDebug if constexpr (atriLog::level != atriLog::LogLevel::Off)
#define dbgvar(var) info(#var": ", var)
}
#endif //COMPILER_LOGGER_H
